'\nThere are a couple of classes documented in here:\n\n- :class:`.BaseName` as an abstact base class for almost everything.\n- :class:`.Name` used in a lot of places\n- :class:`.Completion` for completions\n- :class:`.BaseSignature` as a base class for signatures\n- :class:`.Signature` for :meth:`.Script.get_signatures` only\n- :class:`.ParamName` used for parameters of signatures\n- :class:`.Refactoring` for refactorings\n- :class:`.SyntaxError` for :meth:`.Script.get_syntax_errors` only\n\nThese classes are the much biggest part of the API, because they contain\nthe interesting information about all operations.\n'
_F='class'
_E='function'
_D='os.path'
_C=True
_B=False
_A=None
import re
from pathlib import Path
from typing import Optional
from parso.tree import search_ancestor
from jedi import settings
from jedi import debug
from jedi.inference.utils import unite
from jedi.cache import memoize_method
from jedi.inference.compiled.mixed import MixedName
from jedi.inference.names import ImportName,SubModuleName
from jedi.inference.gradual.stub_value import StubModuleValue
from jedi.inference.gradual.conversion import convert_names,convert_values
from jedi.inference.base_value import ValueSet,HasNoContext
from jedi.api.keywords import KeywordName
from jedi.api import completion_cache
from jedi.api.helpers import filter_follow_imports
def _sort_names_by_start_pos(names):return sorted(names,key=lambda s:s.start_pos or(0,0))
def defined_names(inference_state,value):
	'\n    List sub-definitions (e.g., methods in class).\n\n    :type scope: Scope\n    :rtype: list of Name\n    '
	try:A=value.as_context()
	except HasNoContext:return[]
	filter=next(A.get_filters());B=[A for A in filter.values()];return[Name(inference_state,A)for A in _sort_names_by_start_pos(B)]
def _values_to_definitions(values):return[Name(A.inference_state,A.name)for A in values]
class BaseName:
	'\n    The base class for all definitions, completions and signatures.\n    ';_mapping={'posixpath':_D,'riscospath':_D,'ntpath':_D,'os2emxpath':_D,'macpath':_D,'genericpath':_D,'posix':'os','_io':'io','_functools':'functools','_collections':'collections','_socket':'socket','_sqlite3':'sqlite3'};_tuple_mapping=dict((tuple(A.split('.')),B)for(A,B)in{'argparse._ActionsContainer':'argparse.ArgumentParser'}.items())
	def __init__(A,inference_state,name):A._inference_state=inference_state;A._name=name;'\n        An instance of :class:`parso.python.tree.Name` subclass.\n        ';A.is_keyword=isinstance(A._name,KeywordName)
	@memoize_method
	def _get_module_context(self):return self._name.get_root_context()
	@property
	def module_path(self):
		'\n        Shows the file path of a module. e.g. ``/usr/lib/python3.9/os.py``\n        ';A=self._get_module_context()
		if A.is_stub()or not A.is_compiled():B=self._get_module_context().py__file__();return B
	@property
	def name(self):"\n        Name of variable/function/class/module.\n\n        For example, for ``x = None`` it returns ``'x'``.\n\n        :rtype: str or None\n        ";return self._name.get_public_name()
	@property
	def type(self):
		"\n        The type of the definition.\n\n        Here is an example of the value of this attribute.  Let's consider\n        the following source.  As what is in ``variable`` is unambiguous\n        to Jedi, :meth:`jedi.Script.infer` should return a list of\n        definition for ``sys``, ``f``, ``C`` and ``x``.\n\n        >>> from jedi import Script\n        >>> source = '''\n        ... import keyword\n        ...\n        ... class C:\n        ...     pass\n        ...\n        ... class D:\n        ...     pass\n        ...\n        ... x = D()\n        ...\n        ... def f():\n        ...     pass\n        ...\n        ... for variable in [keyword, f, C, x]:\n        ...     variable'''\n\n        >>> script = Script(source)\n        >>> defs = script.infer()\n\n        Before showing what is in ``defs``, let's sort it by :attr:`line`\n        so that it is easy to relate the result to the source code.\n\n        >>> defs = sorted(defs, key=lambda d: d.line)\n        >>> print(defs)  # doctest: +NORMALIZE_WHITESPACE\n        [<Name full_name='keyword', description='module keyword'>,\n         <Name full_name='__main__.C', description='class C'>,\n         <Name full_name='__main__.D', description='instance D'>,\n         <Name full_name='__main__.f', description='def f'>]\n\n        Finally, here is what you can get from :attr:`type`:\n\n        >>> defs = [d.type for d in defs]\n        >>> defs[0]\n        'module'\n        >>> defs[1]\n        'class'\n        >>> defs[2]\n        'instance'\n        >>> defs[3]\n        'function'\n\n        Valid values for type are ``module``, ``class``, ``instance``, ``function``,\n        ``param``, ``path``, ``keyword``, ``property`` and ``statement``.\n\n        ";A=self;B=A._name.tree_name;C=_B
		if B is not _A:
			D=B.get_definition()
			if D is not _A and D.type=='import_from'and B.is_definition():C=_C
		if isinstance(A._name,SubModuleName)or C:
			for E in A._name.infer():return E.api_type
		return A._name.api_type
	@property
	def module_name(self):"\n        The module name, a bit similar to what ``__name__`` is in a random\n        Python module.\n\n        >>> from jedi import Script\n        >>> source = 'import json'\n        >>> script = Script(source, path='example.py')\n        >>> d = script.infer()[0]\n        >>> print(d.module_name)  # doctest: +ELLIPSIS\n        json\n        ";return self._get_module_context().py__name__()
	def in_builtin_module(B):
		'\n        Returns True, if this is a builtin module.\n        ';A=B._get_module_context().get_value()
		if isinstance(A,StubModuleValue):return any(A.is_compiled()for A in A.non_stub_value_set)
		return A.is_compiled()
	@property
	def line(self):
		'The line where the definition occurs (starting with 1).';A=self._name.start_pos
		if A is _A:return
		return A[0]
	@property
	def column(self):
		'The column where the definition occurs (starting with 0).';A=self._name.start_pos
		if A is _A:return
		return A[1]
	def get_definition_start_position(A):
		'\n        The (row, column) of the start of the definition range. Rows start with\n        1, columns start with 0.\n\n        :rtype: Optional[Tuple[int, int]]\n        '
		if A._name.tree_name is _A:return
		B=A._name.tree_name.get_definition()
		if B is _A:return A._name.start_pos
		return B.start_pos
	def get_definition_end_position(A):
		'\n        The (row, column) of the end of the definition range. Rows start with\n        1, columns start with 0.\n\n        :rtype: Optional[Tuple[int, int]]\n        '
		if A._name.tree_name is _A:return
		B=A._name.tree_name.get_definition()
		if B is _A:return A._name.tree_name.end_pos
		if A.type in(_E,_F):
			C=B.get_last_leaf()
			if C.type=='newline':return C.get_previous_leaf().end_pos
			return C.end_pos
		return B.end_pos
	def docstring(B,raw=_B,fast=_C):
		'\n        Return a document string for this completion object.\n\n        Example:\n\n        >>> from jedi import Script\n        >>> source = \'\'\'\\\n        ... def f(a, b=1):\n        ...     "Document for function f."\n        ... \'\'\'\n        >>> script = Script(source, path=\'example.py\')\n        >>> doc = script.infer(1, len(\'def f\'))[0].docstring()\n        >>> print(doc)\n        f(a, b=1)\n        <BLANKLINE>\n        Document for function f.\n\n        Notice that useful extra information is added to the actual\n        docstring, e.g. function signatures are prepended to their docstrings.\n        If you need the actual docstring, use ``raw=True`` instead.\n\n        >>> print(script.infer(1, len(\'def f\'))[0].docstring(raw=True))\n        Document for function f.\n\n        :param fast: Don\'t follow imports that are only one level deep like\n            ``import foo``, but follow ``from foo import bar``. This makes\n            sense for speed reasons. Completing `import a` is slow if you use\n            the ``foo.docstring(fast=False)`` on every object, because it\n            parses all libraries starting with ``a``.\n        '
		if isinstance(B._name,ImportName)and fast:return''
		A=B._get_docstring()
		if raw:return A
		C=B._get_docstring_signature()
		if C and A:return C+'\n\n'+A
		else:return C+A
	def _get_docstring(A):return A._name.py__doc__()
	def _get_docstring_signature(A):return'\n'.join(A.to_string()for A in A._get_signatures(for_docstring=_C))
	@property
	def description(self):
		"\n        A description of the :class:`.Name` object, which is heavily used\n        in testing. e.g. for ``isinstance`` it returns ``def isinstance``.\n\n        Example:\n\n        >>> from jedi import Script\n        >>> source = '''\n        ... def f():\n        ...     pass\n        ...\n        ... class C:\n        ...     pass\n        ...\n        ... variable = f if random.choice([0,1]) else C'''\n        >>> script = Script(source)  # line is maximum by default\n        >>> defs = script.infer(column=3)\n        >>> defs = sorted(defs, key=lambda d: d.line)\n        >>> print(defs)  # doctest: +NORMALIZE_WHITESPACE\n        [<Name full_name='__main__.f', description='def f'>,\n         <Name full_name='__main__.C', description='class C'>]\n        >>> str(defs[0].description)\n        'def f'\n        >>> str(defs[1].description)\n        'class C'\n\n        ";D=' ';C=self;A=C.type;E=C._name.tree_name
		if A=='param':return A+D+C._name.to_string()
		if A in(_E,_F,'module','instance')or E is _A:
			if A==_E:A='def'
			return A+D+C._name.get_public_name()
		F=E.get_definition(include_setitem=_C)or E;B=F.get_code(include_prefix=_B);B=re.sub('#[^\\n]+\\n',D,B);B=re.sub('\\s+',D,B).strip();return B
	@property
	def full_name(self):
		"\n        Dot-separated path of this object.\n\n        It is in the form of ``<module>[.<submodule>[...]][.<object>]``.\n        It is useful when you want to look up Python manual of the\n        object at hand.\n\n        Example:\n\n        >>> from jedi import Script\n        >>> source = '''\n        ... import os\n        ... os.path.join'''\n        >>> script = Script(source, path='example.py')\n        >>> print(script.infer(3, len('os.path.join'))[0].full_name)\n        os.path.join\n\n        Notice that it returns ``'os.path.join'`` instead of (for example)\n        ``'posixpath.join'``. This is not correct, since the modules name would\n        be ``<module 'posixpath' ...>```. However most users find the latter\n        more practical.\n        ";B=self
		if not B._name.is_value_name:return
		A=B._name.get_qualified_names(include_module_names=_C)
		if A is _A:return
		A=list(A)
		try:A[0]=B._mapping[A[0]]
		except KeyError:pass
		return'.'.join(A)
	def is_stub(A):
		'\n        Returns True if the current name is defined in a stub file.\n        '
		if not A._name.is_value_name:return _B
		return A._name.get_root_context().is_stub()
	def is_side_effect(B):
		'\n        Checks if a name is defined as ``self.foo = 3``. In case of self, this\n        function would return False, for foo it would return True.\n        ';A=B._name.tree_name
		if A is _A:return _B
		return A.is_definition()and A.parent.type=='trailer'
	@debug.increase_indent_cm('goto on name')
	def goto(self,*,follow_imports=_B,follow_builtin_imports=_B,only_stubs=_B,prefer_stubs=_B):
		'\n        Like :meth:`.Script.goto` (also supports the same params), but does it\n        for the current name. This is typically useful if you are using\n        something like :meth:`.Script.get_names()`.\n\n        :param follow_imports: The goto call will follow imports.\n        :param follow_builtin_imports: If follow_imports is True will try to\n            look up names in builtins (i.e. compiled or extension modules).\n        :param only_stubs: Only return stubs for this goto call.\n        :param prefer_stubs: Prefer stubs to Python objects for this goto call.\n        :rtype: list of :class:`Name`\n        ';A=self
		if not A._name.is_value_name:return[]
		B=A._name.goto()
		if follow_imports:B=filter_follow_imports(B,follow_builtin_imports)
		B=convert_names(B,only_stubs=only_stubs,prefer_stubs=prefer_stubs);return[A if B==A._name else Name(A._inference_state,B)for B in B]
	@debug.increase_indent_cm('infer on name')
	def infer(self,*,only_stubs=_B,prefer_stubs=_B):
		"\n        Like :meth:`.Script.infer`, it can be useful to understand which type\n        the current name has.\n\n        Return the actual definitions. I strongly recommend not using it for\n        your completions, because it might slow down |jedi|. If you want to\n        read only a few objects (<=20), it might be useful, especially to get\n        the original docstrings. The basic problem of this function is that it\n        follows all results. This means with 1000 completions (e.g.  numpy),\n        it's just very, very slow.\n\n        :param only_stubs: Only return stubs for this goto call.\n        :param prefer_stubs: Prefer stubs to Python objects for this type\n            inference call.\n        :rtype: list of :class:`Name`\n        ";C=prefer_stubs;B=only_stubs;A=self;assert not(B and C)
		if not A._name.is_value_name:return[]
		D=convert_names([A._name],prefer_stubs=_C);E=convert_values(ValueSet.from_sets(A.infer()for A in D),only_stubs=B,prefer_stubs=C);F=[A.name for A in E];return[A if B==A._name else Name(A._inference_state,B)for B in F]
	def parent(A):
		'\n        Returns the parent scope of this identifier.\n\n        :rtype: Name\n        '
		if not A._name.is_value_name:return
		if A.type in(_E,_F,'param')and A._name.tree_name is not _A:C=A._name.tree_name.get_definition();D=search_ancestor(C,'funcdef','classdef','file_input');B=A._get_module_context().create_value(D).as_context()
		else:B=A._name.parent_context
		if B is _A:return
		while B.name is _A:B=B.parent_context
		return Name(A._inference_state,B.name)
	def __repr__(A):return'<%s %sname=%r, description=%r>'%(A.__class__.__name__,'full_'if A.full_name else'',A.full_name or A.name,A.description)
	def get_line_code(A,before=0,after=0):
		"\n        Returns the line of code where this object was defined.\n\n        :param before: Add n lines before the current line to the output.\n        :param after: Add n lines after the current line to the output.\n\n        :return str: Returns the line(s) of code or an empty string if it's a\n                     builtin.\n        "
		if not A._name.is_value_name:return''
		B=A._name.get_root_context().code_lines
		if B is _A:return''
		C=A._name.start_pos[0]-1;D=max(C-before,0);return''.join(B[D:C+after+1])
	def _get_signatures(A,for_docstring=_B):
		if A._name.api_type=='property':return[]
		if for_docstring and A._name.api_type=='statement'and not A.is_stub():return[]
		if isinstance(A._name,MixedName):return A._name.infer_compiled_value().get_signatures()
		B=convert_names([A._name],prefer_stubs=_C);return[B for A in B for B in A.infer().get_signatures()]
	def get_signatures(A):'\n        Returns all potential signatures for a function or a class. Multiple\n        signatures are typical if you use Python stubs with ``@overload``.\n\n        :rtype: list of :class:`BaseSignature`\n        ';return[BaseSignature(A._inference_state,B)for B in A._get_signatures()]
	def execute(A):'\n        Uses type inference to "execute" this identifier and returns the\n        executed objects.\n\n        :rtype: list of :class:`Name`\n        ';return _values_to_definitions(A._name.infer().execute_with_values())
	def get_type_hint(A):'\n        Returns type hints like ``Iterable[int]`` or ``Union[int, str]``.\n\n        This method might be quite slow, especially for functions. The problem\n        is finding executions for those functions to return something like\n        ``Callable[[int, str], str]``.\n\n        :rtype: str\n        ';return A._name.infer().get_type_hint()
class Completion(BaseName):
	'\n    ``Completion`` objects are returned from :meth:`.Script.complete`. They\n    provide additional information about a completion.\n    '
	def __init__(A,inference_state,name,stack,like_name_length,is_fuzzy,cached_name=_A):super().__init__(inference_state,name);A._like_name_length=like_name_length;A._stack=stack;A._is_fuzzy=is_fuzzy;A._cached_name=cached_name;A._same_name_completions=[]
	def _complete(A,like_name):
		C=''
		if settings.add_bracket_after_function and A.type==_E:C='('
		B=A._name.get_public_name()
		if like_name:B=B[A._like_name_length:]
		return B+C
	@property
	def complete(self):
		"\n        Only works with non-fuzzy completions. Returns None if fuzzy\n        completions are used.\n\n        Return the rest of the word, e.g. completing ``isinstance``::\n\n            isinstan# <-- Cursor is here\n\n        would return the string 'ce'. It also adds additional stuff, depending\n        on your ``settings.py``.\n\n        Assuming the following function definition::\n\n            def foo(param=0):\n                pass\n\n        completing ``foo(par`` would give a ``Completion`` which ``complete``\n        would be ``am=``.\n        "
		if self._is_fuzzy:return
		return self._complete(_C)
	@property
	def name_with_symbols(self):'\n        Similar to :attr:`.name`, but like :attr:`.name` returns also the\n        symbols, for example assuming the following function definition::\n\n            def foo(param=0):\n                pass\n\n        completing ``foo(`` would give a ``Completion`` which\n        ``name_with_symbols`` would be "param=".\n\n        ';return self._complete(_B)
	def docstring(A,raw=_B,fast=_C):
		'\n        Documented under :meth:`BaseName.docstring`.\n        '
		if A._like_name_length>=3:fast=_B
		return super().docstring(raw=raw,fast=fast)
	def _get_docstring(A):
		if A._cached_name is not _A:return completion_cache.get_docstring(A._cached_name,A._name.get_public_name(),lambda:A._get_cache())
		return super()._get_docstring()
	def _get_docstring_signature(A):
		if A._cached_name is not _A:return completion_cache.get_docstring_signature(A._cached_name,A._name.get_public_name(),lambda:A._get_cache())
		return super()._get_docstring_signature()
	def _get_cache(A):return super().type,super()._get_docstring_signature(),super()._get_docstring()
	@property
	def type(self):
		'\n        Documented under :meth:`BaseName.type`.\n        ';A=self
		if A._cached_name is not _A:return completion_cache.get_type(A._cached_name,A._name.get_public_name(),lambda:A._get_cache())
		return super().type
	def get_completion_prefix_length(A):"\n        Returns the length of the prefix being completed.\n        For example, completing ``isinstance``::\n\n            isinstan# <-- Cursor is here\n\n        would return 8, because len('isinstan') == 8.\n\n        Assuming the following function definition::\n\n            def foo(param=0):\n                pass\n\n        completing ``foo(par`` would return 3.\n        ";return A._like_name_length
	def __repr__(A):return'<%s: %s>'%(type(A).__name__,A._name.get_public_name())
class Name(BaseName):
	'\n    *Name* objects are returned from many different APIs including\n    :meth:`.Script.goto` or :meth:`.Script.infer`.\n    '
	def __init__(A,inference_state,definition):super().__init__(inference_state,definition)
	@memoize_method
	def defined_names(self):'\n        List sub-definitions (e.g., methods in class).\n\n        :rtype: list of :class:`Name`\n        ';A=self._name.infer();return sorted(unite(defined_names(self._inference_state,A)for A in A),key=lambda s:s._name.start_pos or(0,0))
	def is_definition(A):
		"\n        Returns True, if defined as a name in a statement, function or class.\n        Returns False, if it's a reference to such a definition.\n        "
		if A._name.tree_name is _A:return _C
		else:return A._name.tree_name.is_definition()
	def __eq__(A,other):B=other;return A._name.start_pos==B._name.start_pos and A.module_path==B.module_path and A.name==B.name and A._inference_state==B._inference_state
	def __ne__(A,other):return not A.__eq__(other)
	def __hash__(A):return hash((A._name.start_pos,A.module_path,A.name,A._inference_state))
class BaseSignature(Name):
	'\n    These signatures are returned by :meth:`BaseName.get_signatures`\n    calls.\n    '
	def __init__(B,inference_state,signature):A=signature;super().__init__(inference_state,A.name);B._signature=A
	@property
	def params(self):'\n        Returns definitions for all parameters that a signature defines.\n        This includes stuff like ``*args`` and ``**kwargs``.\n\n        :rtype: list of :class:`.ParamName`\n        ';return[ParamName(self._inference_state,A)for A in self._signature.get_param_names(resolve_stars=_C)]
	def to_string(A):'\n        Returns a text representation of the signature. This could for example\n        look like ``foo(bar, baz: int, **kwargs)``.\n\n        :rtype: str\n        ';return A._signature.to_string()
class Signature(BaseSignature):
	'\n    A full signature object is the return value of\n    :meth:`.Script.get_signatures`.\n    '
	def __init__(A,inference_state,signature,call_details):B=signature;super().__init__(inference_state,B);A._call_details=call_details;A._signature=B
	@property
	def index(self):'\n        Returns the param index of the current cursor position.\n        Returns None if the index cannot be found in the curent call.\n\n        :rtype: int\n        ';return self._call_details.calculate_index(self._signature.get_param_names(resolve_stars=_C))
	@property
	def bracket_start(self):'\n        Returns a line/column tuple of the bracket that is responsible for the\n        last function call. The first line is 1 and the first column 0.\n\n        :rtype: int, int\n        ';return self._call_details.bracket_leaf.start_pos
	def __repr__(A):return'<%s: index=%r %s>'%(type(A).__name__,A.index,A._signature.to_string())
class ParamName(Name):
	def infer_default(A):'\n        Returns default values like the ``1`` of ``def foo(x=1):``.\n\n        :rtype: list of :class:`.Name`\n        ';return _values_to_definitions(A._name.infer_default())
	def infer_annotation(A,**B):'\n        :param execute_annotation: Default True; If False, values are not\n            executed and classes are returned instead of instances.\n        :rtype: list of :class:`.Name`\n        ';return _values_to_definitions(A._name.infer_annotation(ignore_stars=_C,**B))
	def to_string(A):'\n        Returns a simple representation of a param, like\n        ``f: Callable[..., Any]``.\n\n        :rtype: str\n        ';return A._name.to_string()
	@property
	def kind(self):"\n        Returns an enum instance of :mod:`inspect`'s ``Parameter`` enum.\n\n        :rtype: :py:attr:`inspect.Parameter.kind`\n        ";return self._name.get_kind()