'\nEnvironments are a way to activate different Python versions or Virtualenvs for\nstatic analysis. The Python binary in that environment is going to be executed.\n'
_F='VIRTUAL_ENV'
_E='python.exe'
_D='python'
_C=False
_B=True
_A=None
import os,sys,hashlib,filecmp
from collections import namedtuple
from shutil import which
from jedi.cache import memoize_method,time_cache
from jedi.inference.compiled.subprocess import CompiledSubprocess,InferenceStateSameProcess,InferenceStateSubprocess
import parso
_VersionInfo=namedtuple('VersionInfo','major minor micro')
_SUPPORTED_PYTHONS=['3.12','3.11','3.10','3.9','3.8','3.7','3.6']
_SAFE_PATHS=['/usr/bin','/usr/local/bin']
_CONDA_VAR='CONDA_PREFIX'
_CURRENT_VERSION='%s.%s'%(sys.version_info.major,sys.version_info.minor)
class InvalidPythonEnvironment(Exception):'\n    If you see this exception, the Python executable or Virtualenv you have\n    been trying to use is probably not a correct Python version.\n    '
class _BaseEnvironment:
	@memoize_method
	def get_grammar(self):A='%s.%s'%(self.version_info.major,self.version_info.minor);return parso.load_grammar(version=A)
	@property
	def _sha256(self):
		A=self
		try:return A._hash
		except AttributeError:A._hash=_calculate_sha256_for_file(A.executable);return A._hash
def _get_info():return sys.executable,sys.prefix,sys.version_info[:3]
class Environment(_BaseEnvironment):
	'\n    This class is supposed to be created by internal Jedi architecture. You\n    should not create it directly. Please use create_environment or the other\n    functions instead. It is then returned by that function.\n    ';_subprocess=_A
	def __init__(A,executable,env_vars=_A):A._start_executable=executable;A._env_vars=env_vars;A._get_subprocess()
	def _get_subprocess(A):
		if A._subprocess is not _A and not A._subprocess.is_crashed:return A._subprocess
		try:A._subprocess=CompiledSubprocess(A._start_executable,env_vars=A._env_vars);B=A._subprocess._send(_A,_get_info)
		except Exception as C:raise InvalidPythonEnvironment('Could not get version information for %r: %r'%(A._start_executable,C))
		A.executable=B[0];'\n        The Python executable, matches ``sys.executable``.\n        ';A.path=B[1];'\n        The path to an environment, matches ``sys.prefix``.\n        ';A.version_info=_VersionInfo(*B[2]);"\n        Like :data:`sys.version_info`: a tuple to show the current\n        Environment's Python version.\n        ";return A._subprocess
	def __repr__(A):B='.'.join(str(A)for A in A.version_info);return'<%s: %s in %s>'%(A.__class__.__name__,B,A.path)
	def get_inference_state_subprocess(A,inference_state):return InferenceStateSubprocess(inference_state,A._get_subprocess())
	@memoize_method
	def get_sys_path(self):'\n        The sys path for this environment. Does not include potential\n        modifications from e.g. appending to :data:`sys.path`.\n\n        :returns: list of str\n        ';return self._get_subprocess().get_sys_path()
class _SameEnvironmentMixin:
	def __init__(A):A._start_executable=A.executable=sys.executable;A.path=sys.prefix;A.version_info=_VersionInfo(*sys.version_info[:3]);A._env_vars=_A
class SameEnvironment(_SameEnvironmentMixin,Environment):0
class InterpreterEnvironment(_SameEnvironmentMixin,_BaseEnvironment):
	def get_inference_state_subprocess(A,inference_state):return InferenceStateSameProcess(inference_state)
	def get_sys_path(A):return sys.path
def _get_virtual_env_from_var(env_var=_F):
	'Get virtualenv environment from VIRTUAL_ENV environment variable.\n\n    It uses `safe=False` with ``create_environment``, because the environment\n    variable is considered to be safe / controlled by the user solely.\n    ';A=os.environ.get(env_var)
	if A:
		if os.path.realpath(A)==os.path.realpath(sys.prefix):return _try_get_same_env()
		try:return create_environment(A,safe=_C)
		except InvalidPythonEnvironment:pass
def _calculate_sha256_for_file(path):
	A=hashlib.sha256()
	with open(path,'rb')as B:
		for C in iter(lambda:B.read(filecmp.BUFSIZE),b''):A.update(C)
	return A.hexdigest()
def get_default_environment():
	'\n    Tries to return an active Virtualenv or conda environment.\n    If there is no VIRTUAL_ENV variable or no CONDA_PREFIX variable set\n    set it will return the latest Python version installed on the system. This\n    makes it possible to use as many new Python features as possible when using\n    autocompletion and other functionality.\n\n    :returns: :class:`.Environment`\n    ';A=_get_virtual_env_from_var()
	if A is not _A:return A
	B=_get_virtual_env_from_var(_CONDA_VAR)
	if B is not _A:return B
	return _try_get_same_env()
def _try_get_same_env():
	A=SameEnvironment()
	if not os.path.basename(A.executable).lower().startswith(_D):
		if os.name=='nt':B='Scripts\\python.exe',_E
		else:B='bin/python%s.%s'%(sys.version_info[0],sys.version[1]),'bin/python%s'%sys.version_info[0],'bin/python'
		for D in B:
			C=os.path.join(sys.exec_prefix,D)
			if os.path.isfile(C):return Environment(C)
		return InterpreterEnvironment()
	return A
def get_cached_default_environment():
	A=os.environ.get(_F)or os.environ.get(_CONDA_VAR);B=_get_cached_default_environment()
	if A and os.path.realpath(A)!=os.path.realpath(B.path):_get_cached_default_environment.clear_cache();return _get_cached_default_environment()
	return B
@time_cache(seconds=600)
def _get_cached_default_environment():
	try:return get_default_environment()
	except InvalidPythonEnvironment:return InterpreterEnvironment()
def find_virtualenvs(paths=_A,*,safe=_B,use_environment_vars=_B):
	'\n    :param paths: A list of paths in your file system to be scanned for\n        Virtualenvs. It will search in these paths and potentially execute the\n        Python binaries.\n    :param safe: Default True. In case this is False, it will allow this\n        function to execute potential `python` environments. An attacker might\n        be able to drop an executable in a path this function is searching by\n        default. If the executable has not been installed by root, it will not\n        be executed.\n    :param use_environment_vars: Default True. If True, the VIRTUAL_ENV\n        variable will be checked if it contains a valid VirtualEnv.\n        CONDA_PREFIX will be checked to see if it contains a valid conda\n        environment.\n\n    :yields: :class:`.Environment`\n    ';D=paths
	if D is _A:D=[]
	C=set()
	if use_environment_vars:
		E=_get_virtual_env_from_var()
		if E is not _A:yield E;C.add(E.path)
		F=_get_virtual_env_from_var(_CONDA_VAR)
		if F is not _A:yield F;C.add(F.path)
	for A in D:
		if not os.path.isdir(A):continue
		A=os.path.abspath(A)
		for B in os.listdir(A):
			B=os.path.join(A,B)
			if B in C:continue
			C.add(B)
			try:G=_get_executable_path(B,safe=safe);yield Environment(G)
			except InvalidPythonEnvironment:pass
def find_system_environments(*,env_vars=_A):
	"\n    Ignores virtualenvs and returns the Python versions that were installed on\n    your system. This might return nothing, if you're running Python e.g. from\n    a portable version.\n\n    The environments are sorted from latest to oldest Python version.\n\n    :yields: :class:`.Environment`\n    "
	for A in _SUPPORTED_PYTHONS:
		try:yield get_system_environment(A,env_vars=env_vars)
		except InvalidPythonEnvironment:pass
def get_system_environment(version,*,env_vars=_A):
	"\n    Return the first Python environment found for a string of the form 'X.Y'\n    where X and Y are the major and minor versions of Python.\n\n    :raises: :exc:`.InvalidPythonEnvironment`\n    :returns: :class:`.Environment`\n    ";B=version;A=which(_D+B)
	if A:
		if A==sys.executable:return SameEnvironment()
		return Environment(A)
	if os.name=='nt':
		for A in _get_executables_from_windows_registry(B):
			try:return Environment(A,env_vars=env_vars)
			except InvalidPythonEnvironment:pass
	raise InvalidPythonEnvironment('Cannot find executable python%s.'%B)
def create_environment(path,*,safe=_B,env_vars=_A):
	'\n    Make it possible to manually create an Environment object by specifying a\n    Virtualenv path or an executable path and optional environment variables.\n\n    :raises: :exc:`.InvalidPythonEnvironment`\n    :returns: :class:`.Environment`\n    ';B=env_vars;A=path
	if os.path.isfile(A):_assert_safe(A,safe);return Environment(A,env_vars=B)
	return Environment(_get_executable_path(A,safe=safe),env_vars=B)
def _get_executable_path(path,safe=_B):
	"\n    Returns None if it's not actually a virtual env.\n    "
	if os.name=='nt':A=os.path.join(path,'Scripts',_E)
	else:A=os.path.join(path,'bin',_D)
	if not os.path.exists(A):raise InvalidPythonEnvironment('%s seems to be missing.'%A)
	_assert_safe(A,safe);return A
def _get_executables_from_windows_registry(version):
	import winreg as A;D=['SOFTWARE\\Python\\PythonCore\\{version}\\InstallPath','SOFTWARE\\Wow6432Node\\Python\\PythonCore\\{version}\\InstallPath','SOFTWARE\\Python\\PythonCore\\{version}-32\\InstallPath','SOFTWARE\\Wow6432Node\\Python\\PythonCore\\{version}-32\\InstallPath']
	for E in[A.HKEY_CURRENT_USER,A.HKEY_LOCAL_MACHINE]:
		for B in D:
			B=B.format(version=version)
			try:
				with A.OpenKey(E,B)as F:
					G=A.QueryValueEx(F,'')[0];C=os.path.join(G,_E)
					if os.path.isfile(C):yield C
			except WindowsError:pass
def _assert_safe(executable_path,safe):
	if safe and not _is_safe(executable_path):raise InvalidPythonEnvironment('The python binary is potentially unsafe.')
def _is_safe(executable_path):
	A=os.path.realpath(executable_path)
	if _is_unix_safe_simple(A):return _B
	for B in find_system_environments():
		if B.executable==A:return _B
		if B._sha256==_calculate_sha256_for_file(A):return _B
	return _C
def _is_unix_safe_simple(real_path):
	A=real_path
	if _is_unix_admin():return any(A.startswith(B)for B in _SAFE_PATHS)
	B=os.stat(A).st_uid;return B==0
def _is_unix_admin():
	try:return os.getuid()==0
	except AttributeError:return _C