'\nProjects are a way to handle Python projects within Jedi. For simpler plugins\nyou might not want to deal with projects, but if you want to give the user more\nflexibility to define sys paths and Python interpreters for a project,\n:class:`.Project` is the perfect way to allow for that.\n\nProjects can be saved to disk and loaded again, to allow project definitions to\nbe used across repositories.\n'
_E='module'
_D='__init__.py'
_C=False
_B=True
_A=None
import json
from pathlib import Path
from itertools import chain
from jedi import debug
from jedi.api.environment import get_cached_default_environment,create_environment
from jedi.api.exceptions import WrongVersion
from jedi.api.completion import search_in_module
from jedi.api.helpers import split_search_string,get_module_names
from jedi.inference.imports import load_module_from_path,load_namespace_from_path,iter_module_names
from jedi.inference.sys_path import discover_buildout_paths
from jedi.inference.cache import inference_state_as_method_param_cache
from jedi.inference.references import recurse_find_python_folders_and_files,search_in_file_ios
from jedi.file_io import FolderIO
_CONFIG_FOLDER='.jedi'
_CONTAINS_POTENTIAL_PROJECT='setup.py','.git','.hg','requirements.txt','MANIFEST.in','pyproject.toml'
_SERIALIZER_VERSION=1
def _try_to_skip_duplicates(func):
	def A(*E,**F):
		C=[];D=[]
		for A in func(*E,**F):
			B=A._name.tree_name
			if B is not _A and B in C:continue
			if A.type==_E and A.module_path is not _A:
				if A.module_path in D:continue
				D.append(A.module_path)
			yield A;C.append(B)
	return A
def _remove_duplicates_from_path(path):
	B=set()
	for A in path:
		if A in B:continue
		B.add(A);yield A
class Project:
	'\n    Projects are a simple way to manage Python folders and define how Jedi does\n    import resolution. It is mostly used as a parameter to :class:`.Script`.\n    Additionally there are functions to search a whole project.\n    ';_environment=_A
	@staticmethod
	def _get_config_folder_path(base_path):return base_path.joinpath(_CONFIG_FOLDER)
	@staticmethod
	def _get_json_path(base_path):return Project._get_config_folder_path(base_path).joinpath('project.json')
	@classmethod
	def load(B,path):
		'\n        Loads a project from a specific path. You should not provide the path\n        to ``.jedi/project.json``, but rather the path to the project folder.\n\n        :param path: The path of the directory you want to use as a project.\n        ';A=path
		if isinstance(A,str):A=Path(A)
		with open(B._get_json_path(A))as C:D,E=json.load(C)
		if D==1:return B(**E)
		else:raise WrongVersion('The Jedi version of this project seems newer than what we can handle.')
	def save(B):
		'\n        Saves the project configuration in the project in ``.jedi/project.json``.\n        ';C='path';A=dict(B.__dict__);A.pop('_environment',_A);A.pop('_django',_A);A={A.lstrip('_'):B for(A,B)in A.items()};A[C]=str(A[C]);B._get_config_folder_path(B._path).mkdir(parents=_B,exist_ok=_B)
		with open(B._get_json_path(B._path),'w')as D:return json.dump((_SERIALIZER_VERSION,A),D)
	def __init__(A,path,*,environment_path=_A,load_unsafe_extensions=_C,sys_path=_A,added_sys_path=(),smart_sys_path=_B):
		"\n        :param path: The base path for this project.\n        :param environment_path: The Python executable path, typically the path\n            of a virtual environment.\n        :param load_unsafe_extensions: Default False, Loads extensions that are not in the\n            sys path and in the local directories. With this option enabled,\n            this is potentially unsafe if you clone a git repository and\n            analyze it's code, because those compiled extensions will be\n            important and therefore have execution privileges.\n        :param sys_path: list of str. You can override the sys path if you\n            want. By default the ``sys.path.`` is generated by the\n            environment (virtualenvs, etc).\n        :param added_sys_path: list of str. Adds these paths at the end of the\n            sys path.\n        :param smart_sys_path: If this is enabled (default), adds paths from\n            local directories. Otherwise you will have to rely on your packages\n            being properly configured on the ``sys.path``.\n        ";C=sys_path;B=path
		if isinstance(B,str):B=Path(B).absolute()
		A._path=B;A._environment_path=environment_path
		if C is not _A:C=list(map(str,C))
		A._sys_path=C;A._smart_sys_path=smart_sys_path;A._load_unsafe_extensions=load_unsafe_extensions;A._django=_C;A.added_sys_path=list(map(str,added_sys_path));'The sys path that is going to be added at the end of the '
	@property
	def path(self):'\n        The base path for this project.\n        ';return self._path
	@property
	def sys_path(self):'\n        The sys path provided to this project. This can be None and in that\n        case will be auto generated.\n        ';return self._sys_path
	@property
	def smart_sys_path(self):'\n        If the sys path is going to be calculated in a smart way, where\n        additional paths are added.\n        ';return self._smart_sys_path
	@property
	def load_unsafe_extensions(self):'\n        Wheter the project loads unsafe extensions.\n        ';return self._load_unsafe_extensions
	@inference_state_as_method_param_cache()
	def _get_base_sys_path(self,inference_state):
		A=list(inference_state.environment.get_sys_path())
		try:A.remove('')
		except ValueError:pass
		return A
	@inference_state_as_method_param_cache()
	def _get_sys_path(self,inference_state,add_parent_paths=_B,add_init_paths=_C):
		'\n        Keep this method private for all users of jedi. However internally this\n        one is used like a public method.\n        ';B=inference_state;A=self;D=list(A.added_sys_path);E=[]
		if A._sys_path is _A:F=list(A._get_base_sys_path(B))
		else:F=list(A._sys_path)
		if A._smart_sys_path:
			E.append(str(A._path))
			if B.script_path is not _A:
				D+=map(str,discover_buildout_paths(B,B.script_path))
				if add_parent_paths:
					G=[]
					for C in B.script_path.parents:
						if C==A._path or A._path not in C.parents:break
						if not add_init_paths and C.joinpath(_D).is_file():continue
						G.append(str(C))
					D+=reversed(G)
		if A._django:E.append(str(A._path))
		H=E+F+D;return list(_remove_duplicates_from_path(H))
	def get_environment(A):
		if A._environment is _A:
			if A._environment_path is not _A:A._environment=create_environment(A._environment_path,safe=_C)
			else:A._environment=get_cached_default_environment()
		return A._environment
	def search(A,string,*,all_scopes=_C):"\n        Searches a name in the whole project. If the project is very big,\n        at some point Jedi will stop searching. However it's also very much\n        recommended to not exhaust the generator. Just display the first ten\n        results to the user.\n\n        There are currently three different search patterns:\n\n        - ``foo`` to search for a definition foo in any file or a file called\n          ``foo.py`` or ``foo.pyi``.\n        - ``foo.bar`` to search for the ``foo`` and then an attribute ``bar``\n          in it.\n        - ``class foo.bar.Bar`` or ``def foo.bar.baz`` to search for a specific\n          API type.\n\n        :param bool all_scopes: Default False; searches not only for\n            definitions on the top level of a module level, but also in\n            functions and classes.\n        :yields: :class:`.Name`\n        ";return A._search_func(string,all_scopes=all_scopes)
	def complete_search(A,string,**B):'\n        Like :meth:`.Script.search`, but completes that string. An empty string\n        lists all definitions in a project, so be careful with that.\n\n        :param bool all_scopes: Default False; searches not only for\n            definitions on the top level of a module level, but also in\n            functions and classes.\n        :yields: :class:`.Completion`\n        ';return A._search_func(string,complete=_B,**B)
	@_try_to_skip_duplicates
	def _search_func(self,string,complete=_C,all_scopes=_C):
		M=string;F=self;D=complete;from jedi import Script as R;N=R('',project=F);A=N._inference_state;O=N._get_module_context();debug.dbg('Search for string %s, complete=%s',M,D);J,G=split_search_string(M);E=G[0];S=E+'-stubs';T=recurse_find_python_folders_and_files(FolderIO(str(F._path)));P=[]
		for(H,I)in T:
			if I is _A:
				Q=H.get_base_name()
				if Q==E or Q==S:
					K=H.get_file_io(_D)
					try:C=load_module_from_path(A,K).as_context()
					except FileNotFoundError:
						K=H.get_file_io('__init__.pyi')
						try:C=load_module_from_path(A,K).as_context()
						except FileNotFoundError:C=load_namespace_from_path(A,H).as_context()
				else:continue
			else:
				P.append(I)
				if Path(I.path).name in(E+'.py',E+'.pyi'):C=load_module_from_path(A,I).as_context()
				else:continue
			debug.dbg('Search of a specific module %s',C);yield from search_in_module(A,C,names=[C.name],wanted_type=J,wanted_names=G,complete=D,convert=_B,ignore_imports=_B)
		for L in search_in_file_ios(A,P,E,complete=D):B=get_module_names(L.tree_node,all_scopes=all_scopes);B=[L.create_name(A)for A in B];B=_remove_imports(B);yield from search_in_module(A,L,names=B,wanted_type=J,wanted_names=G,complete=D,ignore_imports=_B)
		U=[A for A in F._get_sys_path(A)if A!=F._path];B=list(iter_module_names(A,O,U));yield from search_in_module(A,O,names=B,wanted_type=J,wanted_names=G,complete=D,convert=_B)
	def __repr__(A):return'<%s: %s>'%(A.__class__.__name__,A._path)
def _is_potential_project(path):
	for A in _CONTAINS_POTENTIAL_PROJECT:
		try:
			if path.joinpath(A).exists():return _B
		except OSError:continue
	return _C
def _is_django_path(directory):
	' Detects the path of the very well known Django library (if used) '
	try:
		with open(directory.joinpath('manage.py'),'rb')as A:return b'DJANGO_SETTINGS_MODULE'in A.read()
	except(FileNotFoundError,IsADirectoryError,PermissionError):return _C
def get_default_project(path=_A):
	'\n    If a project is not defined by the user, Jedi tries to define a project by\n    itself as well as possible. Jedi traverses folders until it finds one of\n    the following:\n\n    1. A ``.jedi/config.json``\n    2. One of the following files: ``setup.py``, ``.git``, ``.hg``,\n       ``requirements.txt`` and ``MANIFEST.in``.\n    ';A=path
	if A is _A:A=Path.cwd()
	elif isinstance(A,str):A=Path(A)
	D=A.absolute();B=_A;C=_A
	for dir in chain([D],D.parents):
		try:return Project.load(dir)
		except(FileNotFoundError,IsADirectoryError,PermissionError):pass
		except NotADirectoryError:continue
		if C is _A:
			if dir.joinpath(_D).exists():continue
			elif not dir.is_file():C=dir
		if _is_django_path(dir):E=Project(dir);E._django=_B;return E
		if B is _A and _is_potential_project(dir):B=dir
	if B is not _A:return Project(B)
	if C is not _A:return Project(C)
	F=A if A.is_dir()else A.parent;return Project(F)
def _remove_imports(names):return[A for A in names if A.tree_name is _A or A.api_type not in(_E,'namespace')]