"\nThis caching is very important for speed and memory optimizations. There's\nnothing really spectacular, just some decorators. The following cache types are\navailable:\n\n- ``time_cache`` can be used to cache something for just a limited time span,\n  which can be useful if there's user interaction and the user cannot react\n  faster than a certain time.\n\nThis module is one of the reasons why |jedi| is not thread-safe. As you can see\nthere are global variables, which are holding the cache information. Some of\nthese variables are being cleaned after every API usage.\n"
import time
from functools import wraps
from typing import Any,Dict,Tuple
from jedi import settings
from parso.cache import parser_cache
_time_caches={}
def clear_time_caches(delete_all=False):
	' Jedi caches many things, that should be completed after each completion\n    finishes.\n\n    :param delete_all: Deletes also the cache that is normally not deleted,\n        like parser cache, which is important for faster parsing.\n    ';global _time_caches
	if delete_all:
		for B in _time_caches.values():B.clear()
		parser_cache.clear()
	else:
		for A in _time_caches.values():
			for(C,(D,E))in list(A.items()):
				if D<time.time():del A[C]
def signature_time_cache(time_add_setting):
	'\n    This decorator works as follows: Call it with a setting and after that\n    use the function with a callable that returns the key.\n    But: This function is only called if the key is not available. After a\n    certain amount of time (`time_add_setting`) the cache is invalid.\n\n    If the given key is None, the function will not be cached.\n    ';D=time_add_setting
	def A(key_func):
		B={};_time_caches[D]=B
		def A(*F,**G):
			E=key_func(*F,**G);C=next(E)
			try:
				H,A=B[C]
				if H>time.time():return A
			except KeyError:pass
			A=next(E);I=getattr(settings,D)
			if C is not None:B[C]=time.time()+I,A
			return A
		return A
	return A
def time_cache(seconds):
	def A(func):
		B={}
		@wraps(func)
		def A(*C,**D):
			E=C,frozenset(D.items())
			try:
				F,A=B[E]
				if time.time()<F+seconds:return A
			except KeyError:pass
			A=func(*C,**D);B[E]=time.time(),A;return A
		A.clear_cache=lambda:B.clear();return A
	return A
def memoize_method(method):
	'A normal memoize function.';A=method
	@wraps(A)
	def B(self,*B,**C):
		G=self.__dict__.setdefault('_memoize_method_dct',{});D=G.setdefault(A,{});E=B,frozenset(C.items())
		try:return D[E]
		except KeyError:F=A(self,*B,**C);D[E]=F;return F
	return B