'\nType inference of Python code in |jedi| is based on three assumptions:\n\n* The code uses as least side effects as possible. Jedi understands certain\n  list/tuple/set modifications, but there\'s no guarantee that Jedi detects\n  everything (list.append in different modules for example).\n* No magic is being used:\n\n  - metaclasses\n  - ``setattr()`` / ``__import__()``\n  - writing to ``globals()``, ``locals()``, ``object.__dict__``\n* The programmer is not a total dick, e.g. like `this\n  <https://github.com/davidhalter/jedi/issues/24>`_ :-)\n\nThe actual algorithm is based on a principle I call lazy type inference.  That\nsaid, the typical entry point for static analysis is calling\n``infer_expr_stmt``. There\'s separate logic for autocompletion in the API, the\ninference_state is all about inferring an expression.\n\nTODO this paragraph is not what jedi does anymore, it\'s similar, but not the\nsame.\n\nNow you need to understand what follows after ``infer_expr_stmt``. Let\'s\nmake an example::\n\n    import datetime\n    datetime.date.toda# <-- cursor here\n\nFirst of all, this module doesn\'t care about completion. It really just cares\nabout ``datetime.date``. At the end of the procedure ``infer_expr_stmt`` will\nreturn the ``date`` class.\n\nTo *visualize* this (simplified):\n\n- ``InferenceState.infer_expr_stmt`` doesn\'t do much, because there\'s no assignment.\n- ``Context.infer_node`` cares for resolving the dotted path\n- ``InferenceState.find_types`` searches for global definitions of datetime, which\n  it finds in the definition of an import, by scanning the syntax tree.\n- Using the import logic, the datetime module is found.\n- Now ``find_types`` is called again by ``infer_node`` to find ``date``\n  inside the datetime module.\n\nNow what would happen if we wanted ``datetime.date.foo.bar``? Two more\ncalls to ``find_types``. However the second call would be ignored, because the\nfirst one would return nothing (there\'s no foo attribute in ``date``).\n\nWhat if the import would contain another ``ExprStmt`` like this::\n\n    from foo import bar\n    Date = bar.baz\n\nWell... You get it. Just another ``infer_expr_stmt`` recursion. It\'s really\neasy. Python can obviously get way more complicated then this. To understand\ntuple assignments, list comprehensions and everything else, a lot more code had\nto be written.\n\nJedi has been tested very well, so you can just start modifying code. It\'s best\nto write your own test first for your "new" feature. Don\'t be scared of\nbreaking stuff. As long as the tests pass, you\'re most likely to be fine.\n\nI need to mention now that lazy type inference is really good because it\nonly *inferes* what needs to be *inferred*. All the statements and modules\nthat are not used are just being ignored.\n'
_B=False
_A=None
import parso
from jedi.file_io import FileIO
from jedi import debug
from jedi import settings
from jedi.inference import imports
from jedi.inference import recursion
from jedi.inference.cache import inference_state_function_cache
from jedi.inference import helpers
from jedi.inference.names import TreeNameDefinition
from jedi.inference.base_value import ContextualizedNode,ValueSet,iterate_values
from jedi.inference.value import ClassValue,FunctionValue
from jedi.inference.syntax_tree import infer_expr_stmt,check_tuple_assignments,tree_name_to_values
from jedi.inference.imports import follow_error_node_imports_if_possible
from jedi.plugins import plugin_manager
class InferenceState:
	def __init__(A,project,environment=_A,script_path=_A):
		C=project;B=environment
		if B is _A:B=C.get_environment()
		A.environment=B;A.script_path=script_path;A.compiled_subprocess=B.get_inference_state_subprocess(A);A.grammar=B.get_grammar();A.latest_grammar=parso.load_grammar(version='3.12');A.memoize_cache={};A.module_cache=imports.ModuleCache();A.stub_module_cache={};A.compiled_cache={};A.inferred_element_counts={};A.mixed_cache={};A.analysis=[];A.dynamic_params_depth=0;A.do_dynamic_params_search=settings.dynamic_params;A.is_analysis=_B;A.project=C;A.access_cache={};A.allow_unsafe_executions=_B;A.flow_analysis_enabled=True;A.reset_recursion_limitations()
	def import_module(A,import_names,sys_path=_A,prefer_stubs=True):return imports.import_module_by_names(A,import_names,sys_path,prefer_stubs=prefer_stubs)
	@staticmethod
	@plugin_manager.decorate()
	def execute(value,arguments):
		B=arguments;A=value;debug.dbg('execute: %s %s',A,B)
		with debug.increase_indent_cm():C=A.py__call__(arguments=B)
		debug.dbg('execute result: %s in %s',C,A);return C
	@property
	@inference_state_function_cache()
	def builtins_module(self):A='builtins';B,=self.import_module((A,),sys_path=[]);return B
	@property
	@inference_state_function_cache()
	def typing_module(self):A,=self.import_module(('typing',));return A
	def reset_recursion_limitations(A):A.recursion_detector=recursion.RecursionDetector();A.execution_recursion_detector=recursion.ExecutionRecursionDetector(A)
	def get_sys_path(A,**B):'Convenience function';return A.project._get_sys_path(A,**B)
	def infer(E,context,name):
		B=name;A=context;D=B.get_definition(import_name_always=True)
		if D is not _A:
			C=D.type;F=C=='classdef'
			if F or C=='funcdef':
				if F:G=ClassValue(E,A,B.parent)
				else:G=FunctionValue.from_context(A,B.parent)
				return ValueSet([G])
			if C=='expr_stmt':
				I=B.parent.type not in('power','trailer')
				if I:return infer_expr_stmt(A,D,B)
			if C=='for_stmt':J=A.infer_node(D.children[3]);K=ContextualizedNode(A,D.children[3]);L=iterate_values(J,K);M=TreeNameDefinition(A,B);return check_tuple_assignments(M,L)
			if C in('import_from','import_name'):return imports.infer_import(A,B)
			if C=='with_stmt':return tree_name_to_values(E,A,B)
			elif C=='param':return A.py__getattribute__(B.value,position=B.end_pos)
			elif C=='namedexpr_test':return A.infer_node(D)
		else:
			H=follow_error_node_imports_if_possible(A,B)
			if H is not _A:return H
		return helpers.infer_call_of_leaf(A,B)
	def parse_and_get_code(C,code=_A,path=_A,use_latest_grammar=_B,file_io=_A,**D):
		B=file_io;A=code
		if A is _A:
			if B is _A:B=FileIO(path)
			A=B.read()
		A=parso.python_bytes_to_unicode(A,encoding='utf-8',errors='replace')
		if len(A)>settings._cropped_file_size:A=A[:settings._cropped_file_size]
		E=C.latest_grammar if use_latest_grammar else C.grammar;return E.parse(code=A,path=path,file_io=B,**D),A
	def parse(A,*B,**C):return A.parse_and_get_code(*B,**C)[0]