'\n- the popular ``_memoize_default`` works like a typical memoize and returns the\n  default otherwise.\n- ``CachedMetaClass`` uses ``_memoize_default`` to do the same with classes.\n'
from functools import wraps
from jedi import debug
_NO_DEFAULT=object()
_RECURSION_SENTINEL=object()
def _memoize_default(default=_NO_DEFAULT,inference_state_is_first_arg=False,second_arg_is_inference_state=False):
	" This is a typical memoization decorator, BUT there is one difference:\n    To prevent recursion it sets defaults.\n\n    Preventing recursion is in this case the much bigger use than speed. I\n    don't think, that there is a big speed difference, but there are many cases\n    where recursion could happen (think about a = b; b = a).\n    ";G=default
	def A(function):
		E=function
		def A(obj,*F,**H):
			B=obj
			if inference_state_is_first_arg:C=B.memoize_cache
			elif second_arg_is_inference_state:C=F[0].memoize_cache
			else:C=B.inference_state.memoize_cache
			try:A=C[E]
			except KeyError:C[E]=A={}
			D=B,F,frozenset(H.items())
			if D in A:return A[D]
			else:
				if G is not _NO_DEFAULT:A[D]=G
				I=E(B,*F,**H);A[D]=I;return I
		return A
	return A
def inference_state_function_cache(default=_NO_DEFAULT):
	def A(func):return _memoize_default(default=default,inference_state_is_first_arg=True)(func)
	return A
def inference_state_method_cache(default=_NO_DEFAULT):
	def A(func):return _memoize_default(default=default)(func)
	return A
def inference_state_as_method_param_cache():
	def A(call):return _memoize_default(second_arg_is_inference_state=True)(call)
	return A
class CachedMetaClass(type):
	'\n    This is basically almost the same than the decorator above, it just caches\n    class initializations. Either you do it this way or with decorators, but\n    with decorators you lose class access (isinstance, etc).\n    '
	@inference_state_as_method_param_cache()
	def __call__(self,*A,**B):return super().__call__(*A,**B)
def inference_state_method_generator_cache():
	'\n    This is a special memoizer. It memoizes generators and also checks for\n    recursion errors and returns no further iterator elemends in that case.\n    '
	def A(function):
		C=function
		@wraps(C)
		def A(obj,*H,**I):
			D=obj;J=D.inference_state.memoize_cache
			try:E=J[C]
			except KeyError:J[C]=E={}
			F=D,H,frozenset(I.items())
			if F in E:G,A=E[F]
			else:G=C(D,*H,**I);A=[];E[F]=G,A
			K=0
			while True:
				try:
					B=A[K]
					if B is _RECURSION_SENTINEL:debug.warning('Found a generator recursion for %s'%D);return
				except IndexError:
					A.append(_RECURSION_SENTINEL);B=next(G,None)
					if B is None:A.pop();return
					A[-1]=B
				yield B;K+=1
		return A
	return A