'\nMakes it possible to do the compiled analysis in a subprocess. This has two\ngoals:\n\n1. Making it safer - Segfaults and RuntimeErrors as well as stdout/stderr can\n   be ignored and dealt with.\n2. Make it possible to handle different Python versions as well as virtualenvs.\n'
_D='replace'
_C=False
_B=True
_A=None
import collections,os,sys,queue,subprocess,traceback,weakref
from functools import partial
from threading import Thread
from jedi._compatibility import pickle_dump,pickle_load
from jedi import debug
from jedi.cache import memoize_method
from jedi.inference.compiled.subprocess import functions
from jedi.inference.compiled.access import DirectObjectAccess,AccessPath,SignatureParam
from jedi.api.exceptions import InternalError
_MAIN_PATH=os.path.join(os.path.dirname(__file__),'__main__.py')
PICKLE_PROTOCOL=4
def _GeneralizedPopen(*C,**A):
	if os.name=='nt':
		try:B=subprocess.CREATE_NO_WINDOW
		except AttributeError:B=134217728
		A['creationflags']=B
	A['close_fds']='posix'in sys.builtin_module_names;return subprocess.Popen(*C,**A)
def _enqueue_output(out,queue_):
	for A in iter(out.readline,b''):queue_.put(A)
def _add_stderr_to_debug(stderr_queue):
	while _B:
		try:A=stderr_queue.get_nowait();A=A.decode('utf-8',_D);debug.warning('stderr output: %s'%A.rstrip('\n'))
		except queue.Empty:break
def _get_function(name):return getattr(functions,name)
def _cleanup_process(process,thread):
	A=process
	try:A.kill();A.wait()
	except OSError:pass
	thread.join()
	for B in[A.stdin,A.stdout,A.stderr]:
		try:B.close()
		except OSError:pass
class _InferenceStateProcess:
	def __init__(A,inference_state):B=inference_state;A._inference_state_weakref=weakref.ref(B);A._inference_state_id=id(B);A._handles={}
	def get_or_create_access_handle(A,obj):
		B=id(obj)
		try:return A.get_access_handle(B)
		except KeyError:D=DirectObjectAccess(A._inference_state_weakref(),obj);C=AccessHandle(A,D,B);A.set_access_handle(C);return C
	def get_access_handle(A,id_):return A._handles[id_]
	def set_access_handle(B,handle):A=handle;B._handles[A.id]=A
class InferenceStateSameProcess(_InferenceStateProcess):
	'\n    Basically just an easy access to functions.py. It has the same API\n    as InferenceStateSubprocess and does the same thing without using a subprocess.\n    This is necessary for the Interpreter process.\n    '
	def __getattr__(A,name):return partial(_get_function(name),A._inference_state_weakref())
class InferenceStateSubprocess(_InferenceStateProcess):
	def __init__(A,inference_state,compiled_subprocess):super().__init__(inference_state);A._used=_C;A._compiled_subprocess=compiled_subprocess
	def __getattr__(A,name):
		B=_get_function(name)
		def C(*C,**D):A._used=_B;E=A._compiled_subprocess.run(A._inference_state_weakref(),B,args=C,kwargs=D);return A._convert_access_handles(E)
		return C
	def _convert_access_handles(B,obj):
		A=obj
		if isinstance(A,SignatureParam):return SignatureParam(*B._convert_access_handles(tuple(A)))
		elif isinstance(A,tuple):return tuple(B._convert_access_handles(A)for A in A)
		elif isinstance(A,list):return[B._convert_access_handles(A)for A in A]
		elif isinstance(A,AccessHandle):
			try:A=B.get_access_handle(A.id)
			except KeyError:A.add_subprocess(B);B.set_access_handle(A)
		elif isinstance(A,AccessPath):return AccessPath(B._convert_access_handles(A.accesses))
		return A
	def __del__(A):
		if A._used and not A._compiled_subprocess.is_crashed:A._compiled_subprocess.delete_inference_state(A._inference_state_id)
class CompiledSubprocess:
	is_crashed=_C
	def __init__(A,executable,env_vars=_A):A._executable=executable;A._env_vars=env_vars;A._inference_state_deletion_queue=collections.deque();A._cleanup_callable=lambda:_A
	def __repr__(A):B=os.getpid();return'<%s _executable=%r, is_crashed=%r, pid=%r>'%(A.__class__.__name__,A._executable,A.is_crashed,B)
	@memoize_method
	def _get_process(self):A=self;debug.dbg('Start environment subprocess %s',A._executable);D=sys.modules['parso'].__file__;E=A._executable,_MAIN_PATH,os.path.dirname(os.path.dirname(D)),'.'.join(str(A)for A in sys.version_info[:3]);B=_GeneralizedPopen(E,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE,env=A._env_vars);A._stderr_queue=queue.Queue();A._stderr_thread=C=Thread(target=_enqueue_output,args=(B.stderr,A._stderr_queue));C.daemon=_B;C.start();A._cleanup_callable=weakref.finalize(A,_cleanup_process,B,C);return B
	def run(A,inference_state,function,args=(),kwargs={}):
		B=function
		while _B:
			try:C=A._inference_state_deletion_queue.pop()
			except IndexError:break
			else:A._send(C,_A)
		assert callable(B);return A._send(id(inference_state),B,args,kwargs)
	def get_sys_path(A):return A._send(_A,functions.get_sys_path,(),{})
	def _kill(A):A.is_crashed=_B;A._cleanup_callable()
	def _send(A,inference_state_id,function,args=(),kwargs={}):
		if A.is_crashed:raise InternalError('The subprocess %s has crashed.'%A._executable)
		D=inference_state_id,function,args,kwargs
		try:pickle_dump(D,A._get_process().stdin,PICKLE_PROTOCOL)
		except BrokenPipeError:A._kill();raise InternalError('The subprocess %s was killed. Maybe out of memory?'%A._executable)
		try:E,F,B=pickle_load(A._get_process().stdout)
		except EOFError as G:
			try:C=A._get_process().stderr.read().decode('utf-8',_D)
			except Exception as H:C='<empty/not available (%r)>'%H
			A._kill();_add_stderr_to_debug(A._stderr_queue);raise InternalError('The subprocess %s has crashed (%r, stderr=%s).'%(A._executable,G,C))
		_add_stderr_to_debug(A._stderr_queue)
		if E:B.args=F,;raise B
		return B
	def delete_inference_state(A,inference_state_id):"\n        Currently we are not deleting inference_state instantly. They only get\n        deleted once the subprocess is used again. It would probably a better\n        solution to move all of this into a thread. However, the memory usage\n        of a single inference_state shouldn't be that high.\n        ";A._inference_state_deletion_queue.append(inference_state_id)
class Listener:
	def __init__(A):A._inference_states={};A._process=_InferenceStateProcess(Listener)
	def _get_inference_state(B,function,inference_state_id):
		C=inference_state_id;from jedi.inference import InferenceState as D
		try:A=B._inference_states[C]
		except KeyError:from jedi import InterpreterEnvironment as E;A=D(project=_A,environment=E());B._inference_states[C]=A
		return A
	def _run(F,inference_state_id,function,args,kwargs):
		D=inference_state_id;C=kwargs;B=function;A=args
		if D is _A:return B(*A,**C)
		elif B is _A:del F._inference_states[D]
		else:
			E=F._get_inference_state(B,D);A=list(A)
			for(I,G)in enumerate(A):
				if isinstance(G,AccessHandle):A[I]=E.compiled_subprocess.get_access_handle(G.id)
			for(J,H)in C.items():
				if isinstance(H,AccessHandle):C[J]=E.compiled_subprocess.get_access_handle(H.id)
			return B(E,*A,**C)
	def listen(D):
		A=sys.stdout;sys.stdout=open(os.devnull,'w');B=sys.stdin;A=A.buffer;B=B.buffer
		while _B:
			try:E=pickle_load(B)
			except EOFError:exit(0)
			try:C=_C,_A,D._run(*E)
			except Exception as F:C=_B,traceback.format_exc(),F
			pickle_dump(C,A,PICKLE_PROTOCOL)
class AccessHandle:
	def __init__(A,subprocess,access,id_):A.access=access;A._subprocess=subprocess;A.id=id_
	def add_subprocess(A,subprocess):A._subprocess=subprocess
	def __repr__(A):
		try:B=A.access
		except AttributeError:B='#'+str(A.id)
		return'<%s of %s>'%(A.__class__.__name__,B)
	def __getstate__(A):return A.id
	def __setstate__(A,state):A.id=state
	def __getattr__(B,name):
		A=name
		if A in('id','access')or A.startswith('_'):raise AttributeError('Something went wrong with unpickling')
		return partial(B._workaround,A)
	def _workaround(B,name,*A,**C):
		"\n        TODO Currently we're passing slice objects around. This should not\n        happen. They are also the only unhashable objects that we're passing\n        around.\n        "
		if A and isinstance(A[0],slice):return B._subprocess.get_compiled_method_return(B.id,name,*A,**C)
		return B._cached_results(name,*A,**C)
	@memoize_method
	def _cached_results(self,name,*A,**B):return self._subprocess.get_compiled_method_return(self.id,name,*A,**B)