'\nDocstrings are another source of information for functions and classes.\n:mod:`jedi.inference.dynamic_params` tries to find all executions of functions,\nwhile the docstring parsing is much easier. There are three different types of\ndocstrings that |jedi| understands:\n\n- `Sphinx <http://sphinx-doc.org/markup/desc.html#info-field-lists>`_\n- `Epydoc <http://epydoc.sourceforge.net/manual-fields.html>`_\n- `Numpydoc <https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt>`_\n\nFor example, the sphinx annotation ``:type foo: str`` clearly states that the\ntype of ``foo`` is ``str``.\n\nAs an addition to parameter searching, this module also provides return\nannotations.\n'
_A='ignore'
import re,warnings
from parso import parse,ParserSyntaxError
from jedi import debug
from jedi.inference.cache import inference_state_method_cache
from jedi.inference.base_value import iterator_to_value_set,ValueSet,NO_VALUES
from jedi.inference.lazy_value import LazyKnownValues
DOCSTRING_PARAM_PATTERNS=['\\s*:type\\s+%s:\\s*([^\\n]+)','\\s*:param\\s+(\\w+)\\s+%s:[^\\n]*','\\s*@type\\s+%s:\\s*([^\\n]+)']
DOCSTRING_RETURN_PATTERNS=[re.compile('\\s*:rtype:\\s*([^\\n]+)',re.M),re.compile('\\s*@rtype:\\s*([^\\n]+)',re.M)]
REST_ROLE_PATTERN=re.compile(':[^`]+:`([^`]+)`')
_numpy_doc_string_cache=None
def _get_numpy_doc_string_cls():
	global _numpy_doc_string_cache
	if isinstance(_numpy_doc_string_cache,(ImportError,SyntaxError)):raise _numpy_doc_string_cache
	from numpydoc.docscrape import NumpyDocString as A;_numpy_doc_string_cache=A;return _numpy_doc_string_cache
def _search_param_in_numpydocstr(docstr,param_str):
	'Search `docstr` (in numpydoc format) for type(-s) of `param_str`.'
	with warnings.catch_warnings():
		warnings.simplefilter(_A)
		try:C=_get_numpy_doc_string_cls()(docstr)._parsed_data['Parameters']
		except Exception:return[]
	for(D,A,E)in C:
		if D==param_str:
			B=re.match('([^,]+(,[^,]+)*?)(,[ ]*optional)?$',A)
			if B:A=B.group(1)
			return list(_expand_typestr(A))
	return[]
def _search_return_in_numpydocstr(docstr):
	'\n    Search `docstr` (in numpydoc format) for type(-s) of function returns.\n    '
	with warnings.catch_warnings():
		warnings.simplefilter(_A)
		try:B=_get_numpy_doc_string_cls()(docstr)
		except Exception:return
	try:C=B._parsed_data['Returns'];C+=B._parsed_data['Yields']
	except Exception:return
	for(D,A,E)in C:
		if not A:A=D
		yield from _expand_typestr(A)
def _expand_typestr(type_str):
	'\n    Attempts to interpret the possible types in `type_str`\n    ';A=type_str
	if re.search('\\bor\\b',A):
		for D in A.split('or'):yield D.split('of')[0].strip()
	elif re.search('\\bof\\b',A):yield A.split('of')[0]
	elif A.startswith('{'):
		C=parse(A,version='3.7').children[0]
		if C.type=='atom':
			for B in getattr(C.children[1],'children',[]):
				if B.type=='number':
					if'.'in B.value:yield'float'
					else:yield'int'
				elif B.type=='string':
					if'b'in B.string_prefix.lower():yield'bytes'
					else:yield'str'
	else:yield A
def _search_param_in_docstr(docstr,param_str):
	"\n    Search `docstr` for type(-s) of `param_str`.\n\n    >>> _search_param_in_docstr(':type param: int', 'param')\n    ['int']\n    >>> _search_param_in_docstr('@type param: int', 'param')\n    ['int']\n    >>> _search_param_in_docstr(\n    ...   ':type param: :class:`threading.Thread`', 'param')\n    ['threading.Thread']\n    >>> bool(_search_param_in_docstr('no document', 'param'))\n    False\n    >>> _search_param_in_docstr(':param int param: some description', 'param')\n    ['int']\n\n    ";B=param_str;A=docstr;D=[re.compile(A%re.escape(B))for A in DOCSTRING_PARAM_PATTERNS]
	for E in D:
		C=E.search(A)
		if C:return[_strip_rst_role(C.group(1))]
	return _search_param_in_numpydocstr(A,B)
def _strip_rst_role(type_str):
	"\n    Strip off the part looks like a ReST role in `type_str`.\n\n    >>> _strip_rst_role(':class:`ClassName`')  # strip off :class:\n    'ClassName'\n    >>> _strip_rst_role(':py:obj:`module.Object`')  # works with domain\n    'module.Object'\n    >>> _strip_rst_role('ClassName')  # do nothing when not ReST role\n    'ClassName'\n\n    See also:\n    http://sphinx-doc.org/domains.html#cross-referencing-python-objects\n\n    ";A=type_str;B=REST_ROLE_PATTERN.match(A)
	if B:return B.group(1)
	else:return A
def _infer_for_statement_string(module_context,string):
	B=module_context;A=string
	if A is None:return[]
	E=re.findall('((?:\\w+\\.)*\\w+)\\.',A);F='\n'.join(f"import {A}"for A in E);A=f"{F}\n{A}";debug.dbg('Parse docstring code %s',A,color='BLUE');G=B.inference_state.grammar
	try:C=G.parse(A,error_recovery=False)
	except ParserSyntaxError:return[]
	try:D=C.children[-2]
	except(AttributeError,IndexError):return[]
	if D.type not in('name','atom','atom_expr'):return[]
	from jedi.inference.docstring_utils import DocstringModule as H;I=H(in_module_context=B,inference_state=B.inference_state,module_node=C,code_lines=[]);return list(_execute_types_in_stmt(I.as_context(),D))
def _execute_types_in_stmt(module_context,stmt):"\n    Executing all types or general elements that we find in a statement. This\n    doesn't include tuple, list and dict literals, because the stuff they\n    contain is executed. (Used as type information).\n    ";A=module_context;B=A.infer_node(stmt);return ValueSet.from_sets(_execute_array_values(A.inference_state,B)for B in B)
def _execute_array_values(inference_state,array):
	"\n    Tuples indicate that there's not just one return value, but the listed\n    ones.  `(str, int)` means that it returns a tuple with both types.\n    ";D='tuple';B=inference_state;A=array;from jedi.inference.value.iterable import SequenceLiteralValue as E,FakeTuple as F,FakeList as G
	if isinstance(A,E)and A.array_type in(D,'list'):
		C=[]
		for H in A.py__iter__():I=ValueSet.from_sets(_execute_array_values(B,A)for A in H.infer());C.append(LazyKnownValues(I))
		J=F if A.array_type==D else G;return{J(B,C)}
	else:return A.execute_annotation()
@inference_state_method_cache()
def infer_param(function_value,param):
	C=param;A=function_value
	def D(docstring):return ValueSet(B for A in _search_param_in_docstr(docstring,C.name.value)for B in _infer_for_statement_string(E,A))
	E=A.get_root_context();F=C.get_parent_function()
	if F.type=='lambdef':return NO_VALUES
	B=D(A.py__doc__())
	if A.is_bound_method()and A.py__name__()=='__init__':B|=D(A.class_context.py__doc__())
	debug.dbg('Found param types for docstring: %s',B,color='BLUE');return B
@inference_state_method_cache()
@iterator_to_value_set
def infer_return_types(function_value):
	A=function_value
	def B(code):
		for B in DOCSTRING_RETURN_PATTERNS:
			A=B.search(code)
			if A:yield _strip_rst_role(A.group(1))
		yield from _search_return_in_numpydocstr(code)
	for C in B(A.py__doc__()):yield from _infer_for_statement_string(A.get_root_context(),C)