"\nOne of the really important features of |jedi| is to have an option to\nunderstand code like this::\n\n    def foo(bar):\n        bar. # completion here\n    foo(1)\n\nThere's no doubt wheter bar is an ``int`` or not, but if there's also a call\nlike ``foo('str')``, what would happen? Well, we'll just show both. Because\nthat's what a human would expect.\n\nIt works as follows:\n\n- |Jedi| sees a param\n- search for function calls named ``foo``\n- execute these calls and check the input.\n"
_B='classdef'
_A=None
from jedi import settings,debug
from jedi.parser_utils import get_parent_scope
from jedi.inference.cache import inference_state_method_cache
from jedi.inference.arguments import TreeArguments
from jedi.inference.param import get_executed_param_names
from jedi.inference.helpers import is_stdlib_path
from jedi.inference.utils import to_list
from jedi.inference.value import instance
from jedi.inference.base_value import ValueSet,NO_VALUES
from jedi.inference.references import get_module_contexts_containing_name
from jedi.inference import recursion
MAX_PARAM_SEARCHES=20
def _avoid_recursions(func):
	def A(function_value,param_index):
		A=function_value;B=A.inference_state
		with recursion.execution_allowed(B,A.tree_node)as C:
			if C:
				B.dynamic_params_depth+=1
				try:return func(A,param_index)
				finally:B.dynamic_params_depth-=1
			return NO_VALUES
	return A
@debug.increase_indent
@_avoid_recursions
def dynamic_param_lookup(function_value,param_index):
	'\n    A dynamic search for param values. If you try to complete a type:\n\n    >>> def func(foo):\n    ...     foo\n    >>> func(1)\n    >>> func("")\n\n    It is not known what the type ``foo`` without analysing the whole code. You\n    have to look for all calls to ``func`` to find out what ``foo`` possibly\n    is.\n    ';E='MAGENTA';A=function_value
	if not A.inference_state.do_dynamic_params_search:return NO_VALUES
	B=A.tree_node;D=A.get_root_context().py__file__()
	if D is not _A and is_stdlib_path(D):return NO_VALUES
	if B.type=='lambdef':
		C=_get_lambda_name(B)
		if C is _A:return NO_VALUES
	else:C=B.name.value
	debug.dbg('Dynamic param search in %s.',C,color=E);F=A.get_root_context();G=_search_function_arguments(F,B,C);H=ValueSet.from_sets(get_executed_param_names(A,B)[param_index].infer()for B in G);debug.dbg('Dynamic param result finished',color=E);return H
@inference_state_method_cache(default=_A)
@to_list
def _search_function_arguments(module_context,funcdef,string_name):
	'\n    Returns a list of param names.\n    ';E=funcdef;B=module_context;A=string_name;F=E
	if A=='__init__':
		C=get_parent_scope(E)
		if C.type==_B:A=C.name.value;F=C
	G=False;H=0;D=B.inference_state
	if settings.dynamic_params_for_other_modules:I=get_module_contexts_containing_name(D,[B],A,limit_reduction=5)
	else:I=[B]
	for J in I:
		for(K,L)in _get_potential_nodes(J,A):
			H+=1
			if H*D.dynamic_params_depth>MAX_PARAM_SEARCHES:return
			M=J.create_context(K)
			for N in _check_name_for_execution(D,M,F,K,L):G=True;yield N
		if G:return
def _get_lambda_name(node):
	A=node.parent
	if A.type=='expr_stmt':
		C=next(A.yield_operators(),_A)
		if C=='=':
			B=A.children[0]
			if B.type=='name':return B.value
def _get_potential_nodes(module_value,func_string_name):
	try:D=module_value.tree_node.get_used_names()[func_string_name]
	except KeyError:return
	for A in D:
		B=A.get_next_leaf();C=B.parent
		if C.type=='trailer'and B=='(':yield(A,C)
def _check_name_for_execution(inference_state,context,compare_node,name,trailer):
	F=context;E=trailer;D=compare_node;C=inference_state;B=name;from jedi.inference.value.function import BaseFunctionExecutionContext as J
	def I(value):
		A=value;D=E.children[1]
		if D==')':D=_A
		B=TreeArguments(C,F,D,E);from jedi.inference.value.instance import InstanceArguments as G
		if A.tree_node.type==_B:H=instance.TreeInstance(C,A.parent_context,A,B);return G(H,B)
		else:
			if A.is_bound_method():B=G(A.instance,B)
			return B
	for A in C.infer(F,B):
		G=A.tree_node
		if D==G:yield I(A)
		elif isinstance(A.parent_context,J)and D.type=='funcdef':
			H=A.parent_context.get_param_names()
			if len(H)!=1:continue
			K=H[0].infer()
			if[A.tree_node for A in K]==[D]:
				L=F.get_root_context();M=A.as_context(I(A));N=_get_potential_nodes(L,H[0].string_name)
				for(B,E)in N:
					if G.start_pos<B.start_pos<G.end_pos:O=M.create_context(B);yield from _check_name_for_execution(C,O,D,B,E)