'\nFilters are objects that you can use to filter names in different scopes. They\nare needed for name resolution.\n'
_B='function'
_A=None
from abc import abstractmethod
from typing import List,MutableMapping,Type
import weakref
from parso.tree import search_ancestor
from parso.python.tree import Name,UsedNamesMapping
from jedi.inference import flow_analysis
from jedi.inference.base_value import ValueSet,ValueWrapper,LazyValueWrapper
from jedi.parser_utils import get_cached_parent_scope,get_parso_cache_node
from jedi.inference.utils import to_list
from jedi.inference.names import TreeNameDefinition,ParamName,AnonymousParamName,AbstractNameDefinition,NameWrapper
_definition_name_cache:0
_definition_name_cache=weakref.WeakKeyDictionary()
class AbstractFilter:
	_until_position=_A
	def _filter(A,names):
		B=names
		if A._until_position is not _A:return[B for B in B if B.start_pos<A._until_position]
		return B
	@abstractmethod
	def get(self,name):raise NotImplementedError
	@abstractmethod
	def values(self):raise NotImplementedError
class FilterWrapper:
	name_wrapper_class:Type[NameWrapper]
	def __init__(A,wrapped_filter):A._wrapped_filter=wrapped_filter
	def wrap_names(A,names):return[A.name_wrapper_class(B)for B in names]
	def get(A,name):return A.wrap_names(A._wrapped_filter.get(name))
	def values(A):return A.wrap_names(A._wrapped_filter.values())
def _get_definition_names(parso_cache_node,used_names,name_key):
	E=used_names;B=parso_cache_node;A=name_key
	if B is _A:C=E.get(A,());return tuple(A for A in C if A.is_definition(include_setitem=True))
	try:D=_definition_name_cache[B]
	except KeyError:D=_definition_name_cache[B]={}
	try:return D[A]
	except KeyError:C=E.get(A,());F=D[A]=tuple(A for A in C if A.is_definition(include_setitem=True));return F
class _AbstractUsedNamesFilter(AbstractFilter):
	name_class=TreeNameDefinition
	def __init__(A,parent_context,node_context=_A):
		D=parent_context;B=node_context
		if B is _A:B=D
		A._node_context=B;A._parser_scope=B.tree_node;C=B.get_root_context();E=C.py__file__()
		if E is _A:A._parso_cache_node=_A
		else:A._parso_cache_node=get_parso_cache_node(C.inference_state.latest_grammar if C.is_stub()else C.inference_state.grammar,E)
		A._used_names=C.tree_node.get_used_names();A.parent_context=D
	def get(A,name):return A._convert_names(A._filter(_get_definition_names(A._parso_cache_node,A._used_names,name)))
	def _convert_names(A,names):return[A.name_class(A.parent_context,B)for B in names]
	def values(A):return A._convert_names(C for B in A._used_names for C in A._filter(_get_definition_names(A._parso_cache_node,A._used_names,B)))
	def __repr__(A):return'<%s: %s>'%(A.__class__.__name__,A.parent_context)
class ParserTreeFilter(_AbstractUsedNamesFilter):
	def __init__(A,parent_context,node_context=_A,until_position=_A,origin_scope=_A):"\n        node_context is an option to specify a second value for use cases\n        like the class mro where the parent class of a new name would be the\n        value, but for some type inference it's important to have a local\n        value of the other classes.\n        ";super().__init__(parent_context,node_context);A._origin_scope=origin_scope;A._until_position=until_position
	def _filter(B,names):A=names;A=super()._filter(A);A=[A for A in A if B._is_name_reachable(A)];return list(B._check_flows(A))
	def _is_name_reachable(B,name):
		A=name.parent
		if A.type=='trailer':return False
		C=A if A.type in('classdef','funcdef')else name;return get_cached_parent_scope(B._parso_cache_node,C)==B._parser_scope
	def _check_flows(A,names):
		for B in sorted(names,key=lambda name:name.start_pos,reverse=True):
			C=flow_analysis.reachability_check(context=A._node_context,value_scope=A._parser_scope,node=B,origin_scope=A._origin_scope)
			if C is not flow_analysis.UNREACHABLE:yield B
			if C is flow_analysis.REACHABLE:break
class _FunctionExecutionFilter(ParserTreeFilter):
	def __init__(A,parent_context,function_value,until_position,origin_scope):super().__init__(parent_context,until_position=until_position,origin_scope=origin_scope);A._function_value=function_value
	def _convert_param(A,param,name):raise NotImplementedError
	@to_list
	def _convert_names(self,names):
		for A in names:
			B=search_ancestor(A,'param')
			if B:yield self._convert_param(B,A)
			else:yield TreeNameDefinition(self.parent_context,A)
class FunctionExecutionFilter(_FunctionExecutionFilter):
	def __init__(A,*B,arguments,**C):super().__init__(*B,**C);A._arguments=arguments
	def _convert_param(A,param,name):return ParamName(A._function_value,name,A._arguments)
class AnonymousFunctionExecutionFilter(_FunctionExecutionFilter):
	def _convert_param(A,param,name):return AnonymousParamName(A._function_value,name)
class GlobalNameFilter(_AbstractUsedNamesFilter):
	def get(A,name):
		try:B=A._used_names[name]
		except KeyError:return[]
		return A._convert_names(A._filter(B))
	@to_list
	def _filter(self,names):
		for A in names:
			if A.parent.type=='global_stmt':yield A
	def values(A):return A._convert_names(C for B in A._used_names.values()for C in A._filter(B))
class DictFilter(AbstractFilter):
	def __init__(A,dct):A._dct=dct
	def get(A,name):
		try:B=A._convert(name,A._dct[name])
		except KeyError:return[]
		else:return list(A._filter([B]))
	def values(A):
		def B():
			for B in A._dct.items():
				try:yield A._convert(*B)
				except KeyError:pass
		return A._filter(B())
	def _convert(A,name,value):return value
	def __repr__(A):B=', '.join(A._dct.keys());return'<%s: for {%s}>'%(A.__class__.__name__,B)
class MergedFilter:
	def __init__(A,*B):A._filters=B
	def get(A,name):return[A for filter in A._filters for A in filter.get(name)]
	def values(A):return[A for filter in A._filters for A in filter.values()]
	def __repr__(A):return'%s(%s)'%(A.__class__.__name__,', '.join(str(A)for A in A._filters))
class _BuiltinMappedMethod(ValueWrapper):
	'``Generator.__next__`` ``dict.values`` methods and so on.';api_type=_B
	def __init__(A,value,method,builtin_func):super().__init__(builtin_func);A._value=value;A._method=method
	def py__call__(A,arguments):return A._method(A._value,arguments)
class SpecialMethodFilter(DictFilter):
	'\n    A filter for methods that are defined in this module on the corresponding\n    classes like Generator (for __next__, etc).\n    '
	class SpecialMethodName(AbstractNameDefinition):
		api_type=_B
		def __init__(A,parent_context,string_name,callable_,builtin_value):A.parent_context=parent_context;A.string_name=string_name;A._callable=callable_;A._builtin_value=builtin_value
		def infer(A):
			for filter in A._builtin_value.get_filters():
				for B in filter.get(A.string_name):C=next(iter(B.infer()));break
				else:continue
				break
			return ValueSet([_BuiltinMappedMethod(A.parent_context,A._callable,C)])
	def __init__(A,value,dct,builtin_value):super().__init__(dct);A.value=value;A._builtin_value=builtin_value;'\n        This value is what will be used to introspect the name, where as the\n        other value will be used to execute the function.\n\n        We distinguish, because we have to.\n        '
	def _convert(A,name,value):return A.SpecialMethodName(A.value,name,value,A._builtin_value)
class _OverwriteMeta(type):
	def __init__(A,name,bases,dct):
		super().__init__(name,bases,dct);B={}
		for C in reversed(A.__bases__):
			try:B.update(C.overwritten_methods)
			except AttributeError:pass
		for D in A.__dict__.values():
			try:B.update(D.registered_overwritten_methods)
			except AttributeError:pass
		A.overwritten_methods=B
class _AttributeOverwriteMixin:
	def get_filters(A,*B,**C):yield SpecialMethodFilter(A,A.overwritten_methods,A._wrapped_value);yield from A._wrapped_value.get_filters(*B,**C)
class LazyAttributeOverwrite(_AttributeOverwriteMixin,LazyValueWrapper,metaclass=_OverwriteMeta):
	def __init__(A,inference_state):A.inference_state=inference_state
class AttributeOverwrite(_AttributeOverwriteMixin,ValueWrapper,metaclass=_OverwriteMeta):0
def publish_method(method_name):
	def A(func):A=func;B=A.__dict__.setdefault('registered_overwritten_methods',{});B[method_name]=A;return A
	return A