'\nPEP 0484 ( https://www.python.org/dev/peps/pep-0484/ ) describes type hints\nthrough function annotations. There is a strong suggestion in this document\nthat only the type of type hinting defined in PEP0484 should be allowed\nas annotations in future python versions.\n'
_E='return'
_D='power'
_C='atom_expr'
_B=False
_A=None
import re
from inspect import Parameter
from parso import ParserSyntaxError,parse
from jedi.inference.cache import inference_state_method_cache
from jedi.inference.base_value import ValueSet,NO_VALUES
from jedi.inference.gradual.base import DefineGenericBaseClass,GenericClass
from jedi.inference.gradual.generics import TupleGenericManager
from jedi.inference.gradual.type_var import TypeVar
from jedi.inference.helpers import is_string
from jedi.inference.compiled import builtin_from_name
from jedi.inference.param import get_executed_param_names
from jedi import debug
from jedi import parser_utils
def infer_annotation(context,annotation):
	'\n    Inferes an annotation node. This means that it inferes the part of\n    `int` here:\n\n        foo: int = 3\n\n    Also checks for forward references (strings)\n    ';C=annotation;B=context;A=B.infer_node(C)
	if len(A)!=1:debug.warning('Inferred typing index %s should lead to 1 object,  not %s'%(C,A));return A
	D=list(A)[0]
	if is_string(D):
		E=_get_forward_reference_node(B,D.get_safe_value())
		if E is not _A:return B.infer_node(E)
	return A
def _infer_annotation_string(context,string,index=_A):
	C=context;A=index;D=_get_forward_reference_node(C,string)
	if D is _A:return NO_VALUES
	B=C.infer_node(D)
	if A is not _A:B=B.filter(lambda value:value.array_type=='tuple'and len(list(value.py__iter__()))>=A).py__simple_getitem__(A)
	return B
def _get_forward_reference_node(context,string):
	C=string;A=context
	try:B=A.inference_state.grammar.parse(C,start_symbol='eval_input',error_recovery=_B)
	except ParserSyntaxError:debug.warning('Annotation not parsed: %s'%C);return
	else:D=A.tree_node.get_root_node();parser_utils.move(B,D.end_pos[0]);B.parent=A.tree_node;return B
def _split_comment_param_declaration(decl_text):
	'\n    Split decl_text on commas, but group generic expressions\n    together.\n\n    For example, given "foo, Bar[baz, biz]" we return\n    [\'foo\', \'Bar[baz, biz]\'].\n\n    ';E='name';B=decl_text
	try:A=parse(B,error_recovery=_B).children[0]
	except ParserSyntaxError:debug.warning('Comment annotation is not valid Python: %s'%B);return[]
	if A.type in[E,_C,_D]:return[A.get_code().strip()]
	C=[]
	try:F=A.children
	except AttributeError:return[]
	else:
		for D in F:
			if D.type in[E,_C,_D]:C.append(D.get_code().strip())
	return C
@inference_state_method_cache()
def infer_param(function_value,param,ignore_stars=_B):
	D=function_value;B=param;A=_infer_param(D,B)
	if ignore_stars or not A:return A
	C=D.inference_state
	if B.star_count==1:E=builtin_from_name(C,'tuple');return ValueSet([GenericClass(E,TupleGenericManager((A,)))])
	elif B.star_count==2:F=builtin_from_name(C,'dict');G=ValueSet([builtin_from_name(C,'str')]),A;return ValueSet([GenericClass(F,TupleGenericManager(G))])
	return A
def _infer_param(function_value,param):
	'\n    Infers the type of a function parameter, using type annotations.\n    ';D=function_value;A=param;F=A.annotation
	if F is _A:
		E=[A for A in A.parent.children if A.type=='param'];I=A.parent.parent;G=parser_utils.get_following_comment_same_line(I)
		if G is _A:return NO_VALUES
		H=re.match('^#\\s*type:\\s*\\(([^#]*)\\)\\s*->',G)
		if not H:return NO_VALUES
		B=_split_comment_param_declaration(H.group(1));C=E.index(A)
		if len(B)!=len(E):debug.warning('Comments length != Params length %s %s',B,E)
		if D.is_bound_method():
			if C==0:return NO_VALUES
			C-=1
		if C>=len(B):return NO_VALUES
		J=B[C];return _infer_annotation_string(D.get_default_param_context(),J)
	K=D.get_default_param_context();return infer_annotation(K,F)
def py__annotations__(funcdef):
	B=funcdef;A={}
	for C in B.get_params():
		D=C.annotation
		if D is not _A:A[C.name.value]=D
	E=B.annotation
	if E:A[_E]=E
	return A
def resolve_forward_references(context,all_annotations):
	B=context
	def A(node):
		A=node
		if A is _A or A.type!='string':return A
		A=_get_forward_reference_node(B,B.inference_state.compiled_subprocess.safe_literal_eval(A.value))
		if A is _A:return
		A=A.children[0];return A
	return{B:A(C)for(B,C)in all_annotations.items()}
@inference_state_method_cache()
def infer_return_types(function,arguments):
	"\n    Infers the type of a function's return value,\n    according to type annotations.\n    ";A=function;B=A.get_default_param_context();D=resolve_forward_references(B,py__annotations__(A.tree_node));C=D.get(_E,_A)
	if C is _A:
		H=A.tree_node;E=parser_utils.get_following_comment_same_line(H)
		if E is _A:return NO_VALUES
		F=re.match('^#\\s*type:\\s*\\([^#]*\\)\\s*->\\s*([^#]*)',E)
		if not F:return NO_VALUES
		return _infer_annotation_string(B,F.group(1).strip()).execute_annotation()
	I=find_unknown_type_vars(B,C);G=infer_annotation(B,C)
	if not I:return G.execute_annotation()
	J=infer_type_vars_for_execution(A,arguments,D);return ValueSet.from_sets(A.define_generics(J)if isinstance(A,(DefineGenericBaseClass,TypeVar))else ValueSet({A})for A in G).execute_annotation()
def infer_type_vars_for_execution(function,arguments,annotation_dict):
	'\n    Some functions use type vars that are not defined by the class, but rather\n    only defined in the function. See for example `iter`. In those cases we\n    want to:\n\n    1. Search for undefined type vars.\n    2. Infer type vars with the execution state we have.\n    3. Return the union of all type vars that have been found.\n    ';C=function;D=C.get_default_param_context();E={};H=get_executed_param_names(C,arguments)
	for B in H:
		try:F=annotation_dict[B.string_name]
		except KeyError:continue
		I=find_unknown_type_vars(D,F)
		if I:
			J=D.infer_node(F);G=B.get_kind();A=B.infer()
			if G is Parameter.VAR_POSITIONAL:A=A.merge_types_of_iterate()
			elif G is Parameter.VAR_KEYWORD:A=A.try_merge('_dict_values')
			merge_type_var_dicts(E,J.infer_type_vars(A))
	return E
def infer_return_for_callable(arguments,param_values,result_values):
	B={}
	for A in param_values:
		if A.array_type=='list':C=_infer_type_vars_for_callable(arguments,A.py__iter__());B.update(C)
	return ValueSet.from_sets(A.define_generics(B)if isinstance(A,(DefineGenericBaseClass,TypeVar))else ValueSet({A})for A in result_values).execute_annotation()
def _infer_type_vars_for_callable(arguments,lazy_params):
	'\n    Infers type vars for the Calllable class:\n\n        def x() -> Callable[[Callable[..., _T]], _T]: ...\n    ';A={}
	for((F,B),C)in zip(arguments.unpack(),lazy_params):D=C.infer();E=B.infer();merge_type_var_dicts(A,D.infer_type_vars(E))
	return A
def merge_type_var_dicts(base_dict,new_dict):
	B=base_dict
	for(C,A)in new_dict.items():
		if A:
			try:B[C]|=A
			except KeyError:B[C]=A
def merge_pairwise_generics(annotation_value,annotated_argument_class):
	"\n    Match up the generic parameters from the given argument class to the\n    target annotation.\n\n    This walks the generic parameters immediately within the annotation and\n    argument's type, in order to determine the concrete values of the\n    annotation's parameters for the current case.\n\n    For example, given the following code:\n\n        def values(mapping: Mapping[K, V]) -> List[V]: ...\n\n        for val in values({1: 'a'}):\n            val\n\n    Then this function should be given representations of `Mapping[K, V]`\n    and `Mapping[int, str]`, so that it can determine that `K` is `int and\n    `V` is `str`.\n\n    Note that it is responsibility of the caller to traverse the MRO of the\n    argument type as needed in order to find the type matching the\n    annotation (in this case finding `Mapping[int, str]` as a parent of\n    `Dict[int, str]`).\n\n    Parameters\n    ----------\n\n    `annotation_value`: represents the annotation to infer the concrete\n        parameter types of.\n\n    `annotated_argument_class`: represents the annotated class of the\n        argument being passed to the object annotated by `annotation_value`.\n    ";B=annotated_argument_class;A={}
	if not isinstance(B,DefineGenericBaseClass):return A
	C=annotation_value.get_generics();D=B.get_generics()
	for(E,F)in zip(C,D):merge_type_var_dicts(A,E.infer_type_vars(F.execute_annotation()))
	return A
def find_type_from_comment_hint_for(context,node,name):return _find_type_from_comment_hint(context,node,node.children[1],name)
def find_type_from_comment_hint_with(context,node,name):
	A=node
	if len(A.children)>4:return[]
	assert len(A.children[1].children)==3,"Can only be here when children[1] is 'foo() as f'";B=A.children[1].children[2];return _find_type_from_comment_hint(context,A,B,name)
def find_type_from_comment_hint_assign(context,node,name):return _find_type_from_comment_hint(context,node,node.children[0],name)
def _find_type_from_comment_hint(context,node,varlist,name):
	B=varlist;A=_A
	if B.type in('testlist_star_expr','exprlist','testlist'):
		A=0
		for C in B.children:
			if C==name:break
			if C.type=='operator':continue
			A+=1
		else:return[]
	D=parser_utils.get_following_comment_same_line(node)
	if D is _A:return[]
	E=re.match('^#\\s*type:\\s*([^#]*)',D)
	if E is _A:return[]
	return _infer_annotation_string(context,E.group(1).strip(),A).execute_annotation()
def find_unknown_type_vars(context,node):
	def D(node):
		B=node
		if B.type in(_C,_D):
			C=B.children[-1]
			if C.type=='trailer'and C.children[0]=='[':
				for E in _unpack_subscriptlist(C.children[1]):D(E)
		else:A[:]=_filter_type_vars(context.infer_node(B),A)
	A=[];D(node);return A
def _filter_type_vars(value_set,found=()):
	B=found;C=list(B)
	for A in value_set:
		if isinstance(A,TypeVar)and A not in B:C.append(A)
	return C
def _unpack_subscriptlist(subscriptlist):
	C='subscript';A=subscriptlist
	if A.type=='subscriptlist':
		for B in A.children[::2]:
			if B.type!=C:yield B
	elif A.type!=C:yield A