"\nRecursions are the recipe of |jedi| to conquer Python code. However, someone\nmust stop recursions going mad. Some settings are here to make |jedi| stop at\nthe right time. You can read more about them :ref:`here <settings-recursion>`.\n\nNext to the internal ``jedi.inference.cache`` this module also makes |jedi| not\nthread-safe, because ``execution_recursion_decorator`` uses class variables to\ncount the function calls.\n\n.. _settings-recursion:\n\nSettings\n~~~~~~~~~~\n\nRecursion settings are important if you don't want extremely\nrecursive python code to go absolutely crazy.\n\nThe default values are based on experiments while completing the |jedi| library\nitself (inception!). But I don't think there's any other Python library that\nuses recursion in a similarly extreme way. Completion should also be fast and\ntherefore the quality might not always be maximal.\n\n.. autodata:: recursion_limit\n.. autodata:: total_function_execution_limit\n.. autodata:: per_function_execution_limit\n.. autodata:: per_function_recursion_limit\n"
_B=False
_A=True
from contextlib import contextmanager
from jedi import debug
from jedi.inference.base_value import NO_VALUES
recursion_limit=15
'\nLike :func:`sys.getrecursionlimit()`, just for |jedi|.\n'
total_function_execution_limit=200
'\nThis is a hard limit of how many non-builtin functions can be executed.\n'
per_function_execution_limit=6
'\nThe maximal amount of times a specific function may be executed.\n'
per_function_recursion_limit=2
'\nA function may not be executed more than this number of times recursively.\n'
class RecursionDetector:
	def __init__(A):A.pushed_nodes=[]
@contextmanager
def execution_allowed(inference_state,node):
	'\n    A decorator to detect recursions in statements. In a recursion a statement\n    at the same place, in the same module may not be executed two times.\n    ';A=node;B=inference_state.recursion_detector.pushed_nodes
	if A in B:debug.warning('catched stmt recursion: %s @%s',A,getattr(A,'start_pos',None));yield _B
	else:
		try:B.append(A);yield _A
		finally:B.pop()
def execution_recursion_decorator(default=NO_VALUES):
	def A(func):
		def A(self,**D):
			A=self;B=A.inference_state.execution_recursion_detector;E=B.push_execution(A)
			try:
				if E:C=default
				else:C=func(A,**D)
			finally:B.pop_execution()
			return C
		return A
	return A
class ExecutionRecursionDetector:
	'\n    Catches recursions of executions.\n    '
	def __init__(A,inference_state):A._inference_state=inference_state;A._recursion_level=0;A._parent_execution_funcs=[];A._funcdef_execution_counts={};A._execution_count=0
	def pop_execution(A):A._parent_execution_funcs.pop();A._recursion_level-=1
	def push_execution(A,execution):
		C=execution;B=C.tree_node;A._recursion_level+=1;A._parent_execution_funcs.append(B);D=C.get_root_context()
		if D.is_builtins_module():return _B
		if A._recursion_level>recursion_limit:debug.warning('Recursion limit (%s) reached',recursion_limit);return _A
		if A._execution_count>=total_function_execution_limit:debug.warning('Function execution limit (%s) reached',total_function_execution_limit);return _A
		A._execution_count+=1
		if A._funcdef_execution_counts.setdefault(B,0)>=per_function_execution_limit:
			if D.py__name__()=='typing':return _B
			debug.warning('Per function execution limit (%s) reached: %s',per_function_execution_limit,B);return _A
		A._funcdef_execution_counts[B]+=1
		if A._parent_execution_funcs.count(B)>per_function_recursion_limit:debug.warning('Per function recursion limit (%s) reached: %s',per_function_recursion_limit,B);return _A
		return _B