"\nA module to deal with stuff like `list.append` and `set.add`.\n\nArray modifications\n*******************\n\nIf the content of an array (``set``/``list``) is requested somewhere, the\ncurrent module will be checked for appearances of ``arr.append``,\n``arr.insert``, etc.  If the ``arr`` name points to an actual array, the\ncontent will be added\n\nThis can be really cpu intensive, as you can imagine. Because |jedi| has to\nfollow **every** ``append`` and check whether it's the right array. However this\nworks pretty good, because in *slow* cases, the recursion detector and other\nsettings will stop this process.\n\nIt is important to note that:\n\n1. Array modifications work only in the current module.\n2. Jedi only checks Array additions; ``list.pop``, etc are ignored.\n"
_A=None
from jedi import debug,settings
from jedi.inference import recursion
from jedi.inference.base_value import ValueSet,NO_VALUES,HelperValueMixin,ValueWrapper
from jedi.inference.lazy_value import LazyKnownValues
from jedi.inference.helpers import infer_call_of_leaf
from jedi.inference.cache import inference_state_method_cache
_sentinel=object()
def check_array_additions(context,sequence):
	' Just a mapper function for the internal _internal_check_array_additions ';A=sequence
	if A.array_type not in('list','set'):return NO_VALUES
	return _internal_check_array_additions(context,A)
@inference_state_method_cache(default=NO_VALUES)
@debug.increase_indent
def _internal_check_array_additions(context,sequence):
	'\n    Checks if a `Array` has "add" (append, insert, extend) statements:\n\n    >>> a = [""]\n    >>> a.append(1)\n    ';Q='update';P='extend';O='add';N='append';H='insert';G='MAGENTA';D=sequence;A=context;from jedi.inference import arguments as R;debug.dbg('Dynamic array search for %s'%D,color=G);I=A.get_root_context()
	if not settings.dynamic_array_additions or I.is_compiled():debug.dbg('Dynamic array search aborted.',color=G);return NO_VALUES
	def S(context,arglist,add_name):
		E=context;B=add_name;A=list(R.TreeArguments(E.inference_state,E,arglist).unpack());C=set()
		if B in[H]:A=A[1:]
		if B in[N,O,H]:
			for(F,D)in A:C.add(D)
		elif B in[P,Q]:
			for(F,D)in A:C|=set(D.infer().iterate())
		return C
	T,settings.dynamic_params_for_other_modules=settings.dynamic_params_for_other_modules,False;U=D.name.string_name=='list';V=[N,P,H]if U else[O,Q];E=set()
	for J in V:
		try:W=I.tree_node.get_used_names()[J]
		except KeyError:continue
		else:
			for B in W:
				K=A.tree_node
				if not K.start_pos<B.start_pos<K.end_pos:continue
				L=B.parent;F=L.parent;X=F.children.index(L)
				try:C=F.children[X+1]
				except IndexError:continue
				else:
					if C.type!='trailer'or C.children[0]!='('or C.children[1]==')':continue
				M=A.create_context(B)
				with recursion.execution_allowed(A.inference_state,F)as Y:
					if Y:
						Z=infer_call_of_leaf(M,B,cut_own_trailer=True)
						if D in Z:E|=S(M,C.children[1],J)
	settings.dynamic_params_for_other_modules=T;debug.dbg('Dynamic array result %s',E,color=G);return E
def get_dynamic_array_instance(instance,arguments):'Used for set() and list() instances.';A=_DynamicArrayAdditions(instance,arguments);from jedi.inference import arguments;return arguments.ValuesArguments([ValueSet([A])])
class _DynamicArrayAdditions(HelperValueMixin):
	"\n    Used for the usage of set() and list().\n    This is definitely a hack, but a good one :-)\n    It makes it possible to use set/list conversions.\n\n    This is not a proper context, because it doesn't have to be. It's not used\n    in the wild, it's just used within typeshed as an argument to `__init__`\n    for set/list and never used in any other place.\n    "
	def __init__(A,instance,arguments):A._instance=instance;A._arguments=arguments
	def py__class__(A):B,=A._instance.inference_state.builtins_module.py__getattribute__('tuple');return B
	def py__iter__(B,contextualized_node=_A):
		A=B._arguments
		try:F,C=next(A.unpack())
		except StopIteration:pass
		else:yield from C.infer().iterate()
		from jedi.inference.arguments import TreeArguments as D
		if isinstance(A,D):E=_internal_check_array_additions(A.context,B._instance);yield from E
	def iterate(A,contextualized_node=_A,is_async=False):return A.py__iter__(contextualized_node)
class _Modification(ValueWrapper):
	def __init__(A,wrapped_value,assigned_values,contextualized_key):super().__init__(wrapped_value);A._assigned_values=assigned_values;A._contextualized_key=contextualized_key
	def py__getitem__(A,*B,**C):return A._wrapped_value.py__getitem__(*B,**C)|A._assigned_values
	def py__simple_getitem__(A,index):
		B=index;C=[A.get_safe_value(_sentinel)for A in A._contextualized_key.infer()]
		if B in C:return A._assigned_values
		return A._wrapped_value.py__simple_getitem__(B)
class DictModification(_Modification):
	def py__iter__(A,contextualized_node=_A):yield from A._wrapped_value.py__iter__(contextualized_node);yield A._contextualized_key
	def get_key_values(A):return A._wrapped_value.get_key_values()|A._contextualized_key.infer()
class ListModification(_Modification):
	def py__iter__(A,contextualized_node=_A):yield from A._wrapped_value.py__iter__(contextualized_node);yield LazyKnownValues(A._assigned_values)