'\nLike described in the :mod:`parso.python.tree` module,\nthere\'s a need for an ast like module to represent the states of parsed\nmodules.\n\nBut now there are also structures in Python that need a little bit more than\nthat. An ``Instance`` for example is only a ``Class`` before it is\ninstantiated. This class represents these cases.\n\nSo, why is there also a ``Class`` class here? Well, there are decorators and\nthey change classes in Python 3.\n\nRepresentation modules also define "magic methods". Those methods look like\n``py__foo__`` and are typically mappable to the Python equivalents ``__call__``\nand others. Here\'s a list:\n\n====================================== ========================================\n**Method**                             **Description**\n-------------------------------------- ----------------------------------------\npy__call__(arguments: Array)           On callable objects, returns types.\npy__bool__()                           Returns True/False/None; None means that\n                                       there\'s no certainty.\npy__bases__()                          Returns a list of base classes.\npy__iter__()                           Returns a generator of a set of types.\npy__class__()                          Returns the class of an instance.\npy__simple_getitem__(index: int/str)   Returns a a set of types of the index.\n                                       Can raise an IndexError/KeyError.\npy__getitem__(indexes: ValueSet)       Returns a a set of types of the index.\npy__file__()                           Only on modules. Returns None if does\n                                       not exist.\npy__package__() -> List[str]           Only on modules. For the import system.\npy__path__()                           Only on modules. For the import system.\npy__get__(call_object)                 Only on instances. Simulates\n                                       descriptors.\npy__doc__()                            Returns the docstring for a value.\n====================================== ========================================\n\n'
_C=True
_B=False
_A=None
from jedi import debug
from jedi.parser_utils import get_cached_parent_scope,expr_is_dotted,function_is_property
from jedi.inference.cache import inference_state_method_cache,CachedMetaClass,inference_state_method_generator_cache
from jedi.inference import compiled
from jedi.inference.lazy_value import LazyKnownValues,LazyTreeValue
from jedi.inference.filters import ParserTreeFilter
from jedi.inference.names import TreeNameDefinition,ValueName
from jedi.inference.arguments import unpack_arglist,ValuesArguments
from jedi.inference.base_value import ValueSet,iterator_to_value_set,NO_VALUES
from jedi.inference.context import ClassContext
from jedi.inference.value.function import FunctionAndClassBase
from jedi.inference.gradual.generics import LazyGenericManager,TupleGenericManager
from jedi.plugins import plugin_manager
class ClassName(TreeNameDefinition):
	def __init__(A,class_value,tree_name,name_context,apply_decorators):super().__init__(name_context,tree_name);A._apply_decorators=apply_decorators;A._class_value=class_value
	@iterator_to_value_set
	def infer(self):
		A=self;from jedi.inference.syntax_tree import tree_name_to_values as C;D=C(A.parent_context.inference_state,A.parent_context,A.tree_name)
		for B in D:
			if A._apply_decorators:yield from B.py__get__(instance=_A,class_value=A._class_value)
			else:yield B
	@property
	def api_type(self):
		A=super().api_type
		if A=='function':
			B=self.tree_name.get_definition()
			if B is _A:return A
			if function_is_property(B):return'property'
		return A
class ClassFilter(ParserTreeFilter):
	def __init__(A,class_value,node_context=_A,until_position=_A,origin_scope=_A,is_instance=_B):B=class_value;super().__init__(B.as_context(),node_context,until_position=until_position,origin_scope=origin_scope);A._class_value=B;A._is_instance=is_instance
	def _convert_names(A,names):return[ClassName(class_value=A._class_value,tree_name=B,name_context=A._node_context,apply_decorators=not A._is_instance)for B in names]
	def _equals_origin_scope(B):
		A=B._origin_scope
		while A is not _A:
			if A==B._parser_scope or A==B.parent_context:return _C
			A=get_cached_parent_scope(B._parso_cache_node,A)
		return _B
	def _access_possible(A,name):return not name.value.startswith('__')or name.value.endswith('__')or A._equals_origin_scope()
	def _filter(B,names):A=names;A=super()._filter(A);return[A for A in A if B._access_possible(A)]
class ClassMixin:
	def is_class(A):return _C
	def is_class_mixin(A):return _C
	def py__call__(A,arguments):
		from jedi.inference.value import TreeInstance as B;from jedi.inference.gradual.typing import TypedDict as C
		if A.is_typeddict():return ValueSet([C(A)])
		return ValueSet([B(A.inference_state,A.parent_context,A,arguments)])
	def py__class__(A):return compiled.builtin_from_name(A.inference_state,'type')
	@property
	def name(self):return ValueName(self,self.tree_node.name)
	def py__name__(A):return A.name.string_name
	@inference_state_method_generator_cache()
	def py__mro__(self):
		A=self;C=[A];yield A
		for E in A.py__bases__():
			for D in E.infer():
				try:F=D.py__mro__
				except AttributeError:'\n                    >>> class Y(lambda: test): pass\n                    Traceback (most recent call last):\n                      File "<stdin>", line 1, in <module>\n                    TypeError: function() argument 1 must be code, not str\n                    >>> class Y(1): pass\n                    Traceback (most recent call last):\n                      File "<stdin>", line 1, in <module>\n                    TypeError: int() takes at most 2 arguments (3 given)\n                    ';debug.warning('Super class of %s is not a class: %s',A,D)
				else:
					for B in F():
						if B not in C:C.append(B);yield B
	def get_filters(A,origin_scope=_A,is_instance=_B,include_metaclasses=_C,include_type_when_class=_C):
		B=is_instance
		if include_metaclasses:
			F=A.get_metaclasses()
			if F:yield from A.get_metaclass_filters(F,B)
		for C in A.py__mro__():
			if C.is_compiled():yield from C.get_filters(is_instance=B)
			else:yield ClassFilter(A,node_context=C.as_context(),origin_scope=origin_scope,is_instance=B)
		if not B and include_type_when_class:
			from jedi.inference.compiled import builtin_from_name as H;D=H(A.inference_state,'type');assert isinstance(D,ClassValue)
			if D!=A:
				I=ValuesArguments([])
				for J in D.py__call__(I):E=J.get_filters();next(E,_A);next(E,_A);G=next(E,_A);assert G is not _A;yield G
	def get_signatures(A):
		B=A.get_metaclasses()
		if B:
			C=A.get_metaclass_signatures(B)
			if C:return C
		D=ValuesArguments([]);E=A.py__call__(D).py__getattribute__('__init__');return[B.bind(A)for B in E.get_signatures()]
	def _as_context(A):return ClassContext(A)
	def get_type_hint(A,add_class_info=_C):
		if add_class_info:return'Type[%s]'%A.py__name__()
		return A.py__name__()
	@inference_state_method_cache(default=_B)
	def is_typeddict(self):
		from jedi.inference.gradual.typing import TypedDictClass as C
		for A in self.py__bases__():
			if not isinstance(A,LazyTreeValue):return _B
			D=A.data
			if not expr_is_dotted(D):return _B
			for B in A.infer():
				if isinstance(B,C):return _C
				try:E=B.is_typeddict
				except AttributeError:return _B
				else:
					if E():return _C
		return _B
	def py__getitem__(A,index_value_set,contextualized_node):
		B=index_value_set;from jedi.inference.gradual.base import GenericClass as C
		if not B:debug.warning('Class indexes inferred to nothing. Returning class instead');return ValueSet([A])
		return ValueSet(C(A,LazyGenericManager(context_of_index=contextualized_node.context,index_value=B))for B in B)
	def with_generics(A,generics_tuple):from jedi.inference.gradual.base import GenericClass as B;return B(A,TupleGenericManager(generics_tuple))
	def define_generics(A,type_var_dict):
		B=type_var_dict;from jedi.inference.gradual.base import GenericClass as C
		def D():
			'\n            The TypeVars in the resulting classes have sometimes different names\n            and we need to check for that, e.g. a signature can be:\n\n            def iter(iterable: Iterable[_T]) -> Iterator[_T]: ...\n\n            However, the iterator is defined as Iterator[_T_co], which means it has\n            a different type var name.\n            '
			for C in A.list_type_vars():yield B.get(C.py__name__(),NO_VALUES)
		if B:return ValueSet([C(A,TupleGenericManager(tuple(D())))])
		return ValueSet({A})
class ClassValue(ClassMixin,FunctionAndClassBase,metaclass=CachedMetaClass):
	api_type='class'
	@inference_state_method_cache()
	def list_type_vars(self):
		A=[];B=self.tree_node.get_super_arglist()
		if B is _A:return[]
		for(D,E)in unpack_arglist(B):
			if D:continue
			from jedi.inference.gradual.annotation import find_unknown_type_vars as F
			for C in F(self.parent_context,E):
				if C not in A:A.append(C)
		return A
	def _get_bases_arguments(A):
		B=A.tree_node.get_super_arglist()
		if B:from jedi.inference import arguments as C;return C.TreeArguments(A.inference_state,A.parent_context,B)
	@inference_state_method_cache(default=())
	def py__bases__(self):
		D='object';A=self;B=A._get_bases_arguments()
		if B is not _A:
			C=[B for(A,B)in B.unpack()if A is _A]
			if C:return C
		if A.py__name__()==D and A.parent_context.is_builtins_module():return[]
		return[LazyKnownValues(A.inference_state.builtins_module.py__getattribute__(D))]
	@plugin_manager.decorate()
	def get_metaclass_filters(self,metaclasses,is_instance):debug.warning('Unprocessed metaclass %s',metaclasses);return[]
	@inference_state_method_cache(default=NO_VALUES)
	def get_metaclasses(self):
		B=self._get_bases_arguments()
		if B is not _A:
			E=[B for(A,B)in B.unpack()if A=='metaclass'];A=ValueSet.from_sets(A.infer()for A in E);A=ValueSet(A for A in A if A.is_class())
			if A:return A
		for F in self.py__bases__():
			for C in F.infer():
				if C.is_class():
					D=C.get_metaclasses()
					if D:return D
		return NO_VALUES
	@plugin_manager.decorate()
	def get_metaclass_signatures(self,metaclasses):return[]