from __future__ import annotations
_F='TRY_ELEMENT_AT'
_E='TIMESTAMP_NTZ'
_D='TIMESTAMP_LTZ'
_C='DATEDIFF'
_B='DATE_ADD'
_A=True
import typing as t
from sqlglot import exp
from sqlglot.dialects.dialect import rename_func,unit_to_var
from sqlglot.dialects.hive import _build_with_ignore_nulls
from sqlglot.dialects.spark2 import Spark2,temporary_storage_provider,_build_as_cast
from sqlglot.helper import ensure_list,seq_get
from sqlglot.transforms import ctas_with_tmp_tables_to_create_tmp_view,remove_unique_constraints,preprocess,move_partitioned_by_to_schema_columns
def _build_datediff(args):
	'\n    Although Spark docs don\'t mention the "unit" argument, Spark3 added support for\n    it at some point. Databricks also supports this variant (see below).\n\n    For example, in spark-sql (v3.3.1):\n    - SELECT DATEDIFF(\'2020-01-01\', \'2020-01-05\') results in -4\n    - SELECT DATEDIFF(day, \'2020-01-01\', \'2020-01-05\') results in 4\n\n    See also:\n    - https://docs.databricks.com/sql/language-manual/functions/datediff3.html\n    - https://docs.databricks.com/sql/language-manual/functions/datediff.html\n    ';A=args;C=None;B=seq_get(A,0);D=seq_get(A,1)
	if len(A)==3:C=exp.var(t.cast(exp.Expression,B).name);B=A[2]
	return exp.DateDiff(this=exp.TsOrDsToDate(this=B),expression=exp.TsOrDsToDate(this=D),unit=C)
def _build_dateadd(args):
	A=args;B=seq_get(A,1)
	if len(A)==2:return exp.TsOrDsAdd(this=seq_get(A,0),expression=B,unit=exp.Literal.string('DAY'))
	return exp.TimestampAdd(this=seq_get(A,2),expression=B,unit=seq_get(A,0))
def _normalize_partition(e):
	'Normalize the expressions in PARTITION BY (<expression>, <expression>, ...)'
	if isinstance(e,str):return exp.to_identifier(e)
	if isinstance(e,exp.Literal):return exp.to_identifier(e.name)
	return e
def _dateadd_sql(self,expression):
	A=expression
	if not A.unit or isinstance(A,exp.TsOrDsAdd)and A.text('unit').upper()=='DAY':return self.func(_B,A.this,A.expression)
	B=self.func(_B,unit_to_var(A),A.expression,A.this)
	if isinstance(A,exp.TsOrDsAdd):
		C=A.return_type
		if not C.is_type(exp.DataType.Type.TIMESTAMP,exp.DataType.Type.DATETIME):B=f"CAST({B} AS {C})"
	return B
class Spark(Spark2):
	SUPPORTS_ORDER_BY_ALL=_A
	class Tokenizer(Spark2.Tokenizer):STRING_ESCAPES_ALLOWED_IN_RAW_STRINGS=False;RAW_STRINGS=[(B+A,A)for A in t.cast(t.List[str],Spark2.Tokenizer.QUOTES)for B in('r','R')]
	class Parser(Spark2.Parser):
		FUNCTIONS={**Spark2.Parser.FUNCTIONS,'ANY_VALUE':_build_with_ignore_nulls(exp.AnyValue),_B:_build_dateadd,'DATEADD':_build_dateadd,'TIMESTAMPADD':_build_dateadd,_C:_build_datediff,'DATE_DIFF':_build_datediff,_D:_build_as_cast(_D),_E:_build_as_cast(_E),_F:lambda args:exp.Bracket(this=seq_get(args,0),expressions=ensure_list(seq_get(args,1)),offset=1,safe=_A)}
		def _parse_generated_as_identity(B):
			A=super()._parse_generated_as_identity()
			if A.expression:return B.expression(exp.ComputedColumnConstraint,this=A.expression)
			return A
	class Generator(Spark2.Generator):
		SUPPORTS_TO_NUMBER=_A;PAD_FILL_PATTERN_IS_REQUIRED=False;SUPPORTS_CONVERT_TIMEZONE=_A;TYPE_MAPPING={**Spark2.Generator.TYPE_MAPPING,exp.DataType.Type.MONEY:'DECIMAL(15, 4)',exp.DataType.Type.SMALLMONEY:'DECIMAL(6, 4)',exp.DataType.Type.UNIQUEIDENTIFIER:'STRING',exp.DataType.Type.TIMESTAMPLTZ:_D,exp.DataType.Type.TIMESTAMPNTZ:_E};TRANSFORMS={**Spark2.Generator.TRANSFORMS,exp.ArrayConstructCompact:lambda self,e:self.func('ARRAY_COMPACT',self.func('ARRAY',*e.expressions)),exp.Create:preprocess([remove_unique_constraints,lambda e:ctas_with_tmp_tables_to_create_tmp_view(e,temporary_storage_provider),move_partitioned_by_to_schema_columns]),exp.PartitionedByProperty:lambda self,e:f"PARTITIONED BY {self.wrap(self.expressions(sqls=[_normalize_partition(A)for A in e.this.expressions],skip_first=_A))}",exp.StartsWith:rename_func('STARTSWITH'),exp.TsOrDsAdd:_dateadd_sql,exp.TimestampAdd:_dateadd_sql,exp.TryCast:lambda self,e:self.trycast_sql(e)if e.args.get('safe')else self.cast_sql(e)};TRANSFORMS.pop(exp.AnyValue);TRANSFORMS.pop(exp.DateDiff);TRANSFORMS.pop(exp.Group)
		def bracket_sql(B,expression):
			A=expression
			if A.args.get('safe'):C=seq_get(B.bracket_offset_expressions(A,index_offset=1),0);return B.func(_F,A.this,C)
			return super().bracket_sql(A)
		def computedcolumnconstraint_sql(A,expression):return f"GENERATED ALWAYS AS ({A.sql(expression,"this")})"
		def anyvalue_sql(A,expression):return A.function_fallback_sql(expression)
		def datediff_sql(A,expression):
			B=expression;C=A.sql(B,'this');D=A.sql(B,'expression')
			if B.unit:return A.func(_C,unit_to_var(B),D,C)
			return A.func(_C,C,D)