'\n.. include:: ../posts/sql_diff.md\n\n----\n'
from __future__ import annotations
_C=False
_B=True
_A=None
import typing as t
from collections import defaultdict
from dataclasses import dataclass
from heapq import heappop,heappush
from sqlglot import Dialect,expressions as exp
from sqlglot.helper import ensure_list
if t.TYPE_CHECKING:from sqlglot.dialects.dialect import DialectType
@dataclass(frozen=_B)
class Insert:'Indicates that a new node has been inserted';expression:exp.Expression
@dataclass(frozen=_B)
class Remove:'Indicates that an existing node has been removed';expression:exp.Expression
@dataclass(frozen=_B)
class Move:"Indicates that an existing node's position within the tree has changed";expression:exp.Expression
@dataclass(frozen=_B)
class Update:'Indicates that an existing node has been updated';source:exp.Expression;target:exp.Expression
@dataclass(frozen=_B)
class Keep:"Indicates that an existing node hasn't been changed";source:exp.Expression;target:exp.Expression
if t.TYPE_CHECKING:from sqlglot._typing import T;Edit=t.Union[Insert,Remove,Move,Update,Keep]
def diff(source,target,matchings=_A,delta_only=_C,copy=_B,**H):
	'\n    Returns the list of changes between the source and the target expressions.\n\n    Examples:\n        >>> diff(parse_one("a + b"), parse_one("a + c"))\n        [\n            Remove(expression=(COLUMN this: (IDENTIFIER this: b, quoted: False))),\n            Insert(expression=(COLUMN this: (IDENTIFIER this: c, quoted: False))),\n            Keep(\n                source=(ADD this: ...),\n                target=(ADD this: ...)\n            ),\n            Keep(\n                source=(COLUMN this: (IDENTIFIER this: a, quoted: False)),\n                target=(COLUMN this: (IDENTIFIER this: a, quoted: False))\n            ),\n        ]\n\n    Args:\n        source: the source expression.\n        target: the target expression against which the diff should be calculated.\n        matchings: the list of pre-matched node pairs which is used to help the algorithm\'s\n            heuristics produce better results for subtrees that are known by a caller to be matching.\n            Note: expression references in this list must refer to the same node objects that are\n            referenced in source / target trees.\n        delta_only: excludes all `Keep` nodes from the diff.\n        copy: whether to copy the input expressions.\n            Note: if this is set to false, the caller must ensure that there are no shared references\n            in the two ASTs, otherwise the diffing algorithm may produce unexpected behavior.\n        kwargs: additional arguments to pass to the ChangeDistiller instance.\n\n    Returns:\n        the list of Insert, Remove, Move, Update and Keep objects for each node in the source and the\n        target expression trees. This list represents a sequence of steps needed to transform the source\n        expression tree into the target one.\n    ';C=target;B=source;A=matchings;A=A or[];I={id(B)for A in A for B in A}
	def D(original,copy):return{id(A):B for(A,B)in zip(original.walk(),copy.walk())if id(A)in I}
	E=B.copy()if copy else B;F=C.copy()if copy else C;G={**D(B,E),**D(C,F)};J=[(G[id(A)],G[id(B)])for(A,B)in A];return ChangeDistiller(**H).diff(E,F,matchings=J,delta_only=delta_only)
UPDATABLE_EXPRESSION_TYPES=exp.Alias,exp.Boolean,exp.Column,exp.DataType,exp.Lambda,exp.Literal,exp.Table,exp.Window
IGNORED_LEAF_EXPRESSION_TYPES=exp.Identifier,
class ChangeDistiller:
	'\n    The implementation of the Change Distiller algorithm described by Beat Fluri and Martin Pinzger in\n    their paper https://ieeexplore.ieee.org/document/4339230, which in turn is based on the algorithm by\n    Chawathe et al. described in http://ilpubs.stanford.edu:8090/115/1/1995-46.pdf.\n    '
	def __init__(A,f=.6,t=.6,dialect=_A):A.f=f;A.t=t;A._sql_generator=Dialect.get_or_raise(dialect).generator()
	def diff(A,source,target,matchings=_A,delta_only=_C):
		B=matchings;B=B or[];C={id(A):id(B)for(A,B)in B}
		if len({B for A in C.items()for B in A})!=2*len(B):raise ValueError('Each node can be referenced at most once in the list of matchings')
		A._source=source;A._target=target;A._source_index={id(A):A for A in A._source.bfs()if not isinstance(A,IGNORED_LEAF_EXPRESSION_TYPES)};A._target_index={id(A):A for A in A._target.bfs()if not isinstance(A,IGNORED_LEAF_EXPRESSION_TYPES)};A._unmatched_source_nodes=set(A._source_index)-set(C);A._unmatched_target_nodes=set(A._target_index)-set(C.values());A._bigram_histo_cache={};D=A._compute_matching_set()|{(A,B)for(A,B)in C.items()};return A._generate_edit_script(D,delta_only)
	def _generate_edit_script(A,matching_set,delta_only):
		E=matching_set;B=[]
		for F in A._unmatched_source_nodes:B.append(Remove(A._source_index[F]))
		for G in A._unmatched_target_nodes:B.append(Insert(A._target_index[G]))
		for(H,I)in E:
			C=A._source_index[H];D=A._target_index[I]
			if not isinstance(C,UPDATABLE_EXPRESSION_TYPES)or C==D:
				B.extend(A._generate_move_edits(C,D,E))
				if not delta_only:B.append(Keep(C,D))
			else:B.append(Update(C,D))
		return B
	def _generate_move_edits(B,source,target,matching_set):
		C=[id(A)for A in _expression_only_args(source)];E=[id(A)for A in _expression_only_args(target)];F=set(_lcs(C,E,lambda l,r:(l,r)in matching_set));D=[]
		for A in C:
			if A not in F and A not in B._unmatched_source_nodes:D.append(Move(B._source_index[A]))
		return D
	def _compute_matching_set(A):
		I=A._compute_leaf_matching_set();J=I.copy();M={id(B):_A for B in A._source.bfs()if id(B)in A._unmatched_source_nodes};K={id(B):_A for B in A._target.bfs()if id(B)in A._unmatched_target_nodes}
		for C in M:
			for B in K:
				D=A._source_index[C];E=A._target_index[B]
				if _is_same_type(D,E):
					F={id(A)for A in _get_leaves(D)};G={id(A)for A in _get_leaves(E)};L=max(len(F),len(G))
					if L:N=sum(1 if A in F and B in G else 0 for(A,B)in I);H=N/L
					else:H=.0
					O=A.t if min(len(F),len(G))>4 else .4
					if H>=.8 or H>=O and A._dice_coefficient(D,E)>=A.f:J.add((C,B));A._unmatched_source_nodes.remove(C);A._unmatched_target_nodes.remove(B);K.pop(B,_A);break
		return J
	def _compute_leaf_matching_set(A):
		D=[];H=list(_get_leaves(A._source));I=list(_get_leaves(A._target))
		for B in H:
			for C in I:
				if _is_same_type(B,C):
					E=A._dice_coefficient(B,C)
					if E>=A.f:heappush(D,(-E,-_parent_similarity_score(B,C),len(D),B,C))
		F=set()
		while D:
			G,G,G,B,C=heappop(D)
			if id(B)in A._unmatched_source_nodes and id(C)in A._unmatched_target_nodes:F.add((id(B),id(C)));A._unmatched_source_nodes.remove(id(B));A._unmatched_target_nodes.remove(id(C))
		return F
	def _dice_coefficient(C,source,target):
		E=target;D=source;A=C._bigram_histo(D);B=C._bigram_histo(E);F=sum(A.values())+sum(B.values())
		if not F:return 1. if D==E else .0
		G=0;I=set(A)&set(B)
		for H in I:G+=min(A[H],B[H])
		return 2*G/F
	def _bigram_histo(A,expression):
		B=expression
		if id(B)in A._bigram_histo_cache:return A._bigram_histo_cache[id(B)]
		D=A._sql_generator.generate(B);F=max(0,len(D)-1);C=defaultdict(int)
		for E in range(F):C[D[E:E+2]]+=1
		A._bigram_histo_cache[id(B)]=C;return C
def _get_leaves(expression):
	A=expression;B=_C
	for C in A.iter_expressions():
		if not isinstance(C,IGNORED_LEAF_EXPRESSION_TYPES):B=_B;yield from _get_leaves(C)
	if not B:yield A
def _is_same_type(source,target):
	C='side';B=target;A=source
	if type(A)is type(B):
		if isinstance(A,exp.Join):return A.args.get(C)==B.args.get(C)
		if isinstance(A,exp.Anonymous):return A.this==B.this
		return _B
	return _C
def _parent_similarity_score(source,target):
	B=target;A=source
	if A is _A or B is _A or type(A)is not type(B):return 0
	return 1+_parent_similarity_score(A.parent,B.parent)
def _expression_only_args(expression):
	A=expression;B=[]
	if A:
		for C in A.args.values():B.extend(ensure_list(C))
	return[A for A in B if isinstance(A,exp.Expression)and not isinstance(A,IGNORED_LEAF_EXPRESSION_TYPES)]
def _lcs(seq_a,seq_b,equal):
	'Calculates the longest common subsequence';G=seq_b;D=seq_a;E=len(D);F=len(G);C=[[_A]*(F+1)for A in range(E+1)]
	for A in range(E+1):
		for B in range(F+1):
			if A==0 or B==0:C[A][B]=[]
			elif equal(D[A-1],G[B-1]):C[A][B]=C[A-1][B-1]+[D[A-1]]
			else:C[A][B]=C[A-1][B]if len(C[A-1][B])>len(C[A][B-1])else C[A][B-1]
	return C[E][F]