'\n## Expressions\n\nEvery AST node in SQLGlot is represented by a subclass of `Expression`.\n\nThis module contains the implementation of all supported `Expression` types. Additionally,\nit exposes a number of helper functions, which are mainly used to programmatically build\nSQL expressions, such as `sqlglot.expressions.select`.\n\n----\n'
from __future__ import annotations
_Aa='occurrence'
_AZ='only_json_types'
_AY='on_condition'
_AX='unique_keys'
_AW='precision'
_AV='retention_period'
_AU='input_format'
_AT='overwrite'
_AS='credentials'
_AR='index_type'
_AQ='condition'
_AP='tablespace'
_AO='allow_null'
_AN='increment'
_AM='concurrently'
_AL='length'
_AK='charset'
_AJ='decimals'
_AI='nullable'
_AH='TABLE'
_AG='distinct'
_AF='joins'
_AE='laterals'
_AD='count'
_AC='settings'
_AB='partition_by'
_AA='quoted'
_A9='local'
_A8='temporary'
_A7='delete'
_A6='serde'
_A5='constraints'
_A4='materialized'
_A3='recursive'
_A2='target'
_A1='option'
_A0='unique'
_z='replace'
_y='keys'
_x='null_handling'
_w='method'
_v='percent'
_u='source'
_t='params'
_s='global'
_r='is_string'
_q='values'
_p='step'
_o='null'
_n='hint'
_m='group'
_l='to'
_k='partition'
_j='path'
_i='return_type'
_h='no'
_g='cluster'
_f='returning'
_e='position'
_d='end'
_c='limit'
_b='parent'
_a='columns'
_Z='table'
_Y='from'
_X='options'
_W='order'
_V='catalog'
_U='on'
_T='using'
_S='db'
_R='exists'
_Q='safe'
_P='where'
_O='start'
_N='offset'
_M='default'
_L='zone'
_K='format'
_J='with'
_I='alias'
_H='unit'
_G='kind'
_F='expression'
_E='expressions'
_D=None
_C='this'
_B=True
_A=False
import datetime,math,numbers,re,textwrap,typing as t
from collections import deque
from copy import deepcopy
from decimal import Decimal
from enum import auto
from functools import reduce
from sqlglot.errors import ErrorLevel,ParseError
from sqlglot.helper import AutoName,camel_to_snake_case,ensure_collection,ensure_list,seq_get,subclasses
from sqlglot.tokens import Token,TokenError
if t.TYPE_CHECKING:from typing_extensions import Self;from sqlglot._typing import E,Lit;from sqlglot.dialects.dialect import DialectType;Q=t.TypeVar('Q',bound='Query');S=t.TypeVar('S',bound='SetOperation')
class _Expression(type):
	def __new__(C,clsname,bases,attrs):B=clsname;A=super().__new__(C,B,bases,attrs);A.key=B.lower();A.__doc__=A.__doc__ or'';return A
SQLGLOT_META='sqlglot.meta'
TABLE_PARTS=_C,_S,_V
COLUMN_PARTS=_C,_Z,_S,_V
class Expression(metaclass=_Expression):
	'\n    The base class for all expressions in a syntax tree. Each Expression encapsulates any necessary\n    context, such as its child expressions, their names (arg keys), and whether a given child expression\n    is optional or not.\n\n    Attributes:\n        key: a unique key for each class in the Expression hierarchy. This is useful for hashing\n            and representing expressions as strings.\n        arg_types: determines the arguments (child nodes) supported by an expression. It maps\n            arg keys to booleans that indicate whether the corresponding args are optional.\n        parent: a reference to the parent expression (or None, in case of root expressions).\n        arg_key: the arg key an expression is associated with, i.e. the name its parent expression\n            uses to refer to it.\n        index: the index of an expression if it is inside of a list argument in its parent.\n        comments: a list of comments that are associated with a given expression. This is used in\n            order to preserve comments when transpiling SQL code.\n        type: the `sqlglot.expressions.DataType` type of an expression. This is inferred by the\n            optimizer, in order to enable some transformations that require type information.\n        meta: a dictionary that can be used to store useful metadata for a given expression.\n\n    Example:\n        >>> class Foo(Expression):\n        ...     arg_types = {"this": True, "expression": False}\n\n        The above definition informs us that Foo is an Expression that requires an argument called\n        "this" and may also optionally receive an argument called "expression".\n\n    Args:\n        args: a mapping used for retrieving the arguments of an expression, given their arg keys.\n    ';key=_F;arg_types={_C:_B};__slots__='args','parent','arg_key','index','comments','_type','_meta','_hash'
	def __init__(A,**B):
		A.args=B;A.parent=_D;A.arg_key=_D;A.index=_D;A.comments=_D;A._type=_D;A._meta=_D;A._hash=_D
		for(C,D)in A.args.items():A._set_parent(C,D)
	def __eq__(A,other):B=other;return type(A)is type(B)and hash(A)==hash(B)
	@property
	def hashable_args(self):return frozenset((B,tuple(_norm_arg(A)for A in A)if type(A)is list else _norm_arg(A))for(B,A)in self.args.items()if not(A is _D or A is _A or type(A)is list and not A))
	def __hash__(A):
		if A._hash is not _D:return A._hash
		return hash((A.__class__,A.hashable_args))
	@property
	def this(self):'\n        Retrieves the argument with key "this".\n        ';return self.args.get(_C)
	@property
	def expression(self):'\n        Retrieves the argument with key "expression".\n        ';return self.args.get(_F)
	@property
	def expressions(self):'\n        Retrieves the argument with key "expressions".\n        ';return self.args.get(_E)or[]
	def text(B,key):
		'\n        Returns a textual representation of the argument corresponding to "key". This can only be used\n        for args that are strings or leaf Expression instances, such as identifiers and literals.\n        ';A=B.args.get(key)
		if isinstance(A,str):return A
		if isinstance(A,(Identifier,Literal,Var)):return A.this
		if isinstance(A,(Star,Null)):return A.name
		return''
	@property
	def is_string(self):'\n        Checks whether a Literal expression is a string.\n        ';return isinstance(self,Literal)and self.args[_r]
	@property
	def is_number(self):'\n        Checks whether a Literal expression is a number.\n        ';A=self;return isinstance(A,Literal)and not A.args[_r]or isinstance(A,Neg)and A.this.is_number
	def to_py(A):'\n        Returns a Python object equivalent of the SQL node.\n        ';raise ValueError(f"{A} cannot be converted to a Python object.")
	@property
	def is_int(self):'\n        Checks whether an expression is an integer.\n        ';return self.is_number and isinstance(self.to_py(),int)
	@property
	def is_star(self):'Checks whether an expression is a star.';A=self;return isinstance(A,Star)or isinstance(A,Column)and isinstance(A.this,Star)
	@property
	def alias(self):
		"\n        Returns the alias of the expression, or an empty string if it's not aliased.\n        ";A=self
		if isinstance(A.args.get(_I),TableAlias):return A.args[_I].name
		return A.text(_I)
	@property
	def alias_column_names(self):
		A=self.args.get(_I)
		if not A:return[]
		return[A.name for A in A.args.get(_a)or[]]
	@property
	def name(self):return self.text(_C)
	@property
	def alias_or_name(self):return self.alias or self.name
	@property
	def output_name(self):'\n        Name of the output column if this expression is a selection.\n\n        If the Expression has no output name, an empty string is returned.\n\n        Example:\n            >>> from sqlglot import parse_one\n            >>> parse_one("SELECT a").expressions[0].output_name\n            \'a\'\n            >>> parse_one("SELECT b AS c").expressions[0].output_name\n            \'c\'\n            >>> parse_one("SELECT 1 + 2").expressions[0].output_name\n            \'\'\n        ';return''
	@property
	def type(self):return self._type
	@type.setter
	def type(self,dtype):
		A=dtype
		if A and not isinstance(A,DataType):A=DataType.build(A)
		self._type=A
	def is_type(A,*B):return A.type is not _D and A.type.is_type(*B)
	def is_leaf(A):return not any(isinstance(A,(Expression,list))for A in A.args.values())
	@property
	def meta(self):
		A=self
		if A._meta is _D:A._meta={}
		return A._meta
	def __deepcopy__(G,memo):
		H=G.__class__();C=[(G,H)]
		while C:
			A,B=C.pop()
			if A.comments is not _D:B.comments=deepcopy(A.comments)
			if A._type is not _D:B._type=deepcopy(A._type)
			if A._meta is not _D:B._meta=deepcopy(A._meta)
			if A._hash is not _D:B._hash=A._hash
			for(E,D)in A.args.items():
				if hasattr(D,_b):C.append((D,D.__class__()));B.set(E,C[-1][-1])
				elif type(D)is list:
					B.args[E]=[]
					for F in D:
						if hasattr(F,_b):C.append((F,F.__class__()));B.append(E,C[-1][-1])
						else:B.append(E,F)
				else:B.args[E]=D
		return H
	def copy(A):'\n        Returns a deep copy of the expression.\n        ';return deepcopy(A)
	def add_comments(A,comments=_D):
		B=comments
		if A.comments is _D:A.comments=[]
		if B:
			for C in B:
				I,*D=C.split(SQLGLOT_META)
				if D:
					for F in''.join(D).split(','):G,*E=F.split('=');H=E[0].strip()if E else _B;A.meta[G.strip()]=H
				A.comments.append(C)
	def pop_comments(A):B=A.comments or[];A.comments=_D;return B
	def append(A,arg_key,value):
		"\n        Appends value to arg_key if it's a list or sets it as a new list.\n\n        Args:\n            arg_key (str): name of the list expression arg\n            value (Any): value to append to the list\n        ";C=value;B=arg_key
		if type(A.args.get(B))is not list:A.args[B]=[]
		A._set_parent(B,C);D=A.args[B]
		if hasattr(C,_b):C.index=len(D)
		D.append(C)
	def set(D,arg_key,value,index=_D,overwrite=_B):
		'\n        Sets arg_key to value.\n\n        Args:\n            arg_key: name of the expression arg.\n            value: value to set the arg to.\n            index: if the arg is a list, this specifies what position to add the value in it.\n            overwrite: assuming an index is given, this determines whether to overwrite the\n                list entry instead of only inserting a new value (i.e., like list.insert).\n        ';E=arg_key;B=value;A=index
		if A is not _D:
			C=D.args.get(E)or[]
			if seq_get(C,A)is _D:return
			if B is _D:
				C.pop(A)
				for F in C[A:]:F.index=F.index-1
				return
			if isinstance(B,list):C.pop(A);C[A:A]=B
			elif overwrite:C[A]=B
			else:C.insert(A,B)
			B=C
		elif B is _D:D.args.pop(E,_D);return
		D.args[E]=B;D._set_parent(E,B,A)
	def _set_parent(D,arg_key,value,index=_D):
		E=arg_key;C=index;A=value
		if hasattr(A,_b):A.parent=D;A.arg_key=E;A.index=C
		elif type(A)is list:
			for(C,B)in enumerate(A):
				if hasattr(B,_b):B.parent=D;B.arg_key=E;B.index=C
	@property
	def depth(self):
		'\n        Returns the depth of this tree.\n        '
		if self.parent:return self.parent.depth+1
		return 0
	def iter_expressions(B,reverse=_A):
		'Yields the key and expression for all arguments, exploding list args.';C=reverse
		for A in reversed(tuple(B.args.values()))if C else B.args.values():
			if type(A)is list:
				for D in reversed(A)if C else A:
					if hasattr(D,_b):yield D
			elif hasattr(A,_b):yield A
	def find(A,*B,bfs=_B):'\n        Returns the first node in this tree which matches at least one of\n        the specified types.\n\n        Args:\n            expression_types: the expression type(s) to match.\n            bfs: whether to search the AST using the BFS algorithm (DFS is used if false).\n\n        Returns:\n            The node which matches the criteria or None if no such node was found.\n        ';return next(A.find_all(*B,bfs=bfs),_D)
	def find_all(B,*C,bfs=_B):
		'\n        Returns a generator object which visits all nodes in this tree and only\n        yields those that match at least one of the specified expression types.\n\n        Args:\n            expression_types: the expression type(s) to match.\n            bfs: whether to search the AST using the BFS algorithm (DFS is used if false).\n\n        Returns:\n            The generator object.\n        '
		for A in B.walk(bfs=bfs):
			if isinstance(A,C):yield A
	def find_ancestor(B,*C):
		'\n        Returns a nearest parent matching expression_types.\n\n        Args:\n            expression_types: the expression type(s) to match.\n\n        Returns:\n            The parent node.\n        ';A=B.parent
		while A and not isinstance(A,C):A=A.parent
		return A
	@property
	def parent_select(self):'\n        Returns the parent select statement.\n        ';return self.find_ancestor(Select)
	@property
	def same_parent(self):'Returns if the parent is the same class as itself.';return type(self.parent)is self.__class__
	def root(B):
		'\n        Returns the root expression of this tree.\n        ';A=B
		while A.parent:A=A.parent
		return A
	def walk(A,bfs=_B,prune=_D):
		'\n        Returns a generator object which visits all nodes in this tree.\n\n        Args:\n            bfs: if set to True the BFS traversal order will be applied,\n                otherwise the DFS traversal will be used instead.\n            prune: callable that returns True if the generator should stop traversing\n                this branch of the tree.\n\n        Returns:\n            the generator object.\n        ';B=prune
		if bfs:yield from A.bfs(prune=B)
		else:yield from A.dfs(prune=B)
	def dfs(D,prune=_D):
		'\n        Returns a generator object which visits all nodes in this tree in\n        the DFS (Depth-first) order.\n\n        Returns:\n            The generator object.\n        ';C=prune;A=[D]
		while A:
			B=A.pop();yield B
			if C and C(B):continue
			for E in B.iter_expressions(reverse=_B):A.append(E)
	def bfs(D,prune=_D):
		'\n        Returns a generator object which visits all nodes in this tree in\n        the BFS (Breadth-first) order.\n\n        Returns:\n            The generator object.\n        ';C=prune;A=deque([D])
		while A:
			B=A.popleft();yield B
			if C and C(B):continue
			for E in B.iter_expressions():A.append(E)
	def unnest(B):
		'\n        Returns the first non parenthesis child or self.\n        ';A=B
		while type(A)is Paren:A=A.this
		return A
	def unalias(A):
		'\n        Returns the inner expression if this is an Alias.\n        '
		if isinstance(A,Alias):return A.this
		return A
	def unnest_operands(A):'\n        Returns unnested operands as a tuple.\n        ';return tuple(A.unnest()for A in A.iter_expressions())
	def flatten(B,unnest=_B):
		'\n        Returns a generator which yields child nodes whose parents are the same class.\n\n        A AND B AND C -> [A, B, C]\n        '
		for A in B.dfs(prune=lambda n:n.parent and type(n)is not B.__class__):
			if type(A)is not B.__class__:yield A.unnest()if unnest and not isinstance(A,Subquery)else A
	def __str__(A):return A.sql()
	def __repr__(A):return _to_s(A)
	def to_s(A):"\n        Same as __repr__, but includes additional information which can be useful\n        for debugging, like empty or missing args and the AST nodes' object IDs.\n        ";return _to_s(A,verbose=_B)
	def sql(A,dialect=_D,**B):'\n        Returns SQL string representation of this tree.\n\n        Args:\n            dialect: the dialect of the output SQL string (eg. "spark", "hive", "presto", "mysql").\n            opts: other `sqlglot.generator.Generator` options.\n\n        Returns:\n            The SQL string.\n        ';from sqlglot.dialects import Dialect as C;return C.get_or_raise(dialect).generate(A,**B)
	def transform(D,fun,*E,copy=_B,**F):
		'\n        Visits all tree nodes (excluding already transformed ones)\n        and applies the given transformation function to each node.\n\n        Args:\n            fun: a function which takes a node as an argument and returns a\n                new transformed node or the same node without modifications. If the function\n                returns None, then the corresponding node will be removed from the syntax tree.\n            copy: if set to True a new tree instance is constructed, otherwise the tree is\n                modified in place.\n\n        Returns:\n            The transformed tree.\n        ';C=_D;A=_D
		for B in(D.copy()if copy else D).dfs(prune=lambda n:n is not A):
			G,H,I=B.parent,B.arg_key,B.index;A=fun(B,*E,**F)
			if not C:C=A
			elif A is not B:G.set(H,A,I)
		assert C;return C.assert_is(Expression)
	@t.overload
	def replace(self,expression):...
	@t.overload
	def replace(self,expression):...
	def replace(A,expression):
		'\n        Swap out this expression with a new expression.\n\n        For example::\n\n            >>> tree = Select().select("x").from_("tbl")\n            >>> tree.find(Column).replace(column("y"))\n            Column(\n              this=Identifier(this=y, quoted=False))\n            >>> tree.sql()\n            \'SELECT y FROM tbl\'\n\n        Args:\n            expression: new node\n\n        Returns:\n            The new expression or expressions.\n        ';B=expression;C=A.parent
		if not C or C is B:return B
		D=A.arg_key;E=C.args.get(D)
		if type(B)is list and isinstance(E,Expression):E.parent.replace(B)
		else:C.set(D,B,A.index)
		if B is not A:A.parent=_D;A.arg_key=_D;A.index=_D
		return B
	def pop(A):'\n        Remove this expression from its AST.\n\n        Returns:\n            The popped expression.\n        ';A.replace(_D);return A
	def assert_is(A,type_):
		'\n        Assert that this `Expression` is an instance of `type_`.\n\n        If it is NOT an instance of `type_`, this raises an assertion error.\n        Otherwise, this returns this expression.\n\n        Examples:\n            This is useful for type security in chained expressions:\n\n            >>> import sqlglot\n            >>> sqlglot.parse_one("SELECT x from y").assert_is(Select).select("z").sql()\n            \'SELECT x, z FROM y\'\n        ';B=type_
		if not isinstance(A,B):raise AssertionError(f"{A} is not {B}.")
		return A
	def error_messages(A,args=_D):
		"\n        Checks if this expression is valid (e.g. all mandatory args are set).\n\n        Args:\n            args: a sequence of values that were used to instantiate a Func expression. This is used\n                to check that the provided arguments don't exceed the function argument limit.\n\n        Returns:\n            A list of error messages for all possible errors that were found.\n        ";D=args;C=[]
		for B in A.args:
			if B not in A.arg_types:C.append(f"Unexpected keyword: '{B}' for {A.__class__}")
		for(B,F)in A.arg_types.items():
			E=A.args.get(B)
			if F and(E is _D or isinstance(E,list)and not E):C.append(f"Required keyword: '{B}' missing for {A.__class__}")
		if D and isinstance(A,Func)and len(D)>len(A.arg_types)and not A.is_var_len_args:C.append(f"The number of provided arguments ({len(D)}) is greater than the maximum number of supported arguments ({len(A.arg_types)})")
		return C
	def dump(A):'\n        Dump this Expression to a JSON-serializable dict.\n        ';from sqlglot.serde import dump;return dump(A)
	@classmethod
	def load(A,obj):'\n        Load a dict (as returned by `Expression.dump`) into an Expression instance.\n        ';from sqlglot.serde import load;return load(obj)
	def and_(A,*B,dialect=_D,copy=_B,**C):'\n        AND this condition with one or multiple expressions.\n\n        Example:\n            >>> condition("x=1").and_("y=1").sql()\n            \'x = 1 AND y = 1\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            dialect: the dialect used to parse the input expression.\n            copy: whether to copy the involved expressions (only applies to Expressions).\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The new And condition.\n        ';return and_(A,*B,dialect=dialect,copy=copy,**C)
	def or_(A,*B,dialect=_D,copy=_B,**C):'\n        OR this condition with one or multiple expressions.\n\n        Example:\n            >>> condition("x=1").or_("y=1").sql()\n            \'x = 1 OR y = 1\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            dialect: the dialect used to parse the input expression.\n            copy: whether to copy the involved expressions (only applies to Expressions).\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The new Or condition.\n        ';return or_(A,*B,dialect=dialect,copy=copy,**C)
	def not_(A,copy=_B):'\n        Wrap this condition with NOT.\n\n        Example:\n            >>> condition("x=1").not_().sql()\n            \'NOT x = 1\'\n\n        Args:\n            copy: whether to copy this object.\n\n        Returns:\n            The new Not instance.\n        ';return not_(A,copy=copy)
	def as_(A,alias,quoted=_D,dialect=_D,copy=_B,**B):return alias_(A,alias,quoted=quoted,dialect=dialect,copy=copy,**B)
	def _binop(D,klass,other,reverse=_A):
		C=klass;A=other;B=D.copy();A=convert(A,copy=_B)
		if not isinstance(B,C)and not isinstance(A,C):B=_wrap(B,Binary);A=_wrap(A,Binary)
		if reverse:return C(this=A,expression=B)
		return C(this=B,expression=A)
	def __getitem__(A,other):return Bracket(this=A.copy(),expressions=[convert(A,copy=_B)for A in ensure_list(other)])
	def __iter__(A):
		if _E in A.arg_types:return iter(A.args.get(_E)or[])
		raise TypeError(f"'{A.__class__.__name__}' object is not iterable")
	def isin(F,*G,query=_D,unnest=_D,copy=_B,**E):
		D=unnest;C=query;B=copy;A=maybe_parse(C,copy=B,**E)if C else _D
		if A and not isinstance(A,Subquery):A=A.subquery(copy=_A)
		return In(this=maybe_copy(F,B),expressions=[convert(A,copy=B)for A in G],query=A,unnest=Unnest(expressions=[maybe_parse(t.cast(ExpOrStr,A),copy=B,**E)for A in ensure_list(D)])if D else _D)
	def between(C,low,high,copy=_B,**B):A=copy;return Between(this=maybe_copy(C,A),low=convert(low,copy=A,**B),high=convert(high,copy=A,**B))
	def is_(A,other):return A._binop(Is,other)
	def like(A,other):return A._binop(Like,other)
	def ilike(A,other):return A._binop(ILike,other)
	def eq(A,other):return A._binop(EQ,other)
	def neq(A,other):return A._binop(NEQ,other)
	def rlike(A,other):return A._binop(RegexpLike,other)
	def div(B,other,typed=_A,safe=_A):A=B._binop(Div,other);A.args['typed']=typed;A.args[_Q]=safe;return A
	def asc(A,nulls_first=_B):return Ordered(this=A.copy(),nulls_first=nulls_first)
	def desc(A,nulls_first=_A):return Ordered(this=A.copy(),desc=_B,nulls_first=nulls_first)
	def __lt__(A,other):return A._binop(LT,other)
	def __le__(A,other):return A._binop(LTE,other)
	def __gt__(A,other):return A._binop(GT,other)
	def __ge__(A,other):return A._binop(GTE,other)
	def __add__(A,other):return A._binop(Add,other)
	def __radd__(A,other):return A._binop(Add,other,reverse=_B)
	def __sub__(A,other):return A._binop(Sub,other)
	def __rsub__(A,other):return A._binop(Sub,other,reverse=_B)
	def __mul__(A,other):return A._binop(Mul,other)
	def __rmul__(A,other):return A._binop(Mul,other,reverse=_B)
	def __truediv__(A,other):return A._binop(Div,other)
	def __rtruediv__(A,other):return A._binop(Div,other,reverse=_B)
	def __floordiv__(A,other):return A._binop(IntDiv,other)
	def __rfloordiv__(A,other):return A._binop(IntDiv,other,reverse=_B)
	def __mod__(A,other):return A._binop(Mod,other)
	def __rmod__(A,other):return A._binop(Mod,other,reverse=_B)
	def __pow__(A,other):return A._binop(Pow,other)
	def __rpow__(A,other):return A._binop(Pow,other,reverse=_B)
	def __and__(A,other):return A._binop(And,other)
	def __rand__(A,other):return A._binop(And,other,reverse=_B)
	def __or__(A,other):return A._binop(Or,other)
	def __ror__(A,other):return A._binop(Or,other,reverse=_B)
	def __neg__(A):return Neg(this=_wrap(A.copy(),Binary))
	def __invert__(A):return not_(A.copy())
IntoType=t.Union[str,t.Type[Expression],t.Collection[t.Union[str,t.Type[Expression]]]]
ExpOrStr=t.Union[str,Expression]
class Condition(Expression):'Logical conditions like x AND y, or simply x'
class Predicate(Condition):'Relationships like x = y, x > 1, x >= y.'
class DerivedTable(Expression):
	@property
	def selects(self):return self.this.selects if isinstance(self.this,Query)else[]
	@property
	def named_selects(self):return[A.output_name for A in self.selects]
class Query(Expression):
	def subquery(B,alias=_D,copy=_B):
		'\n        Returns a `Subquery` that wraps around this query.\n\n        Example:\n            >>> subquery = Select().select("x").from_("tbl").subquery()\n            >>> Select().select("x").from_(subquery).sql()\n            \'SELECT x FROM (SELECT x FROM tbl)\'\n\n        Args:\n            alias: an optional alias for the subquery.\n            copy: if `False`, modify this expression instance in-place.\n        ';A=alias;C=maybe_copy(B,copy)
		if not isinstance(A,Expression):A=TableAlias(this=to_identifier(A))if A else _D
		return Subquery(this=C,alias=A)
	def limit(A,expression,dialect=_D,copy=_B,**B):'\n        Adds a LIMIT clause to this query.\n\n        Example:\n            >>> select("1").union(select("1")).limit(1).sql()\n            \'SELECT 1 UNION SELECT 1 LIMIT 1\'\n\n        Args:\n            expression: the SQL code string to parse.\n                This can also be an integer.\n                If a `Limit` instance is passed, it will be used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `Limit`.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            A limited Select expression.\n        ';return _apply_builder(expression=expression,instance=A,arg=_c,into=Limit,prefix='LIMIT',dialect=dialect,copy=copy,into_arg=_F,**B)
	def offset(A,expression,dialect=_D,copy=_B,**B):'\n        Set the OFFSET expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x").offset(10).sql()\n            \'SELECT x FROM tbl OFFSET 10\'\n\n        Args:\n            expression: the SQL code string to parse.\n                This can also be an integer.\n                If a `Offset` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `Offset`.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_builder(expression=expression,instance=A,arg=_N,into=Offset,prefix='OFFSET',dialect=dialect,copy=copy,into_arg=_F,**B)
	def order_by(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Set the ORDER BY expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x").order_by("x DESC").sql()\n            \'SELECT x FROM tbl ORDER BY x DESC\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If a `Group` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `Order`.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this flattens all the `Order` expression into a single expression.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_child_list_builder(*B,instance=A,arg=_W,append=append,copy=copy,prefix='ORDER BY',into=Order,dialect=dialect,**C)
	@property
	def ctes(self):'Returns a list of all the CTEs attached to this query.';A=self.args.get(_J);return A.expressions if A else[]
	@property
	def selects(self):"Returns the query's projections.";raise NotImplementedError('Query objects must implement `selects`')
	@property
	def named_selects(self):"Returns the output names of the query's projections.";raise NotImplementedError('Query objects must implement `named_selects`')
	def select(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the SELECT expressions.\n\n        Example:\n            >>> Select().select("x", "y").sql()\n            \'SELECT x, y\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this resets the expressions.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Query expression.\n        ';raise NotImplementedError('Query objects must implement `select`')
	def with_(A,alias,as_,recursive=_D,materialized=_D,append=_B,dialect=_D,copy=_B,**B):'\n        Append to or set the common table expressions.\n\n        Example:\n            >>> Select().with_("tbl2", as_="SELECT * FROM tbl").select("x").from_("tbl2").sql()\n            \'WITH tbl2 AS (SELECT * FROM tbl) SELECT x FROM tbl2\'\n\n        Args:\n            alias: the SQL code string to parse as the table name.\n                If an `Expression` instance is passed, this is used as-is.\n            as_: the SQL code string to parse as the table expression.\n                If an `Expression` instance is passed, it will be used as-is.\n            recursive: set the RECURSIVE part of the expression. Defaults to `False`.\n            materialized: set the MATERIALIZED part of the expression.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this resets the expressions.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified expression.\n        ';return _apply_cte_builder(A,alias,as_,recursive=recursive,materialized=materialized,append=append,dialect=dialect,copy=copy,**B)
	def union(A,expression,distinct=_B,dialect=_D,**B):'\n        Builds a UNION expression.\n\n        Example:\n            >>> import sqlglot\n            >>> sqlglot.parse_one("SELECT * FROM foo").union("SELECT * FROM bla").sql()\n            \'SELECT * FROM foo UNION SELECT * FROM bla\'\n\n        Args:\n            expression: the SQL code string.\n                If an `Expression` instance is passed, it will be used as-is.\n            distinct: set the DISTINCT flag if and only if this is true.\n            dialect: the dialect used to parse the input expression.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The new Union expression.\n        ';return union(left=A,right=expression,distinct=distinct,dialect=dialect,**B)
	def intersect(A,expression,distinct=_B,dialect=_D,**B):'\n        Builds an INTERSECT expression.\n\n        Example:\n            >>> import sqlglot\n            >>> sqlglot.parse_one("SELECT * FROM foo").intersect("SELECT * FROM bla").sql()\n            \'SELECT * FROM foo INTERSECT SELECT * FROM bla\'\n\n        Args:\n            expression: the SQL code string.\n                If an `Expression` instance is passed, it will be used as-is.\n            distinct: set the DISTINCT flag if and only if this is true.\n            dialect: the dialect used to parse the input expression.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The new Intersect expression.\n        ';return intersect(left=A,right=expression,distinct=distinct,dialect=dialect,**B)
	def except_(A,expression,distinct=_B,dialect=_D,**B):'\n        Builds an EXCEPT expression.\n\n        Example:\n            >>> import sqlglot\n            >>> sqlglot.parse_one("SELECT * FROM foo").except_("SELECT * FROM bla").sql()\n            \'SELECT * FROM foo EXCEPT SELECT * FROM bla\'\n\n        Args:\n            expression: the SQL code string.\n                If an `Expression` instance is passed, it will be used as-is.\n            distinct: set the DISTINCT flag if and only if this is true.\n            dialect: the dialect used to parse the input expression.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The new Except expression.\n        ';return except_(left=A,right=expression,distinct=distinct,dialect=dialect,**B)
class UDTF(DerivedTable):
	@property
	def selects(self):A=self.args.get(_I);return A.columns if A else[]
class Cache(Expression):arg_types={_C:_B,'lazy':_A,_X:_A,_F:_A}
class Uncache(Expression):arg_types={_C:_B,_R:_A}
class Refresh(Expression):0
class DDL(Expression):
	@property
	def ctes(self):'Returns a list of all the CTEs attached to this statement.';A=self.args.get(_J);return A.expressions if A else[]
	@property
	def selects(self):"If this statement contains a query (e.g. a CTAS), this returns the query's projections.";return self.expression.selects if isinstance(self.expression,Query)else[]
	@property
	def named_selects(self):"\n        If this statement contains a query (e.g. a CTAS), this returns the output\n        names of the query's projections.\n        ";return self.expression.named_selects if isinstance(self.expression,Query)else[]
class DML(Expression):
	def returning(A,expression,dialect=_D,copy=_B,**B):'\n        Set the RETURNING expression. Not supported by all dialects.\n\n        Example:\n            >>> delete("tbl").returning("*", dialect="postgres").sql()\n            \'DELETE FROM tbl RETURNING *\'\n\n        Args:\n            expression: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            Delete: the modified expression.\n        ';return _apply_builder(expression=expression,instance=A,arg=_f,prefix='RETURNING',dialect=dialect,copy=copy,into=Returning,**B)
class Create(DDL):
	arg_types={_J:_A,_C:_B,_G:_B,_F:_A,_R:_A,'properties':_A,_z:_A,'refresh':_A,_A0:_A,'indexes':_A,'no_schema_binding':_A,'begin':_A,_d:_A,'clone':_A,_AM:_A,'clustered':_A}
	@property
	def kind(self):A=self.args.get(_G);return A and A.upper()
class SequenceProperties(Expression):arg_types={_AN:_A,'minvalue':_A,'maxvalue':_A,'cache':_A,_O:_A,'owned':_A,_X:_A}
class TruncateTable(Expression):arg_types={_E:_B,'is_database':_A,_R:_A,'only':_A,_g:_A,'identity':_A,_A1:_A,_k:_A}
class Clone(Expression):arg_types={_C:_B,'shallow':_A,'copy':_A}
class Describe(Expression):arg_types={_C:_B,'style':_A,_G:_A,_E:_A,_k:_A}
class Summarize(Expression):arg_types={_C:_B,_Z:_A}
class Kill(Expression):arg_types={_C:_B,_G:_A}
class Pragma(Expression):0
class Declare(Expression):arg_types={_E:_B}
class DeclareItem(Expression):arg_types={_C:_B,_G:_B,_M:_A}
class Set(Expression):arg_types={_E:_A,'unset':_A,'tag':_A}
class Heredoc(Expression):arg_types={_C:_B,'tag':_A}
class SetItem(Expression):arg_types={_C:_A,_E:_A,_G:_A,'collate':_A,_s:_A}
class Show(Expression):arg_types={_C:_B,'history':_A,'terse':_A,_A2:_A,_N:_A,'starts_with':_A,_c:_A,_Y:_A,'like':_A,_P:_A,_S:_A,'scope':_A,'scope_kind':_A,'full':_A,'mutex':_A,'query':_A,'channel':_A,_s:_A,'log':_A,_e:_A,'types':_A}
class UserDefinedFunction(Expression):arg_types={_C:_B,_E:_A,'wrapped':_A}
class CharacterSet(Expression):arg_types={_C:_B,_M:_A}
class With(Expression):
	arg_types={_E:_B,_A3:_A}
	@property
	def recursive(self):return bool(self.args.get(_A3))
class WithinGroup(Expression):arg_types={_C:_B,_F:_A}
class CTE(DerivedTable):arg_types={_C:_B,_I:_B,'scalar':_A,_A4:_A}
class ProjectionDef(Expression):arg_types={_C:_B,_F:_B}
class TableAlias(Expression):
	arg_types={_C:_A,_a:_A}
	@property
	def columns(self):return self.args.get(_a)or[]
class BitString(Condition):0
class HexString(Condition):0
class ByteString(Condition):0
class RawString(Condition):0
class UnicodeString(Condition):arg_types={_C:_B,'escape':_A}
class Column(Condition):
	arg_types={_C:_B,_Z:_A,_S:_A,_V:_A,'join_mark':_A}
	@property
	def table(self):return self.text(_Z)
	@property
	def db(self):return self.text(_S)
	@property
	def catalog(self):return self.text(_V)
	@property
	def output_name(self):return self.name
	@property
	def parts(self):'Return the parts of a column in order catalog, db, table, name.';return[t.cast(Identifier,self.args[A])for A in(_V,_S,_Z,_C)if self.args.get(A)]
	def to_dot(C):
		'Converts the column into a dot expression.';B=C.parts;A=C.parent
		while A:
			if isinstance(A,Dot):B.append(A.expression)
			A=A.parent
		return Dot.build(deepcopy(B))if len(B)>1 else B[0]
class ColumnPosition(Expression):arg_types={_C:_A,_e:_B}
class ColumnDef(Expression):
	arg_types={_C:_B,_G:_A,_A5:_A,_R:_A,_e:_A}
	@property
	def constraints(self):return self.args.get(_A5)or[]
	@property
	def kind(self):return self.args.get(_G)
class AlterColumn(Expression):arg_types={_C:_B,'dtype':_A,'collate':_A,_T:_A,_M:_A,'drop':_A,'comment':_A,_AO:_A}
class AlterDistStyle(Expression):0
class AlterSortKey(Expression):arg_types={_C:_A,_E:_A,'compound':_A}
class AlterSet(Expression):arg_types={_E:_A,_A1:_A,_AP:_A,'access_method':_A,'file_format':_A,'copy_options':_A,'tag':_A,'location':_A,_A6:_A}
class RenameColumn(Expression):arg_types={_C:_B,_l:_B,_R:_A}
class RenameTable(Expression):0
class SwapTable(Expression):0
class Comment(Expression):arg_types={_C:_B,_G:_B,_F:_B,_R:_A,_A4:_A}
class Comprehension(Expression):arg_types={_C:_B,_F:_B,'iterator':_B,_AQ:_A}
class MergeTreeTTLAction(Expression):arg_types={_C:_B,_A7:_A,'recompress':_A,'to_disk':_A,'to_volume':_A}
class MergeTreeTTL(Expression):arg_types={_E:_B,_P:_A,_m:_A,'aggregates':_A}
class IndexConstraintOption(Expression):arg_types={'key_block_size':_A,_T:_A,'parser':_A,'comment':_A,'visible':_A,'engine_attr':_A,'secondary_engine_attr':_A}
class ColumnConstraint(Expression):
	arg_types={_C:_A,_G:_B}
	@property
	def kind(self):return self.args[_G]
class ColumnConstraintKind(Expression):0
class AutoIncrementColumnConstraint(ColumnConstraintKind):0
class PeriodForSystemTimeConstraint(ColumnConstraintKind):arg_types={_C:_B,_F:_B}
class CaseSpecificColumnConstraint(ColumnConstraintKind):arg_types={'not_':_B}
class CharacterSetColumnConstraint(ColumnConstraintKind):arg_types={_C:_B}
class CheckColumnConstraint(ColumnConstraintKind):arg_types={_C:_B,'enforced':_A}
class ClusteredColumnConstraint(ColumnConstraintKind):0
class CollateColumnConstraint(ColumnConstraintKind):0
class CommentColumnConstraint(ColumnConstraintKind):0
class CompressColumnConstraint(ColumnConstraintKind):arg_types={_C:_A}
class DateFormatColumnConstraint(ColumnConstraintKind):arg_types={_C:_B}
class DefaultColumnConstraint(ColumnConstraintKind):0
class EncodeColumnConstraint(ColumnConstraintKind):0
class ExcludeColumnConstraint(ColumnConstraintKind):0
class EphemeralColumnConstraint(ColumnConstraintKind):arg_types={_C:_A}
class WithOperator(Expression):arg_types={_C:_B,'op':_B}
class GeneratedAsIdentityColumnConstraint(ColumnConstraintKind):arg_types={_C:_A,_F:_A,'on_null':_A,_O:_A,_AN:_A,'minvalue':_A,'maxvalue':_A,'cycle':_A}
class GeneratedAsRowColumnConstraint(ColumnConstraintKind):arg_types={_O:_A,'hidden':_A}
class IndexColumnConstraint(ColumnConstraintKind):arg_types={_C:_A,_E:_A,_G:_A,_AR:_A,_X:_A,_F:_A,'granularity':_A}
class InlineLengthColumnConstraint(ColumnConstraintKind):0
class NonClusteredColumnConstraint(ColumnConstraintKind):0
class NotForReplicationColumnConstraint(ColumnConstraintKind):arg_types={}
class MaskingPolicyColumnConstraint(ColumnConstraintKind):arg_types={_C:_B,_E:_A}
class NotNullColumnConstraint(ColumnConstraintKind):arg_types={_AO:_A}
class OnUpdateColumnConstraint(ColumnConstraintKind):0
class TagColumnConstraint(ColumnConstraintKind):arg_types={_E:_B}
class TransformColumnConstraint(ColumnConstraintKind):0
class PrimaryKeyColumnConstraint(ColumnConstraintKind):arg_types={'desc':_A}
class TitleColumnConstraint(ColumnConstraintKind):0
class UniqueColumnConstraint(ColumnConstraintKind):arg_types={_C:_A,_AR:_A,'on_conflict':_A,'nulls':_A}
class UppercaseColumnConstraint(ColumnConstraintKind):arg_types:t.Dict[str,t.Any]={}
class PathColumnConstraint(ColumnConstraintKind):0
class ProjectionPolicyColumnConstraint(ColumnConstraintKind):0
class ComputedColumnConstraint(ColumnConstraintKind):arg_types={_C:_B,'persisted':_A,'not_null':_A}
class Constraint(Expression):arg_types={_C:_B,_E:_B}
class Delete(DML):
	arg_types={_J:_A,_C:_A,_T:_A,_P:_A,_f:_A,_c:_A,'tables':_A,_g:_A}
	def delete(A,table,dialect=_D,copy=_B,**B):'\n        Create a DELETE expression or replace the table on an existing DELETE expression.\n\n        Example:\n            >>> delete("tbl").sql()\n            \'DELETE FROM tbl\'\n\n        Args:\n            table: the table from which to delete.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            Delete: the modified expression.\n        ';return _apply_builder(expression=table,instance=A,arg=_C,dialect=dialect,into=Table,copy=copy,**B)
	def where(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the WHERE expressions.\n\n        Example:\n            >>> delete("tbl").where("x = \'a\' OR x < \'b\'").sql()\n            "DELETE FROM tbl WHERE x = \'a\' OR x < \'b\'"\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n                Multiple expressions are combined with an AND operator.\n            append: if `True`, AND the new expressions to any existing expression.\n                Otherwise, this resets the expression.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            Delete: the modified expression.\n        ';return _apply_conjunction_builder(*B,instance=A,arg=_P,append=append,into=Where,dialect=dialect,copy=copy,**C)
class Drop(Expression):
	arg_types={_C:_A,_G:_A,_E:_A,_R:_A,_A8:_A,_A4:_A,'cascade':_A,_A5:_A,'purge':_A,_g:_A,_AM:_A}
	@property
	def kind(self):A=self.args.get(_G);return A and A.upper()
class Filter(Expression):arg_types={_C:_B,_F:_B}
class Check(Expression):0
class Changes(Expression):arg_types={'information':_B,'at_before':_A,_d:_A}
class Connect(Expression):arg_types={_O:_A,'connect':_B,'nocycle':_A}
class CopyParameter(Expression):arg_types={_C:_B,_F:_A,_E:_A}
class Copy(DML):arg_types={_C:_B,_G:_B,'files':_B,_AS:_A,_K:_A,_t:_A}
class Credentials(Expression):arg_types={_AS:_A,'encryption':_A,'storage':_A,'iam_role':_A,'region':_A}
class Prior(Expression):0
class Directory(Expression):arg_types={_C:_B,_A9:_A,'row_format':_A}
class ForeignKey(Expression):arg_types={_E:_B,'reference':_A,_A7:_A,'update':_A}
class ColumnPrefix(Expression):arg_types={_C:_B,_F:_B}
class PrimaryKey(Expression):arg_types={_E:_B,_X:_A}
class Into(Expression):arg_types={_C:_A,_A8:_A,'unlogged':_A,'bulk_collect':_A,_E:_A}
class From(Expression):
	@property
	def name(self):return self.this.name
	@property
	def alias_or_name(self):return self.this.alias_or_name
class Having(Expression):0
class Hint(Expression):arg_types={_E:_B}
class JoinHint(Expression):arg_types={_C:_B,_E:_B}
class Identifier(Expression):
	arg_types={_C:_B,_AA:_A,_s:_A,_A8:_A}
	@property
	def quoted(self):return bool(self.args.get(_AA))
	@property
	def hashable_args(self):return self.this,self.quoted
	@property
	def output_name(self):return self.name
class Opclass(Expression):arg_types={_C:_B,_F:_B}
class Index(Expression):arg_types={_C:_A,_Z:_A,_A0:_A,'primary':_A,'amp':_A,_t:_A}
class IndexParameters(Expression):arg_types={_T:_A,'include':_A,_a:_A,'with_storage':_A,_AB:_A,_AP:_A,_P:_A,_U:_A}
class Insert(DDL,DML):
	arg_types={_n:_A,_J:_A,'is_function':_A,_C:_A,_F:_A,'conflict':_A,_f:_A,_AT:_A,_R:_A,'alternative':_A,_P:_A,'ignore':_A,'by_name':_A,'stored':_A,_k:_A,_AC:_A,_u:_A}
	def with_(A,alias,as_,recursive=_D,materialized=_D,append=_B,dialect=_D,copy=_B,**B):'\n        Append to or set the common table expressions.\n\n        Example:\n            >>> insert("SELECT x FROM cte", "t").with_("cte", as_="SELECT * FROM tbl").sql()\n            \'WITH cte AS (SELECT * FROM tbl) INSERT INTO t SELECT x FROM cte\'\n\n        Args:\n            alias: the SQL code string to parse as the table name.\n                If an `Expression` instance is passed, this is used as-is.\n            as_: the SQL code string to parse as the table expression.\n                If an `Expression` instance is passed, it will be used as-is.\n            recursive: set the RECURSIVE part of the expression. Defaults to `False`.\n            materialized: set the MATERIALIZED part of the expression.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this resets the expressions.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified expression.\n        ';return _apply_cte_builder(A,alias,as_,recursive=recursive,materialized=materialized,append=append,dialect=dialect,copy=copy,**B)
class ConditionalInsert(Expression):arg_types={_C:_B,_F:_A,'else_':_A}
class MultitableInserts(Expression):arg_types={_E:_B,_G:_B,_u:_B}
class OnConflict(Expression):arg_types={'duplicate':_A,_E:_A,'action':_A,'conflict_keys':_A,'constraint':_A}
class OnCondition(Expression):arg_types={'error':_A,'empty':_A,_o:_A}
class Returning(Expression):arg_types={_E:_B,'into':_A}
class Introducer(Expression):arg_types={_C:_B,_F:_B}
class National(Expression):0
class LoadData(Expression):arg_types={_C:_B,_A9:_A,_AT:_A,'inpath':_B,_k:_A,_AU:_A,_A6:_A}
class Partition(Expression):arg_types={_E:_B}
class PartitionRange(Expression):arg_types={_C:_B,_F:_B}
class PartitionId(Expression):0
class Fetch(Expression):arg_types={'direction':_A,_AD:_A,_v:_A,'with_ties':_A}
class Grant(Expression):arg_types={'privileges':_B,_G:_A,'securable':_B,'principals':_B,'grant_option':_A}
class Group(Expression):arg_types={_E:_A,'grouping_sets':_A,'cube':_A,'rollup':_A,'totals':_A,'all':_A}
class Cube(Expression):arg_types={_E:_A}
class Rollup(Expression):arg_types={_E:_A}
class GroupingSets(Expression):arg_types={_E:_B}
class Lambda(Expression):arg_types={_C:_B,_E:_B}
class Limit(Expression):arg_types={_C:_A,_F:_B,_N:_A,_E:_A}
class Literal(Condition):
	arg_types={_C:_B,_r:_B}
	@property
	def hashable_args(self):return self.this,self.args.get(_r)
	@classmethod
	def number(A,number):return A(this=str(number),is_string=_A)
	@classmethod
	def string(A,string):return A(this=str(string),is_string=_B)
	@property
	def output_name(self):return self.name
	def to_py(A):
		if A.is_number:
			try:return int(A.this)
			except ValueError:return Decimal(A.this)
		return A.this
class Join(Expression):
	arg_types={_C:_B,_U:_A,'side':_A,_G:_A,_T:_A,_w:_A,_s:_A,_n:_A,'match_condition':_A}
	@property
	def method(self):return self.text(_w).upper()
	@property
	def kind(self):return self.text(_G).upper()
	@property
	def side(self):return self.text('side').upper()
	@property
	def hint(self):return self.text(_n).upper()
	@property
	def alias_or_name(self):return self.this.alias_or_name
	def on(B,*C,append=_B,dialect=_D,copy=_B,**D):
		'\n        Append to or set the ON expressions.\n\n        Example:\n            >>> import sqlglot\n            >>> sqlglot.parse_one("JOIN x", into=Join).on("y = 1").sql()\n            \'JOIN x ON y = 1\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n                Multiple expressions are combined with an AND operator.\n            append: if `True`, AND the new expressions to any existing expression.\n                Otherwise, this resets the expression.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Join expression.\n        ';A=_apply_conjunction_builder(*C,instance=B,arg=_U,append=append,dialect=dialect,copy=copy,**D)
		if A.kind=='CROSS':A.set(_G,_D)
		return A
	def using(B,*C,append=_B,dialect=_D,copy=_B,**D):
		'\n        Append to or set the USING expressions.\n\n        Example:\n            >>> import sqlglot\n            >>> sqlglot.parse_one("JOIN x", into=Join).using("foo", "bla").sql()\n            \'JOIN x USING (foo, bla)\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            append: if `True`, concatenate the new expressions to the existing "using" list.\n                Otherwise, this resets the expression.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Join expression.\n        ';A=_apply_list_builder(*C,instance=B,arg=_T,append=append,dialect=dialect,copy=copy,**D)
		if A.kind=='CROSS':A.set(_G,_D)
		return A
class Lateral(UDTF):arg_types={_C:_B,'view':_A,'outer':_A,_I:_A,'cross_apply':_A}
class MatchRecognizeMeasure(Expression):arg_types={_C:_B,'window_frame':_A}
class MatchRecognize(Expression):arg_types={_AB:_A,_W:_A,'measures':_A,'rows':_A,'after':_A,'pattern':_A,'define':_A,_I:_A}
class Final(Expression):0
class Offset(Expression):arg_types={_C:_A,_F:_B,_E:_A}
class Order(Expression):arg_types={_C:_A,_E:_B,'siblings':_A}
class WithFill(Expression):arg_types={_Y:_A,_l:_A,_p:_A,'interpolate':_A}
class Cluster(Order):0
class Distribute(Order):0
class Sort(Order):0
class Ordered(Expression):arg_types={_C:_B,'desc':_A,'nulls_first':_B,'with_fill':_A}
class Property(Expression):arg_types={_C:_B,'value':_B}
class GrantPrivilege(Expression):arg_types={_C:_B,_E:_A}
class GrantPrincipal(Expression):arg_types={_C:_B,_G:_A}
class AllowedValuesProperty(Expression):arg_types={_E:_B}
class AlgorithmProperty(Property):arg_types={_C:_B}
class AutoIncrementProperty(Property):arg_types={_C:_B}
class AutoRefreshProperty(Property):arg_types={_C:_B}
class BackupProperty(Property):arg_types={_C:_B}
class BlockCompressionProperty(Property):arg_types={'autotemp':_A,'always':_A,_M:_A,'manual':_A,'never':_A}
class CharacterSetProperty(Property):arg_types={_C:_B,_M:_B}
class ChecksumProperty(Property):arg_types={_U:_A,_M:_A}
class CollateProperty(Property):arg_types={_C:_B,_M:_A}
class CopyGrantsProperty(Property):arg_types={}
class DataBlocksizeProperty(Property):arg_types={'size':_A,'units':_A,'minimum':_A,'maximum':_A,_M:_A}
class DataDeletionProperty(Property):arg_types={_U:_B,'filter_col':_A,_AV:_A}
class DefinerProperty(Property):arg_types={_C:_B}
class DistKeyProperty(Property):arg_types={_C:_B}
class DistributedByProperty(Property):arg_types={_E:_A,_G:_B,'buckets':_A,_W:_A}
class DistStyleProperty(Property):arg_types={_C:_B}
class DuplicateKeyProperty(Property):arg_types={_E:_B}
class EngineProperty(Property):arg_types={_C:_B}
class HeapProperty(Property):arg_types={}
class ToTableProperty(Property):arg_types={_C:_B}
class ExecuteAsProperty(Property):arg_types={_C:_B}
class ExternalProperty(Property):arg_types={_C:_A}
class FallbackProperty(Property):arg_types={_h:_B,'protection':_A}
class FileFormatProperty(Property):arg_types={_C:_B}
class FreespaceProperty(Property):arg_types={_C:_B,_v:_A}
class GlobalProperty(Property):arg_types={}
class IcebergProperty(Property):arg_types={}
class InheritsProperty(Property):arg_types={_E:_B}
class InputModelProperty(Property):arg_types={_C:_B}
class OutputModelProperty(Property):arg_types={_C:_B}
class IsolatedLoadingProperty(Property):arg_types={_h:_A,'concurrent':_A,_A2:_A}
class JournalProperty(Property):arg_types={_h:_A,'dual':_A,'before':_A,_A9:_A,'after':_A}
class LanguageProperty(Property):arg_types={_C:_B}
class ClusteredByProperty(Property):arg_types={_E:_B,'sorted_by':_A,'buckets':_B}
class DictProperty(Property):arg_types={_C:_B,_G:_B,_AC:_A}
class DictSubProperty(Property):0
class DictRange(Property):arg_types={_C:_B,'min':_B,'max':_B}
class DynamicProperty(Property):arg_types={}
class OnCluster(Property):arg_types={_C:_B}
class EmptyProperty(Property):arg_types={}
class LikeProperty(Property):arg_types={_C:_B,_E:_A}
class LocationProperty(Property):arg_types={_C:_B}
class LockProperty(Property):arg_types={_C:_B}
class LockingProperty(Property):arg_types={_C:_A,_G:_B,'for_or_in':_A,'lock_type':_B,'override':_A}
class LogProperty(Property):arg_types={_h:_B}
class MaterializedProperty(Property):arg_types={_C:_A}
class MergeBlockRatioProperty(Property):arg_types={_C:_A,_h:_A,_M:_A,_v:_A}
class NoPrimaryIndexProperty(Property):arg_types={}
class OnProperty(Property):arg_types={_C:_B}
class OnCommitProperty(Property):arg_types={_A7:_A}
class PartitionedByProperty(Property):arg_types={_C:_B}
class PartitionBoundSpec(Expression):arg_types={_C:_A,_F:_A,'from_expressions':_A,'to_expressions':_A}
class PartitionedOfProperty(Property):arg_types={_C:_B,_F:_B}
class StreamingTableProperty(Property):arg_types={}
class RemoteWithConnectionModelProperty(Property):arg_types={_C:_B}
class ReturnsProperty(Property):arg_types={_C:_A,'is_table':_A,_Z:_A,_o:_A}
class StrictProperty(Property):arg_types={}
class RowFormatProperty(Property):arg_types={_C:_B}
class RowFormatDelimitedProperty(Property):arg_types={'fields':_A,'escaped':_A,'collection_items':_A,'map_keys':_A,'lines':_A,_o:_A,_A6:_A}
class RowFormatSerdeProperty(Property):arg_types={_C:_B,'serde_properties':_A}
class QueryTransform(Expression):arg_types={_E:_B,'command_script':_B,'schema':_A,'row_format_before':_A,'record_writer':_A,'row_format_after':_A,'record_reader':_A}
class SampleProperty(Property):arg_types={_C:_B}
class SecurityProperty(Property):arg_types={_C:_B}
class SchemaCommentProperty(Property):arg_types={_C:_B}
class SerdeProperties(Property):arg_types={_E:_B,_J:_A}
class SetProperty(Property):arg_types={'multi':_B}
class SharingProperty(Property):arg_types={_C:_A}
class SetConfigProperty(Property):arg_types={_C:_B}
class SettingsProperty(Property):arg_types={_E:_B}
class SortKeyProperty(Property):arg_types={_C:_B,'compound':_A}
class SqlReadWriteProperty(Property):arg_types={_C:_B}
class SqlSecurityProperty(Property):arg_types={'definer':_B}
class StabilityProperty(Property):arg_types={_C:_B}
class TemporaryProperty(Property):arg_types={_C:_A}
class SecureProperty(Property):arg_types={}
class TransformModelProperty(Property):arg_types={_E:_B}
class TransientProperty(Property):arg_types={_C:_A}
class UnloggedProperty(Property):arg_types={}
class ViewAttributeProperty(Property):arg_types={_C:_B}
class VolatileProperty(Property):arg_types={_C:_A}
class WithDataProperty(Property):arg_types={_h:_B,'statistics':_A}
class WithJournalTableProperty(Property):arg_types={_C:_B}
class WithSchemaBindingProperty(Property):arg_types={_C:_B}
class WithSystemVersioningProperty(Property):arg_types={_U:_A,_C:_A,'data_consistency':_A,_AV:_A,_J:_B}
class Properties(Expression):
	arg_types={_E:_B};NAME_TO_PROPERTY={'ALGORITHM':AlgorithmProperty,'AUTO_INCREMENT':AutoIncrementProperty,'CHARACTER SET':CharacterSetProperty,'CLUSTERED_BY':ClusteredByProperty,'COLLATE':CollateProperty,'COMMENT':SchemaCommentProperty,'DEFINER':DefinerProperty,'DISTKEY':DistKeyProperty,'DISTRIBUTED_BY':DistributedByProperty,'DISTSTYLE':DistStyleProperty,'ENGINE':EngineProperty,'EXECUTE AS':ExecuteAsProperty,'FORMAT':FileFormatProperty,'LANGUAGE':LanguageProperty,'LOCATION':LocationProperty,'LOCK':LockProperty,'PARTITIONED_BY':PartitionedByProperty,'RETURNS':ReturnsProperty,'ROW_FORMAT':RowFormatProperty,'SORTKEY':SortKeyProperty};PROPERTY_TO_NAME={B:A for(A,B)in NAME_TO_PROPERTY.items()}
	class Location(AutoName):POST_CREATE=auto();POST_NAME=auto();POST_SCHEMA=auto();POST_WITH=auto();POST_ALIAS=auto();POST_EXPRESSION=auto();POST_INDEX=auto();UNSUPPORTED=auto()
	@classmethod
	def from_dict(B,properties_dict):
		A=[]
		for(C,D)in properties_dict.items():
			E=B.NAME_TO_PROPERTY.get(C.upper())
			if E:A.append(E(this=convert(D)))
			else:A.append(Property(this=Literal.string(C),value=convert(D)))
		return B(expressions=A)
class Qualify(Expression):0
class InputOutputFormat(Expression):arg_types={_AU:_A,'output_format':_A}
class Return(Expression):0
class Reference(Expression):arg_types={_C:_B,_E:_A,_X:_A}
class Tuple(Expression):
	arg_types={_E:_A}
	def isin(E,*F,query=_D,unnest=_D,copy=_B,**D):C=unnest;B=query;A=copy;return In(this=maybe_copy(E,A),expressions=[convert(B,copy=A)for B in F],query=maybe_parse(B,copy=A,**D)if B else _D,unnest=Unnest(expressions=[maybe_parse(t.cast(ExpOrStr,B),copy=A,**D)for B in ensure_list(C)])if C else _D)
QUERY_MODIFIERS={'match':_A,_AE:_A,_AF:_A,'connect':_A,'pivots':_A,'prewhere':_A,_P:_A,_m:_A,'having':_A,'qualify':_A,'windows':_A,'distribute':_A,'sort':_A,_g:_A,_W:_A,_c:_A,_N:_A,'locks':_A,'sample':_A,_AC:_A,_K:_A,_X:_A}
class QueryOption(Expression):arg_types={_C:_B,_F:_A}
class WithTableHint(Expression):arg_types={_E:_B}
class IndexTableHint(Expression):arg_types={_C:_B,_E:_A,_A2:_A}
class HistoricalData(Expression):arg_types={_C:_B,_G:_B,_F:_B}
class Table(Expression):
	arg_types={_C:_A,_I:_A,_S:_A,_V:_A,_AE:_A,_AF:_A,'pivots':_A,'hints':_A,'system_time':_A,'version':_A,_K:_A,'pattern':_A,'ordinality':_A,'when':_A,'only':_A,_k:_A,'changes':_A,'rows_from':_A,'sample':_A}
	@property
	def name(self):
		if isinstance(self.this,Func):return''
		return self.this.name
	@property
	def db(self):return self.text(_S)
	@property
	def catalog(self):return self.text(_V)
	@property
	def selects(self):return[]
	@property
	def named_selects(self):return[]
	@property
	def parts(self):
		'Return the parts of a table in order catalog, db, table.';B=[]
		for C in(_V,_S,_C):
			A=self.args.get(C)
			if isinstance(A,Dot):B.extend(A.flatten())
			elif isinstance(A,Expression):B.append(A)
		return B
	def to_column(B,copy=_B):
		C=B.parts;A=column(*reversed(C[0:4]),fields=C[4:],copy=copy);D=B.args.get(_I)
		if D:A=alias_(A,D.this,copy=copy)
		return A
class SetOperation(Query):
	arg_types={_J:_A,_C:_B,_F:_B,_AG:_A,'by_name':_A,**QUERY_MODIFIERS}
	def select(F,*B,append=_B,dialect=_D,copy=_B,**E):D=dialect;C=append;A=maybe_copy(F,copy);A.this.unnest().select(*B,append=C,dialect=D,copy=_A,**E);A.expression.unnest().select(*B,append=C,dialect=D,copy=_A,**E);return A
	@property
	def named_selects(self):return self.this.unnest().named_selects
	@property
	def is_star(self):return self.this.is_star or self.expression.is_star
	@property
	def selects(self):return self.this.unnest().selects
	@property
	def left(self):return self.this
	@property
	def right(self):return self.expression
class Union(SetOperation):0
class Except(SetOperation):0
class Intersect(SetOperation):0
class Update(Expression):
	arg_types={_J:_A,_C:_A,_E:_B,_Y:_A,_P:_A,_f:_A,_W:_A,_c:_A}
	def table(A,expression,dialect=_D,copy=_B,**B):'\n        Set the table to update.\n\n        Example:\n            >>> Update().table("my_table").set_("x = 1").sql()\n            \'UPDATE my_table SET x = 1\'\n\n        Args:\n            expression : the SQL code strings to parse.\n                If a `Table` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `Table`.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Update expression.\n        ';return _apply_builder(expression=expression,instance=A,arg=_C,into=Table,prefix=_D,dialect=dialect,copy=copy,**B)
	def set_(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the SET expressions.\n\n        Example:\n            >>> Update().table("my_table").set_("x = 1").sql()\n            \'UPDATE my_table SET x = 1\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If `Expression` instance(s) are passed, they will be used as-is.\n                Multiple expressions are combined with a comma.\n            append: if `True`, add the new expressions to any existing SET expressions.\n                Otherwise, this resets the expressions.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n        ';return _apply_list_builder(*B,instance=A,arg=_E,append=append,into=Expression,prefix=_D,dialect=dialect,copy=copy,**C)
	def where(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the WHERE expressions.\n\n        Example:\n            >>> Update().table("tbl").set_("x = 1").where("x = \'a\' OR x < \'b\'").sql()\n            "UPDATE tbl SET x = 1 WHERE x = \'a\' OR x < \'b\'"\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n                Multiple expressions are combined with an AND operator.\n            append: if `True`, AND the new expressions to any existing expression.\n                Otherwise, this resets the expression.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            Select: the modified expression.\n        ';return _apply_conjunction_builder(*B,instance=A,arg=_P,append=append,into=Where,dialect=dialect,copy=copy,**C)
	def from_(A,expression=_D,dialect=_D,copy=_B,**C):
		'\n        Set the FROM expression.\n\n        Example:\n            >>> Update().table("my_table").set_("x = 1").from_("baz").sql()\n            \'UPDATE my_table SET x = 1 FROM baz\'\n\n        Args:\n            expression : the SQL code strings to parse.\n                If a `From` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `From`.\n                If nothing is passed in then a from is not applied to the expression\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Update expression.\n        ';B=expression
		if not B:return maybe_copy(A,copy)
		return _apply_builder(expression=B,instance=A,arg=_Y,into=From,prefix='FROM',dialect=dialect,copy=copy,**C)
	def with_(A,alias,as_,recursive=_D,materialized=_D,append=_B,dialect=_D,copy=_B,**B):'\n        Append to or set the common table expressions.\n\n        Example:\n            >>> Update().table("my_table").set_("x = 1").from_("baz").with_("baz", "SELECT id FROM foo").sql()\n            \'WITH baz AS (SELECT id FROM foo) UPDATE my_table SET x = 1 FROM baz\'\n\n        Args:\n            alias: the SQL code string to parse as the table name.\n                If an `Expression` instance is passed, this is used as-is.\n            as_: the SQL code string to parse as the table expression.\n                If an `Expression` instance is passed, it will be used as-is.\n            recursive: set the RECURSIVE part of the expression. Defaults to `False`.\n            materialized: set the MATERIALIZED part of the expression.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this resets the expressions.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified expression.\n        ';return _apply_cte_builder(A,alias,as_,recursive=recursive,materialized=materialized,append=append,dialect=dialect,copy=copy,**B)
class Values(UDTF):arg_types={_E:_B,_I:_A}
class Var(Expression):0
class Version(Expression):'\n    Time travel, iceberg, bigquery etc\n    https://trino.io/docs/current/connector/iceberg.html?highlight=snapshot#using-snapshots\n    https://www.databricks.com/blog/2019/02/04/introducing-delta-time-travel-for-large-scale-data-lakes.html\n    https://cloud.google.com/bigquery/docs/reference/standard-sql/query-syntax#for_system_time_as_of\n    https://learn.microsoft.com/en-us/sql/relational-databases/tables/querying-data-in-a-system-versioned-temporal-table?view=sql-server-ver16\n    this is either TIMESTAMP or VERSION\n    kind is ("AS OF", "BETWEEN")\n    ';arg_types={_C:_B,_G:_B,_F:_A}
class Schema(Expression):arg_types={_C:_A,_E:_A}
class Lock(Expression):arg_types={'update':_B,_E:_A,'wait':_A}
class Select(Query):
	arg_types={_J:_A,_G:_A,_E:_A,_n:_A,_AG:_A,'into':_A,_Y:_A,**QUERY_MODIFIERS}
	def from_(A,expression,dialect=_D,copy=_B,**B):'\n        Set the FROM expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x").sql()\n            \'SELECT x FROM tbl\'\n\n        Args:\n            expression : the SQL code strings to parse.\n                If a `From` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `From`.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_builder(expression=expression,instance=A,arg=_Y,into=From,prefix='FROM',dialect=dialect,copy=copy,**B)
	def group_by(A,*B,append=_B,dialect=_D,copy=_B,**C):
		'\n        Set the GROUP BY expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x", "COUNT(1)").group_by("x").sql()\n            \'SELECT x, COUNT(1) FROM tbl GROUP BY x\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If a `Group` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `Group`.\n                If nothing is passed in then a group by is not applied to the expression\n            append: if `True`, add to any existing expressions.\n                Otherwise, this flattens all the `Group` expression into a single expression.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        '
		if not B:return A if not copy else A.copy()
		return _apply_child_list_builder(*B,instance=A,arg=_m,append=append,copy=copy,prefix='GROUP BY',into=Group,dialect=dialect,**C)
	def sort_by(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Set the SORT BY expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x").sort_by("x DESC").sql(dialect="hive")\n            \'SELECT x FROM tbl SORT BY x DESC\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If a `Group` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `SORT`.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this flattens all the `Order` expression into a single expression.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_child_list_builder(*B,instance=A,arg='sort',append=append,copy=copy,prefix='SORT BY',into=Sort,dialect=dialect,**C)
	def cluster_by(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Set the CLUSTER BY expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x").cluster_by("x DESC").sql(dialect="hive")\n            \'SELECT x FROM tbl CLUSTER BY x DESC\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If a `Group` instance is passed, this is used as-is.\n                If another `Expression` instance is passed, it will be wrapped in a `Cluster`.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this flattens all the `Order` expression into a single expression.\n            dialect: the dialect used to parse the input expression.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_child_list_builder(*B,instance=A,arg=_g,append=append,copy=copy,prefix='CLUSTER BY',into=Cluster,dialect=dialect,**C)
	def select(A,*B,append=_B,dialect=_D,copy=_B,**C):return _apply_list_builder(*B,instance=A,arg=_E,append=append,dialect=dialect,into=Expression,copy=copy,**C)
	def lateral(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the LATERAL expressions.\n\n        Example:\n            >>> Select().select("x").lateral("OUTER explode(y) tbl2 AS z").from_("tbl").sql()\n            \'SELECT x FROM tbl LATERAL VIEW OUTER EXPLODE(y) tbl2 AS z\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this resets the expressions.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_list_builder(*B,instance=A,arg=_AE,append=append,into=Lateral,prefix='LATERAL VIEW',dialect=dialect,copy=copy,**C)
	def join(N,expression,on=_D,using=_D,append=_B,join_type=_D,join_alias=_D,dialect=_D,copy=_B,**C):
		'\n        Append to or set the JOIN expressions.\n\n        Example:\n            >>> Select().select("*").from_("tbl").join("tbl2", on="tbl1.y = tbl2.y").sql()\n            \'SELECT * FROM tbl JOIN tbl2 ON tbl1.y = tbl2.y\'\n\n            >>> Select().select("1").from_("a").join("b", using=["x", "y", "z"]).sql()\n            \'SELECT 1 FROM a JOIN b USING (x, y, z)\'\n\n            Use `join_type` to change the type of join:\n\n            >>> Select().select("*").from_("tbl").join("tbl2", on="tbl1.y = tbl2.y", join_type="left outer").sql()\n            \'SELECT * FROM tbl LEFT OUTER JOIN tbl2 ON tbl1.y = tbl2.y\'\n\n        Args:\n            expression: the SQL code string to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n            on: optionally specify the join "on" criteria as a SQL string.\n                If an `Expression` instance is passed, it will be used as-is.\n            using: optionally specify the join "using" criteria as a SQL string.\n                If an `Expression` instance is passed, it will be used as-is.\n            append: if `True`, add to any existing expressions.\n                Otherwise, this resets the expressions.\n            join_type: if set, alter the parsed join type.\n            join_alias: an optional alias for the joined source.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            Select: the modified expression.\n        ';M=dialect;L=join_alias;K=join_type;J=append;I=using;D=copy;B=expression;E={'dialect':M,**C}
		try:B=maybe_parse(B,into=Join,prefix='JOIN',**E)
		except ParseError:B=maybe_parse(B,into=(Join,Expression),**E)
		A=B if isinstance(B,Join)else Join(this=B)
		if isinstance(A.this,Select):A.this.replace(A.this.subquery())
		if K:
			F:0;G:0;H:0;F,G,H=maybe_parse(K,into='JOIN_TYPE',**E)
			if F:A.set(_w,F.text)
			if G:A.set('side',G.text)
			if H:A.set(_G,H.text)
		if on:on=and_(*ensure_list(on),dialect=M,copy=D,**C);A.set(_U,on)
		if I:A=_apply_list_builder(*ensure_list(I),instance=A,arg=_T,append=J,copy=D,into=Identifier,**C)
		if L:A.set(_C,alias_(A.this,L,table=_B))
		return _apply_list_builder(A,instance=N,arg=_AF,append=J,copy=D,**C)
	def where(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the WHERE expressions.\n\n        Example:\n            >>> Select().select("x").from_("tbl").where("x = \'a\' OR x < \'b\'").sql()\n            "SELECT x FROM tbl WHERE x = \'a\' OR x < \'b\'"\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n                Multiple expressions are combined with an AND operator.\n            append: if `True`, AND the new expressions to any existing expression.\n                Otherwise, this resets the expression.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            Select: the modified expression.\n        ';return _apply_conjunction_builder(*B,instance=A,arg=_P,append=append,into=Where,dialect=dialect,copy=copy,**C)
	def having(A,*B,append=_B,dialect=_D,copy=_B,**C):'\n        Append to or set the HAVING expressions.\n\n        Example:\n            >>> Select().select("x", "COUNT(y)").from_("tbl").group_by("x").having("COUNT(y) > 3").sql()\n            \'SELECT x, COUNT(y) FROM tbl GROUP BY x HAVING COUNT(y) > 3\'\n\n        Args:\n            *expressions: the SQL code strings to parse.\n                If an `Expression` instance is passed, it will be used as-is.\n                Multiple expressions are combined with an AND operator.\n            append: if `True`, AND the new expressions to any existing expression.\n                Otherwise, this resets the expression.\n            dialect: the dialect used to parse the input expressions.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input expressions.\n\n        Returns:\n            The modified Select expression.\n        ';return _apply_conjunction_builder(*B,instance=A,arg='having',append=append,into=Having,dialect=dialect,copy=copy,**C)
	def window(A,*B,append=_B,dialect=_D,copy=_B,**C):return _apply_list_builder(*B,instance=A,arg='windows',append=append,into=Window,dialect=dialect,copy=copy,**C)
	def qualify(A,*B,append=_B,dialect=_D,copy=_B,**C):return _apply_conjunction_builder(*B,instance=A,arg='qualify',append=append,into=Qualify,dialect=dialect,copy=copy,**C)
	def distinct(C,*A,distinct=_B,copy=_B):'\n        Set the OFFSET expression.\n\n        Example:\n            >>> Select().from_("tbl").select("x").distinct().sql()\n            \'SELECT DISTINCT x FROM tbl\'\n\n        Args:\n            ons: the expressions to distinct on\n            distinct: whether the Select should be distinct\n            copy: if `False`, modify this expression instance in-place.\n\n        Returns:\n            Select: the modified expression.\n        ';B=maybe_copy(C,copy);D=Tuple(expressions=[maybe_parse(A,copy=copy)for A in A if A])if A else _D;B.set(_AG,Distinct(on=D)if distinct else _D);return B
	def ctas(C,table,properties=_D,dialect=_D,copy=_B,**D):
		'\n        Convert this expression to a CREATE TABLE AS statement.\n\n        Example:\n            >>> Select().select("*").from_("tbl").ctas("x").sql()\n            \'CREATE TABLE x AS SELECT * FROM tbl\'\n\n        Args:\n            table: the SQL code string to parse as the table name.\n                If another `Expression` instance is passed, it will be used as-is.\n            properties: an optional mapping of table properties\n            dialect: the dialect used to parse the input table.\n            copy: if `False`, modify this expression instance in-place.\n            opts: other options to use to parse the input table.\n\n        Returns:\n            The new Create expression.\n        ';A=properties;E=maybe_copy(C,copy);F=maybe_parse(table,into=Table,dialect=dialect,**D);B=_D
		if A:B=Properties.from_dict(A)
		return Create(this=F,kind=_AH,expression=E,properties=B)
	def lock(B,update=_B,copy=_B):'\n        Set the locking read mode for this expression.\n\n        Examples:\n            >>> Select().select("x").from_("tbl").where("x = \'a\'").lock().sql("mysql")\n            "SELECT x FROM tbl WHERE x = \'a\' FOR UPDATE"\n\n            >>> Select().select("x").from_("tbl").where("x = \'a\'").lock(update=False).sql("mysql")\n            "SELECT x FROM tbl WHERE x = \'a\' FOR SHARE"\n\n        Args:\n            update: if `True`, the locking type will be `FOR UPDATE`, else it will be `FOR SHARE`.\n            copy: if `False`, modify this expression instance in-place.\n\n        Returns:\n            The modified expression.\n        ';A=maybe_copy(B,copy);A.set('locks',[Lock(update=update)]);return A
	def hint(B,*C,dialect=_D,copy=_B):'\n        Set hints for this expression.\n\n        Examples:\n            >>> Select().select("x").from_("tbl").hint("BROADCAST(y)").sql(dialect="spark")\n            \'SELECT /*+ BROADCAST(y) */ x FROM tbl\'\n\n        Args:\n            hints: The SQL code strings to parse as the hints.\n                If an `Expression` instance is passed, it will be used as-is.\n            dialect: The dialect used to parse the hints.\n            copy: If `False`, modify this expression instance in-place.\n\n        Returns:\n            The modified expression.\n        ';A=maybe_copy(B,copy);A.set(_n,Hint(expressions=[maybe_parse(A,copy=copy,dialect=dialect)for A in C]));return A
	@property
	def named_selects(self):return[A.output_name for A in self.expressions if A.alias_or_name]
	@property
	def is_star(self):return any(A.is_star for A in self.expressions)
	@property
	def selects(self):return self.expressions
UNWRAPPED_QUERIES=Select,SetOperation
class Subquery(DerivedTable,Query):
	arg_types={_C:_B,_I:_A,_J:_A,**QUERY_MODIFIERS}
	def unnest(B):
		'Returns the first non subquery.';A=B
		while isinstance(A,Subquery):A=A.this
		return A
	def unwrap(B):
		A=B
		while A.same_parent and A.is_wrapper:A=t.cast(Subquery,A.parent)
		return A
	def select(B,*C,append=_B,dialect=_D,copy=_B,**D):A=maybe_copy(B,copy);A.unnest().select(*C,append=append,dialect=dialect,copy=_A,**D);return A
	@property
	def is_wrapper(self):'\n        Whether this Subquery acts as a simple wrapper around another expression.\n\n        SELECT * FROM (((SELECT * FROM t)))\n                      ^\n                      This corresponds to a "wrapper" Subquery node\n        ';return all(B is _D for(A,B)in self.args.items()if A!=_C)
	@property
	def is_star(self):return self.this.is_star
	@property
	def output_name(self):return self.alias
class TableSample(Expression):arg_types={_E:_A,_w:_A,'bucket_numerator':_A,'bucket_denominator':_A,'bucket_field':_A,_v:_A,'rows':_A,'size':_A,'seed':_A}
class Tag(Expression):'Tags are used for generating arbitrary sql like SELECT <span>x</span>.';arg_types={_C:_A,'prefix':_A,'postfix':_A}
class Pivot(Expression):
	arg_types={_C:_A,_I:_A,_E:_A,'field':_A,'unpivot':_A,_T:_A,_m:_A,_a:_A,'include_nulls':_A,'default_on_null':_A}
	@property
	def unpivot(self):return bool(self.args.get('unpivot'))
class Window(Condition):arg_types={_C:_B,_AB:_A,_W:_A,'spec':_A,_I:_A,'over':_A,'first':_A}
class WindowSpec(Expression):arg_types={_G:_A,_O:_A,'start_side':_A,_d:_A,'end_side':_A}
class PreWhere(Expression):0
class Where(Expression):0
class Star(Expression):
	arg_types={'except':_A,_z:_A,'rename':_A}
	@property
	def name(self):return'*'
	@property
	def output_name(self):return self.name
class Parameter(Condition):arg_types={_C:_B,_F:_A}
class SessionParameter(Condition):arg_types={_C:_B,_G:_A}
class Placeholder(Condition):
	arg_types={_C:_A,_G:_A}
	@property
	def name(self):return self.this or'?'
class Null(Condition):
	arg_types:t.Dict[str,t.Any]={}
	@property
	def name(self):return'NULL'
	def to_py(A):0
class Boolean(Condition):
	def to_py(A):return A.this
class DataTypeParam(Expression):
	arg_types={_C:_B,_F:_A}
	@property
	def name(self):return self.this.name
class DataType(Expression):
	arg_types={_C:_B,_E:_A,'nested':_A,_q:_A,'prefix':_A,_G:_A,_AI:_A}
	class Type(AutoName):ARRAY=auto();AGGREGATEFUNCTION=auto();SIMPLEAGGREGATEFUNCTION=auto();BIGDECIMAL=auto();BIGINT=auto();BIGSERIAL=auto();BINARY=auto();BIT=auto();BOOLEAN=auto();BPCHAR=auto();CHAR=auto();DATE=auto();DATE32=auto();DATEMULTIRANGE=auto();DATERANGE=auto();DATETIME=auto();DATETIME64=auto();DECIMAL=auto();DECIMAL32=auto();DECIMAL64=auto();DECIMAL128=auto();DOUBLE=auto();ENUM=auto();ENUM8=auto();ENUM16=auto();FIXEDSTRING=auto();FLOAT=auto();GEOGRAPHY=auto();GEOMETRY=auto();HLLSKETCH=auto();HSTORE=auto();IMAGE=auto();INET=auto();INT=auto();INT128=auto();INT256=auto();INT4MULTIRANGE=auto();INT4RANGE=auto();INT8MULTIRANGE=auto();INT8RANGE=auto();INTERVAL=auto();IPADDRESS=auto();IPPREFIX=auto();IPV4=auto();IPV6=auto();JSON=auto();JSONB=auto();LIST=auto();LONGBLOB=auto();LONGTEXT=auto();LOWCARDINALITY=auto();MAP=auto();MEDIUMBLOB=auto();MEDIUMINT=auto();MEDIUMTEXT=auto();MONEY=auto();NAME=auto();NCHAR=auto();NESTED=auto();NULL=auto();NUMMULTIRANGE=auto();NUMRANGE=auto();NVARCHAR=auto();OBJECT=auto();RANGE=auto();ROWVERSION=auto();SERIAL=auto();SET=auto();SMALLINT=auto();SMALLMONEY=auto();SMALLSERIAL=auto();STRUCT=auto();SUPER=auto();TEXT=auto();TINYBLOB=auto();TINYTEXT=auto();TIME=auto();TIMETZ=auto();TIMESTAMP=auto();TIMESTAMPNTZ=auto();TIMESTAMPLTZ=auto();TIMESTAMPTZ=auto();TIMESTAMP_S=auto();TIMESTAMP_MS=auto();TIMESTAMP_NS=auto();TINYINT=auto();TSMULTIRANGE=auto();TSRANGE=auto();TSTZMULTIRANGE=auto();TSTZRANGE=auto();UBIGINT=auto();UINT=auto();UINT128=auto();UINT256=auto();UMEDIUMINT=auto();UDECIMAL=auto();UNION=auto();UNIQUEIDENTIFIER=auto();UNKNOWN=auto();USERDEFINED='USER-DEFINED';USMALLINT=auto();UTINYINT=auto();UUID=auto();VARBINARY=auto();VARCHAR=auto();VARIANT=auto();VECTOR=auto();XML=auto();YEAR=auto();TDIGEST=auto()
	STRUCT_TYPES={Type.NESTED,Type.OBJECT,Type.STRUCT,Type.UNION};ARRAY_TYPES={Type.ARRAY,Type.LIST};NESTED_TYPES={*STRUCT_TYPES,*ARRAY_TYPES,Type.MAP};TEXT_TYPES={Type.CHAR,Type.NCHAR,Type.NVARCHAR,Type.TEXT,Type.VARCHAR,Type.NAME};SIGNED_INTEGER_TYPES={Type.BIGINT,Type.INT,Type.INT128,Type.INT256,Type.MEDIUMINT,Type.SMALLINT,Type.TINYINT};UNSIGNED_INTEGER_TYPES={Type.UBIGINT,Type.UINT,Type.UINT128,Type.UINT256,Type.UMEDIUMINT,Type.USMALLINT,Type.UTINYINT};INTEGER_TYPES={*SIGNED_INTEGER_TYPES,*UNSIGNED_INTEGER_TYPES,Type.BIT};FLOAT_TYPES={Type.DOUBLE,Type.FLOAT};REAL_TYPES={*FLOAT_TYPES,Type.BIGDECIMAL,Type.DECIMAL,Type.DECIMAL32,Type.DECIMAL64,Type.DECIMAL128,Type.MONEY,Type.SMALLMONEY,Type.UDECIMAL};NUMERIC_TYPES={*INTEGER_TYPES,*REAL_TYPES};TEMPORAL_TYPES={Type.DATE,Type.DATE32,Type.DATETIME,Type.DATETIME64,Type.TIME,Type.TIMESTAMP,Type.TIMESTAMPNTZ,Type.TIMESTAMPLTZ,Type.TIMESTAMPTZ,Type.TIMESTAMP_MS,Type.TIMESTAMP_NS,Type.TIMESTAMP_S,Type.TIMETZ}
	@classmethod
	def build(E,dtype,dialect=_D,udt=_A,copy=_B,**B):
		"\n        Constructs a DataType object.\n\n        Args:\n            dtype: the data type of interest.\n            dialect: the dialect to use for parsing `dtype`, in case it's a string.\n            udt: when set to True, `dtype` will be used as-is if it can't be parsed into a\n                DataType, thus creating a user-defined type.\n            copy: whether to copy the data type.\n            kwargs: additional arguments to pass in the constructor of DataType.\n\n        Returns:\n            The constructed DataType object.\n        ";A=dtype;from sqlglot import parse_one as D
		if isinstance(A,str):
			if A.upper()=='UNKNOWN':return DataType(this=DataType.Type.UNKNOWN,**B)
			try:C=D(A,read=dialect,into=DataType,error_level=ErrorLevel.IGNORE)
			except ParseError:
				if udt:return DataType(this=DataType.Type.USERDEFINED,kind=A,**B)
				raise
		elif isinstance(A,DataType.Type):C=DataType(this=A)
		elif isinstance(A,DataType):return maybe_copy(A,copy)
		else:raise ValueError(f"Invalid data type: {type(A)}. Expected str or DataType.Type")
		return DataType(**{**C.args,**B})
	def is_type(B,*D,check_nullable=_A):
		'\n        Checks whether this DataType matches one of the provided data types. Nested types or precision\n        will be compared using "structural equivalence" semantics, so e.g. array<int> != array<float>.\n\n        Args:\n            dtypes: the data types to compare this DataType to.\n            check_nullable: whether to take the NULLABLE type constructor into account for the comparison.\n                If false, it means that NULLABLE<INT> is equivalent to INT.\n\n        Returns:\n            True, if and only if there is a type in `dtypes` which is equal to this DataType.\n        ';E=B.args.get(_AI)
		for F in D:
			A=DataType.build(F,copy=_A,udt=_B);G=A.args.get(_AI)
			if A.expressions or check_nullable and(E or G)or B.this==DataType.Type.USERDEFINED or A.this==DataType.Type.USERDEFINED:C=B==A
			else:C=B.this==A.this
			if C:return _B
		return _A
DATA_TYPE=t.Union[str,DataType,DataType.Type]
class PseudoType(DataType):arg_types={_C:_B}
class ObjectIdentifier(DataType):arg_types={_C:_B}
class SubqueryPredicate(Predicate):0
class All(SubqueryPredicate):0
class Any(SubqueryPredicate):0
class Exists(SubqueryPredicate):0
class Command(Expression):arg_types={_C:_B,_F:_A}
class Transaction(Expression):arg_types={_C:_A,'modes':_A,'mark':_A}
class Commit(Expression):arg_types={'chain':_A,_C:_A,'durability':_A}
class Rollback(Expression):arg_types={'savepoint':_A,_C:_A}
class Alter(Expression):
	arg_types={_C:_B,_G:_B,'actions':_B,_R:_A,'only':_A,_X:_A,_g:_A,'not_valid':_A}
	@property
	def kind(self):A=self.args.get(_G);return A and A.upper()
	@property
	def actions(self):return self.args.get('actions')or[]
class AddConstraint(Expression):arg_types={_E:_B}
class DropPartition(Expression):arg_types={_E:_B,_R:_A}
class ReplacePartition(Expression):arg_types={_F:_B,_u:_B}
class Binary(Condition):
	arg_types={_C:_B,_F:_B}
	@property
	def left(self):return self.this
	@property
	def right(self):return self.expression
class Add(Binary):0
class Connector(Binary):0
class And(Connector):0
class Or(Connector):0
class BitwiseAnd(Binary):0
class BitwiseLeftShift(Binary):0
class BitwiseOr(Binary):0
class BitwiseRightShift(Binary):0
class BitwiseXor(Binary):0
class Div(Binary):arg_types={_C:_B,_F:_B,'typed':_A,_Q:_A}
class Overlaps(Binary):0
class Dot(Binary):
	@property
	def is_star(self):return self.expression.is_star
	@property
	def name(self):return self.expression.name
	@property
	def output_name(self):return self.name
	@classmethod
	def build(B,expressions):
		'Build a Dot object with a sequence of expressions.';A=expressions
		if len(A)<2:raise ValueError('Dot requires >= 2 expressions.')
		return t.cast(Dot,reduce(lambda x,y:Dot(this=x,expression=y),A))
	@property
	def parts(self):
		'Return the parts of a table / column in order catalog, db, table.';C,*A=self.flatten();A.reverse()
		for D in COLUMN_PARTS:
			B=C.args.get(D)
			if isinstance(B,Expression):A.append(B)
		A.reverse();return A
class DPipe(Binary):arg_types={_C:_B,_F:_B,_Q:_A}
class EQ(Binary,Predicate):0
class NullSafeEQ(Binary,Predicate):0
class NullSafeNEQ(Binary,Predicate):0
class PropertyEQ(Binary):0
class Distance(Binary):0
class Escape(Binary):0
class Glob(Binary,Predicate):0
class GT(Binary,Predicate):0
class GTE(Binary,Predicate):0
class ILike(Binary,Predicate):0
class ILikeAny(Binary,Predicate):0
class IntDiv(Binary):0
class Is(Binary,Predicate):0
class Kwarg(Binary):'Kwarg in special functions like func(kwarg => y).'
class Like(Binary,Predicate):0
class LikeAny(Binary,Predicate):0
class LT(Binary,Predicate):0
class LTE(Binary,Predicate):0
class Mod(Binary):0
class Mul(Binary):0
class NEQ(Binary,Predicate):0
class Operator(Binary):arg_types={_C:_B,'operator':_B,_F:_B}
class SimilarTo(Binary,Predicate):0
class Slice(Binary):arg_types={_C:_A,_F:_A}
class Sub(Binary):0
class Unary(Condition):0
class BitwiseNot(Unary):0
class Not(Unary):0
class Paren(Unary):
	@property
	def output_name(self):return self.this.name
class Neg(Unary):
	def to_py(A):
		if A.is_number:return A.this.to_py()*-1
		return super().to_py()
class Alias(Expression):
	arg_types={_C:_B,_I:_A}
	@property
	def output_name(self):return self.alias
class PivotAlias(Alias):0
class PivotAny(Expression):arg_types={_C:_A}
class Aliases(Expression):
	arg_types={_C:_B,_E:_B}
	@property
	def aliases(self):return self.expressions
class AtIndex(Expression):arg_types={_C:_B,_F:_B}
class AtTimeZone(Expression):arg_types={_C:_B,_L:_B}
class FromTimeZone(Expression):arg_types={_C:_B,_L:_B}
class Between(Predicate):arg_types={_C:_B,'low':_B,'high':_B}
class Bracket(Condition):
	arg_types={_C:_B,_E:_B,_N:_A,_Q:_A,'returns_list_for_maps':_A}
	@property
	def output_name(self):
		if len(self.expressions)==1:return self.expressions[0].output_name
		return super().output_name
class Distinct(Expression):arg_types={_E:_A,_U:_A}
class In(Predicate):arg_types={_C:_B,_E:_A,'query':_A,'unnest':_A,'field':_A,'is_global':_A}
class ForIn(Expression):arg_types={_C:_B,_F:_B}
class TimeUnit(Expression):
	'Automatically converts unit arg into a var.';arg_types={_H:_A};UNABBREVIATED_UNIT_NAME={'D':'DAY','H':'HOUR','M':'MINUTE','MS':'MILLISECOND','NS':'NANOSECOND','Q':'QUARTER','S':'SECOND','US':'MICROSECOND','W':'WEEK','Y':'YEAR'};VAR_LIKE=Column,Literal,Var
	def __init__(C,**B):
		A=B.get(_H)
		if isinstance(A,C.VAR_LIKE):B[_H]=Var(this=(C.UNABBREVIATED_UNIT_NAME.get(A.name)or A.name).upper())
		elif isinstance(A,Week):A.set(_C,Var(this=A.this.name.upper()))
		super().__init__(**B)
	@property
	def unit(self):return self.args.get(_H)
class IntervalOp(TimeUnit):
	arg_types={_H:_A,_F:_B}
	def interval(A):return Interval(this=A.expression.copy(),unit=A.unit.copy()if A.unit else _D)
class IntervalSpan(DataType):arg_types={_C:_B,_F:_B}
class Interval(TimeUnit):arg_types={_C:_A,_H:_A}
class IgnoreNulls(Expression):0
class RespectNulls(Expression):0
class HavingMax(Expression):arg_types={_C:_B,_F:_B,'max':_B}
class Func(Condition):
	"\n    The base class for all function expressions.\n\n    Attributes:\n        is_var_len_args (bool): if set to True the last argument defined in arg_types will be\n            treated as a variable length argument and the argument's value will be stored as a list.\n        _sql_names (list): the SQL name (1st item in the list) and aliases (subsequent items) for this\n            function expression. These values are used to map this node to a name during parsing as\n            well as to provide the function's name during SQL string generation. By default the SQL\n            name is set to the expression's class name transformed to snake case.\n    ";is_var_len_args=_A
	@classmethod
	def from_arg_list(A,args):
		B=args
		if A.is_var_len_args:C=list(A.arg_types);E=C[:-1]if A.is_var_len_args else C;F=len(E);D={B:A for(A,B)in zip(B,E)};D[C[-1]]=B[F:]
		else:D={B:A for(A,B)in zip(B,A.arg_types)}
		return A(**D)
	@classmethod
	def sql_names(A):
		if A is Func:raise NotImplementedError('SQL name is only supported by concrete function implementations')
		if'_sql_names'not in A.__dict__:A._sql_names=[camel_to_snake_case(A.__name__)]
		return A._sql_names
	@classmethod
	def sql_name(A):return A.sql_names()[0]
	@classmethod
	def default_parser_mappings(A):return{B:A.from_arg_list for B in A.sql_names()}
class AggFunc(Func):0
class ParameterizedAgg(AggFunc):arg_types={_C:_B,_E:_B,_t:_B}
class Abs(Func):0
class ArgMax(AggFunc):arg_types={_C:_B,_F:_B,_AD:_A};_sql_names=['ARG_MAX','ARGMAX','MAX_BY']
class ArgMin(AggFunc):arg_types={_C:_B,_F:_B,_AD:_A};_sql_names=['ARG_MIN','ARGMIN','MIN_BY']
class ApproxTopK(AggFunc):arg_types={_C:_B,_F:_A,'counters':_A}
class Flatten(Func):0
class Transform(Func):arg_types={_C:_B,_F:_B}
class Anonymous(Func):
	arg_types={_C:_B,_E:_A};is_var_len_args=_B
	@property
	def name(self):A=self;return A.this if isinstance(A.this,str)else A.this.name
class AnonymousAggFunc(AggFunc):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class CombinedAggFunc(AnonymousAggFunc):arg_types={_C:_B,_E:_A,'parts':_B}
class CombinedParameterizedAgg(ParameterizedAgg):arg_types={_C:_B,_E:_B,_t:_B,'parts':_B}
class Hll(AggFunc):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class ApproxDistinct(AggFunc):arg_types={_C:_B,'accuracy':_A};_sql_names=['APPROX_DISTINCT','APPROX_COUNT_DISTINCT']
class Apply(Func):arg_types={_C:_B,_F:_B}
class Array(Func):arg_types={_E:_A,'bracket_notation':_A};is_var_len_args=_B
class ToArray(Func):0
class List(Func):arg_types={_E:_A};is_var_len_args=_B
class Pad(Func):arg_types={_C:_B,_F:_B,'fill_pattern':_A,'is_left':_B}
class ToChar(Func):arg_types={_C:_B,_K:_A,'nlsparam':_A}
class ToNumber(Func):arg_types={_C:_B,_K:_A,'nlsparam':_A,_AW:_A,'scale':_A}
class Columns(Func):arg_types={_C:_B,'unpack':_A}
class Convert(Func):arg_types={_C:_B,_F:_B,'style':_A}
class ConvertTimezone(Func):arg_types={'source_tz':_A,'target_tz':_B,'timestamp':_B}
class GenerateSeries(Func):arg_types={_O:_B,_d:_B,_p:_A,'is_end_exclusive':_A}
class ExplodingGenerateSeries(GenerateSeries):0
class ArrayAgg(AggFunc):arg_types={_C:_B,'nulls_excluded':_A}
class ArrayUniqueAgg(AggFunc):0
class ArrayAll(Func):arg_types={_C:_B,_F:_B}
class ArrayAny(Func):arg_types={_C:_B,_F:_B}
class ArrayConcat(Func):_sql_names=['ARRAY_CONCAT','ARRAY_CAT'];arg_types={_C:_B,_E:_A};is_var_len_args=_B
class ArrayConstructCompact(Func):arg_types={_E:_B};is_var_len_args=_B
class ArrayContains(Binary,Func):_sql_names=['ARRAY_CONTAINS','ARRAY_HAS']
class ArrayContainsAll(Binary,Func):_sql_names=['ARRAY_CONTAINS_ALL','ARRAY_HAS_ALL']
class ArrayFilter(Func):arg_types={_C:_B,_F:_B};_sql_names=['FILTER','ARRAY_FILTER']
class ArrayToString(Func):arg_types={_C:_B,_F:_B,_o:_A};_sql_names=['ARRAY_TO_STRING','ARRAY_JOIN']
class StringToArray(Func):arg_types={_C:_B,_F:_B,_o:_A};_sql_names=['STRING_TO_ARRAY','SPLIT_BY_STRING']
class ArrayOverlaps(Binary,Func):0
class ArraySize(Func):arg_types={_C:_B,_F:_A};_sql_names=['ARRAY_SIZE','ARRAY_LENGTH']
class ArraySort(Func):arg_types={_C:_B,_F:_A}
class ArraySum(Func):arg_types={_C:_B,_F:_A}
class ArrayUnionAgg(AggFunc):0
class Avg(AggFunc):0
class AnyValue(AggFunc):0
class Lag(AggFunc):arg_types={_C:_B,_N:_A,_M:_A}
class Lead(AggFunc):arg_types={_C:_B,_N:_A,_M:_A}
class First(AggFunc):0
class Last(AggFunc):0
class FirstValue(AggFunc):0
class LastValue(AggFunc):0
class NthValue(AggFunc):arg_types={_C:_B,_N:_B}
class Case(Func):
	arg_types={_C:_A,'ifs':_B,_M:_A}
	def when(D,condition,then,copy=_B,**B):A=copy;C=maybe_copy(D,A);C.append('ifs',If(this=maybe_parse(condition,copy=A,**B),true=maybe_parse(then,copy=A,**B)));return C
	def else_(B,condition,copy=_B,**C):A=maybe_copy(B,copy);A.set(_M,maybe_parse(condition,copy=copy,**C));return A
class Cast(Func):
	arg_types={_C:_B,_l:_B,_K:_A,_Q:_A,'action':_A}
	@property
	def name(self):return self.this.name
	@property
	def to(self):return self.args[_l]
	@property
	def output_name(self):return self.name
	def is_type(A,*B):'\n        Checks whether this Cast\'s DataType matches one of the provided data types. Nested types\n        like arrays or structs will be compared using "structural equivalence" semantics, so e.g.\n        array<int> != array<float>.\n\n        Args:\n            dtypes: the data types to compare this Cast\'s DataType to.\n\n        Returns:\n            True, if and only if there is a type in `dtypes` which is equal to this Cast\'s DataType.\n        ';return A.to.is_type(*B)
class TryCast(Cast):0
class Try(Func):0
class CastToStrType(Func):arg_types={_C:_B,_l:_B}
class Collate(Binary,Func):0
class Ceil(Func):arg_types={_C:_B,_AJ:_A};_sql_names=['CEIL','CEILING']
class Coalesce(Func):arg_types={_C:_B,_E:_A,'is_nvl':_A};is_var_len_args=_B;_sql_names=['COALESCE','IFNULL','NVL']
class Chr(Func):arg_types={_E:_B,_AK:_A};is_var_len_args=_B;_sql_names=['CHR','CHAR']
class Concat(Func):arg_types={_E:_B,_Q:_A,'coalesce':_A};is_var_len_args=_B
class ConcatWs(Concat):_sql_names=['CONCAT_WS']
class ConnectByRoot(Func):0
class Count(AggFunc):arg_types={_C:_A,_E:_A,'big_int':_A};is_var_len_args=_B
class CountIf(AggFunc):_sql_names=['COUNT_IF','COUNTIF']
class Cbrt(Func):0
class CurrentDate(Func):arg_types={_C:_A}
class CurrentDatetime(Func):arg_types={_C:_A}
class CurrentTime(Func):arg_types={_C:_A}
class CurrentTimestamp(Func):arg_types={_C:_A,'sysdate':_A}
class CurrentUser(Func):arg_types={_C:_A}
class DateAdd(Func,IntervalOp):arg_types={_C:_B,_F:_B,_H:_A}
class DateSub(Func,IntervalOp):arg_types={_C:_B,_F:_B,_H:_A}
class DateDiff(Func,TimeUnit):_sql_names=['DATEDIFF','DATE_DIFF'];arg_types={_C:_B,_F:_B,_H:_A}
class DateTrunc(Func):
	arg_types={_H:_B,_C:_B,_L:_A}
	def __init__(C,**B):
		A=B.get(_H)
		if isinstance(A,TimeUnit.VAR_LIKE):B[_H]=Literal.string((TimeUnit.UNABBREVIATED_UNIT_NAME.get(A.name)or A.name).upper())
		elif isinstance(A,Week):A.set(_C,Literal.string(A.this.name.upper()))
		super().__init__(**B)
	@property
	def unit(self):return self.args[_H]
class Datetime(Func):arg_types={_C:_B,_F:_A}
class DatetimeAdd(Func,IntervalOp):arg_types={_C:_B,_F:_B,_H:_A}
class DatetimeSub(Func,IntervalOp):arg_types={_C:_B,_F:_B,_H:_A}
class DatetimeDiff(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class DatetimeTrunc(Func,TimeUnit):arg_types={_C:_B,_H:_B,_L:_A}
class DayOfWeek(Func):_sql_names=['DAY_OF_WEEK','DAYOFWEEK']
class DayOfWeekIso(Func):_sql_names=['DAYOFWEEK_ISO','ISODOW']
class DayOfMonth(Func):_sql_names=['DAY_OF_MONTH','DAYOFMONTH']
class DayOfYear(Func):_sql_names=['DAY_OF_YEAR','DAYOFYEAR']
class ToDays(Func):0
class WeekOfYear(Func):_sql_names=['WEEK_OF_YEAR','WEEKOFYEAR']
class MonthsBetween(Func):arg_types={_C:_B,_F:_B,'roundoff':_A}
class LastDay(Func,TimeUnit):_sql_names=['LAST_DAY','LAST_DAY_OF_MONTH'];arg_types={_C:_B,_H:_A}
class Extract(Func):arg_types={_C:_B,_F:_B}
class Timestamp(Func):arg_types={_C:_A,_L:_A,'with_tz':_A}
class TimestampAdd(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class TimestampSub(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class TimestampDiff(Func,TimeUnit):_sql_names=['TIMESTAMPDIFF','TIMESTAMP_DIFF'];arg_types={_C:_B,_F:_B,_H:_A}
class TimestampTrunc(Func,TimeUnit):arg_types={_C:_B,_H:_B,_L:_A}
class TimeAdd(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class TimeSub(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class TimeDiff(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class TimeTrunc(Func,TimeUnit):arg_types={_C:_B,_H:_B,_L:_A}
class DateFromParts(Func):_sql_names=['DATE_FROM_PARTS','DATEFROMPARTS'];arg_types={'year':_B,'month':_B,'day':_B}
class TimeFromParts(Func):_sql_names=['TIME_FROM_PARTS','TIMEFROMPARTS'];arg_types={'hour':_B,'min':_B,'sec':_B,'nano':_A,'fractions':_A,_AW:_A}
class DateStrToDate(Func):0
class DateToDateStr(Func):0
class DateToDi(Func):0
class Date(Func):arg_types={_C:_A,_L:_A,_E:_A};is_var_len_args=_B
class Day(Func):0
class Decode(Func):arg_types={_C:_B,_AK:_B,_z:_A}
class DiToDate(Func):0
class Encode(Func):arg_types={_C:_B,_AK:_B}
class Exp(Func):0
class Explode(Func):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class Inline(Func):0
class ExplodeOuter(Explode):0
class Posexplode(Explode):0
class PosexplodeOuter(Posexplode,ExplodeOuter):0
class Unnest(Func,UDTF):
	arg_types={_E:_B,_I:_A,_N:_A,'explode_array':_A}
	@property
	def selects(self):
		A=super().selects;B=self.args.get(_N)
		if B:A=A+[to_identifier(_N)if B is _B else B]
		return A
class Floor(Func):arg_types={_C:_B,_AJ:_A}
class FromBase64(Func):0
class ToBase64(Func):0
class FromISO8601Timestamp(Func):_sql_names=['FROM_ISO8601_TIMESTAMP']
class GapFill(Func):arg_types={_C:_B,'ts_column':_B,'bucket_width':_B,'partitioning_columns':_A,'value_columns':_A,'origin':_A,'ignore_nulls':_A}
class GenerateDateArray(Func):arg_types={_O:_B,_d:_B,_p:_A}
class GenerateTimestampArray(Func):arg_types={_O:_B,_d:_B,_p:_B}
class Greatest(Func):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class GroupConcat(AggFunc):arg_types={_C:_B,'separator':_A}
class Hex(Func):0
class LowerHex(Hex):0
class Xor(Connector,Func):arg_types={_C:_A,_F:_A,_E:_A}
class If(Func):arg_types={_C:_B,'true':_B,'false':_A};_sql_names=['IF','IIF']
class Nullif(Func):arg_types={_C:_B,_F:_B}
class Initcap(Func):arg_types={_C:_B,_F:_A}
class IsNan(Func):_sql_names=['IS_NAN','ISNAN']
class IsInf(Func):_sql_names=['IS_INF','ISINF']
class JSON(Expression):arg_types={_C:_A,_J:_A,_A0:_A}
class JSONPath(Expression):
	arg_types={_E:_B,'escape':_A}
	@property
	def output_name(self):A=self.expressions[-1].this;return A if isinstance(A,str)else''
class JSONPathPart(Expression):arg_types={}
class JSONPathFilter(JSONPathPart):arg_types={_C:_B}
class JSONPathKey(JSONPathPart):arg_types={_C:_B}
class JSONPathRecursive(JSONPathPart):arg_types={_C:_A}
class JSONPathRoot(JSONPathPart):0
class JSONPathScript(JSONPathPart):arg_types={_C:_B}
class JSONPathSlice(JSONPathPart):arg_types={_O:_A,_d:_A,_p:_A}
class JSONPathSelector(JSONPathPart):arg_types={_C:_B}
class JSONPathSubscript(JSONPathPart):arg_types={_C:_B}
class JSONPathUnion(JSONPathPart):arg_types={_E:_B}
class JSONPathWildcard(JSONPathPart):0
class FormatJson(Expression):0
class JSONKeyValue(Expression):arg_types={_C:_B,_F:_B}
class JSONObject(Func):arg_types={_E:_A,_x:_A,_AX:_A,_i:_A,'encoding':_A}
class JSONObjectAgg(AggFunc):arg_types={_E:_A,_x:_A,_AX:_A,_i:_A,'encoding':_A}
class JSONArray(Func):arg_types={_E:_B,_x:_A,_i:_A,'strict':_A}
class JSONArrayAgg(Func):arg_types={_C:_B,_W:_A,_x:_A,_i:_A,'strict':_A}
class JSONExists(Func):arg_types={_C:_B,_j:_B,'passing':_A,_AY:_A}
class JSONColumnDef(Expression):arg_types={_C:_A,_G:_A,_j:_A,'nested_schema':_A}
class JSONSchema(Expression):arg_types={_E:_B}
class JSONValue(Expression):arg_types={_C:_B,_j:_B,_f:_A,_AY:_A}
class JSONTable(Func):arg_types={_C:_B,'schema':_B,_j:_A,'error_handling':_A,'empty_handling':_A}
class ObjectInsert(Func):arg_types={_C:_B,'key':_B,'value':_B,'update_flag':_A}
class OpenJSONColumnDef(Expression):arg_types={_C:_B,_G:_B,_j:_A,'as_json':_A}
class OpenJSON(Func):arg_types={_C:_B,_j:_A,_E:_A}
class JSONBContains(Binary,Func):_sql_names=['JSONB_CONTAINS']
class JSONExtract(Binary,Func):
	arg_types={_C:_B,_F:_B,_AZ:_A,_E:_A,'variant_extract':_A,'json_query':_A,_A1:_A};_sql_names=['JSON_EXTRACT'];is_var_len_args=_B
	@property
	def output_name(self):return self.expression.output_name if not self.expressions else''
class JSONExtractScalar(Binary,Func):
	arg_types={_C:_B,_F:_B,_AZ:_A,_E:_A};_sql_names=['JSON_EXTRACT_SCALAR'];is_var_len_args=_B
	@property
	def output_name(self):return self.expression.output_name
class JSONBExtract(Binary,Func):_sql_names=['JSONB_EXTRACT']
class JSONBExtractScalar(Binary,Func):_sql_names=['JSONB_EXTRACT_SCALAR']
class JSONFormat(Func):arg_types={_C:_A,_X:_A};_sql_names=['JSON_FORMAT']
class JSONArrayContains(Binary,Predicate,Func):_sql_names=['JSON_ARRAY_CONTAINS']
class ParseJSON(Func):_sql_names=['PARSE_JSON','JSON_PARSE'];arg_types={_C:_B,_F:_A,_Q:_A}
class Least(Func):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class Left(Func):arg_types={_C:_B,_F:_B}
class Right(Func):arg_types={_C:_B,_F:_B}
class Length(Func):arg_types={_C:_B,'binary':_A};_sql_names=['LENGTH','LEN']
class Levenshtein(Func):arg_types={_C:_B,_F:_A,'ins_cost':_A,'del_cost':_A,'sub_cost':_A}
class Ln(Func):0
class Log(Func):arg_types={_C:_B,_F:_A}
class LogicalOr(AggFunc):_sql_names=['LOGICAL_OR','BOOL_OR','BOOLOR_AGG']
class LogicalAnd(AggFunc):_sql_names=['LOGICAL_AND','BOOL_AND','BOOLAND_AGG']
class Lower(Func):_sql_names=['LOWER','LCASE']
class Map(Func):
	arg_types={_y:_A,_q:_A}
	@property
	def keys(self):A=self.args.get(_y);return A.expressions if A else[]
	@property
	def values(self):A=self.args.get(_q);return A.expressions if A else[]
class ToMap(Func):0
class MapFromEntries(Func):0
class ScopeResolution(Expression):arg_types={_C:_A,_F:_B}
class Stream(Expression):0
class StarMap(Func):0
class VarMap(Func):
	arg_types={_y:_B,_q:_B};is_var_len_args=_B
	@property
	def keys(self):return self.args[_y].expressions
	@property
	def values(self):return self.args[_q].expressions
class MatchAgainst(Func):arg_types={_C:_B,_E:_B,'modifier':_A}
class Max(AggFunc):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class MD5(Func):_sql_names=['MD5']
class MD5Digest(Func):_sql_names=['MD5_DIGEST']
class Min(AggFunc):arg_types={_C:_B,_E:_A};is_var_len_args=_B
class Month(Func):0
class AddMonths(Func):arg_types={_C:_B,_F:_B}
class Nvl2(Func):arg_types={_C:_B,'true':_B,'false':_A}
class Normalize(Func):arg_types={_C:_B,'form':_A}
class Overlay(Func):arg_types={_C:_B,_F:_B,_Y:_B,'for':_A}
class Predict(Func):arg_types={_C:_B,_F:_B,'params_struct':_A}
class Pow(Binary,Func):_sql_names=['POWER','POW']
class PercentileCont(AggFunc):arg_types={_C:_B,_F:_A}
class PercentileDisc(AggFunc):arg_types={_C:_B,_F:_A}
class Quantile(AggFunc):arg_types={_C:_B,'quantile':_B}
class ApproxQuantile(Quantile):arg_types={_C:_B,'quantile':_B,'accuracy':_A,'weight':_A}
class Quarter(Func):0
class Rand(Func):_sql_names=['RAND','RANDOM'];arg_types={_C:_A,'lower':_A,'upper':_A}
class Randn(Func):arg_types={_C:_A}
class RangeN(Func):arg_types={_C:_B,_E:_B,'each':_A}
class ReadCSV(Func):_sql_names=['READ_CSV'];is_var_len_args=_B;arg_types={_C:_B,_E:_A}
class Reduce(Func):arg_types={_C:_B,'initial':_B,'merge':_B,'finish':_A}
class RegexpExtract(Func):arg_types={_C:_B,_F:_B,_e:_A,_Aa:_A,'parameters':_A,_m:_A}
class RegexpReplace(Func):arg_types={_C:_B,_F:_B,'replacement':_A,_e:_A,_Aa:_A,'modifiers':_A}
class RegexpLike(Binary,Func):arg_types={_C:_B,_F:_B,'flag':_A}
class RegexpILike(Binary,Func):arg_types={_C:_B,_F:_B,'flag':_A}
class RegexpSplit(Func):arg_types={_C:_B,_F:_B,_c:_A}
class Repeat(Func):arg_types={_C:_B,'times':_B}
class Round(Func):arg_types={_C:_B,_AJ:_A,'truncate':_A}
class RowNumber(Func):arg_types:t.Dict[str,t.Any]={}
class SafeDivide(Func):arg_types={_C:_B,_F:_B}
class SHA(Func):_sql_names=['SHA','SHA1']
class SHA2(Func):_sql_names=['SHA2'];arg_types={_C:_B,_AL:_A}
class Sign(Func):_sql_names=['SIGN','SIGNUM']
class SortArray(Func):arg_types={_C:_B,'asc':_A}
class Split(Func):arg_types={_C:_B,_F:_B,_c:_A}
class Substring(Func):_sql_names=['SUBSTRING','SUBSTR'];arg_types={_C:_B,_O:_A,_AL:_A}
class StandardHash(Func):arg_types={_C:_B,_F:_A}
class StartsWith(Func):_sql_names=['STARTS_WITH','STARTSWITH'];arg_types={_C:_B,_F:_B}
class StrPosition(Func):arg_types={_C:_B,'substr':_B,_e:_A,'instance':_A}
class StrToDate(Func):arg_types={_C:_B,_K:_A,_Q:_A}
class StrToTime(Func):arg_types={_C:_B,_K:_B,_L:_A,_Q:_A}
class StrToUnix(Func):arg_types={_C:_A,_K:_A}
class StrToMap(Func):arg_types={_C:_B,'pair_delim':_A,'key_value_delim':_A,'duplicate_resolution_callback':_A}
class NumberToStr(Func):arg_types={_C:_B,_K:_B,'culture':_A}
class FromBase(Func):arg_types={_C:_B,_F:_B}
class Struct(Func):arg_types={_E:_A};is_var_len_args=_B
class StructExtract(Func):arg_types={_C:_B,_F:_B}
class Stuff(Func):_sql_names=['STUFF','INSERT'];arg_types={_C:_B,_O:_B,_AL:_B,_F:_B}
class Sum(AggFunc):0
class Sqrt(Func):0
class Stddev(AggFunc):_sql_names=['STDDEV','STDEV']
class StddevPop(AggFunc):0
class StddevSamp(AggFunc):0
class Time(Func):arg_types={_C:_A,_L:_A}
class TimeToStr(Func):arg_types={_C:_B,_K:_B,'culture':_A,_L:_A}
class TimeToTimeStr(Func):0
class TimeToUnix(Func):0
class TimeStrToDate(Func):0
class TimeStrToTime(Func):arg_types={_C:_B,_L:_A}
class TimeStrToUnix(Func):0
class Trim(Func):arg_types={_C:_B,_F:_A,_e:_A,'collation':_A}
class TsOrDsAdd(Func,TimeUnit):
	arg_types={_C:_B,_F:_B,_H:_A,_i:_A}
	@property
	def return_type(self):return DataType.build(self.args.get(_i)or DataType.Type.DATE)
class TsOrDsDiff(Func,TimeUnit):arg_types={_C:_B,_F:_B,_H:_A}
class TsOrDsToDateStr(Func):0
class TsOrDsToDate(Func):arg_types={_C:_B,_K:_A,_Q:_A}
class TsOrDsToTime(Func):0
class TsOrDsToTimestamp(Func):0
class TsOrDiToDi(Func):0
class Unhex(Func):0
class UnixDate(Func):0
class UnixToStr(Func):arg_types={_C:_B,_K:_A}
class UnixToTime(Func):arg_types={_C:_B,'scale':_A,_L:_A,'hours':_A,'minutes':_A,_K:_A};SECONDS=Literal.number(0);DECIS=Literal.number(1);CENTIS=Literal.number(2);MILLIS=Literal.number(3);DECIMILLIS=Literal.number(4);CENTIMILLIS=Literal.number(5);MICROS=Literal.number(6);DECIMICROS=Literal.number(7);CENTIMICROS=Literal.number(8);NANOS=Literal.number(9)
class UnixToTimeStr(Func):0
class Uuid(Func):_sql_names=['UUID','GEN_RANDOM_UUID','GENERATE_UUID','UUID_STRING'];arg_types={_C:_A,'name':_A}
class TimestampFromParts(Func):_sql_names=['TIMESTAMP_FROM_PARTS','TIMESTAMPFROMPARTS'];arg_types={'year':_B,'month':_B,'day':_B,'hour':_B,'min':_B,'sec':_B,'nano':_A,_L:_A,'milli':_A}
class Upper(Func):_sql_names=['UPPER','UCASE']
class Corr(Binary,AggFunc):0
class Variance(AggFunc):_sql_names=['VARIANCE','VARIANCE_SAMP','VAR_SAMP']
class VariancePop(AggFunc):_sql_names=['VARIANCE_POP','VAR_POP']
class CovarSamp(Binary,AggFunc):0
class CovarPop(Binary,AggFunc):0
class Week(Func):arg_types={_C:_B,'mode':_A}
class XMLTable(Func):arg_types={_C:_B,'passing':_A,_a:_A,'by_ref':_A}
class Year(Func):0
class Use(Expression):arg_types={_C:_B,_G:_A}
class Merge(DML):arg_types={_C:_B,_T:_B,_U:_B,_E:_B,_J:_A,_f:_A}
class When(Func):arg_types={'matched':_B,_u:_A,_AQ:_A,'then':_B}
class NextValueFor(Func):arg_types={_C:_B,_W:_A}
class Semicolon(Expression):arg_types={}
def _norm_arg(arg):A=arg;return A.lower()if type(A)is str else A
ALL_FUNCTIONS=subclasses(__name__,Func,(AggFunc,Anonymous,Func))
FUNCTION_BY_NAME={B:A for A in ALL_FUNCTIONS for B in A.sql_names()}
JSON_PATH_PARTS=subclasses(__name__,JSONPathPart,(JSONPathPart,))
PERCENTILES=PercentileCont,PercentileDisc
@t.overload
def maybe_parse(sql_or_expression,*,into,dialect=_D,prefix=_D,copy=_A,**A):...
@t.overload
def maybe_parse(sql_or_expression,*,into=_D,dialect=_D,prefix=_D,copy=_A,**A):...
def maybe_parse(sql_or_expression,*,into=_D,dialect=_D,prefix=_D,copy=_A,**D):
	'Gracefully handle a possible string or expression.\n\n    Example:\n        >>> maybe_parse("1")\n        Literal(this=1, is_string=False)\n        >>> maybe_parse(to_identifier("x"))\n        Identifier(this=x, quoted=False)\n\n    Args:\n        sql_or_expression: the SQL code string or an expression\n        into: the SQLGlot Expression to parse into\n        dialect: the dialect used to parse the input expressions (in the case that an\n            input expression is a SQL string).\n        prefix: a string to prefix the sql with before it gets parsed\n            (automatically includes a space)\n        copy: whether to copy the expression.\n        **opts: other options to use to parse the input expressions (again, in the case\n            that an input expression is a SQL string).\n\n    Returns:\n        Expression: the parsed or given expression.\n    ';C=prefix;A=sql_or_expression
	if isinstance(A,Expression):
		if copy:return A.copy()
		return A
	if A is _D:raise ParseError('SQL cannot be None')
	import sqlglot as E;B=str(A)
	if C:B=f"{C} {B}"
	return E.parse_one(B,read=dialect,into=into,**D)
@t.overload
def maybe_copy(instance,copy=_B):...
@t.overload
def maybe_copy(instance,copy=_B):...
def maybe_copy(instance,copy=_B):A=instance;return A.copy()if copy and A else A
def _to_s(node,verbose=_A,level=0):
	'Generate a textual representation of an Expression tree';F=level;B=verbose;A=node;D='\n'+'  '*(F+1);G=f",{D}"
	if isinstance(A,Expression):
		E={C:A for(C,A)in A.args.items()if A is not _D and A!=[]or B}
		if(A.type or B)and not isinstance(A,DataType):E['_type']=A.type
		if A.comments or B:E['_comments']=A.comments
		if B:E['_id']=id(A)
		if A.is_leaf():D='';G=', '
		C=G.join([f"{A}={_to_s(C,B,F+1)}"for(A,C)in E.items()]);return f"{A.__class__.__name__}({D}{C})"
	if isinstance(A,list):C=G.join(_to_s(A,B,F+1)for A in A);C=f"{D}{C}"if C else'';return f"[{C}]"
	return D.join(textwrap.dedent(str(A).strip('\n')).splitlines())
def _is_wrong_expression(expression,into):A=expression;return isinstance(A,Expression)and not isinstance(A,into)
def _apply_builder(expression,instance,arg,copy=_B,prefix=_D,into=_D,dialect=_D,into_arg=_C,**D):
	C=into;B=instance;A=expression
	if _is_wrong_expression(A,C):A=C(**{into_arg:A})
	B=maybe_copy(B,copy);A=maybe_parse(sql_or_expression=A,prefix=prefix,into=C,dialect=dialect,**D);B.set(arg,A);return B
def _apply_child_list_builder(*J,instance,arg,append=_B,copy=_B,prefix=_D,into=_D,dialect=_D,properties=_D,**K):
	D=into;C=properties;B=instance;B=maybe_copy(B,copy);E=[];C={}if C is _D else C
	for A in J:
		if A is not _D:
			if _is_wrong_expression(A,D):A=D(expressions=[A])
			A=maybe_parse(A,into=D,dialect=dialect,prefix=prefix,**K)
			for(F,G)in A.args.items():
				if F==_E:E.extend(G)
				else:C[F]=G
	H=B.args.get(arg)
	if append and H:E=H.expressions+E
	I=D(expressions=E)
	for(F,G)in C.items():I.set(F,G)
	B.set(arg,I);return B
def _apply_list_builder(*A,instance,arg,append=_B,copy=_B,prefix=_D,into=_D,dialect=_D,**D):
	B=maybe_copy(instance,copy);A=[maybe_parse(sql_or_expression=A,into=into,prefix=prefix,dialect=dialect,**D)for A in A if A is not _D];C=B.args.get(arg)
	if append and C:A=C+A
	B.set(arg,A);return B
def _apply_conjunction_builder(*A,instance,arg,into=_D,append=_B,copy=_B,dialect=_D,**G):
	E=instance;B=into;A=[A for A in A if A is not _D and A!='']
	if not A:return E
	C=maybe_copy(E,copy);D=C.args.get(arg)
	if append and D is not _D:A=[D.this if B else D]+list(A)
	F=and_(*A,dialect=dialect,copy=copy,**G);C.set(arg,B(this=F)if B else F);return C
def _apply_cte_builder(instance,alias,as_,recursive=_D,materialized=_D,append=_B,dialect=_D,copy=_B,**B):A=dialect;C=maybe_parse(alias,dialect=A,into=TableAlias,**B);D=maybe_parse(as_,dialect=A,**B);E=CTE(this=D,alias=C,materialized=materialized);return _apply_child_list_builder(E,instance=instance,arg=_J,append=append,copy=copy,into=With,properties={_A3:recursive or _A})
def _combine(expressions,operator,dialect=_D,copy=_B,**C):
	D=[condition(A,dialect=dialect,copy=copy,**C)for A in expressions if A is not _D];A,*B=D
	if B:A=_wrap(A,Connector)
	for E in B:A=operator(this=A,expression=_wrap(E,Connector))
	return A
def _wrap(expression,kind):A=expression;return Paren(this=A)if isinstance(A,kind)else A
def union(left,right,distinct=_B,dialect=_D,copy=_B,**D):'\n    Initializes a syntax tree from one UNION expression.\n\n    Example:\n        >>> union("SELECT * FROM foo", "SELECT * FROM bla").sql()\n        \'SELECT * FROM foo UNION SELECT * FROM bla\'\n\n    Args:\n        left: the SQL code string corresponding to the left-hand side.\n            If an `Expression` instance is passed, it will be used as-is.\n        right: the SQL code string corresponding to the right-hand side.\n            If an `Expression` instance is passed, it will be used as-is.\n        distinct: set the DISTINCT flag if and only if this is true.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy the expression.\n        opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new Union instance.\n    ';C=dialect;B=right;A=left;A=maybe_parse(sql_or_expression=A,dialect=C,copy=copy,**D);B=maybe_parse(sql_or_expression=B,dialect=C,copy=copy,**D);return Union(this=A,expression=B,distinct=distinct)
def intersect(left,right,distinct=_B,dialect=_D,copy=_B,**D):'\n    Initializes a syntax tree from one INTERSECT expression.\n\n    Example:\n        >>> intersect("SELECT * FROM foo", "SELECT * FROM bla").sql()\n        \'SELECT * FROM foo INTERSECT SELECT * FROM bla\'\n\n    Args:\n        left: the SQL code string corresponding to the left-hand side.\n            If an `Expression` instance is passed, it will be used as-is.\n        right: the SQL code string corresponding to the right-hand side.\n            If an `Expression` instance is passed, it will be used as-is.\n        distinct: set the DISTINCT flag if and only if this is true.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy the expression.\n        opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new Intersect instance.\n    ';C=dialect;B=right;A=left;A=maybe_parse(sql_or_expression=A,dialect=C,copy=copy,**D);B=maybe_parse(sql_or_expression=B,dialect=C,copy=copy,**D);return Intersect(this=A,expression=B,distinct=distinct)
def except_(left,right,distinct=_B,dialect=_D,copy=_B,**D):'\n    Initializes a syntax tree from one EXCEPT expression.\n\n    Example:\n        >>> except_("SELECT * FROM foo", "SELECT * FROM bla").sql()\n        \'SELECT * FROM foo EXCEPT SELECT * FROM bla\'\n\n    Args:\n        left: the SQL code string corresponding to the left-hand side.\n            If an `Expression` instance is passed, it will be used as-is.\n        right: the SQL code string corresponding to the right-hand side.\n            If an `Expression` instance is passed, it will be used as-is.\n        distinct: set the DISTINCT flag if and only if this is true.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy the expression.\n        opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new Except instance.\n    ';C=dialect;B=right;A=left;A=maybe_parse(sql_or_expression=A,dialect=C,copy=copy,**D);B=maybe_parse(sql_or_expression=B,dialect=C,copy=copy,**D);return Except(this=A,expression=B,distinct=distinct)
def select(*A,dialect=_D,**B):'\n    Initializes a syntax tree from one or multiple SELECT expressions.\n\n    Example:\n        >>> select("col1", "col2").from_("tbl").sql()\n        \'SELECT col1, col2 FROM tbl\'\n\n    Args:\n        *expressions: the SQL code string to parse as the expressions of a\n            SELECT statement. If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expressions (in the case that an\n            input expression is a SQL string).\n        **opts: other options to use to parse the input expressions (again, in the case\n            that an input expression is a SQL string).\n\n    Returns:\n        Select: the syntax tree for the SELECT statement.\n    ';return Select().select(*A,dialect=dialect,**B)
def from_(expression,dialect=_D,**A):'\n    Initializes a syntax tree from a FROM expression.\n\n    Example:\n        >>> from_("tbl").select("col1", "col2").sql()\n        \'SELECT col1, col2 FROM tbl\'\n\n    Args:\n        *expression: the SQL code string to parse as the FROM expressions of a\n            SELECT statement. If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expression (in the case that the\n            input expression is a SQL string).\n        **opts: other options to use to parse the input expressions (again, in the case\n            that the input expression is a SQL string).\n\n    Returns:\n        Select: the syntax tree for the SELECT statement.\n    ';return Select().from_(expression,dialect=dialect,**A)
def update(table,properties=_D,where=_D,from_=_D,with_=_D,dialect=_D,**D):
	'\n    Creates an update statement.\n\n    Example:\n        >>> update("my_table", {"x": 1, "y": "2", "z": None}, from_="baz_cte", where="baz_cte.id > 1 and my_table.id = baz_cte.id", with_={"baz_cte": "SELECT id FROM foo"}).sql()\n        "WITH baz_cte AS (SELECT id FROM foo) UPDATE my_table SET x = 1, y = \'2\', z = NULL FROM baz_cte WHERE baz_cte.id > 1 AND my_table.id = baz_cte.id"\n\n    Args:\n        properties: dictionary of properties to SET which are\n            auto converted to sql objects eg None -> NULL\n        where: sql conditional parsed into a WHERE statement\n        from_: sql statement parsed into a FROM statement\n        with_: dictionary of CTE aliases / select statements to include in a WITH clause.\n        dialect: the dialect used to parse the input expressions.\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        Update: the syntax tree for the UPDATE statement.\n    ';G=with_;F=from_;E=properties;B=dialect;A=where;C=Update(this=maybe_parse(table,into=Table,dialect=B))
	if E:C.set(_E,[EQ(this=maybe_parse(A,dialect=B,**D),expression=convert(C))for(A,C)in E.items()])
	if F:C.set(_Y,maybe_parse(F,into=From,dialect=B,prefix='FROM',**D))
	if isinstance(A,Condition):A=Where(this=A)
	if A:C.set(_P,maybe_parse(A,into=Where,dialect=B,prefix='WHERE',**D))
	if G:H=[alias_(CTE(this=maybe_parse(C,dialect=B,**D)),A,table=_B)for(A,C)in G.items()];C.set(_J,With(expressions=H))
	return C
def delete(table,where=_D,returning=_D,dialect=_D,**C):
	'\n    Builds a delete statement.\n\n    Example:\n        >>> delete("my_table", where="id > 1").sql()\n        \'DELETE FROM my_table WHERE id > 1\'\n\n    Args:\n        where: sql conditional parsed into a WHERE statement\n        returning: sql conditional parsed into a RETURNING statement\n        dialect: the dialect used to parse the input expressions.\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        Delete: the syntax tree for the DELETE statement.\n    ';E=returning;D=where;B=dialect;A=Delete().delete(table,dialect=B,copy=_A,**C)
	if D:A=A.where(D,dialect=B,copy=_A,**C)
	if E:A=A.returning(E,dialect=B,copy=_A,**C)
	return A
def insert(expression,into,columns=_D,overwrite=_D,returning=_D,dialect=_D,copy=_B,**C):
	'\n    Builds an INSERT statement.\n\n    Example:\n        >>> insert("VALUES (1, 2, 3)", "tbl").sql()\n        \'INSERT INTO tbl VALUES (1, 2, 3)\'\n\n    Args:\n        expression: the sql string or expression of the INSERT statement\n        into: the tbl to insert data to.\n        columns: optionally the table\'s column names.\n        overwrite: whether to INSERT OVERWRITE or not.\n        returning: sql conditional parsed into a RETURNING statement\n        dialect: the dialect used to parse the input expressions.\n        copy: whether to copy the expression.\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        Insert: the syntax tree for the INSERT statement.\n    ';G=returning;F=columns;B=copy;A=dialect;H=maybe_parse(expression,dialect=A,copy=B,**C);D=maybe_parse(into,into=Table,dialect=A,copy=B,**C)
	if F:D=Schema(this=D,expressions=[to_identifier(A,copy=B)for A in F])
	E=Insert(this=D,expression=H,overwrite=overwrite)
	if G:E=E.returning(G,dialect=A,copy=_A,**C)
	return E
def merge(*F,into,using,on,returning=_D,dialect=_D,copy=_B,**B):
	'\n    Builds a MERGE statement.\n\n    Example:\n        >>> merge("WHEN MATCHED THEN UPDATE SET col1 = source_table.col1",\n        ...       "WHEN NOT MATCHED THEN INSERT (col1) VALUES (source_table.col1)",\n        ...       into="my_table",\n        ...       using="source_table",\n        ...       on="my_table.id = source_table.id").sql()\n        \'MERGE INTO my_table USING source_table ON my_table.id = source_table.id WHEN MATCHED THEN UPDATE SET col1 = source_table.col1 WHEN NOT MATCHED THEN INSERT (col1) VALUES (source_table.col1)\'\n\n    Args:\n        *when_exprs: The WHEN clauses specifying actions for matched and unmatched rows.\n        into: The target table to merge data into.\n        using: The source table to merge data from.\n        on: The join condition for the merge.\n        returning: The columns to return from the merge.\n        dialect: The dialect used to parse the input expressions.\n        copy: Whether to copy the expression.\n        **opts: Other options to use to parse the input expressions.\n\n    Returns:\n        Merge: The syntax tree for the MERGE statement.\n    ';E=returning;C=copy;A=dialect;D=Merge(this=maybe_parse(into,dialect=A,copy=C,**B),using=maybe_parse(using,dialect=A,copy=C,**B),on=maybe_parse(on,dialect=A,copy=C,**B),expressions=[maybe_parse(D,dialect=A,copy=C,into=When,**B)for D in F])
	if E:D=D.returning(E,dialect=A,copy=_A,**B)
	return D
def condition(expression,dialect=_D,copy=_B,**A):'\n    Initialize a logical condition expression.\n\n    Example:\n        >>> condition("x=1").sql()\n        \'x = 1\'\n\n        This is helpful for composing larger logical syntax trees:\n        >>> where = condition("x=1")\n        >>> where = where.and_("y=1")\n        >>> Select().from_("tbl").select("*").where(where).sql()\n        \'SELECT * FROM tbl WHERE x = 1 AND y = 1\'\n\n    Args:\n        *expression: the SQL code string to parse.\n            If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expression (in the case that the\n            input expression is a SQL string).\n        copy: Whether to copy `expression` (only applies to expressions).\n        **opts: other options to use to parse the input expressions (again, in the case\n            that the input expression is a SQL string).\n\n    Returns:\n        The new Condition instance\n    ';return maybe_parse(expression,into=Condition,dialect=dialect,copy=copy,**A)
def and_(*A,dialect=_D,copy=_B,**B):'\n    Combine multiple conditions with an AND logical operator.\n\n    Example:\n        >>> and_("x=1", and_("y=1", "z=1")).sql()\n        \'x = 1 AND (y = 1 AND z = 1)\'\n\n    Args:\n        *expressions: the SQL code strings to parse.\n            If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy `expressions` (only applies to Expressions).\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new condition\n    ';return t.cast(Condition,_combine(A,And,dialect,copy=copy,**B))
def or_(*A,dialect=_D,copy=_B,**B):'\n    Combine multiple conditions with an OR logical operator.\n\n    Example:\n        >>> or_("x=1", or_("y=1", "z=1")).sql()\n        \'x = 1 OR (y = 1 OR z = 1)\'\n\n    Args:\n        *expressions: the SQL code strings to parse.\n            If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy `expressions` (only applies to Expressions).\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new condition\n    ';return t.cast(Condition,_combine(A,Or,dialect,copy=copy,**B))
def xor(*A,dialect=_D,copy=_B,**B):'\n    Combine multiple conditions with an XOR logical operator.\n\n    Example:\n        >>> xor("x=1", xor("y=1", "z=1")).sql()\n        \'x = 1 XOR (y = 1 XOR z = 1)\'\n\n    Args:\n        *expressions: the SQL code strings to parse.\n            If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy `expressions` (only applies to Expressions).\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new condition\n    ';return t.cast(Condition,_combine(A,Xor,dialect,copy=copy,**B))
def not_(expression,dialect=_D,copy=_B,**A):'\n    Wrap a condition with a NOT operator.\n\n    Example:\n        >>> not_("this_suit=\'black\'").sql()\n        "NOT this_suit = \'black\'"\n\n    Args:\n        expression: the SQL code string to parse.\n            If an Expression instance is passed, this is used as-is.\n        dialect: the dialect used to parse the input expression.\n        copy: whether to copy the expression or not.\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        The new condition.\n    ';B=condition(expression,dialect=dialect,copy=copy,**A);return Not(this=_wrap(B,Connector))
def paren(expression,copy=_B):'\n    Wrap an expression in parentheses.\n\n    Example:\n        >>> paren("5 + 3").sql()\n        \'(5 + 3)\'\n\n    Args:\n        expression: the SQL code string to parse.\n            If an Expression instance is passed, this is used as-is.\n        copy: whether to copy the expression or not.\n\n    Returns:\n        The wrapped expression.\n    ';return Paren(this=maybe_parse(expression,copy=copy))
SAFE_IDENTIFIER_RE=re.compile('^[_a-zA-Z][\\w]*$')
@t.overload
def to_identifier(name,quoted=_D,copy=_B):...
@t.overload
def to_identifier(name,quoted=_D,copy=_B):...
def to_identifier(name,quoted=_D,copy=_B):
	"Builds an identifier.\n\n    Args:\n        name: The name to turn into an identifier.\n        quoted: Whether to force quote the identifier.\n        copy: Whether to copy name if it's an Identifier.\n\n    Returns:\n        The identifier ast node.\n    ";B=quoted;A=name
	if A is _D:return
	if isinstance(A,Identifier):C=maybe_copy(A,copy)
	elif isinstance(A,str):C=Identifier(this=A,quoted=not SAFE_IDENTIFIER_RE.match(A)if B is _D else B)
	else:raise ValueError(f"Name needs to be a string or an Identifier, got: {A.__class__}")
	return C
def parse_identifier(name,dialect=_D):
	'\n    Parses a given string into an identifier.\n\n    Args:\n        name: The name to parse into an identifier.\n        dialect: The dialect to parse against.\n\n    Returns:\n        The identifier ast node.\n    '
	try:A=maybe_parse(name,dialect=dialect,into=Identifier)
	except(ParseError,TokenError):A=to_identifier(name)
	return A
INTERVAL_STRING_RE=re.compile('\\s*([0-9]+)\\s*([a-zA-Z]+)\\s*')
def to_interval(interval):
	"Builds an interval expression from a string like '1 day' or '5 months'.";C='Invalid interval string.';A=interval
	if isinstance(A,Literal):
		if not A.is_string:raise ValueError(C)
		A=A.this
	B=INTERVAL_STRING_RE.match(A)
	if not B:raise ValueError(C)
	return Interval(this=Literal.string(B.group(1)),unit=Var(this=B.group(2).upper()))
def to_table(sql_path,dialect=_D,copy=_B,**C):
	'\n    Create a table expression from a `[catalog].[schema].[table]` sql path. Catalog and schema are optional.\n    If a table is passed in then that table is returned.\n\n    Args:\n        sql_path: a `[catalog].[schema].[table]` string.\n        dialect: the source dialect according to which the table name will be parsed.\n        copy: Whether to copy a table if it is passed in.\n        kwargs: the kwargs to instantiate the resulting `Table` expression with.\n\n    Returns:\n        A table expression.\n    ';A=sql_path
	if isinstance(A,Table):return maybe_copy(A,copy=copy)
	B=maybe_parse(A,into=Table,dialect=dialect)
	for(D,E)in C.items():B.set(D,E)
	return B
def to_column(sql_path,quoted=_D,dialect=_D,copy=_B,**D):
	'\n    Create a column from a `[table].[column]` sql path. Table is optional.\n    If a column is passed in then that column is returned.\n\n    Args:\n        sql_path: a `[table].[column]` string.\n        quoted: Whether or not to force quote identifiers.\n        dialect: the source dialect according to which the column name will be parsed.\n        copy: Whether to copy a column if it is passed in.\n        kwargs: the kwargs to instantiate the resulting `Column` expression with.\n\n    Returns:\n        A column expression.\n    ';C=quoted;A=sql_path
	if isinstance(A,Column):return maybe_copy(A,copy=copy)
	try:B=maybe_parse(A,into=Column,dialect=dialect)
	except ParseError:return column(*reversed(A.split('.')),quoted=C,**D)
	for(E,F)in D.items():B.set(E,F)
	if C:
		for G in B.find_all(Identifier):G.set(_AA,_B)
	return B
def alias_(expression,alias,table=_A,quoted=_D,dialect=_D,copy=_B,**F):
	"Create an Alias expression.\n\n    Example:\n        >>> alias_('foo', 'bar').sql()\n        'foo AS bar'\n\n        >>> alias_('(select 1, 2)', 'bar', table=['a', 'b']).sql()\n        '(SELECT 1, 2) AS bar(a, b)'\n\n    Args:\n        expression: the SQL code strings to parse.\n            If an Expression instance is passed, this is used as-is.\n        alias: the alias name to use. If the name has\n            special characters it is quoted.\n        table: Whether to create a table alias, can also be a list of columns.\n        quoted: whether to quote the alias\n        dialect: the dialect used to parse the input expression.\n        copy: Whether to copy the expression.\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        Alias: the aliased expression\n    ";D=quoted;C=table;B=alias;A=maybe_parse(expression,dialect=dialect,copy=copy,**F);B=to_identifier(B,quoted=D)
	if C:
		E=TableAlias(this=B);A.set(_I,E)
		if not isinstance(C,bool):
			for G in C:E.append(_a,to_identifier(G,quoted=D))
		return A
	if _I in A.arg_types and not isinstance(A,Window):A.set(_I,B);return A
	return Alias(this=A,alias=B)
def subquery(expression,alias=_D,dialect=_D,**B):"\n    Build a subquery expression that's selected from.\n\n    Example:\n        >>> subquery('select x from tbl', 'bar').select('x').sql()\n        'SELECT x FROM (SELECT x FROM tbl) AS bar'\n\n    Args:\n        expression: the SQL code strings to parse.\n            If an Expression instance is passed, this is used as-is.\n        alias: the alias name to use.\n        dialect: the dialect used to parse the input expression.\n        **opts: other options to use to parse the input expressions.\n\n    Returns:\n        A new Select instance with the subquery expression included.\n    ";C=dialect;A=expression;A=maybe_parse(A,dialect=C,**B).subquery(alias,**B);return Select().from_(A,dialect=C,**B)
@t.overload
def column(col,table=_D,db=_D,catalog=_D,*,fields,quoted=_D,copy=_B):0
@t.overload
def column(col,table=_D,db=_D,catalog=_D,*,fields=_D,quoted=_D,copy=_B):0
def column(col,table=_D,db=_D,catalog=_D,*,fields=_D,quoted=_D,copy=_B):
	"\n    Build a Column.\n\n    Args:\n        col: Column name.\n        table: Table name.\n        db: Database name.\n        catalog: Catalog name.\n        fields: Additional fields using dots.\n        quoted: Whether to force quotes on the column's identifiers.\n        copy: Whether to copy identifiers if passed in.\n\n    Returns:\n        The new Column instance.\n    ";D=fields;B=copy;A=quoted;C=Column(this=to_identifier(col,quoted=A,copy=B),table=to_identifier(table,quoted=A,copy=B),db=to_identifier(db,quoted=A,copy=B),catalog=to_identifier(catalog,quoted=A,copy=B))
	if D:C=Dot.build((C,*(to_identifier(C,quoted=A,copy=B)for C in D)))
	return C
def cast(expression,to,copy=_B,dialect=_D,**D):
	"Cast an expression to a data type.\n\n    Example:\n        >>> cast('x + 1', 'int').sql()\n        'CAST(x + 1 AS INT)'\n\n    Args:\n        expression: The expression to cast.\n        to: The datatype to cast to.\n        copy: Whether to copy the supplied expressions.\n        dialect: The target dialect. This is used to prevent a re-cast in the following scenario:\n            - The expression to be cast is already a exp.Cast expression\n            - The existing cast is to a type that is logically equivalent to new type\n\n            For example, if :expression='CAST(x as DATETIME)' and :to=Type.TIMESTAMP,\n            but in the target dialect DATETIME is mapped to TIMESTAMP, then we will NOT return `CAST(x (as DATETIME) as TIMESTAMP)`\n            and instead just return the original expression `CAST(x as DATETIME)`.\n\n            This is to prevent it being output as a double cast `CAST(x (as TIMESTAMP) as TIMESTAMP)` once the DATETIME -> TIMESTAMP\n            mapping is applied in the target dialect generator.\n\n    Returns:\n        The new Cast instance.\n    ";C=dialect;A=maybe_parse(expression,copy=copy,dialect=C,**D);B=DataType.build(to,copy=copy,dialect=C,**D)
	if isinstance(A,Cast):
		from sqlglot.dialects.dialect import Dialect as H;I=H.get_or_raise(C);E=I.generator_class.TYPE_MAPPING;F=A.to.this;G=B.this;J=E.get(F,F)==E.get(G,G)
		if A.is_type(B)or J:return A
	A=Cast(this=A,to=B);A.type=B;return A
def table_(table,db=_D,catalog=_D,quoted=_D,alias=_D):"Build a Table.\n\n    Args:\n        table: Table name.\n        db: Database name.\n        catalog: Catalog name.\n        quote: Whether to force quotes on the table's identifiers.\n        alias: Table's alias.\n\n    Returns:\n        The new Table instance.\n    ";D=alias;C=catalog;B=table;A=quoted;return Table(this=to_identifier(B,quoted=A)if B else _D,db=to_identifier(db,quoted=A)if db else _D,catalog=to_identifier(C,quoted=A)if C else _D,alias=TableAlias(this=to_identifier(D))if D else _D)
def values(values,alias=_D,columns=_D):
	'Build VALUES statement.\n\n    Example:\n        >>> values([(1, \'2\')]).sql()\n        "VALUES (1, \'2\')"\n\n    Args:\n        values: values statements that will be converted to SQL\n        alias: optional alias\n        columns: Optional list of ordered column names or ordered dictionary of column names to types.\n         If either are provided then an alias is also required.\n\n    Returns:\n        Values: the Values expression object\n    ';B=columns;A=alias
	if B and not A:raise ValueError('Alias is required when providing columns')
	return Values(expressions=[convert(A)for A in values],alias=TableAlias(this=to_identifier(A),columns=[to_identifier(A)for A in B])if B else TableAlias(this=to_identifier(A))if A else _D)
def var(name):
	"Build a SQL variable.\n\n    Example:\n        >>> repr(var('x'))\n        'Var(this=x)'\n\n        >>> repr(var(column('x', table='y')))\n        'Var(this=x)'\n\n    Args:\n        name: The name of the var or an expression who's name will become the var.\n\n    Returns:\n        The new variable node.\n    ";A=name
	if not A:raise ValueError('Cannot convert empty name into var.')
	if isinstance(A,Expression):A=A.name
	return Var(this=A)
def rename_table(old_name,new_name,dialect=_D):'Build ALTER TABLE... RENAME... expression\n\n    Args:\n        old_name: The old name of the table\n        new_name: The new name of the table\n        dialect: The dialect to parse the table.\n\n    Returns:\n        Alter table expression\n    ';A=dialect;B=to_table(old_name,dialect=A);C=to_table(new_name,dialect=A);return Alter(this=B,kind=_AH,actions=[RenameTable(this=C)])
def rename_column(table_name,old_column_name,new_column_name,exists=_D,dialect=_D):'Build ALTER TABLE... RENAME COLUMN... expression\n\n    Args:\n        table_name: Name of the table\n        old_column: The old name of the column\n        new_column: The new name of the column\n        exists: Whether to add the `IF EXISTS` clause\n        dialect: The dialect to parse the table/column.\n\n    Returns:\n        Alter table expression\n    ';A=dialect;B=to_table(table_name,dialect=A);C=to_column(old_column_name,dialect=A);D=to_column(new_column_name,dialect=A);return Alter(this=B,kind=_AH,actions=[RenameColumn(this=C,to=D,exists=exists)])
def convert(value,copy=_A):
	'Convert a python value into an expression object.\n\n    Raises an error if a conversion is not possible.\n\n    Args:\n        value: A python object.\n        copy: Whether to copy `value` (only applies to Expressions and collections).\n\n    Returns:\n        The equivalent expression object.\n    ';B=copy;A=value
	if isinstance(A,Expression):return maybe_copy(A,B)
	if isinstance(A,str):return Literal.string(A)
	if isinstance(A,bool):return Boolean(this=A)
	if A is _D or isinstance(A,float)and math.isnan(A):return null()
	if isinstance(A,numbers.Number):return Literal.number(A)
	if isinstance(A,bytes):return HexString(this=A.hex())
	if isinstance(A,datetime.datetime):
		D=Literal.string(A.isoformat(sep=' '));C=_D
		if A.tzinfo:C=Literal.string(str(A.tzinfo))
		return TimeStrToTime(this=D,zone=C)
	if isinstance(A,datetime.date):E=Literal.string(A.strftime('%Y-%m-%d'));return DateStrToDate(this=E)
	if isinstance(A,tuple):
		if hasattr(A,'_fields'):return Struct(expressions=[PropertyEQ(this=to_identifier(C),expression=convert(getattr(A,C),copy=B))for C in A._fields])
		return Tuple(expressions=[convert(A,copy=B)for A in A])
	if isinstance(A,list):return Array(expressions=[convert(A,copy=B)for A in A])
	if isinstance(A,dict):return Map(keys=Array(expressions=[convert(A,copy=B)for A in A]),values=Array(expressions=[convert(A,copy=B)for A in A.values()]))
	if hasattr(A,'__dict__'):return Struct(expressions=[PropertyEQ(this=to_identifier(A),expression=convert(C,copy=B))for(A,C)in A.__dict__.items()])
	raise ValueError(f"Cannot convert {A}")
def replace_children(expression,fun,*F,**G):
	'\n    Replace children of an expression with the result of a lambda fun(child) -> exp.\n    ';D=expression
	for(H,B)in tuple(D.args.items()):
		E=type(B)is list;I=B if E else[B];A=[]
		for C in I:
			if isinstance(C,Expression):
				for J in ensure_collection(fun(C,*F,**G)):A.append(J)
			else:A.append(C)
		D.set(H,A if E else seq_get(A,0))
def replace_tree(expression,fun,prune=_D):
	'\n    Replace an entire tree with the result of function calls on each node.\n\n    This will be traversed in reverse dfs, so leaves first.\n    If new nodes are created as a result of function calls, they will also be traversed.\n    ';B=list(expression.dfs(prune=prune))
	while B:
		C=B.pop();A=fun(C)
		if A is not C:
			C.replace(A)
			if isinstance(A,Expression):B.append(A)
	return A
def column_table_names(expression,exclude=''):'\n    Return all table names referenced through columns in an expression.\n\n    Example:\n        >>> import sqlglot\n        >>> sorted(column_table_names(sqlglot.parse_one("a.b AND c.d AND c.e")))\n        [\'a\', \'c\']\n\n    Args:\n        expression: expression to find table names.\n        exclude: a table name to exclude\n\n    Returns:\n        A list of unique names.\n    ';return{A for A in(A.table for A in expression.find_all(Column))if A and A!=exclude}
def table_name(table,dialect=_D,identify=_A):
	'Get the full name of a table as a string.\n\n    Args:\n        table: Table expression node or string.\n        dialect: The dialect to generate the table name for.\n        identify: Determines when an identifier should be quoted. Possible values are:\n            False (default): Never quote, except in cases where it\'s mandatory by the dialect.\n            True: Always quote.\n\n    Examples:\n        >>> from sqlglot import exp, parse_one\n        >>> table_name(parse_one("select * from a.b.c").find(exp.Table))\n        \'a.b.c\'\n\n    Returns:\n        The table name.\n    ';B=dialect;A=table;A=maybe_parse(A,into=Table,dialect=B)
	if not A:raise ValueError(f"Cannot parse {A}")
	return'.'.join(A.sql(dialect=B,identify=_B,copy=_A)if identify or not SAFE_IDENTIFIER_RE.match(A.name)else A.name for A in A.parts)
def normalize_table_name(table,dialect=_D,copy=_B):'Returns a case normalized table name without quotes.\n\n    Args:\n        table: the table to normalize\n        dialect: the dialect to use for normalization rules\n        copy: whether to copy the expression.\n\n    Examples:\n        >>> normalize_table_name("`A-B`.c", dialect="bigquery")\n        \'A-B.c\'\n    ';A=dialect;from sqlglot.optimizer.normalize_identifiers import normalize_identifiers as B;return'.'.join(A.name for A in B(to_table(table,dialect=A,copy=copy),dialect=A).parts)
def replace_tables(expression,mapping,dialect=_D,copy=_B):
	'Replace all tables in expression according to the mapping.\n\n    Args:\n        expression: expression node to be transformed and replaced.\n        mapping: mapping of table names.\n        dialect: the dialect of the mapping table\n        copy: whether to copy the expression.\n\n    Examples:\n        >>> from sqlglot import exp, parse_one\n        >>> replace_tables(parse_one("select * from a.b"), {"a.b": "c"}).sql()\n        \'SELECT * FROM c /* a.b */\'\n\n    Returns:\n        The mapped expression.\n    ';C=dialect;B=mapping;B={normalize_table_name(A,dialect=C):B for(A,B)in B.items()}
	def A(node):
		A=node
		if isinstance(A,Table):
			D=normalize_table_name(A,dialect=C);E=B.get(D)
			if E:F=to_table(E,**{A:B for(A,B)in A.args.items()if A not in TABLE_PARTS},dialect=C);F.add_comments([D]);return F
		return A
	return expression.transform(A,copy=copy)
def replace_placeholders(expression,*A,**B):
	'Replace placeholders in an expression.\n\n    Args:\n        expression: expression node to be transformed and replaced.\n        args: positional names that will substitute unnamed placeholders in the given order.\n        kwargs: keyword arguments that will substitute named placeholders.\n\n    Examples:\n        >>> from sqlglot import exp, parse_one\n        >>> replace_placeholders(\n        ...     parse_one("select * from :tbl where ? = ?"),\n        ...     exp.to_identifier("str_col"), "b", tbl=exp.to_identifier("foo")\n        ... ).sql()\n        "SELECT * FROM foo WHERE str_col = \'b\'"\n\n    Returns:\n        The mapped expression.\n    '
	def C(node,args,**C):
		A=node
		if isinstance(A,Placeholder):
			if A.this:
				B=C.get(A.this)
				if B is not _D:return convert(B)
			else:
				try:return convert(next(args))
				except StopIteration:pass
		return A
	return expression.transform(C,iter(A),**B)
def expand(expression,sources,dialect=_D,copy=_B):
	'Transforms an expression by expanding all referenced sources into subqueries.\n\n    Examples:\n        >>> from sqlglot import parse_one\n        >>> expand(parse_one("select * from x AS z"), {"x": parse_one("select * from y")}).sql()\n        \'SELECT * FROM (SELECT * FROM y) AS z /* source: x */\'\n\n        >>> expand(parse_one("select * from x AS z"), {"x": parse_one("select * from y"), "y": parse_one("select * from z")}).sql()\n        \'SELECT * FROM (SELECT * FROM (SELECT * FROM z) AS y /* source: y */) AS z /* source: x */\'\n\n    Args:\n        expression: The expression to expand.\n        sources: A dictionary of name to Queries.\n        dialect: The dialect of the sources dict.\n        copy: Whether to copy the expression during transformation. Defaults to True.\n\n    Returns:\n        The transformed expression.\n    ';D=dialect;B=sources;B={normalize_table_name(A,dialect=D):B for(A,B)in B.items()}
	def E(node):
		A=node
		if isinstance(A,Table):
			C=normalize_table_name(A,dialect=D);F=B.get(C)
			if F:G=F.subquery(A.alias or C);G.comments=[f"source: {C}"];return G.transform(E,copy=_A)
		return A
	return expression.transform(E,copy=copy)
def func(name,*G,copy=_B,dialect=_D,**B):
	'\n    Returns a Func expression.\n\n    Examples:\n        >>> func("abs", 5).sql()\n        \'ABS(5)\'\n\n        >>> func("cast", this=5, to=DataType.build("DOUBLE")).sql()\n        \'CAST(5 AS DOUBLE)\'\n\n    Args:\n        name: the name of the function to build.\n        args: the args used to instantiate the function of interest.\n        copy: whether to copy the argument expressions.\n        dialect: the source dialect.\n        kwargs: the kwargs used to instantiate the function of interest.\n\n    Note:\n        The arguments `args` and `kwargs` are mutually exclusive.\n\n    Returns:\n        An instance of the function of interest, or an anonymous function, if `name` doesn\'t\n        correspond to an existing `sqlglot.expressions.Func` class.\n    ';F=name;C=dialect
	if G and B:raise ValueError("Can't use both args and kwargs to instantiate a function.")
	from sqlglot.dialects.dialect import Dialect as H;C=H.get_or_raise(C);E=[maybe_parse(A,dialect=C,copy=copy)for A in G];B={A:maybe_parse(B,dialect=C,copy=copy)for(A,B)in B.items()};A=C.parser_class.FUNCTIONS.get(F.upper())
	if A:
		if E:
			if'dialect'in A.__code__.co_varnames:D=A(E,dialect=C)
			else:D=A(E)
		elif A.__name__=='from_arg_list':D=A.__self__(**B)
		else:
			A=FUNCTION_BY_NAME.get(F.upper())
			if A:D=A(**B)
			else:raise ValueError(f"Unable to convert '{F}' into a Func. Either manually construct the Func expression of interest or parse the function call.")
	else:B=B or{_E:E};D=Anonymous(this=F,**B)
	for I in D.error_messages(E):raise ValueError(I)
	return D
def case(expression=_D,**C):
	'\n    Initialize a CASE statement.\n\n    Example:\n        case().when("a = 1", "foo").else_("bar")\n\n    Args:\n        expression: Optionally, the input expression (not all dialects support this)\n        **opts: Extra keyword arguments for parsing `expression`\n    ';A=expression
	if A is not _D:B=maybe_parse(A,**C)
	else:B=_D
	return Case(this=B,ifs=[])
def array(*A,copy=_B,dialect=_D,**B):"\n    Returns an array.\n\n    Examples:\n        >>> array(1, 'x').sql()\n        'ARRAY(1, x)'\n\n    Args:\n        expressions: the expressions to add to the array.\n        copy: whether to copy the argument expressions.\n        dialect: the source dialect.\n        kwargs: the kwargs used to instantiate the function of interest.\n\n    Returns:\n        An array expression.\n    ";return Array(expressions=[maybe_parse(A,copy=copy,dialect=dialect,**B)for A in A])
def tuple_(*A,copy=_B,dialect=_D,**B):"\n    Returns an tuple.\n\n    Examples:\n        >>> tuple_(1, 'x').sql()\n        '(1, x)'\n\n    Args:\n        expressions: the expressions to add to the tuple.\n        copy: whether to copy the argument expressions.\n        dialect: the source dialect.\n        kwargs: the kwargs used to instantiate the function of interest.\n\n    Returns:\n        A tuple expression.\n    ";return Tuple(expressions=[maybe_parse(A,copy=copy,dialect=dialect,**B)for A in A])
def true():'\n    Returns a true Boolean expression.\n    ';return Boolean(this=_B)
def false():'\n    Returns a false Boolean expression.\n    ';return Boolean(this=_A)
def null():'\n    Returns a Null expression.\n    ';return Null()
NONNULL_CONSTANTS=Literal,Boolean
CONSTANTS=Literal,Boolean,Null