from __future__ import annotations
_C=False
_B=True
_A=None
import datetime,inspect,logging,re,sys,typing as t
from collections.abc import Collection,Set
from contextlib import contextmanager
from copy import copy
from enum import Enum
from itertools import count
if t.TYPE_CHECKING:from sqlglot import exp;from sqlglot._typing import A,E,T;from sqlglot.expressions import Expression
CAMEL_CASE_PATTERN=re.compile('(?<!^)(?=[A-Z])')
PYTHON_VERSION=sys.version_info[:2]
logger=logging.getLogger('sqlglot')
class AutoName(Enum):
	'\n    This is used for creating Enum classes where `auto()` is the string form\n    of the corresponding enum\'s identifier (e.g. FOO.value results in "FOO").\n\n    Reference: https://docs.python.org/3/howto/enum.html#using-automatic-values\n    '
	def _generate_next_value_(name,_start,_count,_last_values):return name
class classproperty(property):
	'\n    Similar to a normal property but works for class methods\n    '
	def __get__(self,obj,owner=_A):return classmethod(self.fget).__get__(_A,owner)()
def seq_get(seq,index):
	'Returns the value in `seq` at position `index`, or `None` if `index` is out of bounds.'
	try:return seq[index]
	except IndexError:return
@t.overload
def ensure_list(value):...
@t.overload
def ensure_list(value):...
def ensure_list(value):
	"\n    Ensures that a value is a list, otherwise casts or wraps it into one.\n\n    Args:\n        value: The value of interest.\n\n    Returns:\n        The value cast as a list if it's a list or a tuple, or else the value wrapped in a list.\n    "
	if value is _A:return[]
	if isinstance(value,(list,tuple)):return list(value)
	return[value]
@t.overload
def ensure_collection(value):...
@t.overload
def ensure_collection(value):...
def ensure_collection(value):
	"\n    Ensures that a value is a collection (excluding `str` and `bytes`), otherwise wraps it into a list.\n\n    Args:\n        value: The value of interest.\n\n    Returns:\n        The value if it's a collection, or else the value wrapped in a list.\n    "
	if value is _A:return[]
	return value if isinstance(value,Collection)and not isinstance(value,(str,bytes))else[value]
def csv(*args,sep=', '):'\n    Formats any number of string arguments as CSV.\n\n    Args:\n        args: The string arguments to format.\n        sep: The argument separator.\n\n    Returns:\n        The arguments formatted as a CSV string.\n    ';return sep.join(arg for arg in args if arg)
def subclasses(module_name,classes,exclude=()):'\n    Returns all subclasses for a collection of classes, possibly excluding some of them.\n\n    Args:\n        module_name: The name of the module to search for subclasses in.\n        classes: Class(es) we want to find the subclasses of.\n        exclude: Class(es) we want to exclude from the returned list.\n\n    Returns:\n        The target subclasses.\n    ';return[obj for(_,obj)in inspect.getmembers(sys.modules[module_name],lambda obj:inspect.isclass(obj)and issubclass(obj,classes)and obj not in exclude)]
def apply_index_offset(this,expressions,offset):
	"\n    Applies an offset to a given integer literal expression.\n\n    Args:\n        this: The target of the index.\n        expressions: The expression the offset will be applied to, wrapped in a list.\n        offset: The offset that will be applied.\n\n    Returns:\n        The original expression with the offset applied to it, wrapped in a list. If the provided\n        `expressions` argument contains more than one expression, it's returned unaffected.\n    "
	if not offset or len(expressions)!=1:return expressions
	expression=expressions[0];from sqlglot import exp;from sqlglot.optimizer.annotate_types import annotate_types;from sqlglot.optimizer.simplify import simplify
	if not this.type:annotate_types(this)
	if t.cast(exp.DataType,this.type).this not in(exp.DataType.Type.UNKNOWN,exp.DataType.Type.ARRAY):return expressions
	if not expression.type:annotate_types(expression)
	if t.cast(exp.DataType,expression.type).this in exp.DataType.INTEGER_TYPES:logger.info('Applying array index offset (%s)',offset);expression=simplify(expression+offset);return[expression]
	return expressions
def camel_to_snake_case(name):'Converts `name` from camelCase to snake_case and returns the result.';return CAMEL_CASE_PATTERN.sub('_',name).upper()
def while_changing(expression,func):
	'\n    Applies a transformation to a given expression until a fix point is reached.\n\n    Args:\n        expression: The expression to be transformed.\n        func: The transformation to be applied.\n\n    Returns:\n        The transformed expression.\n    '
	while _B:
		for n in reversed(tuple(expression.walk())):n._hash=hash(n)
		start=hash(expression);expression=func(expression)
		for n in expression.walk():n._hash=_A
		if start==hash(expression):break
	return expression
def tsort(dag):
	"\n    Sorts a given directed acyclic graph in topological order.\n\n    Args:\n        dag: The graph to be sorted.\n\n    Returns:\n        A list that contains all of the graph's nodes in topological order.\n    ";result=[]
	for(node,deps)in tuple(dag.items()):
		for dep in deps:
			if dep not in dag:dag[dep]=set()
	while dag:
		current={node for(node,deps)in dag.items()if not deps}
		if not current:raise ValueError('Cycle error')
		for node in current:dag.pop(node)
		for deps in dag.values():deps-=current
		result.extend(sorted(current))
	return result
def open_file(file_name):
	'Open a file that may be compressed as gzip and return it in universal newline mode.'
	with open(file_name,'rb')as f:gzipped=f.read(2)==b'\x1f\x8b'
	if gzipped:import gzip;return gzip.open(file_name,'rt',newline='')
	return open(file_name,encoding='utf-8',newline='')
@contextmanager
def csv_reader(read_csv):
	"\n    Returns a csv reader given the expression `READ_CSV(name, ['delimiter', '|', ...])`.\n\n    Args:\n        read_csv: A `ReadCSV` function call.\n\n    Yields:\n        A python csv reader.\n    ";args=read_csv.expressions;file=open_file(read_csv.name);delimiter=',';args=iter(arg.name for arg in args)
	for(k,v)in zip(args,args):
		if k=='delimiter':delimiter=v
	try:import csv as csv_;yield csv_.reader(file,delimiter=delimiter)
	finally:file.close()
def find_new_name(taken,base):
	'\n    Searches for a new name.\n\n    Args:\n        taken: A collection of taken names.\n        base: Base name to alter.\n\n    Returns:\n        The new, available name.\n    '
	if base not in taken:return base
	i=2;new=f"{base}_{i}"
	while new in taken:i+=1;new=f"{base}_{i}"
	return new
def is_int(text):return is_type(text,int)
def is_float(text):return is_type(text,float)
def is_type(text,target_type):
	try:target_type(text);return _B
	except ValueError:return _C
def name_sequence(prefix):'Returns a name generator given a prefix (e.g. a0, a1, a2, ... if the prefix is "a").';sequence=count();return lambda:f"{prefix}{next(sequence)}"
def object_to_dict(obj,**kwargs):"Returns a dictionary created from an object's attributes.";return{**{k:v.copy()if hasattr(v,'copy')else copy(v)for(k,v)in vars(obj).items()},**kwargs}
def split_num_words(value,sep,min_num_words,fill_from_start=_B):
	'\n    Perform a split on a value and return N words as a result with `None` used for words that don\'t exist.\n\n    Args:\n        value: The value to be split.\n        sep: The value to use to split on.\n        min_num_words: The minimum number of words that are going to be in the result.\n        fill_from_start: Indicates that if `None` values should be inserted at the start or end of the list.\n\n    Examples:\n        >>> split_num_words("db.table", ".", 3)\n        [None, \'db\', \'table\']\n        >>> split_num_words("db.table", ".", 3, fill_from_start=False)\n        [\'db\', \'table\', None]\n        >>> split_num_words("db.table", ".", 1)\n        [\'db\', \'table\']\n\n    Returns:\n        The list of words returned by `split`, possibly augmented by a number of `None` values.\n    ';words=value.split(sep)
	if fill_from_start:return[_A]*(min_num_words-len(words))+words
	return words+[_A]*(min_num_words-len(words))
def is_iterable(value):'\n    Checks if the value is an iterable, excluding the types `str` and `bytes`.\n\n    Examples:\n        >>> is_iterable([1,2])\n        True\n        >>> is_iterable("test")\n        False\n\n    Args:\n        value: The value to check if it is an iterable.\n\n    Returns:\n        A `bool` value indicating if it is an iterable.\n    ';from sqlglot import Expression;return hasattr(value,'__iter__')and not isinstance(value,(str,bytes,Expression))
def flatten(values):
	'\n    Flattens an iterable that can contain both iterable and non-iterable elements. Objects of\n    type `str` and `bytes` are not regarded as iterables.\n\n    Examples:\n        >>> list(flatten([[1, 2], 3, {4}, (5, "bla")]))\n        [1, 2, 3, 4, 5, \'bla\']\n        >>> list(flatten([1, 2, 3]))\n        [1, 2, 3]\n\n    Args:\n        values: The value to be flattened.\n\n    Yields:\n        Non-iterable elements in `values`.\n    '
	for value in values:
		if is_iterable(value):yield from flatten(value)
		else:yield value
def dict_depth(d):
	'\n    Get the nesting depth of a dictionary.\n\n    Example:\n        >>> dict_depth(None)\n        0\n        >>> dict_depth({})\n        1\n        >>> dict_depth({"a": "b"})\n        1\n        >>> dict_depth({"a": {}})\n        2\n        >>> dict_depth({"a": {"b": {}}})\n        3\n    '
	try:return 1+dict_depth(next(iter(d.values())))
	except AttributeError:return 0
	except StopIteration:return 1
def first(it):'Returns the first element from an iterable (useful for sets).';return next(i for i in it)
def merge_ranges(ranges):
	'\n    Merges a sequence of ranges, represented as tuples (low, high) whose values\n    belong to some totally-ordered set.\n\n    Example:\n        >>> merge_ranges([(1, 3), (2, 6)])\n        [(1, 6)]\n    '
	if not ranges:return[]
	ranges=sorted(ranges);merged=[ranges[0]]
	for(start,end)in ranges[1:]:
		last_start,last_end=merged[-1]
		if start<=last_end:merged[-1]=last_start,max(last_end,end)
		else:merged.append((start,end))
	return merged
def is_iso_date(text):
	try:datetime.date.fromisoformat(text);return _B
	except ValueError:return _C
def is_iso_datetime(text):
	try:datetime.datetime.fromisoformat(text);return _B
	except ValueError:return _C
DATE_UNITS={'day','week','month','quarter','year','year_month'}
def is_date_unit(expression):return expression is not _A and expression.name.lower()in DATE_UNITS
K=t.TypeVar('K')
V=t.TypeVar('V')
class SingleValuedMapping(t.Mapping[K,V]):
	'\n    Mapping where all keys return the same value.\n\n    This rigamarole is meant to avoid copying keys, which was originally intended\n    as an optimization while qualifying columns for tables with lots of columns.\n    '
	def __init__(self,keys,value):self._keys=keys if isinstance(keys,Set)else set(keys);self._value=value
	def __getitem__(self,key):
		if key in self._keys:return self._value
		raise KeyError(key)
	def __len__(self):return len(self._keys)
	def __iter__(self):return iter(self._keys)