from __future__ import annotations
_C=True
_B=False
_A=None
import json,logging,typing as t
from dataclasses import dataclass,field
from sqlglot import Schema,exp,maybe_parse
from sqlglot.errors import SqlglotError
from sqlglot.optimizer import Scope,build_scope,find_all_in_scope,normalize_identifiers,qualify
from sqlglot.optimizer.scope import ScopeType
if t.TYPE_CHECKING:from sqlglot.dialects.dialect import DialectType
logger=logging.getLogger('sqlglot')
@dataclass(frozen=_C)
class Node:
	name:str;expression:exp.Expression;source:exp.Expression;downstream:t.List[Node]=field(default_factory=list);source_name:str='';reference_node_name:str=''
	def walk(A):
		yield A
		for B in A.downstream:yield from B.walk()
	def to_html(I,dialect=_A,**J):
		C=dialect;D={};E=[]
		for A in I.walk():
			if isinstance(A.expression,exp.Table):F=f"FROM {A.expression.this}";G=f"<pre>SELECT {A.name} FROM {A.expression.this}</pre>";H=1
			else:F=A.expression.sql(pretty=_C,dialect=C);K=A.source.transform(lambda n:exp.Tag(this=n,prefix='<b>',postfix='</b>')if n is A.expression else n,copy=_B).sql(pretty=_C,dialect=C);G=f"<pre>{K}</pre>";H=0
			B=id(A);D[B]={'id':B,'label':F,'title':G,'group':H}
			for L in A.downstream:E.append({'from':B,'to':id(L)})
		return GraphHTML(D,E,**J)
def lineage(column,sql,schema=_A,sources=_A,dialect=_A,scope=_A,trim_selects=_C,**F):
	'Build the lineage graph for a column of a SQL query.\n\n    Args:\n        column: The column to build the lineage for.\n        sql: The SQL string or expression.\n        schema: The schema of tables.\n        sources: A mapping of queries which will be used to continue building lineage.\n        dialect: The dialect of input SQL.\n        scope: A pre-created scope to use instead.\n        trim_selects: Whether or not to clean up selects by trimming to only relevant columns.\n        **kwargs: Qualification optimizer kwargs.\n\n    Returns:\n        A lineage node.\n    ';E=sources;C=scope;B=column;A=dialect;D=maybe_parse(sql,dialect=A);B=normalize_identifiers.normalize_identifiers(B,dialect=A).name
	if E:D=exp.expand(D,{B:t.cast(exp.Query,maybe_parse(C,dialect=A))for(B,C)in E.items()},dialect=A)
	if not C:D=qualify.qualify(D,dialect=A,schema=schema,**{'validate_qualify_columns':_B,'identify':_B,**F});C=build_scope(D)
	if not C:raise SqlglotError('Cannot build lineage, sql must be SELECT')
	if not any(A.alias_or_name==B for A in C.expression.selects):raise SqlglotError(f"Cannot find column '{B}' in query.")
	return to_node(B,C,A,trim_selects=trim_selects)
def to_node(column,scope,dialect,scope_name=_A,upstream=_A,source_name=_A,reference_node_name=_A,trim_selects=_C):
	O=scope_name;K=source_name;J=trim_selects;I=dialect;E=reference_node_name;D=upstream;C=column;A=scope;F=A.expression.selects[C]if isinstance(C,int)else next((A for A in A.expression.selects if A.alias_or_name==C),exp.Star()if A.expression.is_star else A.expression)
	if isinstance(A.expression,exp.Subquery):
		for B in A.subquery_scopes:return to_node(C,scope=B,dialect=I,upstream=D,source_name=K,reference_node_name=E,trim_selects=J)
	if isinstance(A.expression,exp.SetOperation):
		L=type(A.expression).__name__.upper();D=D or Node(name=L,source=A.expression,expression=F);P=C if isinstance(C,int)else next((B for(B,A)in enumerate(A.expression.selects)if A.alias_or_name==C or A.is_star),-1)
		if P==-1:raise ValueError(f"Could not find {C} in {A.expression}")
		for V in A.union_scopes:to_node(P,scope=V,dialect=I,upstream=D,source_name=K,reference_node_name=E,trim_selects=J)
		return D
	if J and isinstance(A.expression,exp.Select):B=t.cast(exp.Expression,A.expression.select(F,append=_B))
	else:B=A.expression
	G=Node(name=f"{O}.{C}"if O else str(C),source=B,expression=F,source_name=K or'',reference_node_name=E or'')
	if D:D.downstream.append(G)
	W={id(A.expression):A for A in A.subquery_scopes}
	for M in find_all_in_scope(F,exp.UNWRAPPED_QUERIES):
		Q=W.get(id(M))
		if not Q:logger.warning(f"Unknown subquery scope: {M.sql(dialect=I)}");continue
		for L in M.named_selects:to_node(L,scope=Q,dialect=I,upstream=G,trim_selects=J)
	if F.is_star:
		for B in A.sources.values():
			if isinstance(B,Scope):B=B.expression
			G.downstream.append(Node(name=F.sql(comments=_B),source=B,expression=B))
	R=set(find_all_in_scope(F,exp.Column))
	if isinstance(B,exp.UDTF):R|=set(B.find_all(exp.Column));S=[A.expression.parent for A in A.sources.values()if isinstance(A,Scope)and A.is_derived_table]
	else:S=A.derived_tables
	T={A.alias:A.comments[0].split()[1]for A in S if A.comments and A.comments[0].startswith('source: ')}
	for N in R:
		H=N.table;B=A.sources.get(H)
		if isinstance(B,Scope):
			E=_A
			if B.scope_type==ScopeType.DERIVED_TABLE and H not in T:E=H
			elif B.scope_type==ScopeType.CTE:U,X=A.selected_sources.get(H,(_A,_A));E=U.name if U else _A
			to_node(N.name,scope=B,dialect=I,scope_name=H,upstream=G,source_name=T.get(H)or K,reference_node_name=E,trim_selects=J)
		else:B=B or exp.Placeholder();G.downstream.append(Node(name=N.sql(comments=_B),source=B,expression=B))
	return G
class GraphHTML:
	'Node to HTML generator using vis.js.\n\n    https://visjs.github.io/vis-network/docs/network/\n    '
	def __init__(A,nodes,edges,imports=_C,options=_A):B='enabled';A.imports=imports;A.options={'height':'500px','width':'100%','layout':{'hierarchical':{B:_C,'nodeSpacing':200,'sortMethod':'directed'}},'interaction':{'dragNodes':_B,'selectable':_B},'physics':{B:_B},'edges':{'arrows':'to'},'nodes':{'font':'20px monaco','shape':'box','widthConstraint':{'maximum':300}},**(options or{})};A.nodes=nodes;A.edges=edges
	def __str__(A):B=json.dumps(list(A.nodes.values()));C=json.dumps(A.edges);D=json.dumps(A.options);E='<script type="text/javascript" src="https://unpkg.com/vis-data@latest/peer/umd/vis-data.min.js"></script>\n  <script type="text/javascript" src="https://unpkg.com/vis-network@latest/peer/umd/vis-network.min.js"></script>\n  <link rel="stylesheet" type="text/css" href="https://unpkg.com/vis-network/styles/vis-network.min.css" />'if A.imports else'';return f'''<div>
  <div id="sqlglot-lineage"></div>
  {E}
  <script type="text/javascript">
    var nodes = new vis.DataSet({B})
    nodes.forEach(row => row["title"] = new DOMParser().parseFromString(row["title"], "text/html").body.childNodes[0])

    new vis.Network(
        document.getElementById("sqlglot-lineage"),
        {{
            nodes: nodes,
            edges: new vis.DataSet({C})
        }},
        {D},
    )
  </script>
</div>'''
	def _repr_html_(A):return A.__str__()