from __future__ import annotations
import itertools,typing as t
from sqlglot import exp
from sqlglot.helper import is_date_unit,is_iso_date,is_iso_datetime
def canonicalize(expression):
	'Converts a sql expression into a standard form.\n\n    This method relies on annotate_types because many of the\n    conversions rely on type inference.\n\n    Args:\n        expression: The expression to canonicalize.\n    '
	def A(expression):A=expression;A=add_text_to_concat(A);A=replace_date_funcs(A);A=coerce_type(A);A=remove_redundant_casts(A);A=ensure_bools(A,_replace_int_predicate);A=remove_ascending_order(A);return A
	return exp.replace_tree(expression,A)
def add_text_to_concat(node):
	A=node
	if isinstance(A,exp.Add)and A.type and A.type.this in exp.DataType.TEXT_TYPES:A=exp.Concat(expressions=[A.left,A.right])
	return A
def replace_date_funcs(node):
	B='zone';A=node
	if isinstance(A,(exp.Date,exp.TsOrDsToDate))and not A.expressions and not A.args.get(B)and A.this.is_string and is_iso_date(A.this.name):return exp.cast(A.this,to=exp.DataType.Type.DATE)
	if isinstance(A,exp.Timestamp)and not A.args.get(B):
		if not A.type:from sqlglot.optimizer.annotate_types import annotate_types as C;A=C(A)
		return exp.cast(A.this,to=A.type or exp.DataType.Type.TIMESTAMP)
	return A
COERCIBLE_DATE_OPS=exp.Add,exp.Sub,exp.EQ,exp.NEQ,exp.GT,exp.GTE,exp.LT,exp.LTE,exp.NullSafeEQ,exp.NullSafeNEQ
def coerce_type(node):
	A=node
	if isinstance(A,COERCIBLE_DATE_OPS):_coerce_date(A.left,A.right)
	elif isinstance(A,exp.Between):_coerce_date(A.this,A.args['low'])
	elif isinstance(A,exp.Extract)and not A.expression.type.is_type(*exp.DataType.TEMPORAL_TYPES):_replace_cast(A.expression,exp.DataType.Type.DATETIME)
	elif isinstance(A,(exp.DateAdd,exp.DateSub,exp.DateTrunc)):_coerce_timeunit_arg(A.this,A.unit)
	elif isinstance(A,exp.DateDiff):_coerce_datediff_args(A)
	return A
def remove_redundant_casts(expression):
	A=expression
	if isinstance(A,exp.Cast)and A.this.type and A.to.this==A.this.type.this:return A.this
	if isinstance(A,(exp.Date,exp.TsOrDsToDate))and A.this.type and A.this.type.this==exp.DataType.Type.DATE:return A.this
	return A
def ensure_bools(expression,replace_func):
	B=replace_func;A=expression
	if isinstance(A,exp.Connector):B(A.left);B(A.right)
	elif isinstance(A,exp.Not):B(A.this)
	elif isinstance(A,exp.If)and not(isinstance(A.parent,exp.Case)and A.parent.this):B(A.this)
	elif isinstance(A,(exp.Where,exp.Having)):B(A.this)
	return A
def remove_ascending_order(expression):
	B='desc';A=expression
	if isinstance(A,exp.Ordered)and A.args.get(B)is False:A.set(B,None)
	return A
def _coerce_date(a,b):
	for(a,b)in itertools.permutations([a,b]):
		if isinstance(b,exp.Interval):a=_coerce_timeunit_arg(a,b.unit)
		if a.type and a.type.this in exp.DataType.TEMPORAL_TYPES and b.type and b.type.this in exp.DataType.TEXT_TYPES:_replace_cast(b,exp.DataType.Type.DATETIME)
def _coerce_timeunit_arg(arg,unit):
	A=arg
	if not A.type:return A
	if A.type.this in exp.DataType.TEXT_TYPES:
		B=A.name;C=is_iso_date(B)
		if C and is_date_unit(unit):return A.replace(exp.cast(A.copy(),to=exp.DataType.Type.DATE))
		if C or is_iso_datetime(B):return A.replace(exp.cast(A.copy(),to=exp.DataType.Type.DATETIME))
	elif A.type.this==exp.DataType.Type.DATE and not is_date_unit(unit):return A.replace(exp.cast(A.copy(),to=exp.DataType.Type.DATETIME))
	return A
def _coerce_datediff_args(node):
	for A in(node.this,node.expression):
		if A.type.this not in exp.DataType.TEMPORAL_TYPES:A.replace(exp.cast(A.copy(),to=exp.DataType.Type.DATETIME))
def _replace_cast(node,to):node.replace(exp.cast(node.copy(),to=to))
def _replace_int_predicate(expression):
	A=expression
	if isinstance(A,exp.Coalesce):
		for B in A.iter_expressions():_replace_int_predicate(B)
	elif A.type and A.type.this in exp.DataType.INTEGER_TYPES:A.replace(A.neq(0))