from sqlglot import expressions as exp
from sqlglot.optimizer.normalize import normalized
from sqlglot.optimizer.scope import Scope,traverse_scope
def eliminate_joins(expression):
	'\n    Remove unused joins from an expression.\n\n    This only removes joins when we know that the join condition doesn\'t produce duplicate rows.\n\n    Example:\n        >>> import sqlglot\n        >>> sql = "SELECT x.a FROM x LEFT JOIN (SELECT DISTINCT y.b FROM y) AS y ON x.b = y.b"\n        >>> expression = sqlglot.parse_one(sql)\n        >>> eliminate_joins(expression).sql()\n        \'SELECT x.a FROM x\'\n\n    Args:\n        expression (sqlglot.Expression): expression to optimize\n    Returns:\n        sqlglot.Expression: optimized expression\n    ';C=expression
	for A in traverse_scope(C):
		if A.unqualified_columns:continue
		E=A.expression.args.get('joins',[])
		for B in reversed(E):
			D=B.alias_or_name
			if _should_eliminate_join(A,B,D):B.pop();A.remove_source(D)
	return C
def _should_eliminate_join(scope,join,alias):D=alias;C=scope;A=join;B=C.sources.get(D);return isinstance(B,Scope)and not _join_is_used(C,A,D)and(A.side=='LEFT'and _is_joined_on_all_unique_outputs(B,A)or not A.args.get('on')and _has_single_output_row(B))
def _join_is_used(scope,join,alias):
	A=join.args.get('on')
	if A:B={id(A)for A in A.find_all(exp.Column)}
	else:B=set()
	return any(A for A in scope.source_columns(alias)if id(A)not in B)
def _is_joined_on_all_unique_outputs(scope,join):
	A=_unique_outputs(scope)
	if not A:return False
	B,C,B=join_condition(join);D=A-{A.name for A in C};return not D
def _unique_outputs(scope):
	'Determine output columns of `scope` that must have a unique combination per row';A=scope
	if A.expression.args.get('distinct'):return set(A.expression.named_selects)
	B=A.expression.args.get('group')
	if B:
		C=set(B.expressions);D=set();E=set()
		for F in A.expression.selects:
			G=F.unalias()
			if G in C:D.add(G);E.add(F.alias_or_name)
		if not C.difference(D):return E
		else:return set()
	if _has_single_output_row(A):return set(A.expression.named_selects)
	return set()
def _has_single_output_row(scope):A=scope;return isinstance(A.expression,exp.Select)and(all(isinstance(A.unalias(),exp.AggFunc)for A in A.expression.selects)or _is_limit_1(A)or not A.expression.args.get('from'))
def _is_limit_1(scope):A=scope.expression.args.get('limit');return A and A.expression.this=='1'
def join_condition(join):
	'\n    Extract the join condition from a join expression.\n\n    Args:\n        join (exp.Join)\n    Returns:\n        tuple[list[str], list[str], exp.Expression]:\n            Tuple of (source key, join key, remaining predicate)\n    ';D=join.alias_or_name;A=(join.args.get('on')or exp.true()).copy();E=[];F=[]
	def H(condition):
		A=condition;B,C=A.unnest_operands();G=exp.column_table_names(B);H=exp.column_table_names(C)
		if D in G and D not in H:F.append(B);E.append(C);A.replace(exp.true())
		elif D in H and D not in G:F.append(C);E.append(B);A.replace(exp.true())
	if normalized(A):
		A=A if isinstance(A,exp.And)else exp.and_(A,exp.true(),copy=False)
		for B in A.flatten():
			if isinstance(B,exp.EQ):H(B)
	elif normalized(A,dnf=True):
		C=None
		for B in A.flatten():
			I=[A for A in B.flatten()if isinstance(A,exp.EQ)]
			if C is None:C=I
			else:
				G=[]
				for J in I:
					K=[A for A in C if J==A]
					if K:G.append(J);G.extend(K)
				C=G
		for B in C:H(B)
	return E,F,A