_G='expressions'
_F='hint'
_E='where'
_D='order'
_C='from'
_B='joins'
_A=False
from collections import defaultdict
from sqlglot import expressions as exp
from sqlglot.helper import find_new_name
from sqlglot.optimizer.scope import Scope,traverse_scope
def merge_subqueries(expression,leave_tables_isolated=_A):'\n    Rewrite sqlglot AST to merge derived tables into the outer query.\n\n    This also merges CTEs if they are selected from only once.\n\n    Example:\n        >>> import sqlglot\n        >>> expression = sqlglot.parse_one("SELECT a FROM (SELECT x.a FROM x) CROSS JOIN y")\n        >>> merge_subqueries(expression).sql()\n        \'SELECT x.a FROM x CROSS JOIN y\'\n\n    If `leave_tables_isolated` is True, this will not merge inner queries into outer\n    queries if it would result in multiple table selects in a single query:\n        >>> expression = sqlglot.parse_one("SELECT a FROM (SELECT x.a FROM x) CROSS JOIN y")\n        >>> merge_subqueries(expression, leave_tables_isolated=True).sql()\n        \'SELECT a FROM (SELECT x.a FROM x) CROSS JOIN y\'\n\n    Inspired by https://dev.mysql.com/doc/refman/8.0/en/derived-table-optimization.html\n\n    Args:\n        expression (sqlglot.Expression): expression to optimize\n        leave_tables_isolated (bool):\n    Returns:\n        sqlglot.Expression: optimized expression\n    ';B=leave_tables_isolated;A=expression;A=merge_ctes(A,B);A=merge_derived_tables(A,B);return A
UNMERGABLE_ARGS=set(exp.Select.arg_types)-{_G,_C,_B,_E,_D,_F}
SAFE_TO_REPLACE_UNWRAPPED=exp.Column,exp.EQ,exp.Func,exp.NEQ,exp.Paren
def merge_ctes(expression,leave_tables_isolated=_A):
	F=expression;H=traverse_scope(F);G=defaultdict(list)
	for B in H:
		for(C,A)in B.selected_sources.values():
			if isinstance(A,Scope)and A.is_cte:G[id(A)].append((B,A,C))
	I=[A[0]for(B,A)in G.items()if len(A)==1]
	for(B,A,C)in I:
		D=C.find_ancestor(exp.From,exp.Join)
		if _mergeable(B,A,leave_tables_isolated,D):E=C.alias_or_name;_rename_inner_sources(B,A,E);_merge_from(B,A,C,E);_merge_expressions(B,A,E);_merge_joins(B,A,D);_merge_where(B,A,D);_merge_order(B,A);_merge_hints(B,A);_pop_cte(A);B.clear_cache()
	return F
def merge_derived_tables(expression,leave_tables_isolated=_A):
	F=expression
	for A in traverse_scope(F):
		for D in A.derived_tables:
			E=D.find_ancestor(exp.From,exp.Join);C=D.alias_or_name;B=A.sources[C]
			if _mergeable(A,B,leave_tables_isolated,E):_rename_inner_sources(A,B,C);_merge_from(A,B,D,C);_merge_expressions(A,B,C);_merge_joins(A,B,E);_merge_where(A,B,E);_merge_order(A,B);_merge_hints(A,B);A.clear_cache()
	return F
def _mergeable(outer_scope,inner_scope,leave_tables_isolated,from_or_join):
	'\n    Return True if `inner_select` can be merged into outer query.\n\n    Args:\n        outer_scope (Scope)\n        inner_scope (Scope)\n        leave_tables_isolated (bool)\n        from_or_join (exp.From|exp.Join)\n    Returns:\n        bool: True if can be merged\n    ';F='RIGHT';E='FULL';D=inner_scope;C=from_or_join;A=outer_scope;B=D.expression.unnest()
	def G():D=B.find_all(exp.Window);E={A.parent.alias_or_name for A in D};F=C.alias_or_name;G=[A for A in A.columns if A.find_ancestor(exp.Where,exp.Group,exp.Order,exp.Join,exp.Having,exp.AggFunc)];H=[A for A in G if A.table==F and A.name in E];return any(H)
	def H():
		"\n        All columns from the inner select in the ON clause must be from the first FROM table.\n\n        That is, this can be merged:\n            SELECT * FROM x JOIN (SELECT y.a AS a FROM y JOIN z) AS q ON x.a = q.a\n                                         ^^^           ^\n        But this can't:\n            SELECT * FROM x JOIN (SELECT z.a AS a FROM y JOIN z) AS q ON x.a = q.a\n                                         ^^^                  ^\n        "
		if not isinstance(C,exp.Join):return _A
		E=C.alias_or_name;A=C.args.get('on')
		if not A:return _A
		F=[A.name for A in A.find_all(exp.Column)if A.table==E];B=D.expression.args.get(_C)
		if not B:return _A
		G=B.alias_or_name;H={A.alias_or_name:A for A in D.expression.selects};return any(B.table!=G for A in F for B in H[A].find_all(exp.Column))
	def I():
		C=D.expression.parent;B=A.expression.parent
		while B:
			if B is C:return True
			B=B.parent
		return _A
	return isinstance(A.expression,exp.Select)and not A.expression.is_star and isinstance(B,exp.Select)and not any(B.args.get(A)for A in UNMERGABLE_ARGS)and B.args.get(_C)and not A.pivots and not any(A.find(exp.AggFunc,exp.Select,exp.Explode)for A in B.expressions)and not(leave_tables_isolated and len(A.selected_sources)>1)and not(isinstance(C,exp.Join)and B.args.get(_E)and C.side in(E,'LEFT',F))and not(isinstance(C,exp.From)and B.args.get(_E)and any(A.side in(E,F)for A in A.expression.args.get(_B,[])))and not H()and not G()and not I()and not(B.args.get(_D)and A.is_union)
def _rename_inner_sources(outer_scope,inner_scope,alias):
	'\n    Renames any sources in the inner query that conflict with names in the outer query.\n\n    Args:\n        outer_scope (sqlglot.optimizer.scope.Scope)\n        inner_scope (sqlglot.optimizer.scope.Scope)\n        alias (str)\n    ';H='alias';B=inner_scope;F=set(outer_scope.selected_sources);G=F.intersection(set(B.selected_sources));G-={alias}
	for C in G:
		D=find_new_name(F,C);A,J=B.selected_sources[C];E=exp.to_identifier(D)
		if isinstance(A,exp.Subquery):A.set(H,exp.TableAlias(this=E))
		elif isinstance(A,exp.Table)and A.alias:A.set(H,E)
		elif isinstance(A,exp.Table):A.replace(exp.alias_(A,E))
		for I in B.source_columns(C):I.set('table',exp.to_identifier(D))
		B.rename_source(C,D)
def _merge_from(outer_scope,inner_scope,node_to_replace,alias):
	'\n    Merge FROM clause of inner query into outer query.\n\n    Args:\n        outer_scope (sqlglot.optimizer.scope.Scope)\n        inner_scope (sqlglot.optimizer.scope.Scope)\n        node_to_replace (exp.Subquery|exp.Table)\n        alias (str)\n    ';D=inner_scope;C=node_to_replace;B=outer_scope;A=D.expression.args[_C].this;A.set(_B,C.args.get(_B));C.replace(A)
	for F in B.join_hints:
		G=F.find_all(exp.Table)
		for E in G:
			if E.alias_or_name==C.alias_or_name:E.set('this',exp.to_identifier(A.alias_or_name))
	B.remove_source(alias);B.add_source(A.alias_or_name,D.sources[A.alias_or_name])
def _merge_joins(outer_scope,inner_scope,from_or_join):
	'\n    Merge JOIN clauses of inner query into outer query.\n\n    Args:\n        outer_scope (sqlglot.optimizer.scope.Scope)\n        inner_scope (sqlglot.optimizer.scope.Scope)\n        from_or_join (exp.From|exp.Join)\n    ';G=from_or_join;F=inner_scope;A=outer_scope;B=[];H=F.expression.args.get(_B)or[]
	for C in H:B.append(C);A.add_source(C.alias_or_name,F.sources[C.alias_or_name])
	if B:
		D=A.expression.args.get(_B,[])
		if isinstance(G,exp.From):E=0
		else:E=D.index(G)+1
		D[E:E]=B;A.expression.set(_B,D)
def _merge_expressions(outer_scope,inner_scope,alias):
	'\n    Merge projections of inner query into outer query.\n\n    Args:\n        outer_scope (sqlglot.optimizer.scope.Scope)\n        inner_scope (sqlglot.optimizer.scope.Scope)\n        alias (str)\n    ';C=defaultdict(list)
	for B in outer_scope.columns:
		if B.table==alias:C[B.name].append(B)
	for A in inner_scope.expression.expressions:
		D=A.alias_or_name
		if not D:continue
		E=C.get(D,[]);A=A.unalias();F=not isinstance(A,SAFE_TO_REPLACE_UNWRAPPED)
		for B in E:
			if isinstance(B.parent,(exp.Unary,exp.Binary))and F:A=exp.paren(A,copy=_A)
			B.replace(A.copy())
def _merge_where(outer_scope,inner_scope,from_or_join):
	'\n    Merge WHERE clause of inner query into outer query.\n\n    Args:\n        outer_scope (sqlglot.optimizer.scope.Scope)\n        inner_scope (sqlglot.optimizer.scope.Scope)\n        from_or_join (exp.From|exp.Join)\n    ';A=from_or_join;B=inner_scope.expression.args.get(_E)
	if not B or not B.this:return
	C=outer_scope.expression
	if isinstance(A,exp.Join):
		D=C.args.get(_C);E={D.alias_or_name}if D else{}
		for G in C.args[_B]:
			F=G.alias_or_name;E.add(F)
			if F==A.alias_or_name:break
		if exp.column_table_names(B.this)<=E:A.on(B.this,copy=_A);A.set('on',A.args.get('on'));return
	C.where(B.this,copy=_A)
def _merge_order(outer_scope,inner_scope):
	'\n    Merge ORDER clause of inner query into outer query.\n\n    Args:\n        outer_scope (sqlglot.optimizer.scope.Scope)\n        inner_scope (sqlglot.optimizer.scope.Scope)\n    ';A=outer_scope
	if any(A.expression.args.get(B)for B in['group','distinct','having',_D])or len(A.selected_sources)!=1 or any(A.find(exp.AggFunc)for A in A.expression.expressions):return
	A.expression.set(_D,inner_scope.expression.args.get(_D))
def _merge_hints(outer_scope,inner_scope):
	B=outer_scope;A=inner_scope.expression.args.get(_F)
	if not A:return
	C=B.expression.args.get(_F)
	if C:
		for D in A.expressions:C.append(_G,D)
	else:B.expression.set(_F,A)
def _pop_cte(inner_scope):
	'\n    Remove CTE from the AST.\n\n    Args:\n        inner_scope (sqlglot.optimizer.scope.Scope)\n    ';A=inner_scope.expression.parent;B=A.parent
	if len(B.expressions)==1:B.pop()
	else:A.pop()