from __future__ import annotations
import typing as t
from sqlglot import exp
from sqlglot.dialects.dialect import Dialect,DialectType
if t.TYPE_CHECKING:from sqlglot._typing import E
@t.overload
def normalize_identifiers(expression,dialect=None):...
@t.overload
def normalize_identifiers(expression,dialect=None):...
def normalize_identifiers(expression,dialect=None):
	'\n    Normalize identifiers by converting them to either lower or upper case,\n    ensuring the semantics are preserved in each case (e.g. by respecting\n    case-sensitivity).\n\n    This transformation reflects how identifiers would be resolved by the engine corresponding\n    to each SQL dialect, and plays a very important role in the standardization of the AST.\n\n    It\'s possible to make this a no-op by adding a special comment next to the\n    identifier of interest:\n\n        SELECT a /* sqlglot.meta case_sensitive */ FROM table\n\n    In this example, the identifier `a` will not be normalized.\n\n    Note:\n        Some dialects (e.g. DuckDB) treat all identifiers as case-insensitive even\n        when they\'re quoted, so in these cases all identifiers are normalized.\n\n    Example:\n        >>> import sqlglot\n        >>> expression = sqlglot.parse_one(\'SELECT Bar.A AS A FROM "Foo".Bar\')\n        >>> normalize_identifiers(expression).sql()\n        \'SELECT bar.a AS a FROM "Foo".bar\'\n        >>> normalize_identifiers("foo", dialect="snowflake").sql(dialect="snowflake")\n        \'FOO\'\n\n    Args:\n        expression: The expression to transform.\n        dialect: The dialect to use in order to decide how to normalize identifiers.\n\n    Returns:\n        The transformed expression.\n    ';D='case_sensitive';B=dialect;A=expression;B=Dialect.get_or_raise(B)
	if isinstance(A,str):A=exp.parse_identifier(A,dialect=B)
	for C in A.walk(prune=lambda n:n.meta.get(D)):
		if not C.meta.get(D):B.normalize_identifier(C)
	return A