from __future__ import annotations
_B=None
_A=True
import typing as t
from sqlglot import exp
from sqlglot.dialects.dialect import Dialect,DialectType
from sqlglot.optimizer.isolate_table_selects import isolate_table_selects
from sqlglot.optimizer.normalize_identifiers import normalize_identifiers
from sqlglot.optimizer.qualify_columns import pushdown_cte_alias_columns as pushdown_cte_alias_columns_func,qualify_columns as qualify_columns_func,quote_identifiers as quote_identifiers_func,validate_qualify_columns as validate_qualify_columns_func
from sqlglot.optimizer.qualify_tables import qualify_tables
from sqlglot.schema import Schema,ensure_schema
def qualify(expression,dialect=_B,db=_B,catalog=_B,schema=_B,expand_alias_refs=_A,expand_stars=_A,infer_schema=_B,isolate_tables=False,qualify_columns=_A,validate_qualify_columns=_A,quote_identifiers=_A,identify=_A,infer_csv_schemas=False):
	'\n    Rewrite sqlglot AST to have normalized and qualified tables and columns.\n\n    This step is necessary for all further SQLGlot optimizations.\n\n    Example:\n        >>> import sqlglot\n        >>> schema = {"tbl": {"col": "INT"}}\n        >>> expression = sqlglot.parse_one("SELECT col FROM tbl")\n        >>> qualify(expression, schema=schema).sql()\n        \'SELECT "tbl"."col" AS "col" FROM "tbl" AS "tbl"\'\n\n    Args:\n        expression: Expression to qualify.\n        db: Default database name for tables.\n        catalog: Default catalog name for tables.\n        schema: Schema to infer column names and types.\n        expand_alias_refs: Whether to expand references to aliases.\n        expand_stars: Whether to expand star queries. This is a necessary step\n            for most of the optimizer\'s rules to work; do not set to False unless you\n            know what you\'re doing!\n        infer_schema: Whether to infer the schema if missing.\n        isolate_tables: Whether to isolate table selects.\n        qualify_columns: Whether to qualify columns.\n        validate_qualify_columns: Whether to validate columns.\n        quote_identifiers: Whether to run the quote_identifiers step.\n            This step is necessary to ensure correctness for case sensitive queries.\n            But this flag is provided in case this step is performed at a later time.\n        identify: If True, quote all identifiers, else only necessary ones.\n        infer_csv_schemas: Whether to scan READ_CSV calls in order to infer the CSVs\' schemas.\n\n    Returns:\n        The qualified expression.\n    ';C=schema;B=dialect;A=expression;C=ensure_schema(C,dialect=B);A=normalize_identifiers(A,dialect=B);A=qualify_tables(A,db=db,catalog=catalog,schema=C,dialect=B,infer_csv_schemas=infer_csv_schemas)
	if isolate_tables:A=isolate_table_selects(A,schema=C)
	if Dialect.get_or_raise(B).PREFER_CTE_ALIAS_COLUMN:A=pushdown_cte_alias_columns_func(A)
	if qualify_columns:A=qualify_columns_func(A,C,expand_alias_refs=expand_alias_refs,expand_stars=expand_stars,infer_schema=infer_schema)
	if quote_identifiers:A=quote_identifiers_func(A,dialect=B,identify=identify)
	if validate_qualify_columns:validate_qualify_columns_func(A)
	return A