from __future__ import annotations
_J='columns'
_I='coalesce'
_H='expressions'
_G='group'
_F='bigquery'
_E='table'
_D='alias'
_C=True
_B=False
_A=None
import itertools,typing as t
from sqlglot import alias,exp
from sqlglot.dialects.dialect import Dialect,DialectType
from sqlglot.errors import OptimizeError
from sqlglot.helper import seq_get,SingleValuedMapping
from sqlglot.optimizer.annotate_types import TypeAnnotator
from sqlglot.optimizer.scope import Scope,build_scope,traverse_scope,walk_in_scope
from sqlglot.optimizer.simplify import simplify_parens
from sqlglot.schema import Schema,ensure_schema
if t.TYPE_CHECKING:from sqlglot._typing import E
def qualify_columns(expression,schema,expand_alias_refs=_C,expand_stars=_C,infer_schema=_A):
	'\n    Rewrite sqlglot AST to have fully qualified columns.\n\n    Example:\n        >>> import sqlglot\n        >>> schema = {"tbl": {"col": "INT"}}\n        >>> expression = sqlglot.parse_one("SELECT col FROM tbl")\n        >>> qualify_columns(expression, schema).sql()\n        \'SELECT tbl.col AS col FROM tbl\'\n\n    Args:\n        expression: Expression to qualify.\n        schema: Database schema.\n        expand_alias_refs: Whether to expand references to aliases.\n        expand_stars: Whether to expand star queries. This is a necessary step\n            for most of the optimizer\'s rules to work; do not set to False unless you\n            know what you\'re doing!\n        infer_schema: Whether to infer the schema if missing.\n\n    Returns:\n        The qualified expression.\n\n    Notes:\n        - Currently only handles a single PIVOT or UNPIVOT operator\n    ';G=expand_alias_refs;F=expression;E=infer_schema;B=schema;B=ensure_schema(B);H=TypeAnnotator(B);E=B.empty if E is _A else E;D=Dialect.get_or_raise(B.dialect);I=D.PSEUDOCOLUMNS
	for A in traverse_scope(F):
		C=Resolver(A,B,infer_schema=E);_pop_table_column_aliases(A.ctes);_pop_table_column_aliases(A.derived_tables);J=_expand_using(A,C)
		if(B.empty or D.FORCE_EARLY_ALIAS_REF_EXPANSION)and G:_expand_alias_refs(A,C,expand_only_groupby=D.EXPAND_ALIAS_REFS_EARLY_ONLY_IN_GROUP_BY)
		_convert_columns_to_dots(A,C);_qualify_columns(A,C)
		if not B.empty and G:_expand_alias_refs(A,C)
		if not isinstance(A.expression,exp.UDTF):
			if expand_stars:_expand_stars(A,C,J,I,H)
			qualify_outputs(A)
		_expand_group_by(A,D);_expand_order_by(A,C)
		if D==_F:H.annotate_scope(A)
	return F
def validate_qualify_columns(expression):
	"Raise an `OptimizeError` if any columns aren't qualified";E=expression;C=[]
	for A in traverse_scope(E):
		if isinstance(A.expression,exp.Select):
			B=A.unqualified_columns
			if A.external_columns and not A.is_correlated_subquery and not A.pivots:D=A.external_columns[0];F=f" for table: '{D.table}'"if D.table else'';raise OptimizeError(f"Column '{D}' could not be resolved{F}")
			if B and A.pivots and A.pivots[0].unpivot:G=set(_unpivot_columns(A.pivots[0]));B=[A for A in B if A not in G]
			C.extend(B)
	if C:raise OptimizeError(f"Ambiguous columns: {C}")
	return E
def _unpivot_columns(unpivot):
	B=unpivot;C=[];A=B.args.get('field')
	if isinstance(A,exp.In)and isinstance(A.this,exp.Column):C.append(A.this)
	D=(B for A in B.expressions for B in A.find_all(exp.Column));return itertools.chain(C,D)
def _pop_table_column_aliases(derived_tables):
	'\n    Remove table column aliases.\n\n    For example, `col1` and `col2` will be dropped in SELECT ... FROM (SELECT ...) AS foo(col1, col2)\n    '
	for A in derived_tables:
		if isinstance(A.parent,exp.With)and A.parent.recursive:continue
		B=A.args.get(_D)
		if B:B.args.pop(_J,_A)
def _expand_using(scope,resolver):
	U='using';K=resolver;H=scope;E={}
	def P(source_name):
		A=source_name
		for B in K.get_source_columns(A):
			if B not in E:E[B]=A
	Q=list(H.find_all(exp.Join));V={A.alias_or_name for A in Q};I=[A for A in H.selected_sources if A not in V];F={}
	for W in I:P(W)
	for(X,J)in enumerate(Q):
		L=I[-1]
		if L:P(L)
		G=J.alias_or_name;I.append(G);M=J.args.get(U)
		if not M:continue
		R=K.get_source_columns(G);S=[];Y=len(M)
		for A in M:
			A=A.name;B=E.get(A)
			if not B or A not in R:
				if(E and'*'not in E)and R:raise OptimizeError(f"Cannot automatically join: {A}")
			B=B or L
			if X==0 or Y==1:N=exp.column(A,table=B)
			else:
				T=[exp.column(A,table=B)for B in I[:-1]if A in K.get_source_columns(B)]
				if len(T)>1:N=exp.func(_I,*T)
				else:N=exp.column(A,table=B)
			S.append(N.eq(exp.column(A,table=G)));D=F.setdefault(A,{})
			if B not in D:D[B]=_A
			if G not in D:D[G]=_A
		J.args.pop(U);J.set('on',exp.and_(*S,copy=_B))
	if F:
		for C in H.columns:
			if not C.table and C.name in F:
				D=F[C.name];Z=[exp.column(C.name,table=A)for A in D];O=exp.func(_I,*Z)
				if isinstance(C.parent,exp.Select):O=alias(O,alias=C.name,copy=_B)
				H.replace(C,O)
	return F
def _expand_alias_refs(scope,resolver,expand_only_groupby=_B):
	D=scope;A=D.expression
	if not isinstance(A,exp.Select):return
	F={}
	def B(node,resolve_table=_B,literal_index=_B):
		E=literal_index;D=resolve_table;C=node
		if not C or expand_only_groupby and not isinstance(C,exp.Group):return
		for A in walk_in_scope(C,prune=lambda node:node.is_star):
			if not isinstance(A,exp.Column):continue
			G=resolver.get_table(A.name)if D and not A.table else _A;B,J=F.get(A.name,(_A,1));H=B.find(exp.AggFunc)and(A.find_ancestor(exp.AggFunc)and not isinstance(A.find_ancestor(exp.Window,exp.Select),exp.Window))if B else _B
			if G and(not B or H):A.set(_E,G)
			elif not A.table and B and not H:
				if isinstance(B,exp.Literal)and(E or D):
					if E:A.replace(exp.Literal.number(J))
				else:
					A=A.replace(exp.paren(B));I=simplify_parens(A)
					if I is not A:A.replace(I)
	for(H,E)in enumerate(D.expression.selects):
		B(E)
		if isinstance(E,exp.Alias):F[E.alias]=E.this,H+1
	C=D
	while C.is_union:C=C.parent
	if C.is_cte:
		G=C.expression.parent
		if G.find_ancestor(exp.With).recursive:
			for I in G.args[_D].columns or G.this.selects:F.pop(I.output_name,_A)
	B(A.args.get('where'));B(A.args.get(_G),literal_index=_C);B(A.args.get('having'),resolve_table=_C);B(A.args.get('qualify'),resolve_table=_C);D.clear_cache()
def _expand_group_by(scope,dialect):
	B=scope;C=B.expression;A=C.args.get(_G)
	if not A:return
	A.set(_H,_expand_positional_references(B,A.expressions,dialect));C.set(_G,A)
def _expand_order_by(scope,resolver):
	E=resolver;B=scope;F=B.expression.args.get('order')
	if not F:return
	C=F.expressions
	for(A,G)in zip(C,_expand_positional_references(B,(A.this for A in C),E.schema.dialect,alias=_C)):
		for H in A.find_all(exp.AggFunc):
			for D in H.find_all(exp.Column):
				if not D.table:D.set(_E,E.get_table(D.name))
		A.set('this',G)
	if B.expression.args.get(_G):
		I={A.this:exp.column(A.alias_or_name)for A in B.expression.selects}
		for A in C:A=A.this;A.replace(exp.to_identifier(_select_by_pos(B,A).alias)if A.is_int else I.get(A,A))
def _expand_positional_references(scope,expressions,dialect,alias=_B):
	D=scope;B=[];E=_A
	for C in expressions:
		if C.is_int:
			A=_select_by_pos(D,t.cast(exp.Literal,C))
			if alias:B.append(exp.column(A.args[_D].copy()))
			else:
				A=A.this
				if dialect==_F:
					if E is _A:E={A.alias_or_name for A in D.expression.selects if A.alias_or_name in D.selected_sources}
					F=any(A.parts[0].name in E for A in A.find_all(exp.Column))
				else:F=_B
				if isinstance(A,exp.CONSTANTS)or A.find(exp.Explode,exp.Unnest)or F:B.append(C)
				else:B.append(A.copy())
		else:B.append(C)
	return B
def _select_by_pos(scope,node):
	try:return scope.expression.selects[int(node.this)-1].assert_is(exp.Alias)
	except IndexError:raise OptimizeError(f"Unknown output column: {node.name}")
def _convert_columns_to_dots(scope,resolver):
	'\n    Converts `Column` instances that represent struct field lookup into chained `Dots`.\n\n    Struct field lookups look like columns (e.g. "struct"."field"), but they need to be\n    qualified separately and represented as Dot(Dot(...(<table>.<column>, field1), field2, ...)).\n    ';A=scope;F=_B
	for D in itertools.chain(A.columns,A.stars):
		if isinstance(D,exp.Dot):continue
		B=D.table
		if B and B not in A.sources and(not A.parent or B not in A.parent.sources or not A.is_correlated_subquery):
			C,*E=D.parts
			if C.name in A.sources:B=C;C,*E=E
			else:B=resolver.get_table(C.name)
			if B:F=_C;D.replace(exp.Dot.build([exp.column(C,table=B),*E]))
	if F:A.clear_cache()
def _qualify_columns(scope,resolver):
	'Disambiguate columns, ensuring each column specifies a source';D=resolver;C=scope
	for A in C.columns:
		B=A.table;E=A.name
		if B and B in C.sources:
			F=D.get_source_columns(B)
			if F and E not in F and'*'not in F:raise OptimizeError(f"Unknown column: {E}")
		if not B:
			if C.pivots and not A.find_ancestor(exp.Pivot):A.set(_E,exp.to_identifier(C.pivots[0].alias));continue
			B=D.get_table(E)
			if B:A.set(_E,B)
	for G in C.pivots:
		for A in G.find_all(exp.Column):
			if not A.table and A.name in D.all_columns:
				B=D.get_table(A.name)
				if B:A.set(_E,B)
def _expand_struct_stars(expression):
	'[BigQuery] Expand/Flatten foo.bar.* where bar is a struct column';D=expression;B=t.cast(exp.Column,D.find(exp.Column))
	if not B.is_type(exp.DataType.Type.STRUCT):return[]
	B=B.copy();C=exp.ColumnDef(this=B.this,kind=B.type);E=D.parts[1:-1]
	for J in E[1:]:
		for A in t.cast(exp.DataType,C.kind).expressions:
			if not isinstance(A.this,exp.Identifier):return[]
			if A.name==J.name and A.kind.is_type(exp.DataType.Type.STRUCT):C=A;break
		else:return[]
	F=set();G=[]
	for A in t.cast(exp.DataType,C.kind).expressions:
		H=A.name
		if H in F or not isinstance(A.this,exp.Identifier):return[]
		F.add(H);I=A.this.copy();K,*L=[A.copy()for A in itertools.chain(E,[I])];M=exp.column(t.cast(exp.Identifier,K),table=B.args.get(_E),fields=L);G.append(alias(M,I,copy=_B))
	return G
def _expand_stars(scope,resolver,using_column_tables,pseudocolumns,annotator):
	'Expand stars to lists of column selections';R=pseudocolumns;Q=resolver;L=using_column_tables;E=scope;G=[];M={};N={};O={};S=set();Z=Q.schema.dialect;H=_A;J=_A;D=t.cast(t.Optional[exp.Pivot],seq_get(E.pivots,0))
	if isinstance(D,exp.Pivot)and not D.alias_column_names:
		if D.unpivot:
			H=[A.output_name for A in _unpivot_columns(D)];T=D.args.get('field')
			if isinstance(T,exp.In):J={B.output_name for A in T.expressions for B in A.find_all(exp.Column)}
		else:
			J=set(A.output_name for A in D.find_all(exp.Column));H=[A.output_name for A in D.args.get(_J,[])]
			if not H:H=[A.alias_or_name for A in D.expressions]
	U=Z==_F
	if U and any(isinstance(A,exp.Dot)for A in E.stars):annotator.annotate_scope(E)
	for B in E.expression.selects:
		C=[]
		if isinstance(B,exp.Star):C.extend(E.selected_sources);_add_except_columns(B,C,M);_add_replace_columns(B,C,N);_add_rename_columns(B,C,O)
		elif B.is_star:
			if not isinstance(B,exp.Dot):C.append(B.table);_add_except_columns(B.this,C,M);_add_replace_columns(B.this,C,N);_add_rename_columns(B.this,C,O)
			elif U:
				V=_expand_struct_stars(B)
				if V:G.extend(V);continue
		if not C:G.append(B);continue
		for I in C:
			if I not in E.sources:raise OptimizeError(f"Unknown table: {I}")
			F=Q.get_source_columns(I,only_visible=_C);F=F or E.outer_columns
			if R:F=[A for A in F if A.upper()not in R]
			if not F or'*'in F:return
			P=id(I);W=M.get(P)or set();a=O.get(P,{});b=N.get(P,{})
			if D:
				if H and J:K=[A for A in F if A not in J];K.extend(H)
				else:K=D.alias_column_names
				if K:G.extend(alias(exp.column(A,table=D.alias),A,copy=_B)for A in K if A not in W);continue
			for A in F:
				if A in W or A in S:continue
				if A in L and I in L[A]:S.add(A);C=L[A];c=[exp.column(A,table=B)for B in C];G.append(alias(exp.func(_I,*c),alias=A,copy=_B))
				else:X=a.get(A,A);Y=b.get(A)or exp.column(A,table=I);G.append(alias(Y,X,copy=_B)if X!=A else Y)
	if G and isinstance(E.expression,exp.Select):E.expression.set(_H,G)
def _add_except_columns(expression,tables,except_columns):
	A=expression.args.get('except')
	if not A:return
	B={A.name for A in A}
	for C in tables:except_columns[id(C)]=B
def _add_rename_columns(expression,tables,rename_columns):
	A=expression.args.get('rename')
	if not A:return
	B={A.this.name:A.alias for A in A}
	for C in tables:rename_columns[id(C)]=B
def _add_replace_columns(expression,tables,replace_columns):
	A=expression.args.get('replace')
	if not A:return
	B={A.alias:A for A in A}
	for C in tables:replace_columns[id(C)]=B
def qualify_outputs(scope_or_expression):
	'Ensure all output columns are aliased';C=scope_or_expression
	if isinstance(C,exp.Expression):
		B=build_scope(C)
		if not isinstance(B,Scope):return
	else:B=C
	D=[]
	for(E,(A,F))in enumerate(itertools.zip_longest(B.expression.selects,B.outer_columns)):
		if A is _A:break
		if isinstance(A,exp.Subquery):
			if not A.output_name:A.set(_D,exp.TableAlias(this=exp.to_identifier(f"_col_{E}")))
		elif not isinstance(A,exp.Alias)and not A.is_star:A=alias(A,alias=A.output_name or f"_col_{E}",copy=_B)
		if F:A.set(_D,exp.to_identifier(F))
		D.append(A)
	if isinstance(B.expression,exp.Select):B.expression.set(_H,D)
def quote_identifiers(expression,dialect=_A,identify=_C):'Makes sure all identifiers that need to be quoted are quoted.';return expression.transform(Dialect.get_or_raise(dialect).quote_identifier,identify=identify,copy=_B)
def pushdown_cte_alias_columns(expression):
	'\n    Pushes down the CTE alias columns into the projection,\n\n    This step is useful in Snowflake where the CTE alias columns can be referenced in the HAVING.\n\n    Example:\n        >>> import sqlglot\n        >>> expression = sqlglot.parse_one("WITH y (c) AS (SELECT SUM(a) FROM ( SELECT 1 a ) AS x HAVING c > 0) SELECT c FROM y")\n        >>> pushdown_cte_alias_columns(expression).sql()\n        \'WITH y(c) AS (SELECT SUM(a) AS c FROM (SELECT 1 AS a) AS x HAVING c > 0) SELECT c FROM y\'\n\n    Args:\n        expression: Expression to pushdown.\n\n    Returns:\n        The expression with the CTE aliases pushed down into the projection.\n    ';C=expression
	for B in C.find_all(exp.CTE):
		if B.alias_column_names:
			D=[]
			for(E,A)in zip(B.alias_column_names,B.this.expressions):
				if isinstance(A,exp.Alias):A.set(_D,E)
				else:A=alias(A,alias=E)
				D.append(A)
			B.this.set(_H,D)
	return C
class Resolver:
	'\n    Helper for resolving columns.\n\n    This is a class so we can lazily load some things and easily share them across functions.\n    '
	def __init__(A,scope,schema,infer_schema=_C):A.scope=scope;A.schema=schema;A._source_columns=_A;A._unambiguous_columns=_A;A._all_columns=_A;A._infer_schema=infer_schema;A._get_source_columns_cache={}
	def get_table(A,column_name):
		'\n        Get the table for a column name.\n\n        Args:\n            column_name: The column name to find the table for.\n        Returns:\n            The table name if it can be found/inferred.\n        '
		if A._unambiguous_columns is _A:A._unambiguous_columns=A._get_unambiguous_columns(A._get_all_source_columns())
		B=A._unambiguous_columns.get(column_name)
		if not B and A._infer_schema:
			D=tuple(B for(B,A)in A._get_all_source_columns().items()if not A or'*'in A)
			if len(D)==1:B=D[0]
		if B not in A.scope.selected_sources:return exp.to_identifier(B)
		C,F=A.scope.selected_sources.get(B)
		if isinstance(C,exp.Query):
			while C and C.alias!=B:C=C.parent
		E=C.args.get(_D)
		if E:return exp.to_identifier(E.this)
		return exp.to_identifier(B)
	@property
	def all_columns(self):
		'All available columns of all sources in this scope';A=self
		if A._all_columns is _A:A._all_columns={B for A in A._get_all_source_columns().values()for B in A}
		return A._all_columns
	def get_source_columns(A,name,only_visible=_B):
		'Resolve the source columns for a given source `name`.';H=only_visible;D=name;G=D,H
		if G not in A._get_source_columns_cache:
			if D not in A.scope.sources:raise OptimizeError(f"Unknown table: {D}")
			B=A.scope.sources[D]
			if isinstance(B,exp.Table):C=A.schema.column_names(B,H)
			elif isinstance(B,Scope)and isinstance(B.expression,(exp.Values,exp.Unnest)):
				C=B.expression.named_selects
				if A.schema.dialect==_F:
					if B.expression.is_type(exp.DataType.Type.STRUCT):
						for I in B.expression.type.expressions:C.append(I.name)
			else:C=B.expression.named_selects
			E,J=A.scope.selected_sources.get(D)or(_A,_A)
			if isinstance(E,Scope):F=E.expression.alias_column_names
			elif isinstance(E,exp.Expression):F=E.alias_column_names
			else:F=[]
			if F:C=[B or A for(A,B)in itertools.zip_longest(C,F)]
			A._get_source_columns_cache[G]=C
		return A._get_source_columns_cache[G]
	def _get_all_source_columns(A):
		if A._source_columns is _A:A._source_columns={B:A.get_source_columns(B)for(B,C)in itertools.chain(A.scope.selected_sources.items(),A.scope.lateral_sources.items())}
		return A._source_columns
	def _get_unambiguous_columns(L,source_columns):
		'\n        Find all the unambiguous columns in sources.\n\n        Args:\n            source_columns: Mapping of names to source columns.\n\n        Returns:\n            Mapping of column name to source name.\n        ';D=source_columns
		if not D:return{}
		B=list(D.items());E,F=B[0]
		if len(B)==1:return SingleValuedMapping(F,E)
		A={A:E for A in F};G=set(A)
		for(K,H)in B[1:]:
			I=set(H);J=G.intersection(I);G.update(H)
			for C in J:A.pop(C,_A)
			for C in I.difference(J):A[C]=K
		return A