from __future__ import annotations
_F='laterals'
_E='pivots'
_D=False
_C='joins'
_B=True
_A=None
import itertools,logging,typing as t
from collections import defaultdict
from enum import Enum,auto
from sqlglot import exp
from sqlglot.errors import OptimizeError
from sqlglot.helper import ensure_collection,find_new_name,seq_get
logger=logging.getLogger('sqlglot')
TRAVERSABLES=exp.Query,exp.DDL,exp.DML
class ScopeType(Enum):ROOT=auto();SUBQUERY=auto();DERIVED_TABLE=auto();CTE=auto();UNION=auto();UDTF=auto()
class Scope:
	'\n    Selection scope.\n\n    Attributes:\n        expression (exp.Select|exp.SetOperation): Root expression of this scope\n        sources (dict[str, exp.Table|Scope]): Mapping of source name to either\n            a Table expression or another Scope instance. For example:\n                SELECT * FROM x                     {"x": Table(this="x")}\n                SELECT * FROM x AS y                {"y": Table(this="x")}\n                SELECT * FROM (SELECT ...) AS y     {"y": Scope(...)}\n        lateral_sources (dict[str, exp.Table|Scope]): Sources from laterals\n            For example:\n                SELECT c FROM x LATERAL VIEW EXPLODE (a) AS c;\n            The LATERAL VIEW EXPLODE gets x as a source.\n        cte_sources (dict[str, Scope]): Sources from CTES\n        outer_columns (list[str]): If this is a derived table or CTE, and the outer query\n            defines a column list for the alias of this scope, this is that list of columns.\n            For example:\n                SELECT * FROM (SELECT ...) AS y(col1, col2)\n            The inner query would have `["col1", "col2"]` for its `outer_columns`\n        parent (Scope): Parent scope\n        scope_type (ScopeType): Type of this scope, relative to it\'s parent\n        subquery_scopes (list[Scope]): List of all child scopes for subqueries\n        cte_scopes (list[Scope]): List of all child scopes for CTEs\n        derived_table_scopes (list[Scope]): List of all child scopes for derived_tables\n        udtf_scopes (list[Scope]): List of all child scopes for user defined tabular functions\n        table_scopes (list[Scope]): derived_table_scopes + udtf_scopes, in the order that they\'re defined\n        union_scopes (list[Scope, Scope]): If this Scope is for a Union expression, this will be\n            a list of the left and right child scopes.\n    '
	def __init__(A,expression,sources=_A,outer_columns=_A,parent=_A,scope_type=ScopeType.ROOT,lateral_sources=_A,cte_sources=_A,can_be_correlated=_A):A.expression=expression;A.sources=sources or{};A.lateral_sources=lateral_sources or{};A.cte_sources=cte_sources or{};A.sources.update(A.lateral_sources);A.sources.update(A.cte_sources);A.outer_columns=outer_columns or[];A.parent=parent;A.scope_type=scope_type;A.subquery_scopes=[];A.derived_table_scopes=[];A.table_scopes=[];A.cte_scopes=[];A.union_scopes=[];A.udtf_scopes=[];A.can_be_correlated=can_be_correlated;A.clear_cache()
	def clear_cache(A):A._collected=_D;A._raw_columns=_A;A._stars=_A;A._derived_tables=_A;A._udtfs=_A;A._tables=_A;A._ctes=_A;A._subqueries=_A;A._selected_sources=_A;A._columns=_A;A._external_columns=_A;A._join_hints=_A;A._pivots=_A;A._references=_A
	def branch(A,expression,scope_type,sources=_A,cte_sources=_A,lateral_sources=_A,**E):'Branch from the current scope to a new, inner scope';D=lateral_sources;C=sources;B=scope_type;return Scope(expression=expression.unnest(),sources=C.copy()if C else _A,parent=A,scope_type=B,cte_sources={**A.cte_sources,**(cte_sources or{})},lateral_sources=D.copy()if D else _A,can_be_correlated=A.can_be_correlated or B in(ScopeType.SUBQUERY,ScopeType.UDTF),**E)
	def _collect(B):
		B._tables=[];B._ctes=[];B._subqueries=[];B._derived_tables=[];B._udtfs=[];B._raw_columns=[];B._stars=[];B._join_hints=[]
		for A in B.walk(bfs=_D):
			if A is B.expression:continue
			if isinstance(A,exp.Dot)and A.is_star:B._stars.append(A)
			elif isinstance(A,exp.Column):
				if isinstance(A.this,exp.Star):B._stars.append(A)
				else:B._raw_columns.append(A)
			elif isinstance(A,exp.Table)and not isinstance(A.parent,exp.JoinHint):B._tables.append(A)
			elif isinstance(A,exp.JoinHint):B._join_hints.append(A)
			elif isinstance(A,exp.UDTF):B._udtfs.append(A)
			elif isinstance(A,exp.CTE):B._ctes.append(A)
			elif _is_derived_table(A)and isinstance(A.parent,(exp.From,exp.Join,exp.Subquery)):B._derived_tables.append(A)
			elif isinstance(A,exp.UNWRAPPED_QUERIES):B._subqueries.append(A)
		B._collected=_B
	def _ensure_collected(A):
		if not A._collected:A._collect()
	def walk(A,bfs=_B,prune=_A):return walk_in_scope(A.expression,bfs=bfs,prune=_A)
	def find(A,*B,bfs=_B):return find_in_scope(A.expression,B,bfs=bfs)
	def find_all(A,*B,bfs=_B):return find_all_in_scope(A.expression,B,bfs=bfs)
	def replace(A,old,new):'\n        Replace `old` with `new`.\n\n        This can be used instead of `exp.Expression.replace` to ensure the `Scope` is kept up-to-date.\n\n        Args:\n            old (exp.Expression): old node\n            new (exp.Expression): new node\n        ';old.replace(new);A.clear_cache()
	@property
	def tables(self):'\n        List of tables in this scope.\n\n        Returns:\n            list[exp.Table]: tables\n        ';self._ensure_collected();return self._tables
	@property
	def ctes(self):'\n        List of CTEs in this scope.\n\n        Returns:\n            list[exp.CTE]: ctes\n        ';self._ensure_collected();return self._ctes
	@property
	def derived_tables(self):"\n        List of derived tables in this scope.\n\n        For example:\n            SELECT * FROM (SELECT ...) <- that's a derived table\n\n        Returns:\n            list[exp.Subquery]: derived tables\n        ";self._ensure_collected();return self._derived_tables
	@property
	def udtfs(self):'\n        List of "User Defined Tabular Functions" in this scope.\n\n        Returns:\n            list[exp.UDTF]: UDTFs\n        ';self._ensure_collected();return self._udtfs
	@property
	def subqueries(self):"\n        List of subqueries in this scope.\n\n        For example:\n            SELECT * FROM x WHERE a IN (SELECT ...) <- that's a subquery\n\n        Returns:\n            list[exp.Select | exp.SetOperation]: subqueries\n        ";self._ensure_collected();return self._subqueries
	@property
	def stars(self):'\n        List of star expressions (columns or dots) in this scope.\n        ';self._ensure_collected();return self._stars
	@property
	def columns(self):
		'\n        List of columns in this scope.\n\n        Returns:\n            list[exp.Column]: Column instances in this scope, plus any\n                Columns that reference this scope from correlated subqueries.\n        ';A=self
		if A._columns is _A:
			A._ensure_collected();D=A._raw_columns;E=[B for A in itertools.chain(A.subquery_scopes,A.udtf_scopes,(A for A in A.derived_table_scopes if A.can_be_correlated))for B in A.external_columns];F=set(A.expression.named_selects);A._columns=[]
			for C in D+E:
				B=C.find_ancestor(exp.Select,exp.Qualify,exp.Order,exp.Having,exp.Hint,exp.Table,exp.Star)
				if not B or C.table or isinstance(B,exp.Select)or isinstance(B,exp.Table)and not isinstance(B.this,exp.Func)or isinstance(B,exp.Order)and(isinstance(B.parent,exp.Window)or C.name not in F)or isinstance(B,exp.Star)and not C.arg_key=='except':A._columns.append(C)
		return A._columns
	@property
	def selected_sources(self):
		"\n        Mapping of nodes and sources that are actually selected from in this scope.\n\n        That is, all tables in a schema are selectable at any point. But a\n        table only becomes a selected source if it's included in a FROM or JOIN clause.\n\n        Returns:\n            dict[str, (exp.Table|exp.Select, exp.Table|Scope)]: selected sources and nodes\n        ";A=self
		if A._selected_sources is _A:
			C={}
			for(B,D)in A.references:
				if B in C:raise OptimizeError(f"Alias already used: {B}")
				if B in A.sources:C[B]=D,A.sources[B]
			A._selected_sources=C
		return A._selected_sources
	@property
	def references(self):
		A=self
		if A._references is _A:
			A._references=[]
			for C in A.tables:A._references.append((C.alias_or_name,C))
			for B in itertools.chain(A.derived_tables,A.udtfs):A._references.append((B.alias,B if B.args.get(_E)else B.unnest()))
		return A._references
	@property
	def external_columns(self):
		"\n        Columns that appear to reference sources in outer scopes.\n\n        Returns:\n            list[exp.Column]: Column instances that don't reference\n                sources in the current scope.\n        ";A=self
		if A._external_columns is _A:
			if isinstance(A.expression,exp.SetOperation):B,C=A.union_scopes;A._external_columns=B.external_columns+C.external_columns
			else:A._external_columns=[B for B in A.columns if B.table not in A.selected_sources]
		return A._external_columns
	@property
	def unqualified_columns(self):'\n        Unqualified columns in the current scope.\n\n        Returns:\n             list[exp.Column]: Unqualified columns\n        ';return[A for A in self.columns if not A.table]
	@property
	def join_hints(self):
		'\n        Hints that exist in the scope that reference tables\n\n        Returns:\n            list[exp.JoinHint]: Join hints that are referenced within the scope\n        '
		if self._join_hints is _A:return[]
		return self._join_hints
	@property
	def pivots(self):
		A=self
		if not A._pivots:A._pivots=[B for(C,A)in A.references for B in A.args.get(_E)or[]]
		return A._pivots
	def source_columns(A,source_name):'\n        Get all columns in the current scope for a particular source.\n\n        Args:\n            source_name (str): Name of the source\n        Returns:\n            list[exp.Column]: Column instances that reference `source_name`\n        ';return[A for A in A.columns if A.table==source_name]
	@property
	def is_subquery(self):'Determine if this scope is a subquery';return self.scope_type==ScopeType.SUBQUERY
	@property
	def is_derived_table(self):'Determine if this scope is a derived table';return self.scope_type==ScopeType.DERIVED_TABLE
	@property
	def is_union(self):'Determine if this scope is a union';return self.scope_type==ScopeType.UNION
	@property
	def is_cte(self):'Determine if this scope is a common table expression';return self.scope_type==ScopeType.CTE
	@property
	def is_root(self):'Determine if this is the root scope';return self.scope_type==ScopeType.ROOT
	@property
	def is_udtf(self):'Determine if this scope is a UDTF (User Defined Table Function)';return self.scope_type==ScopeType.UDTF
	@property
	def is_correlated_subquery(self):'Determine if this scope is a correlated subquery';return bool(self.can_be_correlated and self.external_columns)
	def rename_source(A,old_name,new_name):'Rename a source in this scope';B=A.sources.pop(old_name or'',[]);A.sources[new_name]=B
	def add_source(A,name,source):'Add a source to this scope';A.sources[name]=source;A.clear_cache()
	def remove_source(A,name):'Remove a source from this scope';A.sources.pop(name,_A);A.clear_cache()
	def __repr__(A):return f"Scope<{A.expression.sql()}>"
	def traverse(D):
		'\n        Traverse the scope tree from this node.\n\n        Yields:\n            Scope: scope instances in depth-first-search post-order\n        ';B=[D];C=[]
		while B:A=B.pop();C.append(A);B.extend(itertools.chain(A.cte_scopes,A.union_scopes,A.table_scopes,A.subquery_scopes))
		yield from reversed(C)
	def ref_count(B):
		'\n        Count the number of times each scope in this tree is referenced.\n\n        Returns:\n            dict[int, int]: Mapping of Scope instance ID to reference count\n        ';A=defaultdict(lambda:0)
		for C in B.traverse():
			for(E,D)in C.selected_sources.values():A[id(D)]+=1
		return A
def traverse_scope(expression):
	'\n    Traverse an expression by its "scopes".\n\n    "Scope" represents the current context of a Select statement.\n\n    This is helpful for optimizing queries, where we need more information than\n    the expression tree itself. For example, we might care about the source\n    names within a subquery. Returns a list because a generator could result in\n    incomplete properties which is confusing.\n\n    Examples:\n        >>> import sqlglot\n        >>> expression = sqlglot.parse_one("SELECT a FROM (SELECT a FROM x) AS y")\n        >>> scopes = traverse_scope(expression)\n        >>> scopes[0].expression.sql(), list(scopes[0].sources)\n        (\'SELECT a FROM x\', [\'x\'])\n        >>> scopes[1].expression.sql(), list(scopes[1].sources)\n        (\'SELECT a FROM (SELECT a FROM x) AS y\', [\'y\'])\n\n    Args:\n        expression: Expression to traverse\n\n    Returns:\n        A list of the created scope instances\n    ';A=expression
	if isinstance(A,TRAVERSABLES):return list(_traverse_scope(Scope(A)))
	return[]
def build_scope(expression):'\n    Build a scope tree.\n\n    Args:\n        expression: Expression to build the scope tree for.\n\n    Returns:\n        The root scope\n    ';return seq_get(traverse_scope(expression),-1)
def _traverse_scope(scope):
	A=scope;B=A.expression
	if isinstance(B,exp.Select):yield from _traverse_select(A)
	elif isinstance(B,exp.SetOperation):yield from _traverse_ctes(A);yield from _traverse_union(A);return
	elif isinstance(B,exp.Subquery):
		if A.is_root:yield from _traverse_select(A)
		else:yield from _traverse_subqueries(A)
	elif isinstance(B,exp.Table):yield from _traverse_tables(A)
	elif isinstance(B,exp.UDTF):yield from _traverse_udtfs(A)
	elif isinstance(B,exp.DDL):
		if isinstance(B.expression,exp.Query):yield from _traverse_ctes(A);yield from _traverse_scope(Scope(B.expression,cte_sources=A.cte_sources))
		return
	elif isinstance(B,exp.DML):
		yield from _traverse_ctes(A)
		for C in find_all_in_scope(B,exp.Query):
			if not isinstance(C.parent,exp.CTE):yield from _traverse_scope(Scope(C,cte_sources=A.cte_sources))
		return
	else:logger.warning("Cannot traverse scope %s with type '%s'",B,type(B));return
	yield A
def _traverse_select(scope):A=scope;yield from _traverse_ctes(A);yield from _traverse_tables(A);yield from _traverse_subqueries(A)
def _traverse_union(scope):
	A=scope;C=_A;E=[A];F=[A.expression.right,A.expression.left]
	while F:
		D=F.pop();B=E[-1];G=B.branch(D,outer_columns=B.outer_columns,scope_type=ScopeType.UNION)
		if isinstance(D,exp.SetOperation):yield from _traverse_ctes(G);E.append(G);F.extend([D.right,D.left]);continue
		for A in _traverse_scope(G):yield A
		if C:E.pop();B.union_scopes=[C,A];C=B;yield B
		else:C=A
def _traverse_ctes(scope):
	A=scope;B={}
	for D in A.ctes:
		E=D.alias;F=A.expression.args.get('with')
		if F and F.recursive:
			G=D.this
			if isinstance(G,exp.SetOperation):B[E]=A.branch(G.this,scope_type=ScopeType.CTE)
		C=_A
		for C in _traverse_scope(A.branch(D.this,cte_sources=B,outer_columns=D.alias_column_names,scope_type=ScopeType.CTE)):yield C
		if C:B[E]=C;A.cte_scopes.append(C)
	A.sources.update(B);A.cte_sources.update(B)
def _is_derived_table(expression):'\n    We represent (tbl1 JOIN tbl2) as a Subquery, but it\'s not really a "derived table",\n    as it doesn\'t introduce a new scope. If an alias is present, it shadows all names\n    under the Subquery, so that\'s one exception to this rule.\n    ';A=expression;return isinstance(A,exp.Subquery)and bool(A.alias or isinstance(A.this,exp.UNWRAPPED_QUERIES))
def _traverse_tables(scope):
	B=scope;C={};D=[];H=B.expression.args.get('from')
	if H:D.append(H.this)
	for M in B.expression.args.get(_C)or[]:D.append(M.this)
	if isinstance(B.expression,exp.Table):D.append(B.expression)
	D.extend(B.expression.args.get(_F)or[])
	for A in D:
		if isinstance(A,exp.Table):
			F=A.name;G=A.alias_or_name
			if F in B.sources and not A.db:
				I=A.args.get(_E)
				if I:C[I[0].alias]=A
				else:C[G]=B.sources[F]
			elif G in C:C[find_new_name(C,F)]=A
			else:C[G]=A
			if A is not B.expression:D.extend(A.this for A in A.args.get(_C)or[])
			continue
		if not isinstance(A,exp.DerivedTable):continue
		if isinstance(A,exp.UDTF):J=C;K=ScopeType.UDTF;L=B.udtf_scopes
		elif _is_derived_table(A):J=_A;K=ScopeType.DERIVED_TABLE;L=B.derived_table_scopes;D.extend(A.this for A in A.args.get(_C)or[])
		else:D.append(A.this);D.extend(A.this for A in A.args.get(_C)or[]);continue
		for E in _traverse_scope(B.branch(A,lateral_sources=J,outer_columns=A.alias_column_names,scope_type=K)):yield E;C[A.alias]=E
		L.append(E);B.table_scopes.append(E)
	B.sources.update(C)
def _traverse_subqueries(scope):
	A=scope
	for D in A.subqueries:
		B=_A
		for C in _traverse_scope(A.branch(D,scope_type=ScopeType.SUBQUERY)):yield C;B=C
		A.subquery_scopes.append(B)
def _traverse_udtfs(scope):
	A=scope
	if isinstance(A.expression,exp.Unnest):C=A.expression.expressions
	elif isinstance(A.expression,exp.Lateral):C=[A.expression.this]
	else:C=[]
	E={}
	for B in C:
		if _is_derived_table(B):
			F=_A
			for D in _traverse_scope(A.branch(B,scope_type=ScopeType.SUBQUERY,outer_columns=B.alias_column_names)):yield D;F=D;E[B.alias]=D
			A.subquery_scopes.append(F)
	A.sources.update(E)
def walk_in_scope(expression,bfs=_B,prune=_A):
	'\n    Returns a generator object which visits all nodes in the syntrax tree, stopping at\n    nodes that start child scopes.\n\n    Args:\n        expression (exp.Expression):\n        bfs (bool): if set to True the BFS traversal order will be applied,\n            otherwise the DFS traversal will be used instead.\n        prune ((node, parent, arg_key) -> bool): callable that returns True if\n            the generator should stop traversing this branch of the tree.\n\n    Yields:\n        tuple[exp.Expression, Optional[exp.Expression], str]: node, parent, arg key\n    ';D=prune;C=expression;B=_D
	for A in C.walk(bfs=bfs,prune=lambda n:B or D and D(n)):
		B=_D;yield A
		if A is C:continue
		if isinstance(A,exp.CTE)or isinstance(A.parent,(exp.From,exp.Join,exp.Subquery))and(_is_derived_table(A)or isinstance(A,exp.UDTF))or isinstance(A,exp.UNWRAPPED_QUERIES):
			B=_B
			if isinstance(A,(exp.Subquery,exp.UDTF)):
				for E in(_C,_F,_E):
					for F in A.args.get(E)or[]:yield from walk_in_scope(F,bfs=bfs)
def find_all_in_scope(expression,expression_types,bfs=_B):
	'\n    Returns a generator object which visits all nodes in this scope and only yields those that\n    match at least one of the specified expression types.\n\n    This does NOT traverse into subscopes.\n\n    Args:\n        expression (exp.Expression):\n        expression_types (tuple[type]|type): the expression type(s) to match.\n        bfs (bool): True to use breadth-first search, False to use depth-first.\n\n    Yields:\n        exp.Expression: nodes\n    ';A=expression
	for A in walk_in_scope(A,bfs=bfs):
		if isinstance(A,tuple(ensure_collection(expression_types))):yield A
def find_in_scope(expression,expression_types,bfs=_B):'\n    Returns the first node in this scope which matches at least one of the specified types.\n\n    This does NOT traverse into subscopes.\n\n    Args:\n        expression (exp.Expression):\n        expression_types (tuple[type]|type): the expression type(s) to match.\n        bfs (bool): True to use breadth-first search, False to use depth-first.\n\n    Returns:\n        exp.Expression: the node which matches the criteria or None if no node matching\n        the criteria was found.\n    ';return next(find_all_in_scope(expression,expression_types,bfs=bfs),_A)