from __future__ import annotations
_E='quarter'
_D='alias'
_C=True
_B=None
_A=False
import datetime,logging,functools,itertools,typing as t
from collections import deque,defaultdict
from functools import reduce
import sqlglot
from sqlglot import Dialect,exp
from sqlglot.helper import first,merge_ranges,while_changing
from sqlglot.optimizer.scope import find_all_in_scope,walk_in_scope
if t.TYPE_CHECKING:from sqlglot.dialects.dialect import DialectType;DateTruncBinaryTransform=t.Callable[[exp.Expression,datetime.date,str,Dialect,exp.DataType],t.Optional[exp.Expression]]
logger=logging.getLogger('sqlglot')
FINAL='final'
TINYINT_MIN=-128
TINYINT_MAX=127
UTINYINT_MIN=0
UTINYINT_MAX=255
class UnsupportedUnit(Exception):0
def simplify(expression,constant_propagation=_A,dialect=_B,max_depth=_B):
	'\n    Rewrite sqlglot AST to simplify expressions.\n\n    Example:\n        >>> import sqlglot\n        >>> expression = sqlglot.parse_one("TRUE AND TRUE")\n        >>> simplify(expression).sql()\n        \'TRUE\'\n\n    Args:\n        expression: expression to simplify\n        constant_propagation: whether the constant propagation rule should be used\n        max_depth: Chains of Connectors (AND, OR, etc) exceeding `max_depth` will be skipped\n    Returns:\n        sqlglot.Expression: simplified expression\n    ';E=max_depth;D=dialect;A=expression;D=Dialect.get_or_raise(D)
	def H(expression,root=_C):
		C=root;B=expression
		if E and isinstance(B,exp.Connector)and not isinstance(B.parent,exp.Connector):
			I=connector_depth(B)
			if I>E:logger.info(f"Skipping simplification because connector depth {I} exceeds max {E}");return B
		if B.meta.get(FINAL):return B
		F=B.args.get('group')
		if F and hasattr(B,'selects'):
			J=set(F.expressions);F.meta[FINAL]=_C
			for K in B.selects:
				for A in K.walk():
					if A in J:K.meta[FINAL]=_C;break
			G=B.args.get('having')
			if G:
				for A in G.walk():
					if A in J:G.meta[FINAL]=_C;break
		A=B;A=rewrite_between(A);A=uniq_sort(A,C);A=absorb_and_eliminate(A,C);A=simplify_concat(A);A=simplify_conditionals(A)
		if constant_propagation:A=propagate_constants(A,C)
		exp.replace_children(A,lambda e:H(e,_A));A=simplify_not(A);A=flatten(A);A=simplify_connectors(A,C);A=remove_complements(A,C);A=simplify_coalesce(A);A.parent=B.parent;A=simplify_literals(A,C);A=simplify_equality(A);A=simplify_parens(A);A=simplify_datetrunc(A,D);A=sort_comparison(A);A=simplify_startswith(A)
		if C:B.replace(A)
		return A
	A=while_changing(A,H);remove_where_true(A);return A
def connector_depth(expression):
	'\n    Determine the maximum depth of a tree of Connectors.\n\n    For example:\n        >>> from sqlglot import parse_one\n        >>> connector_depth(parse_one("a AND b AND c AND d"))\n        3\n    ';A=expression;B=deque([(A,0)]);D=0
	while B:
		A,C=B.pop()
		if not isinstance(A,exp.Connector):continue
		C+=1;D=max(C,D);B.append((A.left,C));B.append((A.right,C))
	return D
def catch(*B):
	'Decorator that ignores a simplification function if any of `exceptions` are raised'
	def A(func):
		def A(expression,*C,**D):
			A=expression
			try:return func(A,*C,**D)
			except B:return A
		return A
	return A
def rewrite_between(expression):
	'Rewrite x between y and z to x >= y AND x <= z.\n\n    This is done because comparison simplification is only done on lt/lte/gt/gte.\n    ';A=expression
	if isinstance(A,exp.Between):
		B=isinstance(A.parent,exp.Not);A=exp.and_(exp.GTE(this=A.this.copy(),expression=A.args['low']),exp.LTE(this=A.this.copy(),expression=A.args['high']),copy=_A)
		if B:A=exp.paren(A,copy=_A)
	return A
COMPLEMENT_COMPARISONS={exp.LT:exp.GTE,exp.GT:exp.LTE,exp.LTE:exp.GT,exp.GTE:exp.LT,exp.EQ:exp.NEQ,exp.NEQ:exp.EQ}
def simplify_not(expression):
	"\n    Demorgan's Law\n    NOT (x OR y) -> NOT x AND NOT y\n    NOT (x AND y) -> NOT x OR NOT y\n    ";C=expression
	if isinstance(C,exp.Not):
		A=C.this
		if is_null(A):return exp.null()
		if A.__class__ in COMPLEMENT_COMPARISONS:return COMPLEMENT_COMPARISONS[A.__class__](this=A.this,expression=A.expression)
		if isinstance(A,exp.Paren):
			B=A.unnest()
			if isinstance(B,exp.And):return exp.paren(exp.or_(exp.not_(B.left,copy=_A),exp.not_(B.right,copy=_A),copy=_A))
			if isinstance(B,exp.Or):return exp.paren(exp.and_(exp.not_(B.left,copy=_A),exp.not_(B.right,copy=_A),copy=_A))
			if is_null(B):return exp.null()
		if always_true(A):return exp.false()
		if is_false(A):return exp.true()
		if isinstance(A,exp.Not):return A.this
	return C
def flatten(expression):
	'\n    A AND (B AND C) -> A AND B AND C\n    A OR (B OR C) -> A OR B OR C\n    ';A=expression
	if isinstance(A,exp.Connector):
		for B in A.args.values():
			C=B.unnest()
			if isinstance(C,A.__class__):B.replace(C)
	return A
def simplify_connectors(expression,root=_C):
	A=expression
	def B(expression,left,right):
		C=expression;B=right;A=left
		if isinstance(C,exp.And):
			if is_false(A)or is_false(B):return exp.false()
			if is_zero(A)or is_zero(B):return exp.false()
			if is_null(A)or is_null(B):return exp.null()
			if always_true(A)and always_true(B):return exp.true()
			if always_true(A):return B
			if always_true(B):return A
			return _simplify_comparison(C,A,B)
		elif isinstance(C,exp.Or):
			if always_true(A)or always_true(B):return exp.true()
			if is_null(A)and is_null(B)or is_null(A)and always_false(B)or always_false(A)and is_null(B):return exp.null()
			if is_false(A):return B
			if is_false(B):return A
			return _simplify_comparison(C,A,B,or_=_C)
		elif isinstance(C,exp.Xor):
			if A==B:return exp.false()
	if isinstance(A,exp.Connector):return _flat_simplify(A,B,root)
	return A
LT_LTE=exp.LT,exp.LTE
GT_GTE=exp.GT,exp.GTE
COMPARISONS=*LT_LTE,*GT_GTE,exp.EQ,exp.NEQ,exp.Is
INVERSE_COMPARISONS={exp.LT:exp.GT,exp.GT:exp.LT,exp.LTE:exp.GTE,exp.GTE:exp.LTE}
NONDETERMINISTIC=exp.Rand,exp.Randn
AND_OR=exp.And,exp.Or
def _simplify_comparison(expression,left,right,or_=_A):
	I=or_;H=right;G=left
	if isinstance(G,COMPARISONS)and isinstance(H,COMPARISONS):
		N,O=G.args.values();P,Q=H.args.values();K={N,O};L={P,Q};M=K&L;J={A for A in M if not _is_constant(A)and not A.find(*NONDETERMINISTIC)}
		if M and J:
			try:A=first(K-J);B=first(L-J)
			except StopIteration:return expression
			if A.is_number and B.is_number:A=A.to_py();B=B.to_py()
			elif A.is_string and B.is_string:A=A.name;B=B.name
			else:
				A=extract_date(A)
				if not A:return
				B=extract_date(B)
				if not B:return
				A,B=cast_as_datetime(A),cast_as_datetime(B)
			for((E,C),(F,D))in itertools.permutations(((G,A),(H,B))):
				if isinstance(E,LT_LTE)and isinstance(F,LT_LTE):return G if(C>D if I else C<=D)else H
				if isinstance(E,GT_GTE)and isinstance(F,GT_GTE):return G if(C<D if I else C>=D)else H
				if not I:
					if isinstance(E,exp.LT)and isinstance(F,GT_GTE):
						if C<=D:return exp.false()
					elif isinstance(E,exp.GT)and isinstance(F,LT_LTE):
						if C>=D:return exp.false()
					elif isinstance(E,exp.EQ):
						if isinstance(F,exp.LT):return exp.false()if C>=D else E
						if isinstance(F,exp.LTE):return exp.false()if C>D else E
						if isinstance(F,exp.GT):return exp.false()if C<=D else E
						if isinstance(F,exp.GTE):return exp.false()if C<D else E
						if isinstance(F,exp.NEQ):return exp.false()if C==D else E
def remove_complements(expression,root=_C):
	'\n    Removing complements.\n\n    A AND NOT A -> FALSE\n    A OR NOT A -> TRUE\n    ';A=expression
	if isinstance(A,AND_OR)and(root or not A.same_parent):
		B=set(A.flatten())
		for C in B:
			if isinstance(C,exp.Not)and C.this in B:return exp.false()if isinstance(A,exp.And)else exp.true()
	return A
def uniq_sort(expression,root=_C):
	'\n    Uniq and sort a connector.\n\n    C AND A AND B AND B -> A AND B AND C\n    ';A=expression
	if isinstance(A,exp.Connector)and(root or not A.same_parent):
		D=tuple(A.flatten())
		if isinstance(A,exp.Xor):E=exp.xor;B=_B;C=tuple((gen(A),A)for A in D)
		else:E=exp.and_ if isinstance(A,exp.And)else exp.or_;B={gen(A):A for A in D};C=tuple(B.items())
		for(F,(G,H))in enumerate(C[1:]):
			if G<C[F][0]:A=E(*(A for(B,A)in sorted(C)),copy=_A);break
		else:
			if B and len(B)<len(D):A=E(*B.values(),copy=_A)
	return A
def absorb_and_eliminate(expression,root=_C):
	'\n    absorption:\n        A AND (A OR B) -> A\n        A OR (A AND B) -> A\n        A AND (NOT A OR B) -> A AND B\n        A OR (NOT A AND B) -> A OR B\n    elimination:\n        (A AND B) OR (A AND NOT B) -> A\n        (A OR B) AND (A OR NOT B) -> A\n    ';D=expression
	if isinstance(D,AND_OR)and(root or not D.same_parent):
		E=exp.Or if isinstance(D,exp.And)else exp.And;I=tuple(D.flatten());F=set();G=defaultdict(list);H=defaultdict(list)
		for A in I:
			F.add(A)
			if not isinstance(A,E):G[A].append({A});continue
			J=set(A.flatten())
			for M in J:G[M].append(J)
			B,C=A.unnest_operands()
			if isinstance(B,exp.Not):H[frozenset((B.this,C))].append((A,C))
			if isinstance(C,exp.Not):H[frozenset((B,C.this))].append((A,B))
		for A in I:
			if not isinstance(A,E):continue
			B,C=A.unnest_operands()
			if isinstance(B,exp.Not)and B.this in F:B.replace(exp.true()if E==exp.And else exp.false());continue
			if isinstance(C,exp.Not)and C.this in F:C.replace(exp.true()if E==exp.And else exp.false());continue
			K=set(A.flatten())
			if any(any(A<K for A in G[A])for A in K):A.replace(exp.false()if E==exp.And else exp.true());continue
			for(N,L)in H[frozenset((B,C))]:A.replace(L);N.replace(L)
	return D
def propagate_constants(expression,root=_C):
	'\n    Propagate constants for conjunctions in DNF:\n\n    SELECT * FROM t WHERE a = b AND b = 5 becomes\n    SELECT * FROM t WHERE a = 5 AND b = 5\n\n    Reference: https://www.sqlite.org/optoverview.html\n    ';A=expression
	if isinstance(A,exp.And)and(root or not A.same_parent)and sqlglot.optimizer.normalize.normalized(A,dnf=_C):
		C={}
		for D in walk_in_scope(A,prune=lambda node:isinstance(node,exp.If)):
			if isinstance(D,exp.EQ):
				E,F=D.left,D.right
				if isinstance(E,exp.Column)and isinstance(F,exp.Literal):C[E]=id(E),F
		if C:
			for B in find_all_in_scope(A,exp.Column):
				G=B.parent;H,I=C.get(B)or(_B,_B)
				if H is not _B and id(B)!=H and not(isinstance(G,exp.Is)and isinstance(G.expression,exp.Null)):B.replace(I.copy())
	return A
INVERSE_DATE_OPS={exp.DateAdd:exp.Sub,exp.DateSub:exp.Add,exp.DatetimeAdd:exp.Sub,exp.DatetimeSub:exp.Add}
INVERSE_OPS={**INVERSE_DATE_OPS,exp.Add:exp.Sub,exp.Sub:exp.Add}
def _is_number(expression):return expression.is_number
def _is_interval(expression):A=expression;return isinstance(A,exp.Interval)and extract_interval(A)is not _B
@catch(ModuleNotFoundError,UnsupportedUnit)
def simplify_equality(expression):
	"\n    Use the subtraction and addition properties of equality to simplify expressions:\n\n        x + 1 = 3 becomes x = 2\n\n    There are two binary operations in the above expression: + and =\n    Here's how we reference all the operands in the code below:\n\n          l     r\n        x + 1 = 3\n        a   b\n    ";B=expression
	if isinstance(B,COMPARISONS):
		A,E=B.left,B.right
		if A.__class__ not in INVERSE_OPS:return B
		if E.is_number:F=_is_number;G=_is_number
		elif _is_date_literal(E):F=_is_date_literal;G=_is_interval
		else:return B
		if A.__class__ in INVERSE_DATE_OPS:A=t.cast(exp.IntervalOp,A);C=A.this;D=A.interval()
		else:A=t.cast(exp.Binary,A);C,D=A.left,A.right
		if not F(C)and G(D):0
		elif not F(D)and G(C):C,D=D,C
		else:return B
		return B.__class__(this=C,expression=INVERSE_OPS[A.__class__](this=E,expression=D))
	return B
def simplify_literals(expression,root=_C):
	A=expression
	if isinstance(A,exp.Binary)and not isinstance(A,exp.Connector):return _flat_simplify(A,_simplify_binary,root)
	if isinstance(A,exp.Neg)and isinstance(A.this,exp.Neg):return A.this.this
	if type(A)in INVERSE_DATE_OPS:return _simplify_binary(A,A.this,A.interval())or A
	return A
NULL_OK=exp.NullSafeEQ,exp.NullSafeNEQ,exp.PropertyEQ
def _simplify_integer_cast(expr):
	A=expr
	if isinstance(A,exp.Cast)and isinstance(A.this,exp.Cast):B=_simplify_integer_cast(A.this)
	else:B=A.this
	if isinstance(A,exp.Cast)and B.is_int:
		C=B.to_py()
		if TINYINT_MIN<=C<=TINYINT_MAX and A.to.this in exp.DataType.SIGNED_INTEGER_TYPES or UTINYINT_MIN<=C<=UTINYINT_MAX and A.to.this in exp.DataType.UNSIGNED_INTEGER_TYPES:return B
	return A
def _simplify_binary(expression,a,b):
	A=expression
	if isinstance(A,COMPARISONS):a=_simplify_integer_cast(a);b=_simplify_integer_cast(b)
	if isinstance(A,exp.Is):
		if isinstance(b,exp.Not):G=b.this;F=_C
		else:G=b;F=_A
		if is_null(G):
			if isinstance(a,exp.Literal):return exp.true()if F else exp.false()
			if is_null(a):return exp.false()if F else exp.true()
	elif isinstance(A,NULL_OK):return
	elif is_null(a)or is_null(b):return exp.null()
	if a.is_number and b.is_number:
		C=a.to_py();D=b.to_py()
		if isinstance(A,exp.Add):return exp.Literal.number(C+D)
		if isinstance(A,exp.Mul):return exp.Literal.number(C*D)
		if isinstance(A,exp.Sub):return exp.Literal.number(C-D)if a.parent is b.parent else _B
		if isinstance(A,exp.Div):
			if isinstance(C,int)and isinstance(D,int)or a.parent is not b.parent:return
			return exp.Literal.number(C/D)
		B=eval_boolean(A,C,D)
		if B:return B
	elif a.is_string and b.is_string:
		B=eval_boolean(A,a.this,b.this)
		if B:return B
	elif _is_date_literal(a)and isinstance(b,exp.Interval):
		E,b=extract_date(a),extract_interval(b)
		if E and b:
			if isinstance(A,(exp.Add,exp.DateAdd,exp.DatetimeAdd)):return date_literal(E+b,extract_type(a))
			if isinstance(A,(exp.Sub,exp.DateSub,exp.DatetimeSub)):return date_literal(E-b,extract_type(a))
	elif isinstance(a,exp.Interval)and _is_date_literal(b):
		a,E=extract_interval(a),extract_date(b)
		if a and b and isinstance(A,exp.Add):return date_literal(a+E,extract_type(b))
	elif _is_date_literal(a)and _is_date_literal(b):
		if isinstance(A,exp.Predicate):
			a,b=extract_date(a),extract_date(b);B=eval_boolean(A,a,b)
			if B:return B
def simplify_parens(expression):
	C=expression
	if not isinstance(C,exp.Paren):return C
	A=C.this;B=C.parent;D=isinstance(B,exp.Predicate)
	if not isinstance(A,exp.Select)and not isinstance(B,exp.SubqueryPredicate)and(not isinstance(B,(exp.Condition,exp.Binary))or isinstance(B,exp.Paren)or not isinstance(A,exp.Binary)and not(isinstance(A,(exp.Not,exp.Is))and D)or isinstance(A,exp.Predicate)and not D or isinstance(A,exp.Add)and isinstance(B,exp.Add)or isinstance(A,exp.Mul)and isinstance(B,exp.Mul)or isinstance(A,exp.Mul)and isinstance(B,(exp.Add,exp.Sub))):return A
	return C
def _is_nonnull_constant(expression):A=expression;return isinstance(A,exp.NONNULL_CONSTANTS)or _is_date_literal(A)
def _is_constant(expression):A=expression;return isinstance(A,exp.CONSTANTS)or _is_date_literal(A)
def simplify_coalesce(expression):
	A=expression
	if isinstance(A,exp.Coalesce)and(not A.expressions or _is_nonnull_constant(A.this))and not isinstance(A.parent,exp.Hint):return A.this
	if not isinstance(A,COMPARISONS):return A
	if isinstance(A.left,exp.Coalesce):B=A.left;C=A.right
	elif isinstance(A.right,exp.Coalesce):B=A.right;C=A.left
	else:return A
	if not _is_constant(C):return A
	for(E,D)in enumerate(B.expressions):
		if _is_constant(D):break
	else:return A
	B.set('expressions',B.expressions[:E]);B=B if B.expressions else B.this;return exp.paren(exp.or_(exp.and_(B.is_(exp.null()).not_(copy=_A),A.copy(),copy=_A),exp.and_(B.is_(exp.null()),type(A)(this=D.copy(),expression=C.copy()),copy=_A),copy=_A))
CONCATS=exp.Concat,exp.DPipe
def simplify_concat(expression):
	'Reduces all groups that contain string literals by concatenating them.';J='coalesce';I='safe';A=expression
	if not isinstance(A,CONCATS)or isinstance(A,exp.ConcatWs)and not A.expressions[0].is_string:return A
	if isinstance(A,exp.ConcatWs):D,*E=A.expressions;F=D.name;C=exp.ConcatWs;G={}
	else:E=A.expressions;F='';C=exp.Concat;G={I:A.args.get(I),J:A.args.get(J)}
	B=[]
	for(K,H)in itertools.groupby(E or A.flatten(),lambda e:e.is_string):
		if K:B.append(exp.Literal.string(F.join(A.name for A in H)))
		else:B.extend(H)
	if len(B)==1 and B[0].is_string:return B[0]
	if C is exp.ConcatWs:B=[D]+B
	elif isinstance(A,exp.DPipe):return reduce(lambda x,y:exp.DPipe(this=x,expression=y),B)
	return C(expressions=B,**G)
def simplify_conditionals(expression):
	'Simplifies expressions like IF, CASE if their condition is statically known.';F='true';E='ifs';A=expression
	if isinstance(A,exp.Case):
		D=A.this
		for C in A.args[E]:
			B=C.this
			if D:B=B.replace(D.pop().eq(B))
			if always_true(B):return C.args[F]
			if always_false(B):
				C.pop()
				if not A.args[E]:return A.args.get('default')or exp.null()
	elif isinstance(A,exp.If)and not isinstance(A.parent,exp.Case):
		if always_true(A.this):return A.args[F]
		if always_false(A.this):return A.args.get('false')or exp.null()
	return A
def simplify_startswith(expression):
	'\n    Reduces a prefix check to either TRUE or FALSE if both the string and the\n    prefix are statically known.\n\n    Example:\n        >>> from sqlglot import parse_one\n        >>> simplify_startswith(parse_one("STARTSWITH(\'foo\', \'f\')")).sql()\n        \'TRUE\'\n    ';A=expression
	if isinstance(A,exp.StartsWith)and A.this.is_string and A.expression.is_string:return exp.convert(A.name.startswith(A.expression.name))
	return A
DateRange=t.Tuple[datetime.date,datetime.date]
def _datetrunc_range(date,unit,dialect):
	"\n    Get the date range for a DATE_TRUNC equality comparison:\n\n    Example:\n        _datetrunc_range(date(2021-01-01), 'year') == (date(2021-01-01), date(2022-01-01))\n    Returns:\n        tuple of [min, max) or None if a value can never be equal to `date` for `unit`\n    ";A=date_floor(date,unit,dialect)
	if date!=A:return
	return A,A+interval(unit)
def _datetrunc_eq_expression(left,drange,target_type):'Get the logical expression for a date range';B=target_type;A=drange;return exp.and_(left>=date_literal(A[0],B),left<date_literal(A[1],B),copy=_A)
def _datetrunc_eq(left,date,unit,dialect,target_type):
	A=_datetrunc_range(date,unit,dialect)
	if not A:return
	return _datetrunc_eq_expression(left,A,target_type)
def _datetrunc_neq(left,date,unit,dialect,target_type):
	B=target_type;A=_datetrunc_range(date,unit,dialect)
	if not A:return
	return exp.and_(left<date_literal(A[0],B),left>=date_literal(A[1],B),copy=_A)
DATETRUNC_BINARY_COMPARISONS={exp.LT:lambda l,dt,u,d,t:l<date_literal(dt if dt==date_floor(dt,u,d)else date_floor(dt,u,d)+interval(u),t),exp.GT:lambda l,dt,u,d,t:l>=date_literal(date_floor(dt,u,d)+interval(u),t),exp.LTE:lambda l,dt,u,d,t:l<date_literal(date_floor(dt,u,d)+interval(u),t),exp.GTE:lambda l,dt,u,d,t:l>=date_literal(date_ceil(dt,u,d),t),exp.EQ:_datetrunc_eq,exp.NEQ:_datetrunc_neq}
DATETRUNC_COMPARISONS={exp.In,*DATETRUNC_BINARY_COMPARISONS}
DATETRUNCS=exp.DateTrunc,exp.TimestampTrunc
def _is_datetrunc_predicate(left,right):return isinstance(left,DATETRUNCS)and _is_date_literal(right)
@catch(ModuleNotFoundError,UnsupportedUnit)
def simplify_datetrunc(expression,dialect):
	"Simplify expressions like `DATE_TRUNC('year', x) >= CAST('2021-01-01' AS DATE)`";G=dialect;A=expression;I=A.__class__
	if isinstance(A,DATETRUNCS):
		J=A.this;L=extract_type(J);C=extract_date(J)
		if C and A.unit:return date_literal(date_floor(C,A.unit.name.lower(),G),L)
	elif I not in DATETRUNC_COMPARISONS:return A
	if isinstance(A,exp.Binary):
		B,D=A.left,A.right
		if not _is_datetrunc_predicate(B,D):return A
		B=t.cast(exp.DateTrunc,B);M=B.this;H=B.unit.name.lower();C=extract_date(D)
		if not C:return A
		return DATETRUNC_BINARY_COMPARISONS[I](M,C,H,G,extract_type(D))or A
	if isinstance(A,exp.In):
		B=A.this;F=A.expressions
		if F and all(_is_datetrunc_predicate(B,A)for A in F):
			B=t.cast(exp.DateTrunc,B);H=B.unit.name.lower();E=[]
			for D in F:
				C=extract_date(D)
				if not C:return A
				K=_datetrunc_range(C,H,G)
				if K:E.append(K)
			if not E:return A
			E=merge_ranges(E);N=extract_type(*F);return exp.or_(*[_datetrunc_eq_expression(B,A,N)for A in E],copy=_A)
	return A
def sort_comparison(expression):
	A=expression
	if A.__class__ in COMPLEMENT_COMPARISONS:
		C,B=A.this,A.expression;D=isinstance(C,exp.Column);E=isinstance(B,exp.Column);F=_is_constant(C);G=_is_constant(B)
		if D and not E or G and not F or isinstance(B,exp.SubqueryPredicate):return A
		if E and not D or F and not G or gen(C)>gen(B):return INVERSE_COMPARISONS.get(A.__class__,A.__class__)(this=B,expression=C)
	return A
JOINS={('',''),('','INNER'),('RIGHT',''),('RIGHT','OUTER')}
def remove_where_true(expression):
	B=expression
	for C in B.find_all(exp.Where):
		if always_true(C.this):C.pop()
	for A in B.find_all(exp.Join):
		if always_true(A.args.get('on'))and not A.args.get('using')and not A.args.get('method')and(A.side,A.kind)in JOINS:A.args['on'].pop();A.set('side',_B);A.set('kind','CROSS')
def always_true(expression):A=expression;return isinstance(A,exp.Boolean)and A.this or isinstance(A,exp.Literal)and not is_zero(A)
def always_false(expression):A=expression;return is_false(A)or is_null(A)or is_zero(A)
def is_zero(expression):A=expression;return isinstance(A,exp.Literal)and A.to_py()==0
def is_complement(a,b):return isinstance(b,exp.Not)and b.this==a
def is_false(a):return type(a)is exp.Boolean and not a.this
def is_null(a):return type(a)is exp.Null
def eval_boolean(expression,a,b):
	A=expression
	if isinstance(A,(exp.EQ,exp.Is)):return boolean_literal(a==b)
	if isinstance(A,exp.NEQ):return boolean_literal(a!=b)
	if isinstance(A,exp.GT):return boolean_literal(a>b)
	if isinstance(A,exp.GTE):return boolean_literal(a>=b)
	if isinstance(A,exp.LT):return boolean_literal(a<b)
	if isinstance(A,exp.LTE):return boolean_literal(a<=b)
def cast_as_date(value):
	A=value
	if isinstance(A,datetime.datetime):return A.date()
	if isinstance(A,datetime.date):return A
	try:return datetime.datetime.fromisoformat(A).date()
	except ValueError:return
def cast_as_datetime(value):
	A=value
	if isinstance(A,datetime.datetime):return A
	if isinstance(A,datetime.date):return datetime.datetime(year=A.year,month=A.month,day=A.day)
	try:return datetime.datetime.fromisoformat(A)
	except ValueError:return
def cast_value(value,to):
	A=value
	if not A:return
	if to.is_type(exp.DataType.Type.DATE):return cast_as_date(A)
	if to.is_type(*exp.DataType.TEMPORAL_TYPES):return cast_as_datetime(A)
def extract_date(cast):
	A=cast
	if isinstance(A,exp.Cast):B=A.to
	elif isinstance(A,exp.TsOrDsToDate)and not A.args.get('format'):B=exp.DataType.build(exp.DataType.Type.DATE)
	else:return
	if isinstance(A.this,exp.Literal):C=A.this.name
	elif isinstance(A.this,(exp.Cast,exp.TsOrDsToDate)):C=extract_date(A.this)
	else:return
	return cast_value(C,B)
def _is_date_literal(expression):return extract_date(expression)is not _B
def extract_interval(expression):
	A=expression
	try:B=int(A.this.to_py());C=A.text('unit').lower();return interval(C,B)
	except(UnsupportedUnit,ModuleNotFoundError,ValueError):return
def extract_type(*C):
	A=_B
	for B in C:
		A=B.to if isinstance(B,exp.Cast)else B.type
		if A:break
	return A
def date_literal(date,target_type=_B):
	A=target_type
	if not A or not A.is_type(*exp.DataType.TEMPORAL_TYPES):A=exp.DataType.Type.DATETIME if isinstance(date,datetime.datetime)else exp.DataType.Type.DATE
	return exp.cast(exp.Literal.string(date),A)
def interval(unit,n=1):
	A=unit;from dateutil.relativedelta import relativedelta as B
	if A=='year':return B(years=1*n)
	if A==_E:return B(months=3*n)
	if A=='month':return B(months=1*n)
	if A=='week':return B(weeks=1*n)
	if A=='day':return B(days=1*n)
	if A=='hour':return B(hours=1*n)
	if A=='minute':return B(minutes=1*n)
	if A=='second':return B(seconds=1*n)
	raise UnsupportedUnit(f"Unsupported unit: {A}")
def date_floor(d,unit,dialect):
	A=unit
	if A=='year':return d.replace(month=1,day=1)
	if A==_E:
		if d.month<=3:return d.replace(month=1,day=1)
		elif d.month<=6:return d.replace(month=4,day=1)
		elif d.month<=9:return d.replace(month=7,day=1)
		else:return d.replace(month=10,day=1)
	if A=='month':return d.replace(month=d.month,day=1)
	if A=='week':return d-datetime.timedelta(days=d.weekday()-dialect.WEEK_OFFSET)
	if A=='day':return d
	raise UnsupportedUnit(f"Unsupported unit: {A}")
def date_ceil(d,unit,dialect):
	A=date_floor(d,unit,dialect)
	if A==d:return d
	return A+interval(unit)
def boolean_literal(condition):return exp.true()if condition else exp.false()
def _flat_simplify(expression,simplifier,root=_C):
	A=expression
	if root or not A.same_parent:
		C=[];B=deque(A.flatten(unnest=_A));G=len(B)
		while B:
			E=B.popleft()
			for F in B:
				D=simplifier(A,E,F)
				if D and D is not A:B.remove(F);B.appendleft(D);break
			else:C.append(E)
		if len(C)<G:return functools.reduce(lambda a,b:A.__class__(this=a,expression=b),C)
	return A
def gen(expression):'Simple pseudo sql generator for quickly generating sortable and uniq strings.\n\n    Sorting and deduping sql is a necessary step for optimization. Calling the actual\n    generator is expensive so we have a bare minimum sql generator here.\n    ';return Gen().gen(expression)
class Gen:
	def __init__(A):A.stack=[];A.sqls=[]
	def gen(A,expression):
		A.stack=[expression];A.sqls.clear()
		while A.stack:
			B=A.stack.pop()
			if isinstance(B,exp.Expression):
				C=f"{B.key}_sql"
				if hasattr(A,C):getattr(A,C)(B)
				elif isinstance(B,exp.Func):A._function(B)
				else:D=B.key.upper();A.stack.append(f"{D} "if A._args(B)else D)
			elif type(B)is list:
				for E in reversed(B):
					if E is not _B:A.stack.extend((E,','))
				if B:A.stack.pop()
			elif B is not _B:A.sqls.append(str(B))
		return''.join(A.sqls)
	def add_sql(A,e):A._binary(e,' + ')
	def alias_sql(A,e):A.stack.extend((e.args.get(_D),' AS ',e.args.get('this')))
	def and_sql(A,e):A._binary(e,' AND ')
	def anonymous_sql(C,e):
		A=e.this
		if isinstance(A,str):B=A.upper()
		elif isinstance(A,exp.Identifier):B=A.this;B=f'"{B}"'if A.quoted else B.upper()
		else:raise ValueError(f"Anonymous.this expects a str or an Identifier, got '{A.__class__.__name__}'.")
		C.stack.extend((')',e.expressions,'(',B))
	def between_sql(A,e):A.stack.extend((e.args.get('high'),' AND ',e.args.get('low'),' BETWEEN ',e.this))
	def boolean_sql(A,e):A.stack.append('TRUE'if e.this else'FALSE')
	def bracket_sql(A,e):A.stack.extend((']',e.expressions,'[',e.this))
	def column_sql(A,e):
		for B in reversed(e.parts):A.stack.extend((B,'.'))
		A.stack.pop()
	def datatype_sql(A,e):A._args(e,1);A.stack.append(f"{e.this.name} ")
	def div_sql(A,e):A._binary(e,' / ')
	def dot_sql(A,e):A._binary(e,'.')
	def eq_sql(A,e):A._binary(e,' = ')
	def from_sql(A,e):A.stack.extend((e.this,'FROM '))
	def gt_sql(A,e):A._binary(e,' > ')
	def gte_sql(A,e):A._binary(e,' >= ')
	def identifier_sql(A,e):A.stack.append(f'"{e.this}"'if e.quoted else e.this)
	def ilike_sql(A,e):A._binary(e,' ILIKE ')
	def in_sql(A,e):A.stack.append(')');A._args(e,1);A.stack.extend(('(',' IN ',e.this))
	def intdiv_sql(A,e):A._binary(e,' DIV ')
	def is_sql(A,e):A._binary(e,' IS ')
	def like_sql(A,e):A._binary(e,' Like ')
	def literal_sql(A,e):A.stack.append(f"'{e.this}'"if e.is_string else e.this)
	def lt_sql(A,e):A._binary(e,' < ')
	def lte_sql(A,e):A._binary(e,' <= ')
	def mod_sql(A,e):A._binary(e,' % ')
	def mul_sql(A,e):A._binary(e,' * ')
	def neg_sql(A,e):A._unary(e,'-')
	def neq_sql(A,e):A._binary(e,' <> ')
	def not_sql(A,e):A._unary(e,'NOT ')
	def null_sql(A,e):A.stack.append('NULL')
	def or_sql(A,e):A._binary(e,' OR ')
	def paren_sql(A,e):A.stack.extend((')',e.this,'('))
	def sub_sql(A,e):A._binary(e,' - ')
	def subquery_sql(A,e):
		A._args(e,2);B=e.args.get(_D)
		if B:A.stack.append(B)
		A.stack.extend((')',e.this,'('))
	def table_sql(A,e):
		A._args(e,4);B=e.args.get(_D)
		if B:A.stack.append(B)
		for C in reversed(e.parts):A.stack.extend((C,'.'))
		A.stack.pop()
	def tablealias_sql(A,e):
		B=e.columns
		if B:A.stack.extend((')',B,'('))
		A.stack.extend((e.this,' AS '))
	def var_sql(A,e):A.stack.append(e.this)
	def _binary(A,e,op):A.stack.extend((e.expression,op,e.this))
	def _unary(A,e,op):A.stack.extend((e.this,op))
	def _function(A,e):A.stack.extend((')',list(e.args.values()),'(',e.sql_name()))
	def _args(G,node,arg_index=0):
		C=arg_index;A=node;B=[];D=list(A.arg_types)[C:]if C else A.arg_types
		for E in D or D:
			F=A.args.get(E)
			if F is not _B:B.append([f":{E}",F])
		if B:G.stack.append(B);return _C
		return _A