from __future__ import annotations
_C=False
_B=True
_A=None
import abc,typing as t
from sqlglot import expressions as exp
from sqlglot.dialects.dialect import Dialect
from sqlglot.errors import SchemaError
from sqlglot.helper import dict_depth,first
from sqlglot.trie import TrieResult,in_trie,new_trie
if t.TYPE_CHECKING:from sqlglot.dialects.dialect import DialectType;ColumnMapping=t.Union[t.Dict,str,t.List]
class Schema(abc.ABC):
	'Abstract base class for database schemas';dialect:DialectType
	@abc.abstractmethod
	def add_table(self,table,column_mapping=_A,dialect=_A,normalize=_A,match_depth=_B):"\n        Register or update a table. Some implementing classes may require column information to also be provided.\n        The added table must have the necessary number of qualifiers in its path to match the schema's nesting level.\n\n        Args:\n            table: the `Table` expression instance or string representing the table.\n            column_mapping: a column mapping that describes the structure of the table.\n            dialect: the SQL dialect that will be used to parse `table` if it's a string.\n            normalize: whether to normalize identifiers according to the dialect of interest.\n            match_depth: whether to enforce that the table must match the schema's depth or not.\n        "
	@abc.abstractmethod
	def column_names(self,table,only_visible=_C,dialect=_A,normalize=_A):"\n        Get the column names for a table.\n\n        Args:\n            table: the `Table` expression instance.\n            only_visible: whether to include invisible columns.\n            dialect: the SQL dialect that will be used to parse `table` if it's a string.\n            normalize: whether to normalize identifiers according to the dialect of interest.\n\n        Returns:\n            The sequence of column names.\n        "
	@abc.abstractmethod
	def get_column_type(self,table,column,dialect=_A,normalize=_A):"\n        Get the `sqlglot.exp.DataType` type of a column in the schema.\n\n        Args:\n            table: the source table.\n            column: the target column.\n            dialect: the SQL dialect that will be used to parse `table` if it's a string.\n            normalize: whether to normalize identifiers according to the dialect of interest.\n\n        Returns:\n            The resulting column type.\n        "
	def has_column(B,table,column,dialect=_A,normalize=_A):"\n        Returns whether `column` appears in `table`'s schema.\n\n        Args:\n            table: the source table.\n            column: the target column.\n            dialect: the SQL dialect that will be used to parse `table` if it's a string.\n            normalize: whether to normalize identifiers according to the dialect of interest.\n\n        Returns:\n            True if the column appears in the schema, False otherwise.\n        ";A=column;C=A if isinstance(A,str)else A.name;return C in B.column_names(table,dialect=dialect,normalize=normalize)
	@property
	@abc.abstractmethod
	def supported_table_args(self):'\n        Table arguments this schema support, e.g. `("this", "db", "catalog")`\n        '
	@property
	def empty(self):'Returns whether the schema is empty.';return _B
class AbstractMappingSchema:
	def __init__(A,mapping=_A):A.mapping=mapping or{};A.mapping_trie=new_trie(tuple(reversed(A))for A in flatten_schema(A.mapping,depth=A.depth()));A._supported_table_args=tuple()
	@property
	def empty(self):return not self.mapping
	def depth(A):return dict_depth(A.mapping)
	@property
	def supported_table_args(self):
		A=self
		if not A._supported_table_args and A.mapping:
			B=A.depth()
			if not B:A._supported_table_args=tuple()
			elif 1<=B<=3:A._supported_table_args=exp.TABLE_PARTS[:B]
			else:raise SchemaError(f"Invalid mapping shape. Depth: {B}")
		return A._supported_table_args
	def table_parts(B,table):
		A=table
		if isinstance(A.this,exp.ReadCSV):return[A.this.name]
		return[A.text(B)for B in exp.TABLE_PARTS if A.text(B)]
	def find(A,table,raise_on_missing=_B,ensure_data_types=_C):
		'\n        Returns the schema of a given table.\n\n        Args:\n            table: the target table.\n            raise_on_missing: whether to raise in case the schema is not found.\n            ensure_data_types: whether to convert `str` types to their `DataType` equivalents.\n\n        Returns:\n            The schema of the target table.\n        ';E=raise_on_missing;D=table;B=A.table_parts(D)[0:len(A.supported_table_args)];F,G=in_trie(A.mapping_trie,B)
		if F==TrieResult.FAILED:return
		if F==TrieResult.PREFIX:
			C=flatten_schema(G)
			if len(C)==1:B.extend(C[0])
			else:
				H=', '.join('.'.join(A)for A in C)
				if E:raise SchemaError(f"Ambiguous mapping for {D}: {H}.")
				return
		return A.nested_get(B,raise_on_missing=E)
	def nested_get(A,parts,d=_A,raise_on_missing=_B):return nested_get(d or A.mapping,*zip(A.supported_table_args,reversed(parts)),raise_on_missing=raise_on_missing)
class MappingSchema(AbstractMappingSchema,Schema):
	'\n    Schema based on a nested mapping.\n\n    Args:\n        schema: Mapping in one of the following forms:\n            1. {table: {col: type}}\n            2. {db: {table: {col: type}}}\n            3. {catalog: {db: {table: {col: type}}}}\n            4. None - Tables will be added later\n        visible: Optional mapping of which columns in the schema are visible. If not provided, all columns\n            are assumed to be visible. The nesting should mirror that of the schema:\n            1. {table: set(*cols)}}\n            2. {db: {table: set(*cols)}}}\n            3. {catalog: {db: {table: set(*cols)}}}}\n        dialect: The dialect to be used for custom type mappings & parsing string arguments.\n        normalize: Whether to normalize identifier names according to the given dialect or not.\n    '
	def __init__(A,schema=_A,visible=_A,dialect=_A,normalize=_B):C=visible;B=schema;A.dialect=dialect;A.visible={}if C is _A else C;A.normalize=normalize;A._type_mapping_cache={};A._depth=0;B={}if B is _A else B;super().__init__(A._normalize(B)if A.normalize else B)
	@classmethod
	def from_mapping_schema(B,mapping_schema):A=mapping_schema;return MappingSchema(schema=A.mapping,visible=A.visible,dialect=A.dialect,normalize=A.normalize)
	def find(C,table,raise_on_missing=_B,ensure_data_types=_C):
		B=ensure_data_types;A=super().find(table,raise_on_missing=raise_on_missing,ensure_data_types=B)
		if B and isinstance(A,dict):A={B:C._to_data_type(A)if isinstance(A,str)else A for(B,A)in A.items()}
		return A
	def copy(A,**B):return MappingSchema(**{'schema':A.mapping.copy(),'visible':A.visible.copy(),'dialect':A.dialect,'normalize':A.normalize,**B})
	def add_table(A,table,column_mapping=_A,dialect=_A,normalize=_A,match_depth=_B):
		"\n        Register or update a table. Updates are only performed if a new column mapping is provided.\n        The added table must have the necessary number of qualifiers in its path to match the schema's nesting level.\n\n        Args:\n            table: the `Table` expression instance or string representing the table.\n            column_mapping: a column mapping that describes the structure of the table.\n            dialect: the SQL dialect that will be used to parse `table` if it's a string.\n            normalize: whether to normalize identifiers according to the dialect of interest.\n            match_depth: whether to enforce that the table must match the schema's depth or not.\n        ";D=normalize;C=dialect;B=A._normalize_table(table,dialect=C,normalize=D)
		if match_depth and not A.empty and len(B.parts)!=A.depth():raise SchemaError(f"Table {B.sql(dialect=A.dialect)} must match the schema's nesting level: {A.depth()}.")
		E={A._normalize_name(B,dialect=C,normalize=D):E for(B,E)in ensure_column_mapping(column_mapping).items()};G=A.find(B,raise_on_missing=_C)
		if G and not E:return
		F=A.table_parts(B);nested_set(A.mapping,tuple(reversed(F)),E);new_trie([F],A.mapping_trie)
	def column_names(A,table,only_visible=_C,dialect=_A,normalize=_A):
		C=A._normalize_table(table,dialect=dialect,normalize=normalize);B=A.find(C)
		if B is _A:return[]
		if not only_visible or not A.visible:return list(B)
		D=A.nested_get(A.table_parts(C),A.visible)or[];return[A for A in B if A in D]
	def get_column_type(A,table,column,dialect=_A,normalize=_A):
		E=normalize;D=dialect;C=column;G=A._normalize_table(table,dialect=D,normalize=E);H=A._normalize_name(C if isinstance(C,str)else C.this,dialect=D,normalize=E);F=A.find(G,raise_on_missing=_C)
		if F:
			B=F.get(H)
			if isinstance(B,exp.DataType):return B
			elif isinstance(B,str):return A._to_data_type(B,dialect=D)
		return exp.DataType.build('unknown')
	def has_column(A,table,column,dialect=_A,normalize=_A):D=normalize;C=dialect;B=column;F=A._normalize_table(table,dialect=C,normalize=D);G=A._normalize_name(B if isinstance(B,str)else B.this,dialect=C,normalize=D);E=A.find(F,raise_on_missing=_C);return G in E if E else _C
	def _normalize(D,schema):
		'\n        Normalizes all identifiers in the schema.\n\n        Args:\n            schema: the schema to normalize.\n\n        Returns:\n            The normalized schema mapping.\n        ';E=schema;F={};C=flatten_schema(E);G="Table {} must match the schema's nesting level: {}."
		for A in C:
			B=nested_get(E,*zip(A,A))
			if not isinstance(B,dict):raise SchemaError(G.format('.'.join(A[:-1]),len(C[0])))
			if not B:raise SchemaError(f"Table {".".join(A[:-1])} must have at least one column")
			if isinstance(first(B.values()),dict):raise SchemaError(G.format('.'.join(A+flatten_schema(B)[0]),len(C[0])))
			H=[D._normalize_name(A,is_table=_B)for A in A]
			for(I,J)in B.items():nested_set(F,H+[D._normalize_name(I)],J)
		return F
	def _normalize_table(D,table,dialect=_A,normalize=_A):
		B=dialect;A=normalize;B=B or D.dialect;A=D.normalize if A is _A else A;C=exp.maybe_parse(table,into=exp.Table,dialect=B,copy=A)
		if A:
			for E in exp.TABLE_PARTS:
				F=C.args.get(E)
				if isinstance(F,exp.Identifier):C.set(E,normalize_name(F,dialect=B,is_table=_B,normalize=A))
		return C
	def _normalize_name(A,name,dialect=_A,is_table=_C,normalize=_A):B=normalize;return normalize_name(name,dialect=dialect or A.dialect,is_table=is_table,normalize=A.normalize if B is _A else B).name
	def depth(A):
		if not A.empty and not A._depth:A._depth=super().depth()-1
		return A._depth
	def _to_data_type(C,schema_type,dialect=_A):
		'\n        Convert a type represented as a string to the corresponding `sqlglot.exp.DataType` object.\n\n        Args:\n            schema_type: the type we want to convert.\n            dialect: the SQL dialect that will be used to parse `schema_type`, if needed.\n\n        Returns:\n            The resulting expression type.\n        ';B=schema_type;A=dialect
		if B not in C._type_mapping_cache:
			A=A or C.dialect;D=Dialect.get_or_raise(A).SUPPORTS_USER_DEFINED_TYPES
			try:E=exp.DataType.build(B,dialect=A,udt=D);C._type_mapping_cache[B]=E
			except AttributeError:F=f" in dialect {A}"if A else'';raise SchemaError(f"Failed to build type '{B}'{F}.")
		return C._type_mapping_cache[B]
def normalize_name(identifier,dialect=_A,is_table=_C,normalize=_B):
	B=dialect;A=identifier
	if isinstance(A,str):A=exp.parse_identifier(A,dialect=B)
	if not normalize:return A
	A.meta['is_table']=is_table;return Dialect.get_or_raise(B).normalize_identifier(A)
def ensure_schema(schema,**B):
	A=schema
	if isinstance(A,Schema):return A
	return MappingSchema(A,**B)
def ensure_column_mapping(mapping):
	A=mapping
	if A is _A:return{}
	elif isinstance(A,dict):return A
	elif isinstance(A,str):B=[A.strip()for A in A.split(',')];return{A.split(':')[0].strip():A.split(':')[1].strip()for A in B}
	elif isinstance(A,list):return{A.strip():_A for A in A}
	raise ValueError(f"Invalid mapping provided: {type(A)}")
def flatten_schema(schema,depth=_A,keys=_A):
	D=schema;B=keys;A=depth;C=[];B=B or[];A=dict_depth(D)-1 if A is _A else A
	for(E,F)in D.items():
		if A==1 or not isinstance(F,dict):C.append(B+[E])
		elif A>=2:C.extend(flatten_schema(F,A-1,B+[E]))
	return C
def nested_get(d,*C,raise_on_missing=_B):
	"\n    Get a value for a nested dictionary.\n\n    Args:\n        d: the dictionary to search.\n        *path: tuples of (name, key), where:\n            `key` is the key in the dictionary to get.\n            `name` is a string to use in the error if `key` isn't found.\n\n    Returns:\n        The value or None if it doesn't exist.\n    "
	for(A,B)in C:
		d=d.get(B)
		if d is _A:
			if raise_on_missing:A='table'if A=='this'else A;raise ValueError(f"Unknown {A}: {B}")
			return
	return d
def nested_set(d,keys,value):
	'\n    In-place set a value for a nested dictionary\n\n    Example:\n        >>> nested_set({}, ["top_key", "second_key"], "value")\n        {\'top_key\': {\'second_key\': \'value\'}}\n\n        >>> nested_set({"top_key": {"third_key": "third_value"}}, ["top_key", "second_key"], "value")\n        {\'top_key\': {\'third_key\': \'third_value\', \'second_key\': \'value\'}}\n\n    Args:\n        d: dictionary to update.\n        keys: the keys that makeup the path to `value`.\n        value: the value to set in the dictionary for the given key path.\n\n    Returns:\n        The (possibly) updated dictionary.\n    ';D=value;B=keys
	if not B:return d
	if len(B)==1:d[B[0]]=D;return d
	A=d
	for C in B[:-1]:
		if C not in A:A=A.setdefault(C,{})
		else:A=A[C]
	A[B[-1]]=D;return d