from __future__ import annotations
_M='offset'
_L='recursive'
_K='this'
_J='order'
_I='with'
_H='from'
_G=True
_F='joins'
_E='on'
_D=None
_C='expressions'
_B='alias'
_A=False
import typing as t
from sqlglot import expressions as exp
from sqlglot.errors import UnsupportedError
from sqlglot.helper import find_new_name,name_sequence
if t.TYPE_CHECKING:from sqlglot._typing import E;from sqlglot.generator import Generator
def preprocess(transforms):
	'\n    Creates a new transform by chaining a sequence of transformations and converts the resulting\n    expression to SQL, using either the "_sql" method corresponding to the resulting expression,\n    or the appropriate `Generator.TRANSFORMS` function (when applicable -- see below).\n\n    Args:\n        transforms: sequence of transform functions. These will be called in order.\n\n    Returns:\n        Function that can be used as a generator transform.\n    ';C=transforms
	def A(self,expression):
		B=self;A=expression;F=type(A)
		try:
			A=C[0](A)
			for G in C[1:]:A=G(A)
		except UnsupportedError as H:B.unsupported(str(H))
		D=getattr(B,A.key+'_sql',_D)
		if D:return D(A)
		E=B.TRANSFORMS.get(type(A))
		if E:
			if F is type(A):
				if isinstance(A,exp.Func):return B.function_fallback_sql(A)
				raise ValueError(f"Expression type {A.__class__.__name__} requires a _sql method in order to be transformed.")
			return E(B,A)
		raise ValueError(f"Unsupported expression type {A.__class__.__name__}.")
	return A
def unnest_generate_date_array_using_recursive_cte(expression):
	K='date';B=expression
	if isinstance(B,exp.Select):
		G=0;H=[]
		for A in B.find_all(exp.Unnest):
			if not isinstance(A.parent,(exp.From,exp.Join))or len(A.expressions)!=1 or not isinstance(A.expressions[0],exp.GenerateDateArray):continue
			I=A.expressions[0];C=I.args.get('start');L=I.args.get('end');J=I.args.get('step')
			if not C or not L or not isinstance(J,exp.Interval):continue
			M=A.args.get(_B);D=M.columns[0]if isinstance(M,exp.TableAlias)else'date_value';C=exp.cast(C,K);O=exp.func('date_add',D,exp.Literal.number(J.name),J.args.get('unit'));N=exp.cast(O,K);E='_generated_dates'+(f"_{G}"if G else'');P=exp.select(C.as_(D));Q=exp.select(N).from_(E).where(N<=exp.cast(L,K));R=P.union(Q,distinct=_A);S=exp.select(D).from_(E);A.replace(S.subquery(E));H.append(exp.alias_(exp.CTE(this=R),E,table=[D]));G+=1
		if H:F=B.args.get(_I)or exp.With();F.set(_L,_G);F.set(_C,[*H,*F.expressions]);B.set(_I,F)
	return B
def unnest_generate_series(expression):
	'Unnests GENERATE_SERIES or SEQUENCE table references.';A=expression;B=A.this
	if isinstance(A,exp.Table)and isinstance(B,exp.GenerateSeries):
		C=exp.Unnest(expressions=[B])
		if A.alias:return exp.alias_(C,alias='_u',table=[A.alias],copy=_A)
		return C
	return A
def unalias_group(expression):
	'\n    Replace references to select aliases in GROUP BY clauses.\n\n    Example:\n        >>> import sqlglot\n        >>> sqlglot.parse_one("SELECT a AS b FROM x GROUP BY b").transform(unalias_group).sql()\n        \'SELECT a AS b FROM x GROUP BY 1\'\n\n    Args:\n        expression: the expression that will be transformed.\n\n    Returns:\n        The transformed expression.\n    ';A=expression
	if isinstance(A,exp.Group)and isinstance(A.parent,exp.Select):
		C={A.alias:B for(B,A)in enumerate(A.parent.expressions,start=1)if isinstance(A,exp.Alias)}
		for B in A.expressions:
			if isinstance(B,exp.Column)and not B.table and B.name in C:B.replace(exp.Literal.number(C.get(B.name)))
	return A
def eliminate_distinct_on(expression):
	"\n    Convert SELECT DISTINCT ON statements to a subquery with a window function.\n\n    This is useful for dialects that don't support SELECT DISTINCT ON but support window functions.\n\n    Args:\n        expression: the expression that will be transformed.\n\n    Returns:\n        The transformed expression.\n    ";C='distinct';A=expression
	if isinstance(A,exp.Select)and A.args.get(C)and A.args[C].args.get(_E)and isinstance(A.args[C].args[_E],exp.Tuple):
		D=A.args[C].pop().args[_E].expressions;G=A.selects;E=find_new_name(A.named_selects,'_row_number');B=exp.Window(this=exp.RowNumber(),partition_by=D);F=A.args.get(_J)
		if F:B.set(_J,F.pop())
		else:B.set(_J,exp.Order(expressions=[A.copy()for A in D]))
		B=exp.alias_(B,E);A.select(B,copy=_A);return exp.select(*G,copy=_A).from_(A.subquery('_t',copy=_A),copy=_A).where(exp.column(E).eq(1),copy=_A)
	return A
def eliminate_qualify(expression):
	"\n    Convert SELECT statements that contain the QUALIFY clause into subqueries, filtered equivalently.\n\n    The idea behind this transformation can be seen in Snowflake's documentation for QUALIFY:\n    https://docs.snowflake.com/en/sql-reference/constructs/qualify\n\n    Some dialects don't support window functions in the WHERE clause, so we need to include them as\n    projections in the subquery, in order to refer to them in the outer filter using aliases. Also,\n    if a column is referenced in the QUALIFY clause but is not selected, we need to include it too,\n    otherwise we won't be able to refer to it in the outer query's WHERE clause. Finally, if a\n    newly aliased projection is referenced in the QUALIFY clause, it will be replaced by the\n    corresponding expression to avoid creating invalid column references.\n    ";J='qualify';A=expression
	if isinstance(A,exp.Select)and A.args.get(J):
		G=set(A.named_selects)
		for E in A.selects:
			if not E.alias_or_name:C=find_new_name(G,'_c');E.replace(exp.alias_(E,C));G.add(C)
		def K(select):
			A=select;B=A.alias_or_name;C=A.args.get(_B)or A.this
			if isinstance(C,exp.Identifier):return exp.column(B,quoted=C.args.get('quoted'))
			return B
		L=exp.select(*list(map(K,A.selects)));F=A.args[J].pop().this;H={A.alias:A.this for A in A.selects if isinstance(A,exp.Alias)};M=exp.Window if A.is_star else(exp.Window,exp.Column)
		for B in F.find_all(M):
			if isinstance(B,exp.Window):
				if H:
					for D in B.find_all(exp.Column):
						I=H.get(D.name)
						if I:D.replace(I)
				C=find_new_name(A.named_selects,'_w');A.select(exp.alias_(B,C),copy=_A);D=exp.column(C)
				if isinstance(B.parent,exp.Qualify):F=D
				else:B.replace(D)
			elif B.name not in A.named_selects:A.select(B.copy(),copy=_A)
		return L.from_(A.subquery(alias='_t',copy=_A),copy=_A).where(F,copy=_A)
	return A
def remove_precision_parameterized_types(expression):
	'\n    Some dialects only allow the precision for parameterized types to be defined in the DDL and not in\n    other expressions. This transforms removes the precision from parameterized types in expressions.\n    ';A=expression
	for B in A.find_all(exp.DataType):B.set(_C,[A for A in B.expressions if not isinstance(A,exp.DataTypeParam)])
	return A
def unqualify_unnest(expression):
	"Remove references to unnest table aliases, added by the optimizer's qualify_columns step.";A=expression;from sqlglot.optimizer.scope import find_all_in_scope as D
	if isinstance(A,exp.Select):
		C={A.alias for A in D(A,exp.Unnest)if isinstance(A.parent,(exp.From,exp.Join))}
		if C:
			for B in A.find_all(exp.Column):
				if B.table in C:B.set('table',_D)
				elif B.db in C:B.set('db',_D)
	return A
def unnest_to_explode(expression,unnest_using_arrays_zip=_G):
	'Convert cross join unnest into lateral view explode.';H=unnest_using_arrays_zip;D=expression
	def I(u,unnest_exprs,has_multi_expr):
		A=unnest_exprs
		if has_multi_expr:
			if not H:raise UnsupportedError('Cannot transpile UNNEST with multiple input arrays')
			B=[exp.Anonymous(this='ARRAYS_ZIP',expressions=A)];u.set(_C,B);return B
		return A
	def J(u,has_multi_expr):
		if u.args.get(_M):return exp.Posexplode
		return exp.Inline if has_multi_expr else exp.Explode
	if isinstance(D,exp.Select):
		G=D.args.get(_H)
		if G and isinstance(G.this,exp.Unnest):A=G.this;B=A.args.get(_B);C=A.expressions;E=len(C)>1;N,*O=I(A,C,E);A.replace(exp.Table(this=J(A,E)(this=N,expressions=O),alias=exp.TableAlias(this=B.this,columns=B.columns)if B else _D))
		for K in D.args.get(_F)or[]:
			F=K.this;L=isinstance(F,exp.Lateral);A=F.this if L else F
			if isinstance(A,exp.Unnest):
				if L:B=F.args.get(_B)
				else:B=A.args.get(_B)
				C=A.expressions;E=len(C)>1;C=I(A,C,E);D.args[_F].remove(K);M=B.columns if B else[]
				for(P,Q)in zip(C,M):D.append('laterals',exp.Lateral(this=J(A,E)(this=P),view=_G,alias=exp.TableAlias(this=B.this,columns=M if H else[Q])))
	return D
def explode_to_unnest(index_offset=0):
	'Convert explode/posexplode into unnest.';N=index_offset
	def A(expression):
		W='CROSS';T='pos';A=expression
		if isinstance(A,exp.Select):
			from sqlglot.optimizer.scope import Scope;I=set(A.named_selects);U={A for(A,B)in Scope(A).references}
			def H(names,name):B=names;A=name;A=find_new_name(B,A);B.add(A);return A
			J=[];K=H(I,T);O=exp.alias_(exp.Unnest(expressions=[exp.GenerateSeries(start=exp.Literal.number(N))]),H(U,'_u'),table=[K])
			for B in list(A.selects):
				E=B.find(exp.Explode)
				if E:
					C='';F=''
					if isinstance(B,exp.Alias):F=B.args[_B];L=B
					elif isinstance(B,exp.Aliases):C=B.aliases[0];F=B.aliases[1];L=B.replace(exp.alias_(B.this,'',copy=_A))
					else:L=B.replace(exp.alias_(B,''));E=L.find(exp.Explode);assert E
					V=isinstance(E,exp.Posexplode);D=E.this
					if isinstance(E,exp.ExplodeOuter):Q=D[0];Q.set('safe',_G);Q.set(_M,_G);D=exp.func('IF',exp.func('ARRAY_SIZE',exp.func('COALESCE',D,exp.Array())).eq(0),exp.array(Q,copy=_A),D)
					if isinstance(D,exp.Column):I.add(D.output_name)
					G=H(U,'_u')
					if not F:
						F=H(I,'col')
						if V:C=H(I,T)
					if not C:C=H(I,T)
					L.set(_B,exp.to_identifier(F));P=O.args[_B].this;X=exp.If(this=exp.column(K,table=P).eq(exp.column(C,table=G)),true=exp.column(F,table=G));E.replace(X)
					if V:R=A.expressions;R.insert(R.index(L)+1,exp.If(this=exp.column(K,table=P).eq(exp.column(C,table=G)),true=exp.column(C,table=G)).as_(C));A.set(_C,R)
					if not J:
						if A.args.get(_H):A.join(O,copy=_A,join_type=W)
						else:A.from_(O,copy=_A)
					M=exp.ArraySize(this=D.copy());J.append(M);A.join(exp.alias_(exp.Unnest(expressions=[D.copy()],offset=exp.to_identifier(C)),G,table=[F]),join_type=W,copy=_A)
					if N!=1:M=M-1
					A.where(exp.column(K,table=P).eq(exp.column(C,table=G)).or_((exp.column(K,table=P)>M).and_(exp.column(C,table=G).eq(M))),copy=_A)
			if J:
				S=exp.Greatest(this=J[0],expressions=J[1:])
				if N!=1:S=S-(1-N)
				O.expressions[0].set('end',S)
		return A
	return A
def add_within_group_for_percentiles(expression):
	'Transforms percentiles by adding a WITHIN GROUP clause to them.';A=expression
	if isinstance(A,exp.PERCENTILES)and not isinstance(A.parent,exp.WithinGroup)and A.expression:B=A.this.pop();A.set(_K,A.expression.pop());C=exp.Order(expressions=[exp.Ordered(this=B)]);A=exp.WithinGroup(this=A,expression=C)
	return A
def remove_within_group_for_percentiles(expression):
	'Transforms percentiles by getting rid of their corresponding WITHIN GROUP clause.';A=expression
	if isinstance(A,exp.WithinGroup)and isinstance(A.this,exp.PERCENTILES)and isinstance(A.expression,exp.Order):B=A.this.this;C=t.cast(exp.Ordered,A.find(exp.Ordered)).this;return A.replace(exp.ApproxQuantile(this=C,quantile=B))
	return A
def add_recursive_cte_column_names(expression):
	"Uses projection output names in recursive CTE definitions to define the CTEs' columns.";A=expression
	if isinstance(A,exp.With)and A.recursive:
		D=name_sequence('_c_')
		for C in A.expressions:
			if not C.args[_B].columns:
				B=C.this
				if isinstance(B,exp.SetOperation):B=B.this
				C.args[_B].set('columns',[exp.to_identifier(A.alias_or_name or D())for A in B.selects])
	return A
def epoch_cast_to_ts(expression):
	"Replace 'epoch' in casts by the equivalent date literal.";A=expression
	if isinstance(A,(exp.Cast,exp.TryCast))and A.name.lower()=='epoch'and A.to.this in exp.DataType.TEMPORAL_TYPES:A.this.replace(exp.Literal.string('1970-01-01 00:00:00'))
	return A
def eliminate_semi_and_anti_joins(expression):
	'Convert SEMI and ANTI joins into equivalent forms that use EXIST instead.';E='ANTI';B=expression
	if isinstance(B,exp.Select):
		for A in B.args.get(_F)or[]:
			D=A.args.get(_E)
			if D and A.kind in('SEMI',E):
				F=exp.select('1').from_(A.this).where(D);C=exp.Exists(this=F)
				if A.kind==E:C=C.not_(copy=_A)
				A.pop();B.where(C,copy=_A)
	return B
def eliminate_full_outer_join(expression):
	'\n    Converts a query with a FULL OUTER join to a union of identical queries that\n    use LEFT/RIGHT OUTER joins instead. This transformation currently only works\n    for queries that have a single FULL OUTER join.\n    ';F='side';A=expression
	if isinstance(A,exp.Select):
		D=[(B,A)for(B,A)in enumerate(A.args.get(_F)or[])if A.side=='FULL']
		if len(D)==1:B=A.copy();A.set('limit',_D);G,C=D[0];E=A.args[_H].alias_or_name,C.alias_or_name;H=C.args.get(_E)or exp.and_(*[exp.column(A,E[0]).eq(exp.column(A,E[1]))for A in C.args.get('using')]);C.set(F,'left');I=exp.select('1').from_(A.args[_H]).where(H);B.args[_F][G].set(F,'right');B=B.where(exp.Exists(this=I).not_());B.args.pop(_I,_D);A.args.pop(_J,_D);return exp.union(A,B,copy=_A,distinct=_A)
	return A
def move_ctes_to_top_level(expression):
	'\n    Some dialects (e.g. Hive, T-SQL, Spark prior to version 3) only allow CTEs to be\n    defined at the top-level, so for example queries like:\n\n        SELECT * FROM (WITH t(c) AS (SELECT 1) SELECT * FROM t) AS subq\n\n    are invalid in those dialects. This transformation can be used to ensure all CTEs are\n    moved to the top level so that the final SQL code is valid from a syntax standpoint.\n\n    TODO: handle name clashes whilst moving CTEs (it can get quite tricky & costly).\n    ';C=expression;A=C.args.get(_I)
	for B in C.find_all(exp.With):
		if B.parent is C:continue
		if not A:A=B.pop();C.set(_I,A)
		else:
			if B.recursive:A.set(_L,_G)
			D=B.find_ancestor(exp.CTE);B.pop()
			if D:E=A.expressions.index(D);A.expressions[E:E]=B.expressions;A.set(_C,A.expressions)
			else:A.set(_C,A.expressions+B.expressions)
	return C
def ensure_bools(expression):
	'Converts numeric values used in conditions into explicit boolean expressions.';A=expression;from sqlglot.optimizer.canonicalize import ensure_bools as B
	def C(node):
		A=node
		if A.is_number or not isinstance(A,exp.SubqueryPredicate)and A.is_type(exp.DataType.Type.UNKNOWN,*exp.DataType.NUMERIC_TYPES)or isinstance(A,exp.Column)and not A.type:A.replace(A.neq(0))
	for D in A.walk():B(D,C)
	return A
def unqualify_columns(expression):
	A=expression
	for B in A.find_all(exp.Column):
		for C in B.parts[:-1]:C.pop()
	return A
def remove_unique_constraints(expression):
	A=expression;assert isinstance(A,exp.Create)
	for B in A.find_all(exp.UniqueColumnConstraint):
		if B.parent:B.parent.pop()
	return A
def ctas_with_tmp_tables_to_create_tmp_view(expression,tmp_storage_provider=lambda e:e):
	A=expression;assert isinstance(A,exp.Create);B=A.args.get('properties');C=any(isinstance(A,exp.TemporaryProperty)for A in(B.expressions if B else[]))
	if A.kind=='TABLE'and C:
		if A.expression:return exp.Create(kind='TEMPORARY VIEW',this=A.this,expression=A.expression)
		return tmp_storage_provider(A)
	return A
def move_schema_columns_to_partitioned_by(expression):
	"\n    In Hive, the PARTITIONED BY property acts as an extension of a table's schema. When the\n    PARTITIONED BY value is an array of column names, they are transformed into a schema.\n    The corresponding columns are removed from the create statement.\n    ";A=expression;assert isinstance(A,exp.Create);E=isinstance(A.this,exp.Schema);F=A.kind in{'TABLE','VIEW'}
	if E and F:
		B=A.find(exp.PartitionedByProperty)
		if B and B.this and not isinstance(B.this,exp.Schema):C=A.this;G={A.name.upper()for A in B.this.expressions};D=[A for A in C.expressions if A.name.upper()in G];C.set(_C,[A for A in C.expressions if A not in D]);B.replace(exp.PartitionedByProperty(this=exp.Schema(expressions=D)));A.set(_K,C)
	return A
def move_partitioned_by_to_schema_columns(expression):
	'\n    Spark 3 supports both "HIVEFORMAT" and "DATASOURCE" formats for CREATE TABLE.\n\n    Currently, SQLGlot uses the DATASOURCE format for Spark 3.\n    ';B=expression;assert isinstance(B,exp.Create);A=B.find(exp.PartitionedByProperty)
	if A and A.this and isinstance(A.this,exp.Schema)and all(isinstance(A,exp.ColumnDef)and A.kind for A in A.this.expressions):
		C=exp.Tuple(expressions=[exp.to_identifier(A.this)for A in A.this.expressions]);D=B.this
		for E in A.this.expressions:D.append(_C,E)
		A.set(_K,C)
	return B
def struct_kv_to_alias(expression):
	'Converts struct arguments to aliases, e.g. STRUCT(1 AS y).';A=expression
	if isinstance(A,exp.Struct):A.set(_C,[exp.alias_(A.expression,A.this)if isinstance(A,exp.PropertyEQ)else A for A in A.expressions])
	return A
def eliminate_join_marks(expression):
	'\n    Remove join marks from an AST. This rule assumes that all marked columns are qualified.\n    If this does not hold for a query, consider running `sqlglot.optimizer.qualify` first.\n\n    For example,\n        SELECT * FROM a, b WHERE a.id = b.id(+)    -- ... is converted to\n        SELECT * FROM a LEFT JOIN b ON a.id = b.id -- this\n\n    Args:\n        expression: The AST to remove join marks from.\n\n    Returns:\n       The AST with join marks removed.\n    ';L=expression;E='join_mark';from sqlglot.optimizer.scope import traverse_scope as W
	for M in W(L):
		B=M.expression;F=B.args.get('where');N=B.args.get(_F)
		if not F or not N:continue
		O=B.args[_H];C={};G={A.alias_or_name:A for A in N}
		for P in M.columns:
			if not P.args.get(E):continue
			H=P.find_ancestor(exp.Predicate,exp.Select);assert isinstance(H,exp.Binary),'Columns can only be marked with (+) when involved in a binary operation';A=H.parent;I=H.pop();Q=[A for A in I.left.find_all(exp.Column)if A.args.get(E)];R=[A for A in I.right.find_all(exp.Column)if A.args.get(E)];assert not(Q and R),'The (+) marker cannot appear in both sides of a binary predicate';S=set()
			for D in Q or R:T=D.table;assert T,f"Column {D} needs to be qualified with a table";D.set(E,_A);S.add(T)
			assert len(S)==1,'Columns of only a single table can be marked with (+) in a given binary predicate';X=G.get(D.table,O).this;J=exp.Join(this=X,on=I,kind='LEFT');U=J.alias_or_name;K=C.get(U)
			if K:K.set(_E,exp.and_(K.args.get(_E),J.args[_E]))
			else:C[U]=J
			if isinstance(A,exp.Binary):
				if A.left is _D:A.replace(A.right)
				else:A.replace(A.left)
		if O.alias_or_name in C:V=G.keys()-C.keys();assert len(V)>=1,'Cannot determine which table to use in the new FROM clause';Y=list(V)[0];B.set(_H,exp.From(this=G[Y].this))
		B.set(_F,list(C.values()))
		if not F.this:F.pop()
	return L