import typing as t
from enum import Enum,auto
key=t.Sequence[t.Hashable]
class TrieResult(Enum):FAILED=auto();PREFIX=auto();EXISTS=auto()
def new_trie(keywords,trie=None):
	'\n    Creates a new trie out of a collection of keywords.\n\n    The trie is represented as a sequence of nested dictionaries keyed by either single\n    character strings, or by 0, which is used to designate that a keyword is in the trie.\n\n    Example:\n        >>> new_trie(["bla", "foo", "blab"])\n        {\'b\': {\'l\': {\'a\': {0: True, \'b\': {0: True}}}}, \'f\': {\'o\': {\'o\': {0: True}}}}\n\n    Args:\n        keywords: the keywords to create the trie from.\n        trie: a trie to mutate instead of creating a new one\n\n    Returns:\n        The trie corresponding to `keywords`.\n    ';A=trie;A={}if A is None else A
	for C in keywords:
		B=A
		for D in C:B=B.setdefault(D,{})
		B[0]=True
	return A
def in_trie(trie,key):
	'\n    Checks whether a key is in a trie.\n\n    Examples:\n        >>> in_trie(new_trie(["cat"]), "bob")\n        (<TrieResult.FAILED: 1>, {\'c\': {\'a\': {\'t\': {0: True}}}})\n\n        >>> in_trie(new_trie(["cat"]), "ca")\n        (<TrieResult.PREFIX: 2>, {\'t\': {0: True}})\n\n        >>> in_trie(new_trie(["cat"]), "cat")\n        (<TrieResult.EXISTS: 3>, {0: True})\n\n    Args:\n        trie: The trie to be searched.\n        key: The target key.\n\n    Returns:\n        A pair `(value, subtrie)`, where `subtrie` is the sub-trie we get at the point\n        where the search stops, and `value` is a TrieResult value that can be one of:\n\n        - TrieResult.FAILED: the search was unsuccessful\n        - TrieResult.PREFIX: `value` is a prefix of a keyword in `trie`\n        - TrieResult.EXISTS: `key` exists in `trie`\n    '
	if not key:return TrieResult.FAILED,trie
	A=trie
	for B in key:
		if B not in A:return TrieResult.FAILED,A
		A=A[B]
	if 0 in A:return TrieResult.EXISTS,A
	return TrieResult.PREFIX,A